unit chains;

{  Unit for defining the data structures for use with chains in Siena.         }
{  This unit was written by Tom A.B. Snijders, December 2005.                  }
{  It uses some building blocks (adapted) of unit Digraph, which uses          }
{  types TDoublyLinkedDoubleInt and TDoublyLinkedIntList based                 }
{  on the unit cLinkedLists written by David Butler,                           }
{  which is part of Delphi Fundamentals.                                       }
{  This latter unit contained an omission; see below, noted TS.                }

interface

uses classes, SysUtils, Math, rangen, slib;


  (* This unit represents the embedded chain in one period.
   The embedded chain is a sequence of ministeps (type TMiniStep).
   Each ministep is defined as the combination of
   Fash  = aspect: network or behavior;
   Fih   = actor making the ministep;
   Fjh   = for Fash = Network, the actor towards whom the tie variable is changed;
   Fvarh = the number of the variable (1 for Network aspect)
   Fdifh = the amount of the change (-1, 0 or 1) (not used for Network aspect).
   For Fash = Network, since only binary variables are considered as yet,
              Fdifh plays no role.
   For Fash = Behavior, Fjh plays no role.
   The vector (Fash, Fih, Fjh, Fvarh, Fdifh) is called the kind of the ministep.
   The kind, disregarding sign(Fdifh), is called the rkind (restricted kind).

   Ministeps are labeled by an index called kindex.
   Ministeps are diagonal if they represent no change: this is the case iff
     Fih = Fjh for the Network aspect, Fdifh = 0 for the Behavior aspect.
   Diagonal ministeps are also labeled by the index called dindex.

   The ministeps are arranged in a chain (type TChain).
   The chain goes from Ffirst (with Ffirst.kindex = 0)
     to Flast (with Flast.kindex = 1).
   Ffirst and Flast are empty ministeps in the sense that
   they are not interpreted as changes in the embedded chain;
   their kind has no meaning.
   Ministeps are called multiple if the chain contains 2 or more of their rkind.
   Multiple ministeps are also labeled by the index called mindex.
   Ffirst and Flast are neither diagonal nor multiple.

   k.predh and k.such are the immediate predecessor and successor
   of ministep k in the time ordering;
   k.predhrkind and k.suchrkind are the ministeps of the same rkind as k
   which are the last earlier and the first later with this property;
   if these do not exist, then they are nil.
   FirstOfRkind(kind) is the first occurrence of this kind,
   and FirstOfRkind(kind) = nil means this kind does not occur at all;
   k.Fnrkind is the total number of ministeps of the same rkind as k.

   Two ministeps k1, k2 are a CCP (consecutive canceling pair)
   if they are of the same rkind, not diagonal, cancel each other's effect,
   have no other of the same rkind in between, and are ordered; *)
(* [[[ *)
(*   and there is at least one element of a different rkind in beween; *)
(* ]]] *)
(*
   This is in order to deal with the double-counting in the
   proposal probabilities in S_ML.
   First elements ("k1") of CCPs are labeled by the index cindex.

   TotNumber is the total number of ministeps including Last but not First;
   kindex runs from 0 to TotNumber.
   DiagNumber is the total number of diagonal ministeps;
   dindex runs from 1 to DiagNumber;
   MultipleNumber is the total number of multiple ministeps;
   mindex runs from 1 to MultipleNumber.
   CCPNumber is the total number of CCPs;

   The ordering of k.time gives the ordering in time of the ministeps.
   The numerical values apart from the ordering are meaningless.

   Of course this is a highly redundant representation, constructed
   for efficient calculations, and all modifications must painstakingly respect
   the consistency of the representation.

   For understanding the use of this unit in S_ML, it is only necessary
   to understand the types TMiniStep and TCHain.

   Model-dependent variables:
   k.rate is the rate of change valid before ministep k;
   mu and sigma2 are the parameters depending on rates
     represented by the so-called symbols in the paper;
   kappa is log(symbol called kappa) plus a constant depending on the data
     at the end of the period.
   k.lprob is the log-probability of the change made at ministep k.
   newrates, newrate1, newrate2, newlprobs, newlprob1, newlprob2
    are provisional values
    that will be used to calculate new values of rates and lprobs
    if a proposed change is accepted.
   totaltime is the time length of the period.
   *)


{                                                                              }
{ Linked items forming a list: TMiniSTep                                       }
{                                                                              }
const     eps = 1e-4;
(* A number that should be within numerical precision for working with reals
   between 0 and 1 *)

type
  TAspect = (Network,Behavior);
  TDif    =  -1 .. 1;

  TMiniStep = class
  protected
  (* The <<kind>> of a Tministep is defined as the combination of
   Fash, Fih, Fjh, Fvarh, Fdifh.
   For Fash = Network, since only binary variables are considered as yet,
              Fdifh plays no role.
   For Fash = Behavior, Fjh plays no role. *)
    Fash        : TAspect;       (* Aspect value *)
    Fih         : integer;       (* i value *)
    Fjh         : integer;       (* j value *)
    Fvarh       : byte;          (* variable number *)
    Fdifh       : TDif;          (* direction of change *)
    Fnrkind     : integer;       (* number of ministeps with same rkind *)
    FSuch       : TMiniStep;     (* Next ministep *)
    FPredh      : TMiniStep;     (* Preceding ministep *)
    FSuchrkind  : TMiniStep;     (* Next ministep with same rkind *)
    FPredhrkind : TMiniStep;     (* Preceding ministep with same rkind *)
(* The following index variables are defined for the purpose of enabling
   fast random draws *)
    FIndex      : integer;       (* Index number; inverse of IndexStep *)
    FMultIndex  : integer;       (* Index number for items with Fnrkind > 1; else -1;
                                  inverse of MultIndexStep *)
    FMultSIndex  : integer;       (* Index number for items with Fnrkind > 1; else -1;
                                     all items of the same rkind have the same number;
                                  inverse of MultSIndexStep *)
    FCCPIndex  : integer;       (* Index number for first elements of CCPs; else -1;
                                  inverse of CCPIndexStep *)
    FDiagIndex: integer;       (* Index number for diagonal items; else -1;
                                  inverse of DiagIndexStep *)
    FTime    : real;           (* Used only for quickly determining ordering *)
    FRate    : real;
    FNewrate : real;
    Flprob   : real;
    Fprob1   : real;
    Fprob2   : real;
    FNewlprob: real;

  public
    constructor Create(const V0: TAspect; const V1, V2: Integer;
                       const V3: byte; const V4: TDif);
    property  ash      : TAspect   read Fash;
    property  ih       : integer   read Fih;
    property  jh       : integer   read Fjh;
    property  varh     : byte  read Fvarh;
    property  Difh     : TDif      read FDifh;
    property  Such     : TMiniStep read FSuch;
    property  Predh    : TMiniStep read FPredh;
    property  Suchrkind : TMiniStep read FSuchrkind;
    property  Predhrkind: TMiniStep read FPredhrkind;
    property  Time     : real      read FTime;
    property  kindex   : integer   read FIndex;
    property  mindex   : integer   read FMultIndex;
    property  msindex  : integer   read FMultSIndex;
    property  cindex   : integer   read FCCPIndex;
    property  dindex   : integer   read FDiagIndex;
    property  nrkind   : integer   read Fnrkind;
    property  rate     : real      read Frate      write FRate;
    property  newrate  : real      read FNewrate   write FNewrate;
    property  lprob    : real      read Flprob     write Flprob;
    property  prob1    : real      read Fprob1     write Fprob1;
    property  prob2    : real      read Fprob2     write Fprob2;
    property  newlprob : real      read Fnewlprob  write Fnewlprob;

    function  Extreme   : boolean;
    function  NotExtreme: boolean;
    function  Diagonal  : Boolean;
    function  FCCP      : Boolean;
    function  KindString: shortstring;
    procedure WriteKind(var outf: TFileStream); overload;
    procedure WriteKind(var outf: textfile); overload;
    procedure WriteStep(var outf: textfile);
    procedure ReadStep(var inf: textfile; var theend: boolean);

    function  InsertBefore(const V0: TAspect; const V1, V2: Integer;
                           const V3: byte; const V4: TDif): TMiniStep;
    procedure DestroyList;  (* Changed by TS from destructor to procedure *)
    procedure Remove;
    procedure Delete;
(* The delete procedures execute the corresponding remove procedure
   and also free the memory *)
  end; (* TMiniStep *)


  function KindsString(const V0: TAspect; const V1, V2: Integer;
                       const V3: byte; const V4: TDif): shortstring;

  function Samerkind(const k: TMiniSTep; const V0: TAspect; const V1, V2: Integer;
                    const V3: byte; const V4: TDif): Boolean; overload;
  function Samerkind(const k1, k2: TMiniSTep): Boolean;  overload;
  function ThisOrder(const k1,k2: TMinistep): boolean;
(* k1 precedes k2 in time *)

(* Types TDoublyLinkedIntMiniStep and TDoublyLinkedIntMiniList provide
   some sparse matrix technology for finding the first ministep
   of a given kind with Aspect = Network *)

  Type TDoublyLinkedIntMiniStep = class
  protected
    FNext : TDoublyLinkedIntMiniStep;
    FPrev : TDoublyLinkedIntMiniStep;

  public
    Value1: Integer;
    Value2: TMiniStep;

    constructor Create(const V1: Integer; const V2: TMiniStep);

    property  Next: TDoublyLinkedIntMiniStep read FNext write FNext;
    property  Prev: TDoublyLinkedIntMiniStep read FPrev write FPrev;

    procedure DestroyList;  (* Changed by TS from destructor to procedure *)
    function  HasNext: Boolean;
    function  HasPrev: Boolean;
    function  Count: Integer;

    procedure Remove;
    procedure InsertAfter(const V1: Integer; const V2: TMiniStep); // used @@@
    procedure InsertBefore(const V1: Integer; const V2: TMiniStep);
    procedure Delete;
(* The delete procedures execute the corresponding remove procedure
   and also free the memory *)
  end; (* TDoublyLinkedIntMiniStep *)

{                                                                              }
{ Linked list with first and last element: TDoublyLinkedIntMiniList                }
{                                                                              }

(* Although this is not implied by the type definition,
   the type TDoublyLinkedIntMiniList is used here in the following way:
   values of Value1 are ordered after Ffirst, i.e.,
   if (it1.Next = it2) and (it1 <> Ffirst) then it1.Value1 < it2.Value1;
   and Ffirst.Value1 is the number of successors of First.
   Used in this way, TDoublyLinkedIntMiniList is interpreted as a (sparse) vector,
   where element with index Value1 has the value Value2,
   all other elements have the value 0,
   and Ffirst.Value1 is the number of nonzero elements. *)

type
  TDoublyLinkedIntMiniList = class
  protected
    FFirst : TDoublyLinkedIntMiniStep;
    FLast  : TDoublyLinkedIntMiniStep;
    FCount : Integer;

  public
    constructor Create;
    destructor Destroy; override;

    property  First: TDoublyLinkedIntMiniStep read FFirst write Ffirst;
    property  Last: TDoublyLinkedIntMiniStep read FLast write FLast;
    property  Count: Integer read FCount;

    procedure DeleteList;

    procedure InsertFront(const V1: Integer; const V2: TMiniStep);

    function VectorValueUp(const i: integer): TMiniStep;
(* The value of element i of the TMiniStep vector represented by vec,
    computed by an upward or a downward search; both yield the same outcome. *)

  end; (* TDoublyLinkedIntMiniList *)

{                                                                              }
{ Linked list of TMiniStep with first and last element: TChain                 }
{                                                                              }

type
  TChain = class
  protected
    FFirst             : TMiniStep;
    FLast              : TMiniStep;
    FTotNumber         : Integer;
    FMultipleNumber    : Integer;
    FMultipleSNumber   : Integer;
    FDiagNumber        : Integer;
    FCCPNumber         : Integer;
    FNodes             : Integer;
    FFirstRkindNetwork : array of TDoublyLinkedIntMiniList;
    FFirstRkindBehavior: array of array of array of TMiniStep;
(* If FFirstRkindNetwork[i].Value1 = j then FFirstRkindNetwork[i].Value2 is the ministep
   with the first occurrence of the rkind  (Network,i,j,.,.).
   For behavior there is no necessity to use sparse matrix techniques,
   and a triple array will do:
   FirstOfRKind(Behavior,i,.,v,d) = FFirstRkindBehavior[i,v-1,abs(d)].
   In both cases, this is the return value
   of the function FirstOfRkind(Network,i,j,.,.).
   The sequence defined by Suchrkind (with inverse Predhrkind)
   starting with FirstOfRkind(i,j) visits all occurrences of this rkind,
   and no other.   *)
    FTotalTime       : real;
    FErrorMes        : shortstring;

  public
    FIndexStep, FMultIndexStep, FMultSIndexStep, FCCPIndexStep, FDiagIndexStep:
                               array of TMiniStep;
(* bijections:
   FIndexStep[h].FIndex = h;
   FMultIndexStep[h].FMultIndex = h;
   FDiagIndexStep[h].FDiagIndex = h;
   FCCPIndexStep[h].FCCPIndex = h;
   many-to-one:
      FMultSIndexStep[h].FMultSIndex = h;
      FMultSIndexStep[h] is always first of its rkind;
      item.FMultSIndex depends only on the rkind of item.    *)
    mu, sigma2, kappa: real;

    constructor Create(const n, nz: integer);
    destructor Destroy; override;

    property  First          : TMiniStep     read FFirst;
    property  Last           : TMiniStep     read FLast;
    property  TotalTime      : real          read FTotalTime;
    property  TotNumber      : Integer       read FTotNumber;
    property  MultipleNumber : Integer       read FMultipleNumber;
    property  MultipleSNumber: Integer       read FMultipleSNumber;
    property  CCPNumber      : Integer       read FCCPNumber;
    property  DiagNumber     : Integer       read FDiagNumber;
    property  Nodes          : Integer       read FNodes;
    property  ErrorMes       : shortstring   read FErrorMes;

    procedure Remove(const Item: TMiniStep);

    procedure Delete(var Item: TMiniStep);
    procedure DeleteList;

    procedure InsertBefore(const it: TMiniStep;const V0: TAspect;
                          const V1, V2: Integer;
                          const V3: byte; const V4: TDif);
    procedure ReTime; (* redefine the time fields, if some of them are so close
                         together that they risk to become indistinguishable. *)

    function FirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): TMiniStep;
(* The first occurrence of this rkind *)
    function NCCPOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): integer;
(* Number of CCPs of this rkind *)
    procedure SetFirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif;
                         const it: TMiniStep);
(* Sets the first occurrence of this rkind to it. *)
    procedure DropFirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif);
(* Sets the first occurrence of this rkind to nil. *)
    function SucRKind(const it: TMiniStep): TMiniStep;
(* it.Suchrkind if this is defined, else Last *)
    function PredRKind(const it: TMiniStep): TMiniStep;
(* it.Predhrkind if this is defined, else First *)
    procedure MakeFCCP(Item: TMiniStep);
(* Makes Item the first element of a CCP *)
    procedure DropFCCP(Item: TMiniStep);
(* Unmakes Item the first element of a CCP *)
    function RandomElement(const version: Byte): TMiniStep;
(* Returns a random draw from the chain, excluding First but permitting Last. *)
   function RandomMultiple(const version: Byte): TMiniStep;
(* Returns a random draw from the multiple elements of the chain *)
   function RandomMultipleS(const version: Byte): TMiniStep;
(* Returns a random draw from the single multiple elements of the chain *)
   function RandomCCP(const version: Byte): TMiniStep;  overload;
   function RandomCCP(const version: Byte; var extraCCP: integer): TMiniStep;
                                                        overload;
(* Returns a random draw from the first elements of CCPs of the chain.
    extraCCP is the change in FCCPNumber that would be caused
    by removing the two ministeps defining this CCP. *)
   function RandomCCPrkind(const version: Byte; const V0: TAspect;
                           const V1, V2: integer; const V3: byte): TMiniStep;
(* returns a random CCP of rkind (V0, V1, V2, V3) *)
//@@@@< procedure RandomPairInCCP geheel veranderd.
   function nk(const V0: TAspect; const V1, V2: integer; const V3: byte;
               const thenumber: integer; var it: TMinistep;
               const it1, it2: TMiniStep; const exclude, getelement: boolean): integer;
(* if getelement = false, then
   nk returns n(u) = Number of elements of K(u) = set of those ministeps
   among first and the ministeps of rkind (v0,v1,v2,v3,[0 for v0 = net, 1 for v0 = beh])
   which are not immediately followed by a ministep of this kind;
   if exclude, then it1 and it2 are excluded from the chain.
   if getelement = true, then nk returns nk = thenumber
   and it = the element of K(u) with number thenumber;
   if there is no such element, then nk > thenumber and it = nil.
    *)
   procedure RandomPairInCCP(const version: Byte; const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif;
                         var it1, it2: TMiniStep; var pit: real; var ExtraCCP: integer);
(*
   Returns a random draw of two ordered ministeps it1 <= it2 such that, when
   ministeps of kind V0-V3 will be inserted before them,
   these new ministeps form a CCP.
   See description in the paper.
   A random value for V4 is also returned: V4 must be inserted before it1,
   -V4 before it2; if it1 = it2, then first V4, then -V4.
>@@@@
   pit is the probability of choosing the realized outcome.
   ExtraCCP is the number of additional CCPS that would be formed by this insertion. *)
   function RandomDiag(const version: Byte): TMiniStep;
(* Returns a random draw from the diagonal elements of the chain *)
    function Numberrkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): integer;
(* The number of elements with this value of rkind *)

  end;

//var tf: textfile; // @@@@

//acac: boolean;
//brbr: shortstring;

implementation

{                                                                              }
{ TMiniStep                                                                    }
{                                                                              }

constructor TMiniStep.Create(const V0: TAspect; const V1, V2: Integer;
                           const V3: byte; const V4: TDif);
begin
  inherited Create;
  Fash := V0;
  Fih := V1;
  Fjh := V2;
  Fvarh := V3;
  Fdifh := V4;
(* Defaults overriding this: *)
  if (V0 = Network) then begin Fvarh := 1; Fdifh := 0; end;
  if (V0 = Behavior) then Fjh := 0;
  Fnrkind := -1;
  FIndex := -1;
  FMultIndex := -1;
  FMultSIndex := -1;
  FCCPIndex := -1;
  FDiagIndex := -1;
  Flprob := 0.0;
  Frate := 0.0;
  Fprob1 := 0.0;
  Fprob2 := 0.0;
(* This is a seemingly very incomplete procedure;
  it is called only in InsertBefore which is called only in TChain.InsertBefore;
  a number of essential assignments are done in TChain.InsertBefore. *)
end;


//destructor TMiniStep.DestroyList;
procedure TMiniStep.DestroyList;
var N : TMiniStep;
begin
{  While Assigned(FSuch) do
    begin
      N := FSuch;
      FSuch := N.Such;
//  @@@
      N.Free; freeandnil???
    end;
}    
  inherited Destroy;
end;

procedure TMiniStep.Delete;
begin
  Remove;
  FreeAndNil(Self);
end;

procedure TMiniStep.Remove;
begin
  if Assigned(Such) then
    Such.FPredh := FPredh;
  if Assigned(Predh) then
    FPredh.FSuch := FSuch;
  if Assigned(Suchrkind) then
    Suchrkind.FPredhrkind := FPredhrkind;
  if Assigned(Predhrkind) then
    FPredhrkind.FSuchrkind := FSuchrkind;
end;

function TMiniStep.Diagonal: Boolean;
begin
  if Fash = Network then
    Result := ((Fih = Fjh) and (Fih >= 1))
   else Result := (Fdifh = 0);
end;

function TMiniStep.FCCP: Boolean;
(* First element of a CCP *)
(* [[[ *)
(* The functionality is duplicated in RandomCCP - second overload version. *)
begin
(* old: chainversion 2 *)
{
  if Assigned(FSuchrkind) then
    begin
    if Fash = Network then
      Result := (Fih <> Fjh)
     else Result := (Fdifh <> 0) and (FSuchrkind.Fdifh + Fdifh = 0);
    if Assigned(FPredhrkind) then
(* Check that Self is not immediately preceded by its own rkind @@@@ *)
// @@@@      if (FPredhrkind = FPredh) and (FSuchrkind = FSuch) then Result := false;
      if (FPredhrkind = FPredh) then Result := false; // @@@@
    end
   else Result := false;
 }
(* new : chainversion 3: *)
  if Assigned(FSuchrkind) then
    begin
    if Fash = Network then
      Result := ((Fih <> Fjh) and (FSuchrkind <> Fsuch))
     else Result := ((Fdifh <> 0) and (FSuchrkind.Fdifh + Fdifh = 0)
                      and (FSuchrkind <> Fsuch));
    end
   else Result := false;
end;
(* ]]] *)

function TMiniStep.KindString: shortstring;
var s1: shortstring;
begin
  str(kindex:4,s1);
  Result := s1+' '+ KindsString(Fash, Fih, Fjh, Fvarh, Fdifh); // @@@@
end;          
(* ]]] *)

procedure TMiniStep.WriteKind(var outf: textfile);
begin
  write(outf,KindString,'   ',predh.kindex:4,' ',
          such.kindex:4,'   ',nrkind:4,' ',time:8:4,'  ',rate:9:4,' ',lprob:9:4,
          mindex:4,' ',msindex:4,' ',dindex:4,' ',cindex:4);  // @@@@
end;

procedure TMiniStep.WriteKind(var outf: TFileStream);
var x : string;
begin
  x := format('%s%s%4d%s%4d%s%4d%s%8.4f%s%9.4f%s%9.4f%4d%s%4d%s%4d%s%4d',[KindString,'   ',predh.kindex,' ',
          such.kindex,'   ',nrkind,' ',time,'  ',rate,' ',lprob,
          mindex,' ',msindex,' ',dindex,' ',cindex]);  // @@@@
  outf.Write(pchar(x)^,length(x));
end;

procedure TMiniStep.WriteStep(var outf: textfile);
begin
  if (Fash = Network) then
    writeln(outf,'N ',Fih:4,' ',Fjh:4,' ',Fvarh:2,' ',Fdifh:2)
   else writeln(outf,'B ',Fih:4,' ',Fjh:4,' ',Fvarh:2,' ',Fdifh:2);
end;

procedure TMiniStep.ReadStep(var inf: textfile; var theend: boolean);
var a: char;
begin
  read(inf, a);
  case a of
  'N': begin
       readln(inf,Fih,Fjh);
       theend := false;
       end;
  'B': begin
       readln(inf,Fih,Fjh,Fvarh,Fdifh);
       theend := false;
       end;
    else theend := true;
   end; (* case *)
end;


function TMiniStep.InsertBefore(const V0: TAspect; const V1, V2: Integer;
                           const V3: byte; const V4: TDif): TMiniStep;
var Item: TMiniStep;
begin
  Item := TMiniStep.Create(V0, V1, V2, V3, V4);
  Item.FSuch := self;
  Item.FPredh := FPredh;
  if Assigned(FPredh) then FPredh.FSuch := Item;  (* Added by TS, this was not included in the original cLinkedLists *)
(* FPredhrkind, FSuchrkind is arranged elsewhere because it also involves FFirstRkindNetwork *)
  Item.FTime := 0.5*(self.FTime + FPredh.FTime);
  FPredh := Item;
  Result := Item;
end;

function TMiniStep.Extreme: boolean;
begin
  Result := (kindex <= 1);
end;

function TMiniStep.NotExtreme: boolean;
begin
  Result := (kindex >= 2)
end;


function KindsString(const V0: TAspect; const V1, V2: Integer;
                       const V3: byte; const V4: TDif): shortstring;
var s1,s2,s3,s4: shortstring;
begin
  str(V1:3,s1);
  str(V2:3,s2);
  str(V3:3,s3);
  str(V4:3,s4);
  if (V0 = Network) then Result := ' Network  '+s1+' '+s2+' '+s3+' '+s4+' '
                    else Result := ' Behavior '+s1+' '+s2+' '+s3+' '+s4+' ';
end;

function Samerkind(const k: TMiniSTep; const V0: TAspect; const V1, V2: Integer;
                    const V3: byte; const V4: TDif): Boolean;
begin
  if (k = nil) then Result := false
  else
  Result := (V0 = k.ash) and
           (((V0 = Network) and (V1 = k.ih) and (V2 = k.jh))
        or ((V0 = Behavior) and (V1 = k.ih) and (V3 = k.Fvarh)
                           and (abs(V4) = abs(k.Difh))));
end;

function Samerkind(const k1, k2: TMiniSTep): Boolean;
begin
  if (k1 = nil) or (k2 = nil) then
    Result := false
   else
   Result := (k1.ash = k2.ash) and
          (((k1.ash = Network) and (k1.ih = k2.ih) and (k1.jh = k2.jh))
       or ((k1.ash = Behavior) and (k1.ih = k2.ih) and (k1.Fvarh = k2.Fvarh)
                           and (abs(k1.Difh) = abs(k2.Difh))));
end;

function ThisOrder(const k1,k2: TMinistep): boolean;
begin
  Result := (k1.FTime < k2.FTime);
end;


{                                                                              }
{ TDoublyLinkedIntMiniStep                                                     }
{                                                                              }

function TDoublyLinkedIntMiniStep.HasNext: Boolean;
begin
  Result := Assigned(Next);
end;


function TDoublyLinkedIntMiniStep.Count: Integer;
var N : TDoublyLinkedIntMiniStep;
begin
  Result := 1;
  N := FNext;
  While Assigned(N) do
    begin
      Inc(Result);
      N := N.Next;
    end;
end;

function TDoublyLinkedIntMiniStep.HasPrev: Boolean;
begin
  Result := Assigned(FPrev);
end;


procedure TDoublyLinkedIntMiniStep.Delete;
begin
  Remove;
  FreeAndNil(Self);
end;

procedure TDoublyLinkedIntMiniStep.Remove;
begin
  if Assigned(Next) then
    Next.Prev := FPrev;
  if Assigned(Prev) then
    Prev.Next := FNext;
end;

procedure TDoublyLinkedIntMiniStep.InsertAfter
          (const V1: Integer; const V2: TMiniStep);
var Item: TDoublyLinkedIntMiniStep;
begin
  Item := TDoublyLinkedIntMiniStep.Create(V1, V2);
  Item.Next := FNext;
  Item.Prev := self;
  if Assigned(FNext) then FNext.Prev := Item;  (* Added by TS, this was not included in the original cLinkedLists *)
  FNext := Item;
end;

procedure TDoublyLinkedIntMiniStep.InsertBefore
          (const V1: Integer; const V2: TMiniStep);
var Item: TDoublyLinkedIntMiniStep;
begin
  Item := TDoublyLinkedIntMiniStep.Create(V1, V2);
  Item.Next := self;
  Item.Prev := FPrev;
  if Assigned(FPrev) then FPrev.Next := Item;  (* Added by TS, this was not included in the original cLinkedLists *)
  FPrev := Item;
end;


//destructor TDoublyLinkedIntMiniStep.DestroyList;
procedure TDoublyLinkedIntMiniStep.DestroyList;
var N : TDoublyLinkedIntMiniStep;
begin
  While Assigned(FNext) do
    begin
      N := FNext;
      FNext := N.Next;
      N.Free;
    end;
  inherited Destroy;
end;

constructor TDoublyLinkedIntMiniStep.Create(const V1: Integer; const V2: TMiniStep);
begin
  inherited Create;
  Value1 := V1;
  Value2 := V2;
end;


{                                                                              }
{ TDoublyLinkedIntMiniList                                                         }
{                                                                              }
Constructor TDoublyLinkedIntMiniList.Create;  (* added by TS *)
begin
  inherited Create;
  FCount := 0;
end;

Destructor TDoublyLinkedIntMiniList.Destroy;
begin
  DeleteList;
  inherited Destroy;
end;

procedure TDoublyLinkedIntMiniList.InsertFront
          (const V1: Integer; const V2: TMiniStep);
var Item: TDoublyLinkedIntMiniStep;
begin
  if not Assigned(FFirst) then
    begin
      Item := TDoublyLinkedIntMiniStep.Create(V1, V2);
      FFirst := Item;
      FLast := Item;
      Item.Prev := nil;
      Item.Next := nil;
      end else
    begin                (* This is not allowed in Chains. *)
      FFirst.InsertBefore(V1, V2);
      FFirst := FFirst.Prev;
    end;
  Inc(FCount);
end;

procedure TDoublyLinkedIntMiniList.DeleteList;
begin
  if Assigned(FFirst) then begin
    FFirst.DestroyList;
    Ffirst.Free;         (* Added by TS, this was not included in the original cLinkedLists *)
   end;
  FFirst := nil;
  FLast := nil;
end;

function TDoublyLinkedIntMiniList.VectorValueUp(const i: integer): TMiniStep;
(* The value of element i of the TMiniStep vector represented by vec. *)
var it: TDoublyLinkedIntMiniStep;
begin
  Result := nil;
  it := Ffirst;
  while it.HasNext do begin
   it := it.Next;
   case sign(it.Value1 - i) of
    -1: ;
    0: begin Result := it.Value2; exit; end;
    1: exit;
    end; (* case *)
   end; (* while *)
end; (* VectorValueUp *)


{                                                                              }
{ TChain                                                                       }
{                                                                              }

Constructor TChain.Create(const n, nz: integer);
var Item: TMiniStep;
 i,j,d: integer;
begin
  inherited Create;
  FTotNumber := 1;
  FDiagNumber := 0;
  FCCPNumber := 0;
  FMultipleNumber := 0;
  FMultipleSNumber := 0;
  FNodes := n;
  FErrorMes := '';
  Item := TMiniStep.Create(Network,0,0,0,0);
  FFirst := Item;
  FFirst.FPredh := Ffirst;
  FFirst.FTime := 0.0;
  Item := TMiniStep.Create(Network,0,0,0,0);
  FLast := Item;
  FLast.FPredh := Ffirst;
  FLast.FSuch := FLast;
  FFirst.FSuch := FLast;
  FLast.FTime := 1.0;
  SetLength(FFirstRkindNetwork,n+1);
  for i:=n downto 0 do begin
    FFirstRkindNetwork[i] := TDoublyLinkedIntMiniList.Create;
    FFirstRkindNetwork[i].InsertFront(0,Ffirst);  // @@@ Ffirst or nil?
   end;
  if (nz > 0) then SetLength(FFirstRkindBehavior,n+1,nz,2);
  for i:=0 to n do for j:=0 to nz-1 do for d := 0 to 1 do
    FFirstRkindBehavior[i,j,d] := nil;
  FTotalTime := 1.0;
  SetLength(FIndexStep,2);
  SetLength(FMultIndexStep,2);
  SetLength(FMultSIndexStep,2);
  SetLength(FCCPIndexStep,2);
  SetLength(FDiagIndexStep,2);
  FIndexStep[0] := Ffirst;
  FIndexStep[1] := FLast;
  Ffirst.FIndex := 0;
  FLast.FIndex := 1;
  mu := 0.0;
  sigma2 := 0.0;
  kappa := -999.0;
end;

Destructor TChain.Destroy;
var i: integer;
begin
  DeleteList;
  for i:=Fnodes downto 0 do FFirstRkindNetwork[i].DeleteList;
  SetLength(FFirstRkindNetwork,0);
  SetLength(FIndexStep,0);
  SetLength(FMultIndexStep,0);
  SetLength(FMultSIndexStep,0);
  SetLength(FCCPIndexStep,0);
  SetLength(FDiagIndexStep,0);
  SetLength(FFirstRkindBehavior,0,0,0);
  inherited Destroy;
end;

procedure TChain.MakeFCCP(Item: TMiniStep);
begin
  if (Item.cindex < 0) then
    begin
    Inc(FCCPNumber);
    if (high(FCCPIndexStep) < FCCPNumber) then
      SetLength(FCCPIndexStep,Succ(FCCPNumber));
    Item.FCCPIndex := FCCPNumber;
    FCCPindexStep[FCCPNumber] := Item;
    end;
end;

procedure TChain.DropFCCP(Item: TMiniStep);
begin
  if (Item.FCCPIndex > 0) then
    begin
    if (Item.FCCPIndex < FCCPNumber) then
    begin
     FCCPIndexStep[Item.FCCPIndex] := FCCPIndexStep[FCCPNumber];
     FCCPIndexStep[Item.FCCPIndex].FCCPIndex := Item.FCCPIndex;
    end;
    Dec(FCCPNumber);
    Item.FCCPIndex := -1;
    end;
end;

procedure TChain.InsertBefore(const it: TMiniStep;
                             const V0: TAspect; const V1, V2: Integer;
                             const V3: byte; const V4: TDif);
var Item, Item1, Item2: TMiniStep;
    numberrkind, msn: integer;
    diag: boolean;
begin
  if (not assigned(it)) then begin
    FErrorMes := 'Tchain.Insertbefore entered for non-assigned pointer.';
    exit;
   end;

//if acac then brbr := brbr + ' x';

  if ((it.FTime - (it.Predh).FTime) < eps) then ReTime;
  Item := it.InsertBefore(V0, V1, V2, V3, V4);
  diag := ((V0 = Network) and (V1 = V2)) or ((V0 = Behavior) and (V4 = 0));

(* New Index sequence *)
  Inc(FTotNumber);
  if (high(FIndexStep) < FTotNumber) then SetLength(FIndexStep,Succ(FTotNumber));
  Item.FIndex := FTotNumber;
  FindexStep[FTotNumber] := Item;

  if diag then begin
(* New Diag Sequence *)
    Inc(FDiagNumber);
    if (high(FDiagIndexStep) < FDiagNumber) then
      SetLength(FDiagIndexStep,Succ(FDiagNumber));
    Item.FDiagIndex := FDiagNumber;
    FDiagIndexStep[FDiagNumber] := Item;
   end;


   if Assigned(FirstOfRkind(V0,V1,V2,V3,V4)) then
     begin
(* There was already at least one other occurrence of (rkind) *)
     Item1 := FirstOfRkind(V0,V1,V2,V3,V4);
     numberrkind := Succ(Item1.Fnrkind);
(* numberrkind = new number of occurrences of (rkind) >= 2 *)
(* Determine new Mult Sequence *)
     if (numberrkind <= 2) then
        begin
(* FirstOfRkind(V1,V2) as well as Item must be included in the MultSequence *)
        Inc(FMultipleNumber,2);
        if (high(FMultIndexStep) < FMultipleNumber)
          then SetLength(FMultIndexStep,Succ(FMultipleNumber));
        Item1.FMultIndex := Pred(FMultipleNumber);
        FMultIndexStep[Pred(FMultipleNumber)] := Item1;
        Item.FMultIndex := FMultipleNumber;
        FMultIndexStep[FMultipleNumber] := Item;
(* Similar for the single multiple numbers: *)
        if (not diag) then
          begin
          Inc(FMultipleSNumber);
          if (high(FMultSIndexStep) < FMultipleSNumber)
            then SetLength(FMultSIndexStep,Succ(FMultipleSNumber));
          Item1.FMultSIndex := FMultipleSNumber;
          if (Item1.Time < Item.Time) then FMultSIndexStep[FMultipleSNumber] := Item1
           else FMultSIndexStep[FMultipleSNumber] := Item;
          Item.FMultSIndex := FMultipleSNumber;
          end;
        end (* if (numberrkind <= 2) *)
       else (* (numberrkind >= 3) *)
        begin
(* Item must be included in the Mult Sequence *)
        Inc(FMultipleNumber);
        if (high(FMultIndexStep) < FMultipleNumber)
          then SetLength(FMultIndexStep,Succ(FMultipleNumber));
        Item.FMultIndex := FMultipleNumber;
        FMultIndexStep[FMultipleNumber] := Item;
(* and in the mults sequence *)
        if (not diag) then
          begin
          msn := Item1.FMultSIndex;
          Item.FMultSIndex := msn;
          if (Item1.Time > Item.Time)
            then FMultSIndexStep[msn] := Item;
          end;
        end; (* else , i.e. (numberrkind >= 3) *)

(* Determine new Fnrkind values *)
     Item.Fnrkind := numberrkind;
(*  Item1 still is the old FirstOfRkind(V0,V1,V2,V3,V4) *)
     Item1.Fnrkind := numberrkind;
     while (assigned(Item1.FSuchrkind)) do
       begin
       Item1 := Item1.FSuchrkind;
       Item1.Fnrkind := numberrkind;
       end;

(* Determine if Item must be the new FirstOfRkind(V0,V1,V2,V3,V4)
   and determine the place of Item in the rkind sequence *)
     Item1 := FirstOfRkind(V0,V1,V2,V3,V4);
     Item2 := nil;
     if ThisOrder(Item, Item1) then
       begin
       Item1.FPredhrkind := Item;
       Item.FSuchrkind := Item1;
       SetFirstOfRkind(V0,V1,V2,V3,V4,Item);
       Item.FPredhrkind := nil;
       end
      else (* other order: Item is not the first *)
       begin
(* discover place in the (rkind) order where Item is to be put:
   bracket such that Item2 < Item < Item1 *)
       if (numberrkind <= 2) then
         begin
(* numberrkind must be 2 and Item is not the first *)
         Item1.FSuchrkind := Item;
         Item.FPredhrkind := Item1;
         Item.FSuchrkind := nil;
         end
        else
         begin (* Numberrkind >= 3 *)
         while (ThisOrder(Item1, Item) and assigned(Item1.FSuchrkind)) do
           begin
           Item2 := Item1;
           Item1 := Item1.FSuchrkind;
           end;
         if (not Assigned(Item2)) then
           begin
           FErrorMes := 'Error in TChain.InsertBefore; no last element of this rkind found ';
           exit;
           end;
         if ThisOrder(Item1, Item) then
           begin
(* the end was reached, the highest Item1 is still earlier than Item *)
           Item1.FSuchrkind := Item;
           Item.FPredhrkind := Item1;
           Item.FSuchrkind := nil;
           end
          else
           begin (*  Item2 < Item < Item1 *)
           Item2.FSuchrkind := Item;
           Item.FPredhrkind := Item2;
           Item.FSuchrkind := Item1;
           Item1.FPredhrkind := Item;
           end;
         end; (* if numberrkind >= 3 *)
       end; (* else: other order *)
(* Determine if CCPs are affected. *)
{     if (V0 = Network) then
       begin if (V1 <> V2) then
         begin
         if Assigned(Item.FSuchrkind)  then MakeFCCP(Item);
         if assigned(Item.FPredhrkind) then MakeFCCP(Item.FPredhrkind);
         end;
       end
      else
       begin if (abs(V4) = 1) then
         begin
         if assigned(Item.FSuchrkind) then
           if (Item.Difh +Item.FSuchrkind.Difh = 0)
             then MakeFCCP(Item) else DropFCCP(Item);
         if assigned(Item.FPredhrkind) then
           if (Item.Difh +Item.FPredhrkind.Difh = 0)
             then MakeFCCP(Item.FPredhrkind) else DropFCCP(Item.FPredhrkind);
         end;
@@@       end;}
     if Item.FCCP then MakeFCCP(Item);
     if assigned(Item.FPredhrkind) then
       if Item.FPredhrkind.FCCP
         then MakeFCCP(Item.FPredhrkind) else DropFCCP(Item.FPredhrkind);
// @@@@ regels naar later verschoven
     end (* if assigned FirstOfRkind *)
    else
     (* there earlier was no occurrence of this rkind *)
     begin
      Item.Fnrkind := 1;
//brbr := brbr+item.KindString+' before '+it.KindString;
//if acac then exit;
      SetFirstOfRkind(V0,V1,V2,V3,V4,Item);
      Item.FSuchrkind := nil;
      Item.FPredhrkind := nil;
     end;

   if Item.FSuch.FCCP then MakeFCCP(Item.FSuch) else DropFCCP(Item.FSuch); // @@@@
   if Item.FPredh.FCCP then MakeFCCP(Item.FPredh) else DropFCCP(Item.FPredh); // @@@@
end;

procedure TChain.Remove(const Item: TMiniStep);
var i,j,numberrkind,ccind,msn: integer;
 Item1,Item2,Item3,Item4: TMiniStep;
 asp: TAspect;
 v: byte;
 d: TDif;
begin
  if (not assigned(item)) then begin
    FErrorMes := 'Tchain.Remove entered for non-assigned pointer.';
    exit;
   end;

  if Item.Extreme then begin
    FErrorMes := 'Tchain.Remove entered for extreme ministep.';
    exit;
   end;

  if (Item.FIndex < FTotNumber) then begin
(* rearrange so that new sequence of index values runs from 0 to
   the new, decreased, value of TotNumber *)
     FIndexStep[Item.FIndex] := FIndexStep[FTotNumber];
     FIndexStep[Item.FIndex].FIndex := Item.FIndex;
   end;
  Dec(FTotNumber);

  if (Item.Diagonal) then begin
    if (Item.FDiagIndex < FDiagNumber) then begin
(* Similar, for diagonal index values *)
       FDiagIndexStep[Item.FDiagIndex] := FDiagIndexStep[FDiagNumber];
       FDiagIndexStep[Item.FDiagIndex].FDiagIndex := Item.FDiagIndex;
     end;
    Dec(FDiagNumber);
   end;

  if FFirst = Item then FFirst := Item.Such;
  if FLast = Item then FLast := Item.Predh;

  numberrkind := Item.Fnrkind;
  asp := Item.Fash;
  i   := Item.Fih;
  j   := Item.Fjh;
  v   := Item.varh;
  d   := Item.difh;

(* Determine new Mult Sequence and FirstOfRkind and Fnrkind values *)
  case numberrkind of
    1: begin
         DropFirstOfRkind(asp,i,j,v,d);
       end;
    2: begin
         if (Item = FirstOfRkind(asp,i,j,v,d)) then
           begin
           Item1 := Item.FSuchrkind;
           SetFirstOfRkind(asp,i,j,v,d,Item1);
           if (Item.FCCPIndex > 0) then DropFCCP(Item);
           end
          else
           begin
           Item1 := Item.FPredhrkind;
           if (Item1.FCCPIndex > 0) then DropFCCP(Item1);
           end;
(* Item1 is FirstOfRkind(i,j) in the new situation,
  and ccind is the cindex of the old FirstOfRkind(i,j).
  Note that the last of Item and Item1 cannot be a FCCP. *)
         if (not assigned(item1)) then begin
           FErrorMes := 'TChain.Remove item '+inttostr(Item.kindex)+
                        'reports two ministeps of rkind '+Item.KindString+
                        ', but no other is found ';
           exit;
          end;
         Item1.Fnrkind := 1;
         Item1.FPredhrkind := nil;
         Item1.FSuchrkind := nil;
(* Item and Item1 will be deleted from the Mult Sequence.
   rearrange so that new sequence of Mult Index values runs from 0
   to the new, decreased, value of MultipleNumber *)
         FMultIndexStep[Item.FMultIndex] := FMultIndexStep[FMultipleNumber];
         FMultIndexStep[Item.FMultIndex].FMultIndex := Item.FMultIndex;
         FMultIndexStep[Item1.FMultIndex] := FMultIndexStep[Pred(FMultipleNumber)];
         FMultIndexStep[Item1.FMultIndex].FMultIndex := Item1.FMultIndex;
         Item1.FMultIndex := -1;
         Item1.FMultSIndex := -1;
         Dec(FMultipleNumber,2);
(* also from the MultS sequence *)
         if (not Item.Diagonal) then
           begin
           msn := Item.FMultSIndex;
           if (msn <> FMultipleSNumber) then
             begin
             Item4 := FMultSIndexStep[FMultipleSNumber];
             FMultSIndexStep[msn] := Item4;
             while assigned(Item4) do
               begin
               Item4.FMultSIndex := msn;
               Item4 := Item4.Suchrkind;
               end;
             end;
           Dec(FMultipleSNumber);
           end;
       end; (* case 2 *)
    else begin
(* rearrange so that new sequence of Mult Index values runs from 0 to
   the new, decreased, value of MultipleNumber *)
         if (Item.FMultIndex < FMultipleNumber) then begin
           FMultIndexStep[Item.FMultIndex] := FMultIndexStep[FMultipleNumber];
           FMultIndexStep[Item.FMultIndex].FMultIndex := Item.FMultIndex;
          end;
         Dec(FMultipleNumber);
(* Determine new (rkind) sequence  *)
         if Assigned(Item.FPredhrkind) then
           Item.FPredhrkind.FSuchrkind := Item.FSuchrkind
          else
           begin
           Item1 := Item.FSuchrkind;
           SetFirstOfRkind(asp,i,j,v,d,Item1);
(* if Item is first of its - multiple - kind,
   then point the MultS index to item.suchrkind: *)
           if (not Item.Diagonal) then FMultSIndexStep[Item.FMultSIndex] := Item1;
           end;
         if Assigned(Item.FSuchrkind) then Item.FSuchrkind.FPredhrkind := Item.FPredhrkind;
(* Determine if CCPs are affected. *)
{         if (asp = Network) then
           begin if (i <> j) then
             begin
             if Assigned(Item.FSuchrkind)  then DropFCCP(Item)
              else if assigned(Item.FPredhrkind) then DropFCCP(Item.FPredhrkind);
             end;
           end
          else (* asp = Behavior *)
           begin if (abs(d) = 1) then
             begin
             if assigned(Item.FSuchrkind) then
               if (d +Item.FSuchrkind.Difh = 0) then DropFCCP(Item);
             if assigned(Item.FPredhrkind) then
               begin
               if (d +Item.FPredhrkind.Difh = 0)
                 then DropFCCP(Item.FPredhrkind);
               if assigned(Item.FSuchrkind) then
                if (Item.FSuchrkind.Difh +Item.FPredhrkind.Difh = 0)
                 then MakeFCCP(Item.FPredhrkind);
               end;
             end;
           end;
}
(* Determine new Fnrkind values *)
         Item1 := FirstOfRkind(asp,i,j,v,d);
         Item1.Fnrkind := Pred(numberrkind);
         while assigned(Item1.FSuchrkind) do begin
           Item1 := Item1.FSuchrkind;
           Item1.Fnrkind := Pred(numberrkind);
          end;
       end; (* case else *)
    end; (* case *)

  Item1 := Item.FPredhrkind;
  Item2 := Item.FPredh;
  Item3 := Item.FSuch;

  if Item.FCCP then DropFCCP(Item);
  Item.Remove;

  if assigned(Item1) then
    begin
    if Item1.FCCP then MakeFCCP(Item1) else DropFCCP(Item1);
    end;
  if assigned(Item2) then
    begin
    if Item2.FCCP then MakeFCCP(Item2) else DropFCCP(Item2);
    end;
  if assigned(Item3) then
    begin
    if Item3.FCCP then MakeFCCP(Item3) else DropFCCP(Item3);
    end;
end;

procedure TChain.Delete(var Item: TMiniStep);
begin
  Remove(Item);
  Item.Free;
end;

procedure TChain.DeleteList;
begin
  if Assigned(FFirst) then begin
    FFirst.DestroyList;
    Ffirst.Free;         (* Added by TS, this was not included in the original cLinkedLists *)
   end;
// @@@ die FFirst.destroylist is nog niet goed
  FFirst := nil;
  FLast := nil;
end;

function TChain.FirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): TMiniStep;
(* The first occurrence of rkind (Network,i,j,.,.) *)
var it: TDoublyLinkedIntMiniStep;
begin
  Result := nil;
  if (V0 = Network) then
    begin
    it := FFirstRkindNetwork[V1].Ffirst;
    while it.HasNext do
      begin
      it := it.Next;
      case sign(it.Value1 - V2) of
        -1: ;
        0: begin Result := it.Value2; exit; end;
        1: exit; (* There is none, result remains nil *)
        end; (* case *)
      end; (* while *)
    end
   else
    begin
    if (V4 = 0) then Result := FFirstRkindBehavior[V1,V3-1,0]
    else Result := FFirstRkindBehavior[V1,V3-1,1];
    end;
end;

(* [[[ *)
function Tchain.NCCPOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): integer;
(* Number of CCPs of this rkind *)
var nc: integer;
  k : TMiniStep;
begin
  if assigned(FirstOfRkind(V0,V1,V2,V3,V4)) then
    begin
    nc := 0;
    k := FirstOfRkind(V0,V1,V2,V3,V4);
    repeat
      if k.FCCP then inc(nc);
      k := k.FSuchrkind;
    until not assigned(k);
    Result := nc;
    end
   else result := 0;
end;
(* ]]] *)

procedure TChain.SetFirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif;
                         const it: TMiniStep);
(* Sets the first occurrence of rkind (Network,i,j,.,.) to it. *)
var itt: TDoublyLinkedIntMiniStep;
begin
  if (not assigned(it)) then
    begin
    FErrorMes := 'Tchain.SetFirstOfRkind entered for non-assigned pointer.';
    exit;
    end;
  if (V0 = Network) then
    begin
    itt := FFirstRkindNetwork[V1].Ffirst;
(* Search for the highest Value1 less than V2 *)
    while itt.HasNext and (itt.FNext.Value1 < V2) do itt := itt.Next;
    if (itt.HasNext and (itt.FNext.Value1 <= V2))
(* itt.FNext.Value1 must be equal to V2 *)
     then itt.FNext.Value2 := it else itt.insertafter(V2,it);
    end
   else
    begin
    if (V4 = 0) then FFirstRkindBehavior[V1,V3-1,0] := it
     else FFirstRkindBehavior[V1,V3-1,1] := it;
    end;
end;

procedure TChain.DropFirstOfRkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif);
(* Sets the first occurrence of rkind (Network,i,j,.,.) to nil. *)
var it: TDoublyLinkedIntMiniStep;
begin
  if (V0 = Network) then
    begin
    it := FFirstRkindNetwork[V1].Ffirst;
    while it.HasNext do
      begin
       it := it.Next;
         case sign(it.Value1 - V2) of
          -1: ;
           0: begin it.Delete; exit; end;
           1: exit; (* there was no FirstOfRkind to begin with *)
         end; (* case *)
      end; (* while *)
    end else
    begin
    if (V4 = 0) then FFirstRkindBehavior[V1,V3-1,0] := nil
     else FFirstRkindBehavior[V1,V3-1,1] := nil;
    end;
end;

function TChain.SucRKind(const it: TMiniStep): TMiniStep;
(* it.Suchrkind if this is defined, else Last *)
begin
  if assigned(it.Suchrkind) then Result := it.Suchrkind else Result := Last;
end;

function TChain.PredRKind(const it: TMiniStep): TMiniStep;
(* it.Predhrkind if this is defined, else First *)
begin
  if assigned(it.Predhrkind) then Result := it.Predhrkind else Result := First;
end;


function TChain.RandomElement(const version: Byte): TMiniStep;
(* Returns a random draw from the chain, excluding First *)
var j: integer;
begin
  j := trunc(rando(version)*TotNumber)+1;
  if (j > TotNumber) then j := TotNumber;
  if (j < 1) then j := 1;
  result := FIndexStep[j];
end; (* RandomElement *)

function TChain.RandomMultiple(const version: Byte): TMiniStep;
(* Returns a random draw from the multiple elements of the chain *)
var j: integer;
begin
  j := trunc(rando(version)*MultipleNumber)+1;
  if (j > MultipleNumber) then j := MultipleNumber;
  if (j < 1) then j := 1;
  result := FMultIndexStep[j];
end; (* RandomMultiple *)

function TChain.RandomMultipleS(const version: Byte): TMiniStep;
(* Returns a random draw from the single multiple elements of the chain *)
var j: integer;
begin
  if (MultipleSNumber >= 1) then
    begin
    j := trunc(rando(version)*MultipleSNumber)+1;
    if (j > MultipleNumber) then j := MultipleSNumber;
    if (j < 1) then j := 1;
    result := FMultSIndexStep[j];
    end else result := nil;
end; (* RandomMultipleS *)

function TChain.RandomCCP(const version: Byte): TMiniStep;
(* Returns a random draw from the first elements of CCPs of the chain *)
var j: integer;
begin
  j := trunc(rando(version)*CCPNumber)+1;
  if (j > CCPNumber) then j := CCPNumber;
  if (j < 1) then j := 1;
  result := FCCPIndexStep[j];
end; (* RandomCCP *)

(* [[[ *)
function TChain.RandomCCP(const version: Byte; var extraCCP: integer): TMiniStep;
(* Returns a random draw from the first elements of CCPs of the chain.
    extraCCP is the change in FCCPNumber that would be caused
    by removing the two ministeps defining this CCP. *)
var j, count: integer;
    item1, item2, item3, item4: TMinistep;
begin
  j := trunc(rando(version)*CCPNumber)+1;
  if (j > CCPNumber) then j := CCPNumber;
  if (j < 1) then j := 1;
  item2 := FCCPIndexStep[j];
  Result := item2;

  item1 := item2.FPredhrkind; (* may be nil *)
  item3 := item2.FSuchrkind;  (* cannot be nil *)
  item4 := item3.FSuchrkind;  (* may be nil *)

(* For the following, the functionality of FCCP is used.
   If item 2 and item3 are removed, then item2 is lost as a FCCP,
   item3 perhaps, and the FCCP status of item1 may change.
   count = extraccp *)

  if item3.FCCP then count := -2 else count := -1;
  if Assigned(item1) then
    begin
    if (item1.Fash = Network) then
      begin
      if (item1.Fih <> item1.Fjh) then  (* holds always *)
        begin
        if (item1.Fsuch <> item2) then dec(count); (* item1 is FCCP *)
        if Assigned(item4) then inc(count);(* item1 will become FCCP *)
        end;
      end
     else (* item1.Fash = Behavior *)
      begin
      j := item1.Fdifh;
      if (j <> 0) then        (* holds always *)
        begin
        if ((item1.Fsuch <> item2) and (item2.Difh + j = 0)) then dec(count); (* item1 is FCCP *)
        if (Assigned(item4) and (item4.Difh + j = 0)) then inc(count);(* item1 will become FCCP *)
        end;
      end;
    end;  (*  if Assigned(item1) *)
  extraCCP := count;
end; (* RandomCCP *)


function TChain.RandomCCPrkind(const version: Byte; const V0: TAspect;
                           const V1, V2: integer; const V3: byte): TMiniStep;
var it: TMiniStep;
    nc: integer;
    ccps: array of TMiniStep;
begin
  it := FirstOfRkind(V0,V1,V2,V3,1);
  if (not assigned(it)) then
    begin
    result := nil;
    exit;
    end
   else
    begin
    nc := 0;
    SetLength(ccps,4);
    ccps[0] := nil;
    repeat
      if it.FCCP then
        begin
        inc(nc);
        if (high(ccps) < nc) then SetLength(ccps,succ(nc));
        ccps[nc] := it;
        end;
      it := it.FSuchrkind;
    until not assigned(it);
    if (nc <= 1) then
      result := ccps[nc]
       else result := ccps[succ(randomd(version,nc))];
    end;
end; (* RandomCCPrkind *)
(* ]]] *)

function TChain.nk(const V0: TAspect; const V1, V2: integer; const V3: byte;
               const thenumber: integer; var it: TMinistep;
               const it1, it2: TMiniStep; const exclude, getelement: boolean): integer;
(* if getelement = false, then
   nk returns n(u) = Number of elements of K(u) = set of those ministeps
   among first and the ministeps of rkind (v0,v1,v2,v3,[0 for v0 = net, 1 for v0 = beh])
   which are not immediately followed by a ministep of this kind;
   if exclude, then it1 and it2 are excluded from the chain.
   if getelement = true, then nk returns nk = thenumber
   and it = the element of K(u) with number thenumber;
   if there is no such element, then nk > thenumber and it = nil.
    *)
var it0,itn: TMiniStep;
  nit: integer;
  V4: Tdif;
function nexth(const itt: Tministep): Tministep;
(* itt.such, but if exclude then it1 and it2 are skipped *)
var ith: Tministep;
begin
  ith := itt.such;
  if exclude then
    begin
    if (ith = it1) or (ith = it2) then
      begin
      ith := ith.such;
      if (ith = it1) or (ith = it2) then ith := ith.such;
      end;
    end;
  nexth := ith;
end; (* nexth *)
function nexthr(const itt: Tministep): Tministep;
(* itt.suchrkind, but if exclude then it1 and it2 are skipped *)
var ith: Tministep;
begin
  ith := itt.suchrkind;
  if exclude then
    begin
    if (ith = it1) or (ith = it2) then
      begin
      ith := ith.suchrkind;
      if (ith = it1) or (ith = it2) then ith := ith.suchrkind;
      end;
    end;
  nexthr := ith;
end; (* nexthr *)
begin  (* nk *)
  if (V0 = Network) then V4 := 0 else V4 := 1;
(* Go through set K(u) and count number of elements n(u) : *)
  itn := nexth(First);
  if SameRkind(itn,V0,V1,V2,V3,V4) then nit := 0 else nit := 1;
  if getelement then
    if (nit = thenumber) then
    begin
    it := First;
    result := nit;
    exit;
    end;
  if assigned(FirstOfRkind(V0,V1,V2,V3,V4)) then
    begin
    it0 := FirstOfRkind(V0,V1,V2,V3,V4);
    if exclude then
      begin
      if (it0 = it1) or (it0 = it2) then it0 := nexthr(it0);
      end;
    repeat
      itn := nexth(it0);
      if ((itn <> last) and (not SameRkind(itn,V0,V1,V2,V3,V4))) then inc(nit);
      if getelement then
        if (nit = thenumber) then
        begin
        it := it0;
        result := nit;
        exit;
        end;
      it0 := nexthr(it0);
    until (not assigned(it0));
    end;
  Result := nit;
  if getelement then it := nil; (* this indicates a failure to find the element *)
end;

procedure TChain.RandomPairInCCP(const version: Byte; const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif;
                         var it1, it2: TMiniStep; var pit: real; var ExtraCCP: integer);
(* Returns a random draw of two ordered ministeps it1 <= it2 such that, when
   ministeps of kind V0-V3 will be inserted before them,
   these new ministeps form a CCP.
   See description in the paper.
   pit is the probability of choosing the realized outcome.
   ExtraCCP is the number of additional CCPS that would be formed by this insertion. *)
var it,itt,ith: TMiniStep;
  nit, nn, hit, nitt, hit1, hit2, h: integer;
  dum1, dum2: integer;
begin
(* Pieces of this copied/adapted to Try_Cancel *)
  if assigned(FirstOfRkind(V0,V1,V2,V3,V4)) then
    begin
(* First search for a random "interval" [it,itt]
   in which it and itt are both of rkind (V0,V1,V2,V3,V4) or Extreme,
   and no other ministeps of this rkind are included;
   runs of several ministeps of the same rkind are aggregated for this purpose. *)
(* Correspondence with paper: it = u_a, itt = u_b, nit = n(u),
   nn = b-a-1, it1=u_c, it2=u_d.   *)
    nit := nk(V0,V1,V2,V3,0,it,it1,it2,false,false);
//writeln(tf);
//writeln(tf,kindsstring(V0,V1,V2,V3,0));
//writeln(tf,'nit = ',nit);
(* Select a random one among these nit intervals *)
    case nit of
      0: (* very exceptional: all ministeps are of the same rkind(V0,V1,V2,V3);
            this implies that in the data, there is change in at most one variable. *)
         begin
         it1 := First.Such;
         it2 := it1;
         pit := 1.0;
         ExtraCCP := 0;
         exit;
         end;
      1: (* The only ministeps of this rkind occur immediately after First
            and immediately before last. *)
         begin
         it := First;
         while SameRkind(it.such,V0,V1,V2,V3,V4) do it := it.such;
         itt := Last;
         while SameRkind(itt.predh,V0,V1,V2,V3,V4) do itt := itt.predh;
         end;
      else
(* nit >= 2*)
         begin
         hit := randomd(version,nit)+1;
//writeln(tf,'hit = ',hit);
(* values from 1 to nit *)
(* get it = element number hit from set K(u) *)
         nitt := nk(V0,V1,V2,V3,hit,it,it1,it2,false,true);
         if (nitt <> hit) then
           begin
           FErrorMes := 'Element '+inttostr(hit)+' not found  in K(u) of size'+inttostr(nitt);
           exit;
           end;
         if (it = First) then itt := FirstOfRkind(V0,V1,V2,V3,V4) else itt := SucRKind(it);
         end;
       end; (* case nit *)
(* Count nn = b-(a+)+1 = the number of ministeps between it and itt.
   This is at least 1.  *)
if (itt = it.such) then
begin
FErrorMes := 'itt = it.such in RandomPairInCCP';
//write(tf,'it : '); it.WriteKind(tf); writeln(tf);
//write(tf,'itt: '); itt.WriteKind(tf); writeln(tf);
exit;
end;
    nn := 0;
    if (it = first) then ith := first else ith := it.Such;
    repeat
    ith := ith.Such;
    Inc(nn);
if (nn > TotNUmber) then begin                   //    @@@
FerrorMes := ' infinite loop in RandomPairInCCP';
exit;
end;
    until (ith = itt);
//writeln(tf,'nn = ',nn);
(* Choose it1 and it2 randomly from these nn positions;
   if necessary reorder so that it1 <= it2 *)
    if (nn <= 1) (* then nn = 1 *) then
      begin
      it1 := itt;
      it2 := itt;
      pit := 1.0/nit;
      end
     else (* nn >= 2 *)
      begin
      hit1 := randomd(version,nn)+1;
      hit2 := randomd(version,nn)+1;
//writeln(tf,'hit1 = ',hit1,' hit2 = ',hit2);
      if (hit2 < hit1) then
        begin
        h := hit1;
        hit1 := hit2;
        hit2 := h;
        end;
      if (it = first) then it1 := first else it1 := it.Such;
      for h:=1 to hit1 do it1 := it1.Such;
      it2 := it1;
      for h:=hit1 + 1 to hit2 do it2 := it2.Such;
      if (hit1 = hit2) then pit := 1.0/(sqr(nn)*nit)
        else pit := 2.0/(sqr(nn)*nit);
      end;
(* Count extra number of CCPs formed:
   note that this is in a condition with nrkind >= 1. *)
    if (V0 = Network) then
      begin
      if (itt = Last) then
        begin
        if assigned(it.Predhrkind) then
          begin
          if (it.Predhrkind <> it.Predh) then ExtraCCP := 2 else ExtraCCP := 1;
          end else ExtraCCP := 2;
//if assigned(it.Predhrkind) then writeln(tf,it.kindex,' ',it.predh.kindex,' ',it.predhrkind.kindex)
// else writeln(tf,it.kindex,' ',it.predh.kindex,' --- ');
//write(tf,extraccp);                                 // @@@@
          // always the new it1.predh, the condition indicates that it also becomes an FCCP
        end
       else
(* itt <> Last; the FCCP status of it does not change;
   only the new it1.predh and it2.predh need to be considered;
   it is possible that itt loses its FCCP status *)
        begin
        if (it2 = itt) then
          begin
          if itt.FCCP then ExtraCCP := 0 else ExtraCCP := 1;
          end
         else ExtraCCP := 1;
        if (it1 <> it2) then Inc(ExtracCP);
        end;
      end (* if V0 = Network *)
     else
      begin (* V0 = behavior *)
(* This is similar to the checks for V0 = Network, but the additional
   requirement for a CCP is that the difh values of subsequent ministeps
   som to 0.
   The new sequence of this rkind will have, after the insertion,
   difh values:  it.difh,  V4,  -V4,  itt.difh   *)
      if (itt = Last) then
        begin
        if ((predrkind(it) <> it.Predh) or (it = first.such)) and (it.Difh + V4 = 0)
          then ExtraCCP := 2 else ExtraCCP := 1;
//writeln(tf,' last : ',extraCCP);
//it.WriteKind(tf); writeln(tf,' V4 = ',V4);
          // always the new it1.predh, the condition indicates that it also becomes an FCCP
        end
       else
(* itt <> Last  *)
        begin
        (* new insert before it1 always is a FCCP: *)
        ExtraCCP := 1;
        (* it may loose or get the FCCP status: *)
        if it.FCCP then
          begin
          if (it.Difh + V4 <> 0) then dec(ExtraCCP);
          end
         else
          begin
          if ((predrkind(it) <> it.Predh) or (it = first.such)) and (it.Difh + V4 = 0) then inc(ExtraCCP);
          end;
         (* new insert before it2 : *)
        if (itt.Difh - V4 = 0) and (it1<>it2) then inc(ExtraCCP);
        (* itt : *)
        if itt.FCCP and (it2 = itt) then dec(ExtraCCP);
        end;
      end; (* if V0 = Behavior *)
    end
   else  (*  not assigned(FirstOfRkind(V0,V1,V2,V3,V4)) *)
    begin
    it1 := RandomElement(version);
    it2 := RandomElement(version);
      if ThisOrder(it2,it1) then
      begin
      it := it1;
      it1 := it2;
      it2 := it;
      end;
     if (it1 = it2) then
       begin
       pit := 1.0/sqr(TotNumber);
       end
      else pit := 2.0/sqr(TotNumber);
    ExtraCCP := 1;
    end;
(* Check if it1 and it2 themselves are of a different kind and can become FCCPs. *)
  if ((it1 <> itt) and (not it1.diagonal)) then
    begin
    if (assigned(it1.Suchrkind) and assigned(it1.Predhrkind)) then
      if ((it1.Predhrkind = it1.predh) and (it1.Fdifh + it1.Suchrkind.Fdifh = 0)) then inc(ExtraCCP);
    end;
  if ((it1 <> it2) and (it2 <> itt) and (not it2.diagonal)) then
    begin
    if (assigned(it2.Suchrkind) and assigned(it2.Predhrkind)) then
      begin
      if ((it2.Predhrkind = it2.predh)  and (it2.Fdifh + it2.Suchrkind.Fdifh = 0)) then inc(ExtraCCP);
      end;
    end;
//writeln(tf,' end: ',extraccp);
end;

function TChain.RandomDiag(const version: Byte): TMiniStep;
(* Returns a random draw from the diagonal elements of the chain *)
var j: integer;
begin
  j := trunc(rando(version)*DiagNumber)+1;
  if (j > DiagNumber) then j := DiagNumber;
  if (j < 1) then j := 1;
  result := FDiagIndexStep[j];
end; (* RandomMultiple *)


procedure TChain.ReTime;
var nn: integer; t, step: real;
    it:TMiniStep;
begin
  FErrorMes := '';
  FFirst.FTime := 0.0;
  it := FFirst;
  t := 0.0;
  nn := 0;
  step := TotalTime/(TotNumber + 1);
  if (step < eps) then FErrorMes := 'Mismatch between step and TotNumber in ReTime.';
  repeat
    inc(nn);
    it := it.FSuch;
    t := t+step;
    if it.NotExtreme then it.FTime := t;
   until (it.Extreme or (nn >= 30000));
  FLast.FTime := TotalTime;
  if (t > TotalTime) then FErrorMes := 'Too many successions in ReTime.';
  if (nn >= 30000) then FErrorMes := FErrorMes + 'nn = 30000 in ReTime';
end; (* ReTime *)

function TChain.Numberrkind(const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: TDif): integer;
(* The number of elements with this value of rkind *)
begin
  if (not assigned(FirstOfRkind(V0,V1,V2,V3,V4))) then Result := 0
    else Result := FirstOfRkind(V0,V1,V2,V3,V4).Fnrkind;
end;

begin
//acac := false;
//brbr := '';
end.

