unit s_mcmc_II;
(*Entire unit by MS:
Functions and procedures useful to estimate Bayesian probability models.
Most of the functions and procedures is scattered across the units
<S_ml>, <S_mcmc>, <S_est>, <S_re> , but the intention is to bring most of
them together, with the single exception of the those in <S_mcmc>.
As in <S_mcmc>, the emphasis is on general procedures, which got no
place elsewhere, while specific functions and procedures may be placed
in units such as <S_ml> and <S_re>.*)

(*--------------------------------------------------------------------------------------------------*)

interface

(*--------------------------------------------------------------------------------------------------*)

uses classes, eight, Math, rangen, slib, s_dat, s_base, s_re, s_matrices, s_mcmc, s_constants, Sysutils;

var
  presw:
    real;
  MI_posterior_theta, (*First posterior moment of parameter vector.*)
  lastsample, (*Last accepted candidate of the non-rate parameter vector, before (!) linear transformation.*)
  protheta: (*Stores initial parameter vector, obtained by either the maximum likelihood method or the method of moments.*) 
    pvec;
  ptheta, (*Variance-covariance matrix of prior density corresponding to parameter vector.*)
  psigma, (*Variance-covariance matrix of prior density corresponding to variance-covariance matrix.*)
  MII_posterior_theta, (*Second posterior moment of parameter vector.*)
  ictheta, (*Inverse variance-covariance matrix of the candidate-generating density of the parameter vector.*)
  choltheta: (*Cholesky factor of the variance-covariance-matrix of the candidate-generating density corresponding to parameter vector.*)
    pmatrix;
  gamma_p, (*Hyperparameters of the Gamma prior probability density.*)
  mcI_theta: (*Used to store first moments of parameter vector sampled from the posterior density.*)
    rrarray;
  mcII_theta: (*Used to store second moments of parameter vector sampled from the posterior density.*)
    rrrarray;
    
(*----------*)
(*Set priors*)
(*----------*)

procedure SetPrior(var input: shortstring; var dim: byte; var psigma: pmatrix; var ascii: TFileStream; var output: shortstring);
(*Set the scale parameter of the prior probability density.*)

procedure InitialActionsPriorTheta(const dim1: byte; var gamma_p: rrarray; var dim2: byte; var psigma: pmatrix; var ascii: TFileStream);
(*Some initial actions for the prior probability density (multivariate Gaussian) of the parameter vector.*)

procedure InitialActionsPriorSigma(var pdegrees: integer; var dim: byte; var psigma: pmatrix; var ascii: TFileStream);
(*Some initial actions for the prior probability density (Wishart) of the variance-covariance matrix of the random effects.*)

(*---------------------------*)
(*Posterior integration tools*)
(*---------------------------*)
(*Here, only some Gibbs / Metropolis-Hastings algorithms can be found, most of them are located in units <S_ml>, <S_re>.*)

procedure CandidateTheta1(var dim: byte; var vector1, vector2: pvec);
(*Generates candidate for rate parameter.*)

function CandidateTheta2(const dim: byte; const cholfac: pmatrix; const scale: real;
var vector1, vector10: pvec; const vector2: pvec): real;
(*Generates candidate for network and behavior objective and gratification function parameter.*)

procedure MH_CovRE_Step(var dim: byte; var i_sigma, sigma, cholfac: pmatrix; var n: integer; var re: rrarray;
var i_omega: pmatrix; var phi: integer; var blackbox: blackarray);
(*Gibbs sampler and Metropolis-Hastings step to sample an (inverse) variance-covariance matrix corresponding of actor-dependent random effects.*)

(*-----------------------*)
(*Output-related routines*)
(*-----------------------*)

procedure Store_MCMC_Sample(var Bayesian, randomeffects: boolean; var ldim1, ldim2, actors: integer; var dim1, dim2: byte;
var MI_posterior_theta: pvec; var MII_posterior_theta: pmatrix; var sumre, sumposre, mcI_theta: rrarray;
var mcII_theta: rrrarray; var sumsigma: pmatrix; var mc_re, mc_posre, mc_covre: rrrarray);
(*Average and store MCMC samples, and clean up.*)

procedure Output_MII_theta(var Bayesian: boolean; var ascii: TFileStream; var dim: byte; var matrix1, matrix2: pmatrix; var vector1: pvec; var file1: TFileStream);
(*In the Bayesian approach, output the second moment of the parameter vector, considered as vector-valued random variable.*)

(*--------------------------------------------------------------------------------------------------*)

implementation

(*--------------------------------------------------------------------------------------------------*)

procedure SetPrior(var input: shortstring; var dim: byte; var psigma: pmatrix; var ascii: TFileStream; var output: shortstring);
(*Set the scale parameter of the prior probability density.*)
var x: string;
var
  i, j: byte;
  i1, i2: integer;
  det, u: real;
  s: shortstring;
  f: textfile;
  cov, cholfactor: pmatrix;
begin
s := input;
if FileExists(s) then (*User-determined.*)
  begin
  ReadFile(f,s);
  for i := 1 to dim do
    begin
    for j := 1 to i do
      begin
      read(f,u);
      cov[i,j] := u;
      cov[j,i] := cov[i,j];
      end;
    readln(f);
    end;
  end
else (*Program-determined.*)
  begin
  for i := 1 to dim do
    begin
    cov[i,i] := 1e+10;
    for j := 1 to (i - 1) do
      begin
      cov[i,j] := 0;
      cov[j,i] := cov[i,j];
      end;
    end;
  end;
x := output+eol;
ascii.Write(pchar(x)^,length(x));
i1 := 14;
i2 := 4;
MatrixWriteTriangle(dim,ascii,cov,i1,i2); (*Writes real valued matrix of dimension to ASCII file.*)
Cholesky(dim,cholfactor,cov,det);
MatrixI(dim,det,psigma,cov,cholfactor);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialActionsPriorTheta(const dim1: byte; var gamma_p: rrarray; var dim2: byte; var psigma: pmatrix; var ascii: TFileStream);
(*Some initial actions for the prior probability density (multivariate Gaussian) of the parameter vector.*)
var
  i: byte;
  s1, s2: shortstring;
  x: string;
begin
SetArray(gamma_p,dim1,2); (*<gamma_p> stores the hyperparameters of the Gamma prior densities of the rate function parameters.*)
for i := 1 to dim1 do (*Default.*)
  begin
  gamma_p[i,1] := 1;
  gamma_p[i,2] := 1e-10;
  end;
Heading(2,ascii,'Bayesian model: prior specification');
x := 'Gamma prior p.d.f. of the rate parameters:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to dim1 do
  begin
  x := format('%s%d%s',['Rate parameter ',i,':   ']);
  ascii.Write(pchar(x)^,length(x));
  x := format('%s%10.10f%s%10.10f%s',['Gamma(',gamma_p[i,1],', ',gamma_p[i,2],')'])+eol;
  ascii.Write(pchar(x)^,length(x));
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
s1 := 'PriorTheta.in';
s2 := 'Variance-covariance matrix of the multivariate Gaussian prior p.d.f. of the parameter vector (centered at origin):';
SetPrior(s1,dim2,psigma,ascii,s2); (*Set the scale parameter of the prior probability density (Multivariate Normal) of the parameter vector.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialActionsPriorSigma(var pdegrees: integer; var dim: byte; var psigma: pmatrix; var ascii: TFileStream);
(*Some initial actions for the prior probability density (Wishart) of the variance-covariance matrix of the random effects.*)
var s1, s2: shortstring;
begin
pdegrees := dim; (*Degrees of freedom of the prior Wishart probability density.*)
s1 := 'PriorSigma.in';
s2 := 'Scale matrix of the Wishart prior p.d.f. of the variance-covariance matrix of the actor-dependent random effects (df = ' + inttostr(pdegrees) + '):';
SetPrior(s1,dim,psigma,ascii,s2); (*Set the scale parameter of the prior probability density (Wishart) of the variance-covariance-matrix of the random effects.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CandidateTheta1(var dim: byte; var vector1, vector2: pvec);
(*Generates candidate for rate parameter.*)
var
  i: byte;
  u: real;
begin
for i := 1 to dim do (*Candidate for the (vector valued) rate parameter.*)
  begin
  u := vector1[i]; (*Initial values are the moment estimates of the model without the actor-dependent random effects.*)
  vector2[i] := Gamma(u); (*Generate candidate from the Gamma distribution.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CandidateTheta2(const dim: byte; const cholfac: pmatrix; const scale: real;
var vector1, vector10: pvec; const vector2: pvec): real;
(*Generates candidate for network and behavior objective and gratification function parameter.*)
var u: real;
begin
(*Remark:
- <cdensity1> refers to the candidate-generating density.
- concerning <reflectiontheta>: in case <cdensity1> is
  * 2: then the Markov chain is reflected about <reflectiontheta>.
  * 3: then the the mean of the candidate-generating density is <reflectiontheta>.*)
u := CandidateMH(dim,cholfac,scale,vector1,cdensity1,vector10,vector2,reflectiontheta);
CandidateTheta2 := u;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MH_CovRE_Step(var dim: byte; var i_sigma, sigma, cholfac: pmatrix; var n: integer; var re: rrarray;
var i_omega: pmatrix; var phi: integer; var blackbox: blackarray);
(*Gibbs sampler and Metropolis-Hastings step to sample an (inverse) variance-covariance matrix corresponding of actor-dependent random effects.*)
var
  df: integer;
  logcratio, det, det1, det2, logmvnratio, logpratio, proposal, present, propw, propmvn, presmvn, pr: real;
  accept: boolean;
  candidate, cholfactor, w, invsigma, sum, i_sum, sample, sampleinvsigma: pmatrix;
  densities: R_Ar;
begin
df := phi + n; (*Number of degrees of freedom of the full conditional Wishart density, where <phi> is the number of degrees of freedom of the Wishart prior density.*)
if gibbs2 then
  begin
  EstimateCovIII(dim,n,re,w); (*Compute <w>.*)
  MatrixSum(dim,dim,sum,w,i_omega); (*Compute <sum = w + i_omega>, where <i_omega> is the inverse scale matrix of the Wishart prior density.*)
  Cholesky(dim,cholfactor,sum,det);
  MatrixI(dim,det,i_sum,sum,cholfactor); (*Compute <i_sum = inverse(sum)>.*)
  Cholesky(dim,cholfactor,i_sum,det); (*Compute the Cholesky factor <cholfactor> of <i_sum>: required for sampling <i_sigma>.*)
  SampleW(dim,df,cholfactor,i_sigma); (*Sample <i_sigma>.*)
  Cholesky(dim,cholfactor,i_sigma,det);
  MatrixI(dim,det,sigma,i_sigma,cholfactor); (*Compute <sigma = inverse(i_sigma)>.*)
  Cholesky(dim,cholfac,sigma,det); (*Compute the Cholesky factor <cholfac> of <sigma>: required for sampling <re>.*)
  blackbox[5] := blackbox[5] + 1; (*Acceptance rate: for reasons of completeness stored.*)
  end
else (*Metropolis-Hastings sampler: still something wrong.*)
  begin
  (*-------------------*)
  (*Generate candidate.*)
  (*-------------------*)
  Cholesky(dim,cholfactor,sigma,det);
  MatrixI(dim,det,invsigma,sigma,cholfactor); (*The inverse (!) variance-covariance matrix of the random effects is inverted to obtain the variance-covariance matrix.*)
  Cholesky(dim,cholfactor,invsigma,det1); (*Cholesky decomposition of inverse variance-covariance matrix of random effects.*)
  SampleW(dim,df,cholfactor,sample); (*Generate a sample inverse variance-covariance matrix of the actor-dependent random effects.*)
  SetMatrix(dim,dim,sampleinvsigma,invsigma);
  ScalarProd(dim,sampleinvsigma,df);
  SetMatrix(dim,dim,candidate,sample);
  ScalarDiv(dim,candidate,df);
  Cholesky(dim,cholfactor,candidate,det2);
  (*------------------------------------------------------------------*)
  (*Evaluate ratio of joint multivariate normal probability densities.*)
  (*------------------------------------------------------------------*)
  presmvn := ((n / 2) * ln(det1)) + EvaluateMVN(densities,n,dim,re,invsigma); (*...and used to evaluate the joint multivariate probability density of the random effects.*)
  propmvn := ((n / 2) * ln(det2)) + EvaluateMVN(densities,n,dim,re,candidate); (*...and used to evaluate the joint multivariate probability density of the random effects.*)
  logmvnratio := propmvn - presmvn; (*Ratio of joint normal probability densities on natural logarithmic scale.*)
  (*-------------------------------------------------*)
  (*Evaluate ratio of candidate-generating densities.*)
  (*-------------------------------------------------*)
  proposal := EvaluateW(invsigma,df,dim,sample); (*Evaluate the Wishart density of the candidate with the present variance-covariance matrix of the random effects as parameter.*)
  present :=  EvaluateW(candidate,df,dim,sampleinvsigma); (*Evaluate the Wishart density of the present variance-covariance matrix of the random effects with the candidate as parameter.*)
  logcratio := present - proposal; (*Ratio of candidate-generating Wishart densities on natural logarithmic scale.*)
  (*----------------------------------*)
  (*Evaluate ratio of prior densities.*)
  (*----------------------------------*)
  propw := EvaluateW(i_omega,phi,dim,sample); (*Evaluate the Wishart probability density (on natural logarithmic scale), which is the conjugate prior probability density of the variance-covariance matrix of the random effects.*)
  presw := EvaluateW(i_omega,phi,dim,sampleinvsigma); (*Evaluate the Wishart probability density (on natural logarithmic scale), which is the conjugate prior probability density of the variance-covariance matrix of the random effects.*)
  logpratio := propw - presw; (*Ratio of prior probability densities on logarithmic scale.*)
  logpratio := logpratio + logcratio;
  (*------------------------------------------------*)
  (*Decide whether the candidate is accepted or not.*)
  (*------------------------------------------------*)
  accept := DecisionMH(pr,logmvnratio,logpratio,blackbox); (*Decide whether or not variance-covariance candidate is to be accepted.*)
  if accept then
    begin
    blackbox[5] := blackbox[5] + 1; (*Keep track of acceptance rate.*)
    Cholesky(dim,cholfactor,sample,det2);
    MatrixI(dim,det2,sigma,sample,cholfactor); (*Set variance-covariance matrix to proposed matrix.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Store_MCMC_Sample(var Bayesian, randomeffects: boolean; var ldim1, ldim2, actors: integer; var dim1, dim2: byte;
var MI_posterior_theta: pvec; var MII_posterior_theta: pmatrix; var sumre, sumposre, mcI_theta: rrarray;
var mcII_theta: rrrarray; var sumsigma: pmatrix; var mc_re, mc_posre, mc_covre: rrrarray);
(*Average and store MCMC samples, and clean up.*)
var
  i: byte;
  divisor, number: integer;
  zero: real;
  condition: boolean;
begin
(*
<ldim1> iteration number in <S_est>
<ldim2> total number of iterations in <S_ml>
<i x ldim2> batch size.
*)
i := 10; (*If this value is changed, then <it5> in <S_est> must be changed as well.*)
condition := (Bayesian or randomeffects);
if (Multiple(ldim1,i) and condition) then (*Test whether <ldim1> is a multiple of <i>.*)
  begin
  zero := 0;
  number := trunc(ldim1 / i); (*Number of the sample.*)
  (*Average.*)
  divisor := i * ldim2; (*Batch size.*)
  if Bayesian then
    begin
    ScalarDiv(dim1,MI_posterior_theta,divisor);
    ScalarDiv(dim1,MII_posterior_theta,divisor);
    (*Store.*)
    VectorStore(dim1,mcI_theta,number,MI_posterior_theta); (*Store parameter vector to evaluate lag <lag> autocorrelation function.*)
    MatrixStore(dim1,mcII_theta,number,MII_posterior_theta); (*Store parameter vector to evaluate lag <lag> autocorrelation function.*)
    (*Clean up.*)
    SetVectorValue(dim1,MI_posterior_theta,zero); (*Used by Bayesian algorithm to sum over the first moment of the sampled parameter vectors.*)
    SetMatrixValue(dim1,MII_posterior_theta,zero); (*Used by Bayesian algorithm to sum over the second moment of the sampled parameter vectors.*) 
    end;
  if randomeffects then (*Variance-covariance matrix of actor-dependent random effects.*)
    begin
    ScalarDiv(actors,dim2,sumre,divisor); (*Average: actor-dependent random effects.*)
    MatrixStore(actors,dim2,mc_re,number,sumre); (*Store.*)
    SetMatrixValue(actors,dim2,sumre,zero); (*Clean up.*)
    ScalarDiv(actors,dim2,sumposre,divisor); (*Average: positive value of actor-dependent random effects.*)
    MatrixStore(actors,dim2,mc_posre,number,sumposre); (*Store.*)
    SetMatrixValue(actors,dim2,sumposre,zero); (*Clean up.*)
    if Bayesian then
      begin
      ScalarDiv(dim2,sumsigma,divisor); (*Average: (inverse) variance-covariance matrix of actor-dependents.*)
      MatrixStore(dim2,mc_covre,number,sumsigma); (*Store.*)
      SetMatrixValue(dim2,sumsigma,zero); (*Clean up.*)
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Output_MII_theta(var Bayesian: boolean; var ascii: TFileStream; var dim: byte; var matrix1, matrix2: pmatrix; var vector1: pvec; var file1: TFileStream);
(*In the Bayesian approach, output the second moment of the parameter vector, considered as vector-valued random variable.*)
var
  h, hh: integer;
  i: byte;
  u: real;
  condition: boolean;
  c1, c2: pmatrix;
  vec1: pvec;
  x: string;
begin
if Bayesian then
  begin
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Posterior moment II of theta:'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  MatrixWriteTriangle(dim,ascii,matrix1);
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := '...and associated Monte Carlo s.e.:'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  MatrixWriteTriangle(dim,ascii,matrix2);
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Posterior variance-covariance matrix of theta (correlations below diagonal):'+eol;
  ascii.Write(pchar(x)^,length(x));
  VectorOProd(dim,c1,vector1); (*Computes vector outer product of estimated first posterior moment.*)
  MatrixMinus(dim,dim,c2,matrix1,c1);
  h := dim;
  Jacobi(h,c2,vec1,c1,hh);
  CorrelationMatrix(dim,c2);
  MatrixWrite(ascii,dim,dim,c2);
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  {----------------------}
  {-Standard deviations--}
  {----------------------}
  x := 'Posterior standard deviations: '+eol;
  ascii.Write(pchar(x)^,length(x));
  for i := 1 to dim do
    begin
    if (c2[i,i] > 0) then u := sqrt(c2[i,i])
    else u := -1000;
    x := format('%11.4f%s',[u,' ']);
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
  {-------------------------------------------------------------------}
  {-Check whether variance-covariance matrix is positive-semi-definite-}
  {-------------------------------------------------------------------}
  condition := false;
  x := 'Additional information (efficiency, stability etc.) in '+CurrentProjectName+'.log.'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  (*To check positive-definiteness.*)
  file1.Write(pchar(eol)^,length(eol));//writeln(file1);
  x := 'Latent roots of variance-covariance matrix: ';
  file1.Write(pchar(x)^,length(x));
  for i := 1 to dim do begin
    x := format('%11.4f%s',[vec1[i],' ']);
    file1.Write(pchar(x)^,length(x));
  end;
  file1.Write(pchar(eol)^,length(eol));file1.Write(pchar(eol)^,length(eol));//writeln(file1); writeln(file1);
  for i := 1 to dim do
    begin
    if (vec1[i] <= 0) then condition := true;
    end;
  if condition then
    begin
    x := 'W A R N I N G:   posterior variance-covariance matrix is not positive-definite.'+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

end.
