(*S_phase3
Entire unit originally by MS:
File contains routines needed in phase 3, which make extensive use
of the object <ModelSpecification>.
In the first place, the file contains routines to
(1) estimate derivatives in case of
- conditional and unconditional moment estimation
- of network evolution and behavior evolution models
- with constant and non-constant rate functions
- with utility and endowment functions
- and with changes enforced or not.
(2) other routines concerned with (partitioning, learning about) the parameter vector.
Order of variables in cdscore corresponds to order described in S_Dat,
search for 'Order of statistic types'.
*)

(* Changed by TS 29-12-06:
  functions NumberRateParameters - NumberActionEParameters replaced by variables
   that were already defined in S_Base.
   04-01-07: method estimator_d = 4 added.
   17-07-08: endowment effects behavior corrected.
             added variables aan and aa3 and skipped some local variables a
              so that less time is lost by creating dynamic arrays.
    *)
(*--------------------------------------------------------------------------------------------------*)

var aan, aa3: R_Ar;

function ModelSpecification.GetRE(const function_i: byte; const ldim, actor: integer): real;
(*MS: Get the corresponding random effect.*)
var h: integer;
begin
(*Remark:
- <function_i>
  <1>: objective function (evaluation function),
  <2>: gratification function (endowment function),
  other values not admissible.
- <ldim> is the number of the effect,
- <actor> is the actor initiating the change.
*)
GetRE := 0; (*Default.*)
if randomeffects then (*Random effects model.*)
  begin
  h := res[function_i,ldim]; (*<h = 0> if the corresponding coefficient is no random coefficient.*)
  if (h <> 0) then GetRE := RE[actor,h]; (*Random coefficient.*)
  end;
end; (*MS: GetRE*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.ComputeCoefficient(const theta: pvec; const actor, dim: integer; const actorRE: rrarray): real;
(*MS: Compute coefficient, which may be fixed or random.*)
var k: integer;
begin
(*Note:
- <thetarandom> indicates whether or not the coefficient is random,
- the boolean <randomeffects> is therefore redundant.
*)
if thetarandom[dim] then (*Random coefficient.*)
  begin
  k := re_position[dim]; (*Position of the corresponding random effect in <actorRE>.*)
  ComputeCoefficient := theta[dim] + actorRE[actor,k]; (*Random coefficient.*)
  end
else ComputeCoefficient := theta[dim]; (*Fixed coefficient.*)
end; (*MS: ComputeCoefficient*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.Multinomial(const ldim1, ldim2: integer; const theta: pvec; const actorRE: rrarray; const actor, alternative: integer; const statistic: rrarray): real;
(*MS: Evaluate multinomial choice probability (up to the the normalizing constant) on natural logarithmic scale.*)
var
  h: integer;
  coefficient, u: real;
begin
u := 0;
for h := ldim1 to ldim2 do (*Coefficient <h>.*)
  begin
  coefficient := ComputeCoefficient(theta,actor,h,actorRE); (*MS: Compute coefficient, which may be fixed or random.*)
  u := u + (coefficient * statistic[alternative,h]);
  end;
Multinomial := u;
end; (*MS: Multinomial*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.ChangeTheta(var outf: TFileStream; var theta: pvec);
(*MS: to estimate / test misspecified models, manipulate the value of the (data-generating) parameter
by utilizing user input.*)
var
 i, j: byte;
 u: real;
 x: string;
begin
x := 'Data-generating model'+eol;
outf.Write(pchar(x)^,length(x));
x := '---------------------'+eol;
outf.Write(pchar(x)^,length(x));
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
filename := 'TrueParameterValue.in';
ReadFile(ascii,filename);
(*First parameter:*)
readln(ascii,i);
readln(ascii,u);
theta[i] := u;
x := format('%s%d%s%4.4f',['True (data-generating) value of parameter ',i,
  ' is ',theta[i]])+eol;
outf.Write(pchar(x)^,length(x));
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
(*Second parameter:*)
readln(ascii,j);
readln(ascii,u);
theta[j] := u;
x:= format('%s%d%s%4.4f',['True (data-generating) value of parameter ',j,
' is ',theta[j]])+eol;
outf.Write(pchar(x)^,length(x));
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
x := 'The true (data-generating) values of the remaining parameters are equal to their initial values.'+eol;
outf.Write(pchar(x)^,length(x));
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
closefile(ascii);
end; (*MS: ChangeTheta.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.Sign_n(i, j: integer): real;
(*MS: The sign of the change in statistic which results from one particular change in the network.*)
begin
if yn(i,j) then Sign_n := -1.0 (*Arc exists in the present network, hence the minus sign.*)
else Sign_n := +1.0;
end; (*MS: Signa_n.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.SoLiS: boolean;
(*MS: Data, model, or test "Specification with Limited Support".*)
begin
if (symmetric or signed or (modeltype in [2, 3]) or (homogeneous > 0)) then SoLiS := true
else SoLiS := false;
end; (*MS: SoLiS.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.Initialise_d_est(var ascii: TFileStream;
                     const phase: byte; const total_n: cardinal; const dim: byte;
                     const theta: pvec; const override_default: boolean);
(*MS: method of moments: initialise derivatives estimation.*)
var x : string;
begin
n := nmax;
d_estAB := false; (*Default.*)
d_estA := false; (*Default.*)
d_estB := false; (*Default.*)
FinDiff_method := false; (*Default.*)
(*Comment: in case of ERGM, both booleans above are false.*)
if not(maxlike or Bayesian or (observations = 1)) then (*Moment estimation.*)
  begin
  (*Note:
  - <SoLiS>: = data, model, or test "Specification with Limited Support".
  - <d_estimator = 0>: user input, indicating that the user desires the crude method.*)
  x := format('%s%d%s',['Phase ',phase,':'])+eol;
  ascii.Write(pchar(x)^,length(x));
  if (SoLiS or (d_estimator = 0) or override_default) then
    begin
    FinDiff_method := true;
    if override_default then
      begin
      x := 'Derivative estimation: overriding Siena default choice...'+eol;
      ascii.Write(pchar(x)^,length(x));
      end;
    end
  else
    begin
    d_estAB := true; (*Estimate derivatives by the likelihood ratio (LR) method
                     of Schweinberger and Snijders (2006).*)
    d_estA := (d_estimator in [2..4]);
    d_estB := (d_estimator = 1);
    d_store := nil; (*Initialise derivatives.*)
    s_store := nil; (*Initialise complete-data score.*)
    u_store := nil; (*Initialise statistic.*)
    sc_sum := nil;   (*Initialise sum of scores *)
    st_sum := nil;  (*Initialise sum of statistics *)
    d_sum := nil;   (*Initialise sum of scores * statistics cross products *)
    if d_estA then
      begin
      (*<total_n> is the total number of scheduled iterations.*)
      SetLength(d_store,total_n+1,sqr(dim)+1);
      SetLength(s_store,total_n+1,dim+1);
      SetLength(u_store,total_n+1,dim+1);
      end;
    if d_estB then
      begin
      SetLength(d_sum,dim+1,dim+1,observations);
      SetLength(sc_sum,dim+1,observations);
      SetLength(st_sum,dim+1,observations);
      SetArray(aa3,4);
      SetArray(aan,nmax);
      end;
    SetArray(total_rate,nza,n); (*Initialise vector of total, actor-dependent rates.*)
    Initialise_D(dim);
    changestat := nil; (*<changestat> is created only once, and never reset.*)
    SetLength(changestat,6+1,ccmax+1,n+1);
      (*Explanation: dimension one refers to
        network functions:
        1 = network rate function
        2 = network objective function
        3 = network endowment function
        behavior functions:
        4 = behavior rate function
        5 = behavior objective function
        6 = behavior endowment function.*)
    SetVectorValue(pmax,theta_c,0);
    SetVector(dim,theta_c,theta); (*Set <theta_c = theta>; <theta_c> is the copy of <theta> in <S_base>.*)
    end;
  if FinDiff_method then
    begin
    case phase of
      1: begin
        x := 'Partial derivatives are estimated with the finite difference method.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      3: begin
        x :='Partial derivatives (and hence, standard errors) are estimated with the finite difference method.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'Partial derivatives are estimated with the finite difference method.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      end;
    end;
  if d_estAB then
    begin
    case phase of
      1: begin
        x := format('%s%d%s',['Partial derivatives are estimated with likelihood ratio method ',d_estimator,'.'])+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      3: begin
        x := format('%s%d%s',['Partial derivatives (and hence, standard errors) are estimated with likelihood ratio method ',d_estimator,'.'])+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else begin
        x := format('%s%d%s',['Partial derivatives are estimated with likelihood ratio method ',d_estimator,'.'])+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      end;
    end;
  end;
end; (*MS: Initialise_d_est.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_rate_rho(const dimension: byte): real;
(*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
begin
(*Explanation:
total_rate[0,dimension] = sum of actor-dependent rate functions for dependent dimension <dimension> in period <period>
lambda1[dimension,period] = basic rate parameter for dependent dimension <dimension> in period <period>.
*)
d_rate_rho := total_rate[dimension,0] / lambda1[dimension,thisperiod]; (*Paper: <q / r>.*)
end; (*MS: d_rate_rho.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_rate_alpha(var statistic: R_Ar; const dimension: byte): real;
(*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
var
  i: integer;
  sum: real;
begin
(*Explanation:
statistic = statistic that corresponds to the coordinate
total_rate = rate function.
*)
sum := 0;
(*Observe: the rate function of non-active actors is defined to be <0>.
Hence the summation over the full actor set is admissible.*)
for i := 1 to n do
  begin
  if StructActive[i] then sum := sum + (statistic[i] * total_rate[dimension,i]);
  end;
d_rate_alpha := sum;
end; (*MS: d_rate_alpha.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_rate(const rho: boolean; const dimension: byte; var statistic: R_Ar): real;
(*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
var u: real;
begin
if rho then u := d_rate_rho(dimension) (*Basic rate parameter.*)
else u := d_rate_alpha(statistic,dimension);
d_rate := u;
end; (*MS: d_rate.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_loglambda_rho(const dimension: byte): real;
(*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
begin
(*Explanation:
lambda1[dimension,period] = basic rate parameter for dependent dimension <dimension> in period <period>.
*)
d_loglambda_rho := 1 / lambda1[dimension,thisperiod]; (*Paper: <1 / r>.*)
end; (*MS: d_loglambda_rho.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_loglambda_alpha(var statistic: R_Ar; const ego: integer): real;
(*MS: Evaluates the partial derivative of the rate function with respect to coordinate <k> of the rate parameter.*)
begin
(*Explanation:
ego = actor that changed something
statistic = statistic that corresponds to the coordinate.
*)
d_loglambda_alpha := statistic[ego];
end; (*MS: d_loglambda_alpha.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.d_loglambda(const rho: boolean; const dimension: byte; var statistic: R_Ar; const ego: integer): real;
(*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
var u: real;
begin
if rho then u := d_loglambda_rho(dimension)
else u := d_loglambda_alpha(statistic,ego);
d_loglambda := u;
end; (*MS: d_loglambda.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.ratio_na: real;
(*MS: Evaluates, in the case where both the network and some behavioral dimensions change,
some function of the rates, which is needed in derivative estimation.*)
var
  h, variable: byte;
  denominator, numerator, rate_total: real;
begin
rate_total := rates_na[0]; (*Compute total rate: rate of network change plus...*)
for h := 1 to nza do (*...rate of behavior change, which is 0 in case there is no behavior evolution.*)
  begin
  rate_total := rate_total + rates_na[h];
  end;
variable := trunc(rates_na[0]); (*<variable> corresponds to the dimension that changed; truncation is trouble-free because the stored information is of type <byte>.*)
numerator := rate_total - rates_na[variable]; (*Numerator: difference in rate between total rate and rate of the dimension that changed.*)
denominator := rate_total * rates_na[variable]; (*Denominator: product of total rate and rate of the dimension that changed.*)
ratio_na := denominator / numerator; (*Ratio, which is a multiplicative constant in the complete-data score.*)
end; (*MS: ratio_na.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreRates(lambda_n: real; lambda_a: real; changed_dimension: byte);
(*MS: Stores the rates of network change and behavior change (for each behavioral dimension).*)
var i: integer;
begin
i := 1; (*Default: network evolution, no behavior evolution.*)
if (nza > 0) then i := i + nza; (*Network and behavior evolution.*)
SetArray(rates_na,i); (*Note: the <nil> operator is applied.*)
rates_na[0] := changed_dimension; (*Coordinate <0> indicates what dimension was changed.*)
rates_na[1] := lambda_n; (*Coordinate <1> is the rate of network change.*)
(*Code below to be updated...*)
if (nza > 0) then (*Network and behavior evolution.*)
  begin
  for i := 1 to nza do rates_na[i+1] := lambda_a; (*Coordinate <i + 1> is the rate of change of behavioral dimension <i>.*)
  end;
end; (*MS: StoreRates.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.CompleteStat;
(*MS: Complete the statistic, needed in derivative estimation.*)
var
  i: integer;
  count, position, h, k: byte;
begin
(*Important remarks:
for each outcome variable (network and behavioral outcome variables),
the statistic is set to 1 for the present period, and 0 otherwise:
- when there is conditioning on one of the outcome variables,
then no harm is done, since in the sequel, computations are carried out
if the corresponding rate parameters are included in the model,
and the corresponding boolean indicators are false
for the outcome variable on which one conditions.
- when there is more than one outcome variable, the statistic is set
to 1 for the present period, and 0 otherwise, for each outcome variable:
that is necessary to determine the derivative of the conditional probability
that the event occurs on the present dimension,
given that some event occurs on of the dimensions.
Note that the statistics included in the (multiple) rate function(s)
(constants, actor-bound covariates, and digraph statistics)
are evaluated and stored in <changestat> no matter on what dimension the event occurs
(indeed, they are computed in order to evaluate the rate functions
and to determine the dimension on which some event is to occur).
- <changestat> is not set to 0 after each event;
it hence it is necessary to reset <changestat.>
*)
for h := 0 to nza do
  begin
  for i := 1 to n do
    begin
    if (h = 0) then position := 0 (*Network event.*)
    else position := NetworkEffects_l + ((h - 1) * ActionEffects_l); (*Behavior event.*)
    for k := 1 to (observations - 1) do
      begin
      count := position + k;
      if (k = thisperiod) then changestat[1,count,i] := 1
      else changestat[1,count,i] := 0;
      end;
    end;
  end;
end; (*MS: CompleteStat.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.Prepare_Score(var rates: array of real): real;
(*MS: Prepare the evaluation of the complete-data score.*)
var
  h: byte;
  sum: real;
begin
CompleteStat; (*Complete the statistic corresponding to the basic rate parameters.*)
sum := 0;
for h := 0 to nza do
  begin
  sum := sum + rates[h];
  end;
Prepare_Score := sum; (*Total rate of change: parameter of negative exponential distribution of the joint Markov process.*)
end; (*MS: Prepare_Score.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.CompleteScore;
(*MS: unconditional MoM: at the end of each period, complete the complete-data score corresponding
to the rate parameters.*)
var
  count, dimension: byte;
  remaining_time: real;
begin
remaining_time := 1 - last_time[0,thisperiod]; (*<last_time> is the time point of the last change opportunity.*)
StoreScoreIII(nza+1,remaining_time,n+1); (*Network rate function parameters; first argument: trick for exploiting <StoreScoreIII>; last argument: arbitrary.*)
if (nza > 0) then
  begin
  count := ActiveNetworkEffects_l + ActiveNetworkEffects_f + ActiveNetworkEffects_g; (*Keeps track of right position in the score vector.*)
  for dimension := 1 to nza do
    begin
    remaining_time := 1 - last_time[dimension,thisperiod]; (*<last_time> is the time point of the last change opportunity.*)
    StoreScoreIV_k(dimension,count,remaining_time,false,n+1);
    (* Action rate function parameters; last but one argument: false; last argument: arbitary.*)
    (* StoreScoreIV_k increases parameter count. *)
    end;
  end;
end; (*MS: CompleteScore.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScoreI(var b: array of real; ego, alter: integer);
(*MS: Store the complete-data score corresponding to the change made:
      network objective and endownment function.*)
var
  count, j, h: integer;
  u: real;
begin
(*Explanation:
a = vector valued statistic
b = multinomial probability vector.*)
count := ActiveNetworkEffects_l; (*Keeps track of right position in the score vector.*)
(*----------------------------*)
(*-Network objective function-*)
(*----------------------------*)
for h := 1 to NetworkEffects_f do
  begin
  if Effects_f[h] then
    begin
    inc(count);
    for j := 1 to n do aan[j] := changestat[2,h,j];
// Note that by the use of SetChasZero in S_Base, all changestat values for non-permitted actors are 0.
    aan[ego] := 0; (*<Ego> is the actor with the opportunity to change something: if <ego> chose <ego>, the statistic would not change.*)
    u := VectorIProd(n,aan,b); (*Evaluate inner product <a' b>.*)
    u := aan[alter] - u; (*Complete-data score corresponding to parameter coordinate <h>.*)
    cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSI N.O.    ',i:3,u:12:6,';  period = ',thisperiod);
    end;
  end;
(*----------------------------*)
(*-Network endowment function-*)
(*----------------------------*)
for h := 1 to NetworkEffects_g do
  begin
  if Effects_g[h] then
    begin
    inc(count);
    for j := 1 to n do aan[j] := changestat[3,h,j];
    aan[ego] := 0; (*<Ego> is the actor with the opportunity to change something: if <ego> chose <ego>, the statistic would not change.*)
    u := VectorIProd(n,aan,b); (*Evaluate inner product <a' b>.*)
    u := aan[alter] - u; (*Complete-data score corresponding to parameter coordinate <h>.*)
    cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSI N.E.    ',i:3,u:12:6,';  period = ',thisperiod);
    end;
  end;
end; (*MS: StoreScoreI.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScoreII(var b: array of real; var changed: byte;
                                  var direction: integer; var minp,plup: boolean);
(*MS: Store the complete-data score corresponding to the change made:
      action objective and endowment function.*)
var
  count, h, number, position: integer;
  u: real;
begin
(*Explanation:
aa3 = vector valued statistic
b = multinomial probability vector
changed = the behavior which is changed
direction = direction of change.*)
count := BeforeActionParameters_f[changed]; (*Keeps track of right position in the score vector.*)
number := 3; (*At most three changes possible per action variable: upward (if possible), no change, downward (if possible).*)
(*---------------------------*)
(*-Action objective function-*)
(*---------------------------*)
for h := 1 to ActionEffects_f do
  begin
  position := NetworkEffects_f + ((changed - 1) * ActionEffects_f) + h;
  (*There are <changed - 1> action variables preceeding the present one.*)
  if Effects_f[position] then (*Parameter included.*)
    begin
    inc(count);
    aa3[1] := changestat[5,h,0]; (*Downwards.*)
    aa3[2] := 0; (*No change: <a> = change in statistic!*)
    aa3[3] := changestat[5,h,2]; (*Upwards.*)
    u := VectorIProd(number,aa3,b); (*Evaluate inner product <a' b>.*)
    u := aa3[direction] - u; (*Complete-data score corresponding to parameter coordinate <position>.*)
    cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSII A.O.   ',i:3,u:12:6,';  period = ',thisperiod);
    end;
  end;
//if (count <> BeforeActionParameters_f[changed+1]) then writeln; // check TS
(*---------------------------*)
(*-Action endowment function-*)
(*---------------------------*)
count := BeforeActionParameters_g[changed]; (*Keeps track of right position in the score vector.*)
for h := 1 to ActionEffects_g do
  begin
  position := NetworkEffects_g + ((changed - 1) * ActionEffects_g) + h; (*There are <changed - 1> action variables preceeding the present one.*)
  if Effects_g[position] then (*Parameter included.*)
    begin
    inc(count);
    aa3[1] := changestat[6,h,0]; (*Downwards.*)
    aa3[2] := 0; (*No change: <a> = change in statistic!*)
    aa3[3] := changestat[6,h,2]; (*Upwards.*)
    u := VectorIProd(number,aa3,b); (*Evaluate inner product <a' b>.*)
    u := aa3[direction] - u; (*Complete-data score corresponding to parameter coordinate <position>.*)
    cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSII A.E.   ',i:3,u:12:6,';  period = ',thisperiod);
    end;
  end;
//if (count <> BeforeActionParameters_g[changed+1]) then writeln; // check TS
end; (*MS: StoreScoreII.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScoreIII(const variable: byte;
                                           const htime: real; const ego: integer);
(*MS: Evaluate complete-data score corresponding to the network rate function parameters.*)
var
  count, first: integer;
  dimension, h: byte;
  u: real;
  condition, condition1, condition3, rho: boolean;
begin
(*-----------------------*)
(*-Network rate function-*)
(*-----------------------*)
count := 0; (*Keeps track of right position in the score vector.*)
condition1 := (ConditioningNetworkVariable <> 1); (*No conditioning on network distance.*)
if condition1 then first := 1 else first := observations;
for h := first to NetworkEffects_l do
  begin
  if Effects_l[h] then
    begin
    inc(count);
    rho := (h < observations);
      (*Indicates whether or not the rate parameter under consideration is a basic rate parameter.*)
    condition3 := (h = thisperiod); (*Basic rate parameter of this period.*)
    if (condition3 or not(rho)) then (*Basic rate parameter of this period or non-basic rate parameter.*)
      begin
      dimension := 0;
      condition := (dimension = variable);
      u := E_Score_Rates(rho,h,dimension,htime,condition,ego);
      cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSIII N.R.  ',i:3,u:12:6,';  period = ',thisperiod,'; time = ',htime:8:4);
      end;
    end;
  end;
end; (*MS: StoreScoreIII.*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.E_Score_Rates(const rho: boolean;
                    const parameter, dimension: byte; const htime: real;
                    const condition: boolean; const ego: integer): real;
(*MS: Evaluate the complete-data score corresponding to the parameters in the rate functions
of the behavioral outcome variables.*)
var
  j: integer;
  d, prod, u: real;
  a: R_Ar;
begin
if not(rho) then (*Statistic <a>: required for non-basic rate parameters.*)
  begin
  SetArray(a,n);
  for j := 1 to n do a[j] := changestat[1,parameter,j];
  end;
d := d_rate(rho,dimension,a);
(*<d> refers to derivative,
  and <rate> to the parameter of the negative exponential distribution of the holding times.*)
prod := d * htime;
if condition then
(*The outcome variable under consideration is the one on which the event occurred.*)
  begin
  u := d_loglambda(rho,dimension,a,ego);
(*<d> refers to the derivative, and <loglambda> to the logarithm of the rate function.*)
  u := u - prod;
  end
else u := - prod; (*The last event has passed...in the present time interval.*)
E_Score_Rates := u;
a := nil;
end; (*MS: E_Score_Rates.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScoreIV_k(const dimension: byte; var count: byte; const htime: real;
const condition: boolean; const ego: integer);
(*MS: Evaluate complete-data score corresponding to the behavior rate function parameters.*)
var
  first, ll: integer;
  h: byte;
  u: real;
  condition1, condition2, condition3, rho: boolean;
begin
(*Comment:
Evaluate the complete-data score corresponding to the parameters of the rate function
of the given behavioral outcome variable <dimension>.*)
condition1 := (ConditioningActionVariable <> dimension); (*No conditioning on distance of action variable under consideration.*)
if condition1 then first := 1 else first := observations;
for ll := first to ActionEffects_l do (*Given the outcome variable <dimension> do:*)
  begin
  h := NetworkEffects_l + ((dimension - 1) * ActionEffects_l) + ll;
  (*There are <dimension - 1> action variables preceeding the action variable of interest.*)
  if Effects_l[h] then
    begin
    inc(count);
    rho := (ll < observations);
      (*Indicates whether or not the rate parameter under consideration is a basic rate parameter.*)
    condition3 := (ll = thisperiod); (*Basic rate parameter of this period.*)
    if (condition3 or not(rho)) then (*Basic rate parameter of this period or non-basic rate parameter.*)
      begin
      u := E_Score_Rates(rho,h,dimension,htime,condition,ego); (*Evaluate score.*)
      cdscore[count,thisperiod] := cdscore[count,thisperiod] + u; (*Store score.*)
//writeln(exf,'SSIV_k A.R. ',i:3,u:12:6,';  period = ',thisperiod,'; time = ',htime:8:4);
      end;
    end;
  end;
end; (*MS: StoreScoreIV_k.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScoreIV(const variable: byte; const htime: real; const ego: integer);
(*MS: Evaluate complete-data score corresponding to the behavior rate function parameters.*)
var
  count, dimension: byte;
  condition: boolean;
begin
if (nza > 0) then
  begin
  count := ActiveNetworkEffects_l + ActiveNetworkEffects_f + ActiveNetworkEffects_g; (*Keeps track of right position in the score vector.*)
  for dimension := 1 to nza do
    begin
    condition := (dimension = variable);
    StoreScoreIV_k(dimension,count,htime,condition,ego);
    end;
  end;
end; (*MS: StoreScoreIV.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.Initialise_D(const dim: byte);
(*MS: Initialise arrays needed to estimate the derivatives of the estimating moment function.*)
(* Adapted, TS *)
begin
cdscore := nil;
last_time := nil;
SetLength(cdscore,dim+1,observations);
SetLength(last_time,nza+1,observations); (*Initialise vector of time points at which last change happend.*)
end; (*MS: Initialise_D.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StoreScore(var dim: byte; var statistic: pvec; var d_store: rrarray);
(*MS: The integrand, needed to estimate the derivatives matrix, is evaluated at the present Monte Carlo sample,
and stored.*)
var
  h, i, dimension: integer;
  statistic_centered: pvec;
  vec: R_Ar;
  integrand: rrarray;
  s: real;
begin
(*Moment estimation:
- unconditional: there is an additional term in the complete-data score
- conditional: there is no additional term in the complete-data score,
since the length of time interval is variable.
The statistics and rate functions needed below have been evaluated in <Runstep>
in order to sample the (last) holding time(s).*)
(*----------------------------*)
(*-Step 1: evaluate integrand-*)
(*----------------------------*)
for h := 1 to dim do
  begin
  statistic_centered[h] := statistic[h] - moments[h]; (*<moments> is the observed moment.*)
  end;
d_centring := true; (*Default.*)
if not(d_centring) then statistic_centered := statistic; (*Statistic not centred.*)
VectorOSProd(dim,observations-1,integrand,statistic_centered,cdscore); (*Evaluate the outer product <integrand = statistic * transpose(score)>.*)
VecOperator(dim,vec,integrand); (*The <vec> operator transforms the outer product <integrand> into the vector <vec>.*)
(*-------------------------*)
(*-Step 2: store integrand-*)
(*-------------------------*)
dimension := sqr(dim);
for i := 1 to dimension do (*Store information...*)
  begin
  d_store[phaseI_III,i] := vec[i]; (*The first dimension corresponds to the iteration number in phase 3.*)
  end;
for i := 1 to dim do
  begin
  s := 0.0;
  for h:=observations-1 downto 1 do s := s + cdscore[i,h];
  s_store[phaseI_III,i] := s; (*The first dimension corresponds to the iteration number in phase 3.*)
  end;
for i := 1 to dim do u_store[phaseI_III,i] := statistic[i]; (*The first dimension corresponds to the iteration number in phase 3.*)
end; (*MS: StoreScore.*)
(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.SumScore
          (var dim: byte; const ob: integer; var statistic: pvec);
  (* TS: analogous to StoreScore for d_estimator = 1;
         but needs to be called separately for each period.
         ob = period. *)
var i,j: byte;
begin
  for i:=dim downto 1 do
    begin
    for j:=dim downto 1 do
      d_sum[i,j,ob] := d_sum[i,j,ob] + statistic[i]*cdscore[j,ob];
    sc_sum[i,ob] := sc_sum[i,ob] + cdscore[i,ob];
    st_sum[i,ob] := st_sum[i,ob] + statistic[i];
    end;
end; (* SumScore*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.Estimate_D_EU(var cdim: cardinal; var dim: byte; var d_store: rrarray; var matrix1: pmatrix);
(*MS: Estimate the derivatives of <E(U | theta)> evaluated at the moment estimate of the parameter.*)
var i: integer;
begin
i := cdim;
ScalarDiv(dim,d_store,i); (*Average to obtain the Monte Carlo estimate of the derivatives.*)
SetMatrix(dim,matrix1,d_store); (*Set <dfra := d_store>.*)
end; (*MS: Estimate_D_EU.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.JacobianMatrix(var dim: byte; var ascii, outf: TFileStream;
                               var matrix1: pmatrix; var warning: integer);
(*MS: Moment estimation: estimate Jacobian matrix by various score function / likelihood ratio based methods:
see Schweinberger and Snijders (2005).
 TS: 04-01-07: case d_estB (d_estimator=1) added,
     the old d_estimator=1 changed to d_estimator=4.
     This is like the new d_estimator=4 (= old d_estimator=1), but with a more efficient combination
     across periods. 
 TS: warning is input/output indicating problems with matrix inversion *)
var
  i, j, h, k, l, count, dimension1, samples: integer;
  dimension2, number: byte;
  resultjm: real;
  large_version, condition, givewarnings: boolean;
  name: shortstring;
  error: shortstring;
  f: textfile;
  correction, minimum_n, C_n, sum_C, sum_S, sum_U, sum_US, US_n: R_Ar;
  basic_parameter: array of boolean;
  inverse, variance: pmatrix;
  integrand, isigma11, minimum, sigma11, sigma21, S, C_centered, US, US_centered, U: rrarray;
  x: string;

  procedure writeoutmat(var mm: pmatrix);
  var i,j: byte;
  x : string;
  begin
    for i:=1 to dim
    do begin
      for j:=1 to dim
      do begin
        x := format('%12.4f%s',[mm[i,j],' ']);
        outf.Write(pchar(x)^,length(x));
        (* if ((j mod 6) = 0) then writeln(cf); *)
      end;
      (* if ((pp mod 6) <> 0) then writeln(cf); *)
      outf.Write(pchar(eol)^,length(eol));//writeln(cf);
    end;
    outf.Write(pchar(eol)^,length(eol));//writeln(cf);
  end; (* writeoutmat *)
begin (* JacobianMatrix *)
if d_estB then
  begin
  for i:=1 to dim do for j:=1 to dim do
    begin
    resultjm := 0.0;
    for h:= observations-1 downto 1 do
      resultjm := resultjm + (d_sum[i,j,h] - (st_sum[i,h]*sc_sum[j,h]/phaseI_III))/phaseI_III;
    matrix1[i,j] := resultjm;
    end;
  end
else (* not d_estB *)
begin
(* this begin goes on almost to the end of JacobianMatrix *)
(*Explanation:
- samples: number of sample points (iterations) = sample size
- US: integrand, stored for each sample point
- C: control variate, which is composed of
  * S: complete-data score, stored for each sample point
  * U: statistic, stored for each sample point.
  If <large_version = false>, then the score is used as a control variate,
  otherwise both the score and the statistic are used as control variates.
- dimension1: the dimension of the vectorized Jacobian matrix, which is <dim * dim>
- dimension2: the dimension of the control variate, which depends on <large_version>,
  and is either <dim> or <dim + dim>.
Note that the information is stored in <vec> form.
Three estimators:
 1: plain:
    * <d_centering = false>: without reparametrization (centering)
    * <d_centering = true>: with reparametrization (centering).
 2: with variance reduction, based on the given Monte Carlo sample:
    * <large_version = false>: with the score as control variate
    * <large_version = true>: with the score and the statistic as control variate.
 3: with variance reduction, based on another Monte Carlo sample (input required).
(*----------------------------*)
(*-Estimate 1: plain estimate-*)
(*----------------------------*)
(*Average of the integrand over sample points.*)
dimension1 := sqr(dim);
(*To make the code more lucid:*)
samples := phaseI_III;
SetArray(US,samples,dimension1);
SetArray(S,samples,dim);
SetArray(U,samples,dim);
for i := 1 to samples do
  begin
  for h := 1 to dimension1 do US[i,h] := d_store[i,h];
  for h := 1 to dim do
    begin
    S[i,h] := S_store[i,h];
    U[i,h] := U_store[i,h];
    end;
  end;
SetArray(sum_US,dimension1);
for i := 1 to samples do
  begin
  for h := 1 to dimension1 do
    begin
    (*
    write(ascii,US[i,h]:20:10);
    *)
    sum_US[h] := sum_US[h] + US[i,h];
    end;
  (*
  writeln(ascii);
  *)
  end;
(*
for i := 1 to samples do
  begin
  for h := 1 to dim do write(ascii,U[i,h]:20:10);
  writeln(ascii);
  end;
for i := 1 to samples do
  begin
  for h := 1 to dim do write(ascii,S[i,h]:20:10);
  writeln(ascii);
  end;
*)
ScalarDiv(dimension1,sum_US,samples); (*Average the integrand.*)
SetArray(US_centered,samples,dimension1);
for i := 1 to samples do (*Center the integrand.*)
  begin
  for h := 1 to dimension1 do US_centered[i,h] := US[i,h] - sum_US[h];
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := format('%s%d%s',['Derivative estimator ',d_estimator,' used (statistics ']);
ascii.Write(pchar(x)^,length(x));
if not(d_centring) then ascii.write(pchar(string('not '))^,4);
x := 'centred).'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
if (d_estimator = 2) then
  begin
  x := 'Initial, plain derivative estimate without variance reduction method:'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  i := 0;
  repeat
  Inc(i);
  for h := 1 to dim do (*Comment: the original matrix was transformed by the <vec> operator.*)
    begin
    count := (h - 1) * dim; (*<h - 1> column vectors preceed the column of interest.*)
    count := count + i; (*Get the element in the <i>-th row and the <h>-th column.*)
    x := format('%11.4f',[sum_US[count]]);
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  until (i = dim);
  (*--------------------------------------------------------*)
  (*Preparation: estimate the (variance-)covariance matrices*)
  (*--------------------------------------------------------*)
  (*Comment: The (variance-)covariance matrices are estimated from the same Monte Carlo sample.*)
  (*Step 1: estimate <Sigma11>.*)
  large_version := false;
  if large_version then dimension2 := 2 * dim
  else dimension2 := dim;
  SetArray(sigma11,dimension2);
  SetArray(sum_S,dim);
  SetArray(sum_U,dim);
  SetArray(sum_C,dimension2);
  SetArray(C_centered,samples,dimension2);
  SetArray(C_n,dimension2);
  for i := 1 to samples do
    begin
    for h := 1 to dim do
      begin
      sum_S[h] := sum_S[h] + S[i,h];
      sum_U[h] := sum_U[h] + U[i,h];
      end;
    end;
  ScalarDiv(dim,sum_S,samples); (*Average the complete-data score.*)
  ScalarDiv(dim,sum_U,samples); (*Average the statistic.*)
  count := 0;
  for h := 1 to dim do (*Control variate: complete-data score.*)
    begin
    inc(count);
    sum_C[count] := sum_S[h];
    end;
  if large_version then (*Additional control variate: statistic.*)
    begin
    for h := 1 to dim do
      begin
      inc(count);
      sum_C[count] := sum_U[h];
      end;
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Average score: ';
  ascii.Write(pchar(x)^,length(x));
  for h := 1 to dim do begin
    x := format('%20.10f',[sum_S[h]]);
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Average statistic: ';
  ascii.Write(pchar(x)^,length(x));
  for h := 1 to dim do begin
    x := format('%20.10f',[sum_U[h]]);
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  for i := 1 to samples do
    begin
    count := 0;
    for h := 1 to dim do (*Control variate: complete-data score.*)
      begin
      inc(count);
      C_centered[i,count] := S[i,h] - sum_S[h];
      end;
    if large_version then (*Additional control variate: statistic.*)
      begin
      for h := 1 to dim do
        begin
        inc(count);
        C_centered[i,count] := U[i,h] - sum_U[h];
        end;
      end;
    end;
  for i := 1 to samples do
    begin
    for h := 1 to dimension2 do C_n[h] := C_centered[i,h]; (*Important: the centered (!) version!*)
    VectorOProd(dimension2,integrand,C_n,C_n); (*Outer product <C_n * transpose(C_n)>.*)
    MatrixAdd(dimension2,sigma11,integrand); (*Sum up...*)
    end;
  i := samples - 1;
  ScalarDiv(dimension2,sigma11,i);
  MakeStaticI(dimension2,variance,sigma11); (*Transform dynamic array into static array: <variance = sigma11>.*)
  error := '';
  Pinverse(dimension2,variance,inverse,error); (*Evaluate <inverse(Sigma11)>.*)
  MakeDynamic(dimension2,isigma11,inverse); (*Transform static array into dynamic array: <isigma11 = inverse>.*)
  (*Step 2: estimate <Sigma21>.*)
  SetArray(sigma21,dimension1,dimension2);
  SetArray(US_n,dimension1);
  for i := 1 to samples do
    begin
    for h := 1 to dimension1 do US_n[h] := US_centered[i,h]; (*Important: the centered (!) version!*)
    for h := 1 to dimension2 do C_n[h] := C_centered[i,h]; (*Important: the centered (!) version!*)
    VectorOProd(dimension1,dimension2,integrand,US_n,C_n); (*Outer product <US_n * transpose(C_n)>.*)
    MatrixAdd(dimension1,dimension2,sigma21,integrand); (*Sum up...*)
    end;
  i := samples - 1;
  ScalarDiv(dimension1,dimension2,sigma21,i);
  (*-------------------------------*)
  (*-Estimate 2: advanced estimate-*)
  (*-------------------------------*)
  (*Variance reduction method: complete-data score and statistic as control variates, see Schweinberger and Snijders (2005).*)
  MatrixProd(minimum,dimension1,dimension2,sigma21,isigma11); (*Evaluate the MLE of <Sigma21 * inverse(Sigma11)>.*)
  name := pname + 'Min.in';
  (*Comment:
  The minimizer is estimated even when <estimator = 3>:
  while this may be redundant, it is safer as the input file may not exist.*)
  if (d_estimator = 3) then (*Minimizer estimated from another Monte Carlo sample.*)
    begin
    if FileExists(name) then
      begin
      MatrixRead(f,name,minimum,dimension1,dimension2); (*Otherwise, estimate II is substituted.*)
      ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
      x := 'Minimizer: user input.'+eol;
      ascii.Write(pchar(x)^,length(x));
      end;
    end
  else (*<d_estimator = 2>.*)
    begin
    condition := false;
    if condition then
      begin
      i := 12;
      count := 6;
      OpenFile(f,name);
      MatrixWrite(dimension1,dimension2,f,minimum,i,count);
      closefile(f);
      end;
    end;
  SetArray(correction,dimension1);
  SetArray(minimum_n,dimension2);
  for j := 1 to dimension1 do (*Row <j> of the matrix that minimizes the generalised variance of the derivative estimator.*)
    begin
    for h := 1 to dimension2 do minimum_n[h] := minimum[j,h]; (*The <j>-th row of the matrix.*)
    resultjm := VectorIProd(dimension2,minimum_n,sum_C); (*For given sample point: inner product <row(j) * control variate>.*)
    correction[j] := resultjm;
    end;
  VectorMinus(dimension1,sum_US,correction); (*For given sample point: <mean US = mean US - correction>.*)
  end;
(*--------*)
(*-Output-*)
(*--------*)
count := 0;
for i := 1 to dim do (*Transform into matrix form.*)
  begin
  for j := 1 to dim do
    begin
    Inc(count);
    matrix1[j,i] := sum_US[count]; (*Column by column...*)
    end;
  end;
(*Since it is known that the cross-derivatives of the rates statistics with respect to other rate parameters are vanish,
it is sensible to replace the score function / likelihood ratio estimates of the derivatives by the known values 0.*)
number := ActiveNetworkeffects_l + ActiveActionEffects_l;
basic_parameter := nil;
SetLength(basic_parameter,number+1); (*Identifier: basic rate parameter or not.*)
(*Identify the basic rate parameters, starting with the network, and continuing with the action variables:*)
if (conditioningnetworkvariable = 0) then (*Check whether it is conditioned on the network distance.*)
  begin
  for i := 1 to (observations - 1) do basic_parameter[i] := true;
  end;
count := ActiveNetworkEffects_l;
if (nza > 0) then (*Action variables.*)
  begin
  for h := 1 to nza do (*For each action variable...*)
    begin
    if (h <> conditioningactionvariable) then (*...check whether it is conditioned on its distance...*)
      begin
      for i := (count + 1) to (count + observations - 1) do basic_parameter[i] := true; (*Set true for basic rate parameters of action variable <h>.*)
      end;
    i := 0;  (*Count the total number of rate parameters for action variable <h>, regardless of whether there is conditioning or not.*)
    for l := 1 to ActionEffects_l do (*Given the outcome variable <h> do:*)
      begin
      k := NetworkEffects_l + ((h - 1) * ActionEffects_l) + l; (*There are <variable - 1> action variables preceeding the action variable of interest.*)
      if Effects_l[k] then inc(i);
      end;
    count := count + i; (*Total number of rate parameters, up to and including action variable <h>.*)
    end;
  end;
for i := 1 to number do (*For each row in the rate parameter block of the Jacobian matrix...*)
  begin
  if basic_parameter[i] then (*...check whether the row corresponds to some basic rate parameter...*)
    begin
    for j := 1 to number do (*...if true: for each column in the rate parameter block of the Jacobian matrix...*)
      begin
      if basic_parameter[j] then (*...check whether the column corresponds to some basic rate parameter...*)
        begin
        if (i <> j) then matrix1[i,j] := 0; (*...if true: replace non-diagonal elements: i.e. the cross-derivative by 0.*)
        end;
      end;
    end;
  end;
Output_d_multiple(dimension1,sum_US); (*Multiple sequences output: Jacobian matrix.*)
end; (* if not d_estB *)
givewarnings := (warning > 0);
warning := 0;
for i := 1 to dim do (*To avoid runtime errors when inverting the matrix,
                       check diagonal values:*)
  begin
  if (matrix1[i,i] <= 0) then
    begin
    resultjm := 1;
    warning := i;
    x := format('%s%d%s',['Warning: diagonal value ',i,' is non-positive.']);
    ascii.Write(pchar(x)^,length(x));
    if givewarnings then begin
      x := '..corresponding derivatives deleted.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end
       else ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    if givewarnings then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    if givewarnings then begin
      x := format('%s%d%s%s%s%s',['W A R N I N G: standard error for estimate ',i,' below is unreliable...see file ',CurrentProjectName + '.cck','.'])+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    for j := 1 to dim do (*Delete corresponding column and row, and set the diagonal element equal to <resultjm>.*)
      begin
      matrix1[i,j] := 0;
      matrix1[j,i] := 0;
      end;
    matrix1[i,i] := resultjm;
    end;
  end;
if (warning > 0) then
  begin
  writeoutmat(matrix1);
  end;
if (d_estimator in [2..3]) then
  begin
  x := 'Advanced derivative estimate with variance reduction method'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := 'using the complete-data score';
  ascii.Write(pchar(x)^,length(x));
  if large_version then begin
    x := ' and the first moment of the statistic';
    ascii.Write(pchar(x)^,length(x));
  end;
  x := ' as control variate(s):'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(x)^,length(x));//writeln(ascii);
  end;
  aan := nil;
  aa3 := nil;
(*
Comment:
When it is suspected that the correlations are close to unity, the variance-covariance matrix
need not be estimated...and hence computation time could be saved. See Rubinstein (1989).
*)
end; (*MS: JacobianMatrix.*)

(*--------------------------------------------------------------------------------------------------*)

(*MS: Comment:
For the Independent Arcs model, some procedures follow which
- implement the score function / likelihood ratio method of derivative estimation,
- evaluate the derivatives analytically.
It is helpful to look at the Independent Arcs version of the source code in <.../programs/IA>.
*)

(*--------------------------------------------------------------------------------------------------*)

function ModelSpecification.Ratio(const maximum_outdegree, outdegree: integer; const theta: pvec): extended;
(*Compute <(part1 - part2) / (part1 + part2)>,
where
  <part1 = (maximum_outdegree - outdegree) * exp(theta[2])>
and
  <part2 = outdegree * exp(- theta[2])>,
where <outdegree> may represent an arc from <i> to <j>, the outdegree of <i>, or the total outdegree, depending on the input,
and <maximum_outdegree> represents the maximum the outdegree can attain.
*)
var part1, part2: extended;
begin
part1 := (maximum_outdegree - outdegree) * expo(theta[2]);
part2 := outdegree * expo(- theta[2]);
Ratio := (part1 - part2) / (part1 + part2);
end; (*MS: Ratio*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.CD_Score_IA(const i, j: integer; const q: real; const theta: pvec; const holding_time: real);
(*Complete-data score: Independent Arcs (IA) model.*)
var
  h, sum,
  x_ij_old, x_pp_old: integer;
  d_q_ij, d_q,
  part1, part2, resultsc: extended;
begin
sum := 0;
for h := 1 to n do sum := sum + outd_yn(h);
if yn(i,j) then
  begin
  x_ij_old := 0;
  x_pp_old := sum - 1; (*Total outdegree evaluated at the old digraph.*)
  end
else
  begin
  x_ij_old := 1;
  x_pp_old := sum + 1;  (*Total outdegree evaluated at the old digraph.*)
  end;
(*-----------------*)
(*Score: <theta[1]>*)
(*-----------------*)
d_q_ij := 1 / theta[1]; (*Derivative of <ln q(i,j)>.*)
d_q := q / theta[1]; (*Derivative of <q>.*)
resultsc := d_q_ij - (d_q * holding_time);
cdscore[1,thisperiod] := cdscore[1,thisperiod] + resultsc;
(*-----------------*)
(*Score: <theta[2]>*)
(*-----------------*)
d_q_ij := Ratio(1,x_ij_old,theta); (*Derivative of <ln q(i,j)>: function of old (!) digraph.*)
part1 := (n * (n - 1)) - x_pp_old;
part1 := part1 * expo(theta[2]);
part2 := x_pp_old * expo(- theta[2]);
d_q := part1 - part2;
d_q := (theta[1] * d_q) / (n - 1); (*Derivative of <q>: function of old (!) digraph.*)
resultsc := d_q_ij - (d_q * holding_time);
cdscore[2,thisperiod] := cdscore[2,thisperiod] + resultsc;
end; (*MS: CD_Score_IA*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.CompleteScore_IA(const theta: pvec);
(*MS: Complete complete-data score: Independent Arcs (IA) model.*)
var
  h, sum, x_pp_last: integer;
  d_q, part1, part2, remaining_time: extended;
begin
sum := 0;
for h := 1 to n do sum := sum + outd_yn(h);
x_pp_last := sum;  (*Total outdegree evaluated at the last digraph.*)
remaining_time := 1 - last_time[0,thisperiod]; (*<last_time> is the time point of the last change.*)
part1 := (n * (n - 1)) - x_pp_last;
part1 := part1 * expo(theta[2]);
part2 := x_pp_last * expo(- theta[2]);
(*-----------------*)
(*Score: <theta[1]>*)
(*-----------------*)
d_q := (part1 + part2) / (n - 1); (*Derivative of <q>: equivalent to <q / theta[1]>.*)
cdscore[1,thisperiod] := cdscore[1,thisperiod] - (d_q * remaining_time);
(*-----------------*)
(*Score: <theta[2]>*)
(*-----------------*)
d_q := part1 - part2;
d_q := (theta[1] * d_q) / (n - 1); (*Derivative of <q>: function of old (!) digraph.*)
cdscore[2,thisperiod] := cdscore[2,thisperiod] - (d_q * remaining_time);
end; (*MS: CompleteScore_IA*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.Jacobian_IA(const theta: pvec; const M_0, M_1: integer);
(*MS: Exact (!) Jacobian matrix of estimating moment function: Independent Arcs (IA) model;
note that the Jacobian matrix is not estimated here, because it can be evaluated analytically;
see the paper on estimating Jacobian matrices by Schweinberger and Snijders.*)
var
  i, j: byte;
  factor1, factor2, factor3,
  q, h, part1, part2, p, r,
  term1, term2,
  var_M_01, var_M_11,
  xi_0, xi_1,
  xi_01, xi_02, xi_11, xi_12: real;
  s: shortstring;
  f: textfile;
  Delta, Variance: rrarray;
begin
(*Basic functions:*)
r := theta[1] / (n - 1);
q := expo(theta[2]) + expo(- theta[2]);
p := expo(theta[2]) / q;
h := expo( - (theta[1] * q) /  (n - 1));
(*Derivatives with respect to <theta[1]>:*)
xi_01 := (expo(theta[2]) * h) / (n - 1);
xi_11 := - (expo(- theta[2]) * h) / (n - 1);
(*Derivatives with respect to <theta[2]>:*)
factor1 := 2 / sqr(q);
factor2 := 1 - expo(- r * q);
term1 := factor1 * factor2;
factor1 := p * r * expo(- r * q);
factor2 := expo(theta[2]) - expo(- theta[2]);
term2 := factor1 * factor2;
xi_02 := term1 + term2;
(*<term1> was calculated above.*)
factor1 := 1 - p;
(*<factor2> was calculated above.*)
factor3 := r * expo(- r * q);
term2 := factor1 * factor2 * factor3;
xi_12 := term1 - term2;
(*Jacobian matrix:*)
SetArray(Delta,2,2);
part1 := M_0 * xi_01;
part2 := M_1 * xi_11;
Delta[1,1] := part1 - part2;
Delta[2,1] := part1 + part2;
part1 := M_0 * xi_02;
part2 := M_1 * xi_12;
Delta[1,2] := part1 - part2;
Delta[2,2] := part1 + part2;
(*Output:*)
s := 'Jacobian_IA.out';
OpenFile(f,s);
for i := 1 to 2 do
  begin
  for j := 1 to 2 do write(f,Delta[i,j]:20:10);
  writeln(f);
  end;
closefile(f);
(*Variance-covariance matrix of statistic:*)
xi_0 := p * (1 - expo(- r * q));
xi_1 := p + ((1 - p) * expo(- r * q));
var_M_01 := M_0 * xi_0 * (1 - xi_0);
var_M_11 := M_1 * xi_1 * (1 - xi_1);
SetArray(Variance,2,2);
Variance[1,1] := var_M_01 + var_M_11;
Variance[2,2] := Variance[1,1];
Variance[2,1] := var_M_01 - var_M_11;
Variance[1,2] := Variance[2,1];
s := 'Variance_IA.out';
OpenFile(f,s);
for i := 1 to 2 do
  begin
  for j := 1 to 2 do write(f,Variance[i,j]:20:10);
  writeln(f);
  end;
closefile(f);
Delta := nil;
Variance := nil;
end; (*MS: Jacobian_IA*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.Output_d_multiple(var dimension: integer; var estimate: R_Ar);
(*MS: Multiple sequences output: Jacobian matrix.*)
var
  h: byte;
  s: shortstring;
  f: textfile;
begin
if (sequences > 1) then (*Multiple sequences output.*)
  begin
  s := pname + 'JacobianM.m';
  if (run = 1) then OpenFile(f,s)
  else AppendToFile(f,s);
  for h := 1 to dimension do write(f,estimate[h]:20:10);
  writeln(f);
  closefile(f);
  end;
end; (*MS: Output_d_multiple.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.StatisticsSpecial(var fun: pvec);
(*MS: Evaluate the statistic for the outlier candidates using the simulated digraph.*)
var
 i, j, sum: integer;
begin
(*The procedure <Statistics>, in which this procedure is nested, evaluates the moment for given period <k>.*)
if (homogeneous < 20) then (*Homogeneous actors: outlier.*)
  begin
  sum := 0;
  i := 0;
  repeat
  i := i + 1; (*Identify focal actor <i>.*)
  if outlier[i] then
    begin
    for j := 1 to n do
      begin
      if yn(i,j) then
        begin
        if (homogeneous = 11) then sum := sum + 1 (*outdegree*)
        else if (homogeneous = 12) then
          begin
          if yn(j,i) then sum := sum + 1; (*Reciprocity*)
          end;
        end;
      end;
    end;
  until (i = n);
  end;
fun[NumberFunctions] := sum;
end; (*MS: StatisticsSpecial.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.ObservedMomentPeriod(var ystat: pVec; var outf: TFileStream);
(*MS: computes the observed first moments of specified statistics for period <period>.*)
var
i: integer;
statt: pvec;
ordes: intvec;
begin
Initialise_actors(period+1);
Initialise_y(period+1); (*Resulting digraph and behaviour*)
for i:=1 to pmax do statt[i] := 0.0;
NetworkStatistics(statt,ordes,period+1,outf);
ActionStatistics(statt,ordes,period+1,outf);
for i:=1 to pmax do ystat[i] := statt[i];
for i:=1 to pmax do statt[i] := 0.0;
end; (*MS: ObservedMomentPeriod.*)

(*--------------------------------------------------------------------------------------------------*)

procedure ModelSpecification.PPChecks(const iteration: integer);
(*MS: posterior predictive checks: store simulated networks and behavior and compute functions of data.
Note: it is preferable to do the simulations during running time,
as the integration is with respect to all unobservables (parameters and latent data: paths and random effects),
and doing the simulations afterwards would imply that parameters, paths, and random effects had to be written to files and the inputted again;
the administrative effort would be huge.*)
var
  i, j, h: integer;
  s: shortstring;
  f: textfile;
begin
s := 'PPC' + inttostr(iteration) + '.dat';
OpenFile(f,s);
for i := 1 to n do (*Needs to be extended for more than two time points.*)
  begin
  for j := 1 to n do
    begin
    if yn(i,j) then write(f,' 1')
    else write(f,' 0');
    end;
  writeln(f);
  end;
closefile(f);
end; (*MS: PPChecks.*)

(*--------------------------------------------------------------------------------------------------*)

