unit s_test;
(*Entire unit by MS:
- evaluates Neyman-Rao generalised score test statistic
- outputs the test results
- contains routines to Monte Carlo simulate digraphs according to some given
  probability model, used to study the sampling distributions of statistics:
  (1) statistics used to estimate parameters and (2) test statistics.*)

(*--------------------------------------------------------------------------------------------------*)

interface

(*--------------------------------------------------------------------------------------------------*)

uses classes, s_constants, eight, slib, s_dat, s_base, s_matrices, Sysutils;

var
  obs, line, moments: integer;
  num: cardinal;
  uu, param: real;
  test_n: byte; (*<test_n> is the number of restrictions on the parameter that is to be tested.*)
  zeta: pvec;  (*<zeta> hands over the information contained in the local variable <fra> to <S_test>.*)
  delta, omikron, covtheta: pmatrix; (*<delta>, <omikron>, <covtheta> hand over the information
  contained in the local variables <dd>, <c>, and <c> to <S_test>. Used in the matrix routines below. Static
  array type <pmatrix> is defined in unit <Slib>.*)
  testresult, testresult_o: R_Ar; (*<testresult> stores the p-value that corresponds to the generalised score test with one
  degree of freedom; and to the p-values if the number of degrees of freedom is larger than one, in which case
  the desired test is carried out as well as the corresponding one degree of freedom tests;
  <testresult_o> contains the value of the test statistic in the non-quadratic form, which may be used for on.*)
  one_step_est: R_Ar; (*One-step estimator of the parameter by the generalised score test.*)

procedure Permutate(var dim: byte; var fixed: pBVec; var theta: pvec; var zeta: pvec);
(*Permutes the arrays according to <theta[1]> and <theta[2]>.*)

procedure RandomU(var uu: real);
(*Random draws from U[0,1].*)

procedure StoreAlternative;

procedure StoreOriginal;

procedure AlternativeModel;

procedure OriginalModel;

procedure ModelReSpecification(var line: integer; var param: real);
(*Respecifies the model by rewriting the <CurrentProjectName.mo> file.*)

procedure WriteSeed(CurrentProjectName: string; var modelseed: integer);
(*If <simulate> is enabled, write <seed> to the <CurrentProjectName.mo> file.*)

procedure ObservedMomentActors(var obs: integer; var homogeneous: byte; var moments: integer);
(*Calculating the observed moments if outlier test is to be performed.*)

function MinusOutliers(var line: integer): real;
(*Partition simulated value of statistic into the contribution of the
"normal" subset and the "deviating" subset of actors: compute the latter.*)

procedure PrepareHTests(var ascii: TFileStream);
(*Prepare homogeneity tests.*)

procedure InstabilityAnalysis(var dim: byte; var test: BBarray; var constant: pBVec; var covtheta: pmatrix; var ascii: TFileStream; var covZ: pmatrix);
(*Check instability.*)

procedure InitialiseTest(var test: BBarray; var dim: byte; var delta, d, omikron, msf: pmatrix);
(*Prepares arrays that are used in case score tests are carried out.*)

function TestInitialise(var Bayesian: boolean; var pe, pu, dim: byte; var test: BBarray; var test_CovRE: BBarray): boolean;
(*Initialises: checks whether test is admitted and how many restrictions are imposed on the parameters etc.*)

function EvaluateTestStatistic(var pu, pe, dim: byte; var test: BBarray; var zeta: pVec; var delta: pmatrix;
var omikron: pmatrix; var maxlike: boolean; var one_sided: real; var test_covre: BBarray): real;
(*Parameters: <theta[1]> is the nuisance parameter, <theta[2]> is the parameter of interest.
The restriction <H_0: theta[2] = 0> is to be tested against <H_1: theta[2] <> 0> by computing
- maximum likelihood: Lagrange multiplier / score test (Rao, 1948)
- unconditional / conditional moment estimation: generalized Neyman-Rao test (Neyman, 1959, Basawa, 1985).*)

procedure One_step_estimator(const maxlike: boolean; const dim: byte; var delta, omikron: pmatrix; const statistics: pvec; var one_step_est: R_Ar);
(*Evaluates one-step estimator of the parameter which was subjected to score testing.*)

procedure ScoreTest(var Bayesian: boolean; var dim: byte; var test: BBarray;
var test_CovRE: BBarray; var zeta: pVec; var delta, omikron: pmatrix;
var maxlike: boolean; var run: integer; var testresult: R_Ar; var test_n: byte; var one_step_est: R_Ar);
(*Parameters: <theta[1]> is the nuisance parameter, <theta[2]> is the parameter of interest.
The restriction <H_0: theta[2] = 0> is to be tested against <H_1: theta[2] <> 0> by computing
- maximum likelihood: Lagrange multiplier / score test (Rao, 1948)
- unconditional / conditional moment estimation: generalized Neyman-Rao test (Neyman, 1959, Basawa ,1985).
A one-step estimator is computed at the end.*)

procedure TestOutput(var test_n: byte; var ascii: TFileStream; var maxlike: boolean; var dim: byte; var test: BBarray;
var Description: ppString; var homogeneous: byte; var theta: pvec; var one_step_est: R_Ar);
(*Checks whether a (generalised) score test was executed and---if yes---outputs the results.*)

procedure CompleteSFiles(var run: integer; var cconditional: boolean; var observations: integer;
var nza: integer; var dim: byte; var fixed: pBVec; var theta: pvec);
(*Multiple sequences: write parameter estimates to file.*)

procedure CompleteTFiles(const sequences: integer; var ascii: textfile);
(*Complete file containing the sampling distribution of test statistics.*)

(*--------------------------------------------------------------------------------------------------*)

implementation

(*--------------------------------------------------------------------------------------------------*)

procedure Permutate(var dim: byte; var fixed: pBVec; var theta: pvec; var zeta: pvec);
(*Permutes the arrays according to <theta[1]> and <theta[2]>.*)
type
  intarray = array of integer;
  realarray = array of real;
var
  i, j, ios: integer;
  no: intarray;
  ntheta, nzeta: realarray;
begin
setlength(no,dim+1); (*New Order according to <theta[1]> and <theta[2]>.*)
setlength(ntheta,dim+1);
setlength(nzeta,dim+1);
j := 0;
for i := 1 to dim do
  begin
  if (fixed[i] = false) then (*Proceed down the boolean vector <fixed>, look what coordinates are to be tested, and write these coordinates to <z2>, the remaining to <z1>.*)
    begin
    j := j + 1;
    no[j] := i;
    end;
  end;
for i := 1 to dim do
  begin
  if fixed[i] then (*That is, coordinate <i> is to be tested.*)
    begin
    j := j + 1;
    no[j] := i;
    end;
  end;
for i := 1 to dim do (*Now - since we have the new order - we can re-order the arrays.*)
  begin
  ios := no[i];
  ntheta[i] := theta[ios];
  nzeta[i] := zeta[ios];
  end;
for i := 1 to dim do
  begin
  theta[i] := ntheta[i];
  zeta[i] := nzeta[i];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure RandomU(var uu: real);
(*Random draws from U[0,1].*)
begin
uu := random(100000000);
uu := (uu+1) / 100000000;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure AlternativeModel;
(*Present model specification is changed into original one.*)
var
  s, ss: string;
  file1: textfile;
begin
s := 'Alternative.mo';
ss := paramstr(1)+'.mo';
assignfile(file1,ss);
if fileexists(ss) then erase(file1);
renamefile(s,ss);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure OriginalModel;
(*Present model specification is changed into original one.*)
var
  s, ss: string;
  file1: textfile;
begin
s := 'Original.mo';
ss := paramstr(1)+'.mo';
assignfile(file1,ss);
if fileexists(ss) then erase(file1);
renamefile(s,ss);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure StoreAlternative;
(*Present model specification is changed into original one.*)
var
  s, ss: string;
  file1: textfile;
begin
s := paramstr(1)+'.mo';
ss := 'Alternative.mo';
assignfile(file1,ss);
if fileexists(ss) then erase(file1);
renamefile(s,ss);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure StoreOriginal;
(*Present model specification is changed into original one.*)
var 
  s, ss: string;
  file1: textfile;
begin
s := paramstr(1)+'.mo';
ss := 'Original.mo';
assignfile(file1,ss);
if fileexists(ss) then erase(file1);
renamefile(s,ss);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ModelReSpecification(var line: integer; var param: real);
(*Respecifies the model by rewriting the <CurrentProjectName.mo> file.*)
var
  i: integer;
  s: string;
  original, duplicate1, duplicate2: textfile;
begin
assignfile(original,paramstr(1)+'.mo');
assignfile(duplicate1,'Alternative.mo');
assignfile(duplicate2,'Original.mo');
reset(original);
rewrite(duplicate1);
rewrite(duplicate2);
i := 0;
while not(eof(original)) do
  begin
  i := i + 1;
  readln(original,s);
  if (i <> line) then writeln(duplicate1,s) else writeln(duplicate1,'1 0   ',param:6:6,' 0    0');
  writeln(duplicate2,s);
  end;
closefile(original);
closefile(duplicate1);
closefile(duplicate2);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteSeed(CurrentProjectName: string; var modelseed: integer);
(*If <simulate> is enabled, write <seed> to the <CurrentProjectName.mo> file.*)
var
  count: integer;
  condition: boolean;
  s, ss: string;
  original, duplicate: textfile;
begin
assignfile(original,CurrentProjectName+'.mo');
assignfile(duplicate,'Original.mo');
reset(original);
rewrite(duplicate);
condition := false;
count := 0;
while not(eof(original)) do
  begin
  readln(original,s);
  if (s = '@4 [options : ]') then condition := true; (*Start to count...*)
  if condition then Inc(count);
  if (count <> 16) then writeln(duplicate,s)
  else writeln(duplicate,modelseed:4,' [random seed: 0 randomize, other value: user input]');
  end;
closefile(original);
closefile(duplicate);
Erase(original);
s := 'Original.mo';
ss := CurrentProjectName+'.mo';
Renamefile(s,ss); (*Overwrite the original file by its duplicate.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ObservedMomentActors(var obs: integer; var homogeneous: byte; var moments: integer);
(*Calculating the observed moments if outlier test is to be performed.*)
type
  threearray = array of array of array of byte;
var
  i, j, k, nn, sum: integer;
  f: textfile;
  s: string;
  ss: char;
  x: threearray;
begin
s := paramstr(1) + '.d01';
assignfile(f,s);
reset(f);
nn := 0;
read(f,ss);
while (ss = '0') or (ss = '1') do (*To compute the number of actors.*)
  begin
  nn := nn + 1;
  read(f,ss);
  end;
closefile(f);
x := nil; (*Reset array to be sure...*)
setlength(x,nn+1,nn+1,nn+1);
for i := 2 to obs do
  begin
  s := paramstr(1) + '.d0' + inttostr(i);
  assignfile(f,s);
  reset(f);
  for j := 1 to nn do
    begin
    for k := 1 to nn do
      begin
      read(f,ss);
      if (ss = '1') then x[i,j,k] := 1;
      end;
    readln(f);
    end;
  closefile(f);
  end;
sum := 0; (*Observed ObservedMomentActors.*)
j := 0;
repeat
j := j + 1;
if outlier[j] then
  begin
  if (homogeneous = 11) then (*Out-degree*)
    begin
    for i := 2 to obs do
      begin
      for k := 1 to nn do sum := sum + x[i,j,k];
      end;
    end
  else if (homogeneous = 12) then (*Reciprocity*)
    begin
    for i := 2 to obs do
      begin
      for k := 1 to nn do
        begin
        if (x[i,j,k] = 1) then
          begin
          if (x[i,k,j] = 1) then sum := sum + 1;
          end;
        end;
      end;
    end;
  end;
until (j = nn);
moments := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

function MinusOutliers(var line: integer): real;
(*Partition simulated value of statistic into the contribution of the
"normal" subset and the "deviating" subset of actors: compute the latter.*)
var i, j, sum: integer;
begin
sum := 0;
if (homogeneous < 20) then
  begin (*Computations for given digraph.*)
  for i := 1 to line do
    begin
    if outlier[i] then
      begin
      for j := 1 to line do
        begin
        if (homogeneous = 11) then (*Out-degree*)
          begin
          if yn(i,j) then sum := sum + 1;
          end
        else if (homogeneous = 12) then (*Reciprocity*)
          begin
          if yn(i,j) then
            begin
            if yn(j,i) then sum := sum + 1;
            end;
          end;
        end;
      end;
    end;
  end;
MinusOutliers := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure PrepareHTests(var ascii: TFileStream);
(*Prepare homogeneity tests.*)
var x : string;
begin
Heading(1,ascii,'Testing homogeneity assumptions');
x := 'Homogeneity assumption to be tested regarding parameter ';
ascii.Write(pchar(x)^,length(x));
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InstabilityAnalysis(var dim: byte; var test: BBarray; var constant: pBVec; var covtheta: pmatrix; var ascii: TFileStream; var covZ: pmatrix);
(*Check instability.*)
var
  i, j, k: integer;
  count, one, two: byte;
  cond_n, ma, mi, u: real;
  eigen: pvec;
  covth, v1: pmatrix;
  x: string;
begin
count := 0;
one := 0;
two := 0;
for i := 1 to dim do
  begin
  if not(test[i] or constant[i]) then
    begin
    Inc(one);
    Inc(count); (*Number of nuisance parameters.*)
    two := 0;
    for j := 1 to dim do
      begin
      if not(test[j] or constant[j]) then
        begin
        Inc(two);
        covth[one,two] := covtheta[i,j];
        end;
      end;
    end;
  end;
MatrixNorm(count,covth); (*Makes the lengths of the column vectors 1.*)
u := 0;
SetMatrixValue(dim,v1,u);
Jacobi(count,covth,eigen,v1,i); (*Procecure returns the eigenvalues of the matrix, stored in <eigen>.*)
ma := abs(eigen[1]);
mi := abs(eigen[1]);
for k := 1 to count do
  begin
  if (abs(eigen[k]) > ma) then ma := abs(eigen[k]);
  if (abs(eigen[k]) < mi) then mi := abs(eigen[k]);
  end;
if (mi <> 0) then cond_n := ma/mi; (*Condition number of the matrix.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Instability analysis'+eol;
ascii.Write(pchar(x)^,length(x));
x := '--------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
(*--------------------------------------------------*)
(*-Variance-covariance matrix of parameter estimate-*)
(*--------------------------------------------------*)
x := 'Variance-covariance matrix of parameter estimates';
ascii.Write(pchar(x)^,length(x));
if boolean1 then begin
  x := ' (without coordinates that are kept constant):'+eol;
  ascii.Write(pchar(x)^,length(x));
end
else begin
  x := ':'+eol;
  ascii.Write(pchar(x)^,length(x));
end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
(*Criterion value <400> suggested by Belsley, Kuh & Welsch (1980), according to Greene (2003), 'Econometric Analysis', p. 58.*)
x := format('%s%4.4f%s',['Condition number = ',cond_n,' '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Eigen values  ';
ascii.Write(pchar(x)^,length(x));
for k := 1 to count do begin
  x := format('%6.6f%s',[eigen[k],'  ']); (*Necessary because the dimension of <c> is unknown to the program in the present unit.*)
  ascii.Write(pchar(x)^,length(x));
end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
(*------------------------------------------*)
(*-Variance-covariance matrix of statistics-*)
(*------------------------------------------*)
(*All coordinates are important, whether the corresponding parameters are restricted or not.*)
MatrixNorm(dim,covZ); (*Makes the lengths of the column vectors 1.*)
u := 0;
SetMatrixValue(dim,v1,u);
Jacobi(dim,covZ,eigen,v1,i); (*Procecure returns the eigenvalues of the matrix, stored in <eigen>.*)
ma := abs(eigen[1]);
mi := abs(eigen[1]);
for k := 1 to dim do
  begin
  if (abs(eigen[k]) > ma) then ma := abs(eigen[k]);
  if (abs(eigen[k]) < mi) then mi := abs(eigen[k]);
  end;
if (mi <> 0) then cond_n := ma/mi; (*Condition number of the matrix.*)
x := 'Variance-covariance matrix of X:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
(*Criterion value <400> suggested by Belsley, Kuh & Welsch (1980), according to Greene (2003), 'Econometric Analysis', p. 58.*)
x := format('%s%4.4f%s',['Condition number = ',cond_n,' '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x :='Eigen values  ';
ascii.Write(pchar(x)^,length(x));
for k := 1 to dim do begin
  x := format('%6.6f%s',[eigen[k],'  ']); (*Necessary because the dimension of <c> is unknown to the program in the present unit.*)
  ascii.Write(pchar(x)^,length(x));
end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
x := 'Date and time: '+DateTimeToStr(Now)+eol; (*Sometimes useful to check the running time.*)
ascii.Write(pchar(x)^,length(x));
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialiseTest(var test: BBarray; var dim: byte; var delta, d, omikron, msf: pmatrix);
(*Prepares arrays that are used in case score tests are carried out.*)
var 
  i, j, k: byte;
  condition: boolean;
begin
condition := false; (*Check whether score tests are to be executed.*)
i := 0;
repeat
  inc(i);
  if test[i] then condition := true;
until (condition or (i = dim));
if condition then (*Score tests to be executed.*)
  begin
  for j := 1 to dim do
    begin
    for k := 1 to dim do
      begin
      delta[j,k] := d[j,k]; (*Derivatives of EU.*)
      omikron[j,k] := msf[j,k]; (*Variance-covariance matrix of U.*)
      end;
    end;
  for j := 2 to dim do
    begin
    for k := 1 to (j - 1) do omikron[j,k] := omikron[k,j]; (*Necessary because below diagonal elements are correlations in the original matrix.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function TestInitialise(var Bayesian: boolean; var pe, pu, dim: byte; var test: BBarray; var test_CovRE: BBarray): boolean;
(*Initialises: checks whether test is admitted and how many restrictions are imposed on the parameters etc.*)
var i: byte;
begin
TestInitialise := false; (*Default.*)
if Bayesian then (*Bayesian domain: tests are senseless.*)
  begin
  pe := 0;
  pu := dim;
  end
else (*Frequentist domain: tests do make sense.*)
  begin
  pe := 0;
  pu := 0;
  for i := 1 to dim do (*<pu> is the dimension of the nuisance parameter, <pe> is the dimension of the parameter of interest.*)
    begin
    if test[i] then inc(pe) (*Restrictions on the parameter are indicated by <test>.*)
    else inc(pu);
    end;
  end;
if (homogeneous > 0) then (*Actor / period homogeneity test.*)
  begin
  pe := 1; (*One-parameter test admissible only.*)
  pu := dim - 1;
  end;
if test_CovRE[0] then
  begin
  (*Comment: testing restrictions on both <theta> and the variance-covariance matrix of
  the actor-dependent random effects is not implemented, and is not sensible anyway.*)
  pe := 0;
  pu := dim;
  end;
if ((pe > 0) or test_CovRE[0]) then TestInitialise := true; (*Score test.*) 
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateTestStatistic(var pu, pe, dim: byte; var test: BBarray; var zeta: pVec; var delta: pmatrix;
var omikron: pmatrix; var maxlike: boolean; var one_sided: real; var test_covre: BBarray): real;
(*Parameters: <theta[1]> is the nuisance parameter, <theta[2]> is the parameter of interest.
The restriction <H_0: theta[2] = 0> is to be tested against <H_1: theta[2] <> 0> by computing
- maximum likelihood: Lagrange multiplier / score test (Rao, 1948)
- unconditional / conditional moment estimation: generalized Neyman-Rao test (Neyman, 1959, Basawa, 1985).*)
var
  i, j, one, two, oneone, onetwo, twoone, twotwo: byte;
  cvalue: real;
  f: textfile;
  s: shortstring;
  error: shortstring;
  oh, ov, ne, z1, z2: R_Ar;
  rg, d12, d21, d22, sigma11, sigma12, sigma21, sigma22, v1, v2, v3, v4, v5, v6, v7, v8, v9: rrarray;
  d11, id11, vav, v10: pmatrix; (*Static arrays.*)
begin
(*------------*)
(*-Initialize-*)
(*------------*)
(*Set the lengths of the dynamic, local arrays.*)
SetArray(oh,pe);
SetArray(ov,pe); (*<ov> is a vector-valued function orthogonal to <z1>.*)
SetArray(ne,pe);
SetArray(z1,pu); (*We will partition the statistics vector <z> into <z1> and <z2> according to <theta[1]> and <theta[2]>.*)
SetArray(z2,pe);
SetArray(sigma11,pu,pu);
SetArray(sigma12,pu,pe);
SetArray(d12,pu,pe);
SetArray(d21,pe,pu);
SetArray(sigma21,pe,pu);
SetArray(d22,pe,pe);
SetArray(sigma22,pe,pe);
(*The remaining arrays are created outside the present procedure...*)
rg := nil; v1 := nil; v2 := nil; v3 := nil; v4 := nil; v5 := nil; v6 := nil; v7 := nil; v8 := nil; v9 := nil;
(*Reset local arrays when executing multiple sequences; Delphi seems to keep its values in the memory even though the arrays are local.*)
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    d11[i,j] := 0; id11[i,j] := 0; vav[i,j] := 0; v10[i,j] := 0;
    end;
  end;
(*Partition <E[Z|theta] - z> and <D = d_E[Z|theta] / d_theta> according to <theta[1]> and <theta[2]> and store the relevant parts in <z1>, <z2>, <d11>, <d21> to make the calculations more convenient.*)
one := 0;
two := 0;
for i := 1 to dim do
  begin
  if test[i] then
    begin
    two := two + 1;
    z2[two] := zeta[i];
    twoone := 0;
    for j := 1 to dim do
      begin
      if (test[j] = false) then
        begin
        twoone := twoone + 1;
        d21[two,twoone] := delta[i,j];
        end;
      end;
    end
  else (*That is, <test[i]> is false.*)
    begin
    one := one + 1;
    z1[one] := zeta[i];
    oneone := 0;
    for j := 1 to dim do
      begin
      if (test[j] = false) then
        begin
        oneone := oneone + 1;
        d11[one,oneone] := delta[i,j];
        end;
      end;
    end;
  end;
(*Partition <Cov(Z)> according to <theta[1]> and <theta[2]> and store the parts in <sigma11>, <sigma12>, <sigma21>, <sigma22>.*)
one := 0;
two := 0;
for i := 1 to dim do (*That is, coordinate <i> is restricted.*)
  begin
  if test[i] then
    begin
    two := two + 1;
    twoone := 0;
    twotwo := 0;
    for j := 1 to dim do
      begin
      if test[j] then
        begin
        twotwo := twotwo + 1;
        sigma22[two,twotwo] := omikron[i,j];
        d22[two,twotwo] := delta[i,j]; (*ML estimation: needed.*)
        end
      else (*That is, <test[i]> is true but <test[j]> is false.*)
        begin
        twoone := twoone + 1;
        sigma21[two,twoone] := omikron[i,j];
        end;
      end;
    end
  else (*That is, coordinate <i> is unrestricted.*)
    begin
    one := one + 1;
    oneone := 0;
    onetwo := 0;
    for j := 1 to dim do
      begin
      if test[j] then (*That is, <test[i]> is false but <test[j]> is true.*)
        begin
        onetwo := onetwo + 1;
        sigma12[one,onetwo] := omikron[i,j];
        end
      else (*That is, both <test[i]> and <test[j]> are false.*)
        begin
        oneone := oneone + 1;
        sigma11[one,oneone] := omikron[i,j];
        end;
      end;
    end;
  end;
PInverse(pu,d11,id11,error); (*<Compute <ID_11 = inverse(D_11)>.*)
MatrixProd(pe,pu,rg,d21,id11); (*Determine <rg = D_21 * inverse(D_11)>.*)
if maxlike then (*Maximum likelihood estimation.*)
  begin
  (*----------------*)
  (*Rao's score test*)
  (*----------------*)
  (*Important remark: in ML estimation, <D> is not the derivative of the estimating function, but the incomplete-data information matrix.*)
  for i := 1 to pe do ov[i] := - z2[i]; (*<z2> is the part of the Fisher score vector that corresponds to the parameter of interest;
  the minus sign is required because <S_est> stores minus the Fisher score vector.*)
  MatrixT(pe,pu,d12,d21); (*Compute <transpose(D_21)>.*)
  MatrixP(pe,pu,pe,v1,rg,d12); (*Compute <D_21 * inverse(D_11) * D_12>; note that <D_21 * inverse(D_11)> is of order <pe x pu>, <D_12> of order <pu x pe>, and the resulting matrix <v1> of order <pe x pe>.*)
  MatrixMinus(pe,pe,v9,d22,v1); (*Compute <D_22 - D_21 * inverse(D_11) * D_12>.*)
  end
else (*Moment estimation.*)
  begin
  (*---------------------------------*)
  (*Generalised Neyman-Rao score test*)
  (*---------------------------------*)
  (*Comment: when testing statistical hypotheses where nuisance parameters are present under <H_0> and <H_1>, and standard solutions
  (e.g, sufficient statistics, pivotal statistics) are not at the disposal, it is necessary to make the scores of the parameter
  of interest orthogonal to the "scores" of the nuisance parameter, and then use the orthogonalized "scores" to test the restrictions
  on the parameter of interest.  See Neyman (1959); generalized by Basawa (1985). Note that "score" is set between parentheses,
  because we cannot evaluate the Fisher score function.*)
  for i := 1 to pe do (*Compute <ov>, the vector that is orthogonal to <z1>. Not done in <MatrixProd> procedure due to the dynamic arrays used here.*)
    begin
    for j := 1 to pu do oh[i] := oh[i] + (rg[i,j] * z1[j]);
    end;
  for i := 1 to pe do ov[i] := (z2[i] - oh[i]);
  (*--------------------------------------------*)
  (*-Inverse variance-covariance matrix of <ov>-*)
  (*--------------------------------------------*)
  MatrixProd(pe,pu,v1,rg,sigma11); (*Step 1: compute <v1 = D_21 * inverse(D_11) * Sigma_11>.*)
  MatrixMinus(pe,pu,v2,sigma21,v1); (*Step 2: compute <v2 = Sigma_21 - [D_21 * inverse(D_11) * Sigma_11]>.*)
  MatrixT(pu,pu,v3,id11); (*Step 3: compute <v3 = transpose(inverse(D_11))>.*)
  MatrixT(pe,pu,v4,d21); (*Step 4: compute <v4 = transpose(D_21)>.*)
  MatrixProd(pe,pu,v5,v2,v3); (*Step 5: compute <v5 = (Sigma_21 - [D_21 * inverse(D_11) * Sigma_11]) * transpose(inverse(D_11))>.*)
  (*Note that in the two following matrix products, the first matrix is of order <pe x pu>, the second of order <pu x pe>, and the resulting matrix of order <pe x pe>.*)
  MatrixP(pe,pu,pe,v6,v5,v4); (*Step 6: compute <v6 = (Sigma_21 - [D_21 * inverse(D_11) * Sigma_11]) * transpose(inverse(D_11)) * transpose(D_21)>.*)
  MatrixP(pe,pu,pe,v7,rg,sigma12); (*Step 7: compute <v7 = D_21 * inverse(D_11) * Sigma_12>.*)
  MatrixMinus(pe,pe,v8,sigma22,v7); (*Step 8: compute <v8 = Sigma_22 - D_21 * inverse(D_11) * Sigma_12>.*)
  MatrixMinus(pe,pe,v9,v8,v6); (*Step 9: compute <v9 = v8 - v6>.*)
  end;
MakeStaticII(pe,v10,v9); (*<v10> is the static equivalent of the variance-covariance matrix <v9>; required by <Pinverse>.*)
PInverse(pe,v10,vav,error); (*Compute the inverse variance-covariance matrix of <ov>, that is <vav := inverse(v9)>, dimension = <pe>.*)
(*----------------*)
(*-Test statistic-*)
(*----------------*)
(*Evaluate the test statistic, which, under <H_0>, is asymptotically (central) chi-square distributed with <df = pe>.*)
for i := 1 to pe do
  begin
  for j := 1 to pe do ne[i] := ne[i] + (ov[j] * vav[j,i]);
  end;
cvalue := VectorIProd(pe,ne,ov); (*<cvalue> is the value of the test statistic.*)
if (cvalue < 0) then cvalue := 0; (*<cvalue> must be non-negative; it is negative if and only if some variances are negative...*)
if (pe = 1) then (*One degree of freedom: thus there is the additional option of one-sided tests:*)
  begin
  if (vav[1,1] >= 0) then one_sided :=  ov[1] * sqrt(vav[1,1]) (*<ov> is the testing function, while <vav> represents its inverse (!) variance.*)
  else one_sided := 0;
  if not(maxlike) then one_sided := - one_sided;
  (*Method of moments framework: the test is based on <E(U|theta) - u>, where <E(U|theta)> is a monotone, increasing function of <theta>:
  - if the true parameter value is higher than asserted by the null hypothesis (e.g., positive rather than zero),
    then the test statistic will be negative (expectation too low).
  - if the true parameter value is lower than asserted by the null hypothesis (e.g., negative rather than zero),
    then the test statistic will be positive (expectation too high).
  Therefore, the test statistic should be multiplied by <-1> to avoid confusion among users.*)
  end;
(*Comment: some simulation output, code to be deleted in the future.*)
if simulate then 
  begin
  s := 'Statistic.out';
  if FileExists(s) then AppendToFile(f,s)
  else OpenFile(f,s);
  for i := 1 to dim do write(f,zeta[i]:20:10);
  writeln(f);
  closefile(f);
  s := 'Variance.out';
  if FileExists(s) then AppendToFile(f,s)
  else OpenFile(f,s);
  write(f,'Variance test statistic: ');
  for i := 1 to pe do
    begin
    for j := 1 to pe do write(f,v9[i,j]:20:10);
    writeln(f);
    end;
  writeln(f);   
  writeln(f,'Variance Z: ');
  for i := 1 to dim do
    begin
    for j := 1 to dim do write(f,omikron[i,j]:20:10);
    writeln(f);
    end;
  writeln(f);
  writeln(f,'Derivatives: ');
  for i := 1 to dim do
    begin
    for j := 1 to dim do write(f,delta[i,j]:20:10);
    writeln(f);
    end;
  writeln(f);
  closefile(f);
  end;
(*------------------------------*)
(*Actor-dependent random effects*)
(*------------------------------*)
(*Comment: restrictions on the variance-covariance matrix of the actor-dependent random effects to be tested:
assuming that no other restrictions on <theta> are to be tested at the same time.*)
if test_covre[0] then
  begin
  cvalue := 0; (*To be implemented.*)
  end;
EvaluateTestStatistic := cvalue; (*Value of test statistic.*)
(*Clean up...*)
oh := nil; ov := nil; ne := nil; z1 := nil; z2 := nil;
rg := nil; d12 := nil; d21 := nil; d22 := nil; sigma11 := nil; sigma12 := nil; sigma21 := nil; sigma22 := nil; v1 := nil; v2 := nil; v3 := nil; v4 := nil; v5 := nil; v6 := nil; v7 := nil; v8 := nil; v9 := nil;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure One_step_estimator(const maxlike: boolean; const dim: byte; var delta, omikron: pmatrix; const statistics: pvec; var one_step_est: R_Ar);
(*Evaluates one-step estimator of the parameter which was subjected to score testing.*)
var
  i, j: byte;
  sum: real;
  error: shortstring;
  matrix, inverse: pmatrix;
begin
(*Note: in the ML framework, <dfra> in <S_est> first represents minus the derivative of the estimating function (minus the conditional expectation of the complete-data Hessian),
but then is transformed into the incomplete-data Fisher information matrix by subtracting the conditional variance-covariance matrix of the complete-data Fisher score:
<dfra* = dfra - msf>.
To obtain minus the derivative of the estimating function, required by the Taylor series-based one-step estimator,
<dfra* + msf = dfra>,
where <dfra* = delta> and <msf = omikron>.*)
if maxlike then MatrixSum(dim,dim,matrix,delta,omikron) (*ML: minus the derivative of the estimating function.*)
else SetMatrix(dim,dim,matrix,delta); (*MoM: <delta> is the derivative of the estimating function.*)
PInverse(dim,matrix,inverse,error);
SetArray(one_step_est,dim);
for i := 1 to dim do
  begin
  sum := 0;
  for j := 1 to dim do sum := sum + (inverse[i,j] * statistics[j]);
  one_step_est[i] := - sum; (*The value under the null hypothesis will be added later on.*)
  (*ML framework: the minus sign is absorbed by minus the conditional expectation of the Hessian given the panel data
    so that the increment should be added, not subtracted;
    however, the input, taken from <S_est>, is minus the Fisher score vector.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ScoreTest(var Bayesian: boolean; var dim: byte; var test: BBarray;
var test_CovRE: BBarray; var zeta: pVec; var delta, omikron: pmatrix;
var maxlike: boolean; var run: integer; var testresult: R_Ar; var test_n: byte; var one_step_est: R_Ar);
(*Parameters: <theta[1]> is the nuisance parameter, <theta[2]> is the parameter of interest.
The restriction <H_0: theta[2] = 0> is to be tested against <H_1: theta[2] <> 0> by computing
- maximum likelihood: Lagrange multiplier / score test (Rao, 1948)
- unconditional / conditional moment estimation: generalized Neyman-Rao test (Neyman, 1959, Basawa ,1985).
A one-step estimator is computed at the end.*)
var
  i, j, h, dim_i, dim_n, hdim_n, hdim_i, htotal, l: byte;
  chisquare, one_sided, u: real;
  test_desired: boolean;
  f: textfile;
  statistics: pvec;
  d_U, covU: pmatrix;
  tested: BBarray;
  delete_rc: pBVec;
begin
if (dim > ccmax) then exit;

test_desired := TestInitialise(Bayesian,dim_i,dim_n,dim,test,test_CovRE); (*Checks whether and how many tests are desired and admissible.*)
if test_desired then
  chisquare := EvaluateTestStatistic(dim_n,dim_i,dim,test,zeta,delta,omikron,maxlike,one_sided,test_covre)
   else exit;
    (*If there is more than one degree of freedom, then one degree of freedom tests will be carried out now:*)
(*Check whether an omnibus test was carried out: in that case, carry out in addition the single-parameter tests one-by-one*)
for i := 0 to dim do tested[i] := false;
testresult := nil;
testresult_o := nil;
SetLength(testresult,pmax+1); (*Two-sided multi- and one-parameter tests.*)
SetLength(testresult_o,pmax+1); (*One-sided one-parameter tests.*)
testresult[0] := chisquare; (*Chi-square value of desired test.*)

if (dim_i = 1) then testresult_o[1] := one_sided (*Dimension of the parameter of primary interest.*)
else (*One-parameter tests (one- and two-sided).*)
  begin
  (*Important note: the arrays that communicate the important information from <S_est> to S_test>
  are not changed when evaluating the test statistic. Thus, the values of the arrays need not be
  stored when evaluating the one degrees of freedom test statistics one-by-one in turn.*)
  for i := 1 to dim do
    begin
    tested[i] := test[i]; (*Save original vector.*)
    test[i] := false;
    end;
  j := 0;
  for i := 1 to dim do (*Carry out single-parameter tests one-by-one.*)
    begin
    if tested[i] then
      begin
      (*Delete rows and columns corresponding to other non-estimated parameter coordinates one-by-one.*)
      u := 0;
      h := pmax;
      SetMatrixValue(h,d_U,u);
      SetMatrixValue(h,covU,u);
      (*The following array is necessary to communicate to the matrix / vector delete procedures what rows and colums are to be deleted.*)
      for h := 1 to dim do
        begin
        if (tested[h] and (h <> i)) then delete_rc[h] := true
        else delete_rc[h] := false;
        end;
      MatrixDelete(dim,delete_rc,d_U,delta); (*Delete row <i> and column <i> of the derivatives matrix.*)
      MatrixDelete(dim,delete_rc,covU,omikron); (*Delete row <i> and column <i> of the variance-covariance matrix.*)
      VectorDelete(dim,delete_rc,statistics,zeta); (*Delete coordinate <i> of the statistics vector>.*)
      (*The coordinate that is to be tested is shifted down by <j> units, where <j> is the number of parameter coordinates that are restricted as well and which are down the array.*)
      l := i - j;
      test[l] := true; (*The restriction that is to be tested.*)
      htotal := dim_n + 1; (*Important: the dimension of the parameter is the number of nuisance parameters plus one.*)
      hdim_n := dim_n;
      hdim_i := 1;
      chisquare := EvaluateTestStatistic(hdim_n,hdim_i,htotal,test,statistics,d_U,covU,maxlike,one_sided,test_covre);
      inc(j); (*The number of one-parameter tests that has been carried out up to now.*)
      testresult[j] := chisquare; (*Two-sided test.*)
      testresult_o[j] := one_sided; (*One-sided test.*)
      test[l] := false; (*Reset.*)
      end;
    end;
  for i := 1 to dim do test[i] := tested[i]; (*Restore for possible later use.*)
  end;
test_n := dim_i; (*Communicates number of restricted parameter coordinates to other procedures.*)
(*In case the sampling distribution of the test statistic is studied by Monte Carlo simulation.*)
assignfile(f,CurrentProjectName+'Score.m'); (*File can be used to plot the distribution of the test statistic.*)
if (run = 1) then rewrite(f)
else append(f);
write(f,testresult[0]:20:10);
if (dim_i > 1) then
  begin
  for i := 1 to dim_i do write(f,testresult[i]:20:10);
  end;
writeln(f);
closefile(f);
One_step_estimator(maxlike,dim,delta,omikron,zeta,one_step_est); (*One-step estimates.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure TestOutput(var test_n: byte; var ascii: TFileStream; var maxlike: boolean; var dim: byte; var test: BBarray;
var Description: ppString; var homogeneous: byte; var theta: pvec; var one_step_est: R_Ar);
(*Checks whether a (generalised) score test was executed and---if yes---outputs the results.*)
var
  i, j, k: integer;
  pvalue: real;
  s, st1, st2, x: string;
begin
if (test_n > 0) then (*Restrictions on parameter were tested.*)
  begin
  if maxlike then (*Rao's score test.*)
    begin
    Heading(2,ascii,'Score test <c>');
    end
  else (*Generalised Neyman-Rao score test.*)
    begin
    Heading(2,ascii,'Generalised score test <c>');
    end;
  x := 'Testing the goodness-of-fit of the model restricted by'+eol;
  ascii.Write(pchar(x)^,length(x));
  j := 0;
  for k := 1 to dim do
    begin
    if test[k] then (*Coordinate is restricted by null hypothesis.*)
      begin
      j := j + 1;
      s := Description[k];
      if (homogeneous = 0) then begin
        x := format('%s%d%s%s%s%8.4f',[' (',j,')   ',Alignstring(s,50),' =',theta[k]])+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else if (homogeneous = 11) then begin
        x := format('%s%d%s',[' (',j,')   u: out-degree (specified actor(s))   '])+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else if (homogeneous = 12) then begin
        x := format('%s%d%s',[' (',j,')   u: reciprocity (specified actor(s))    '])+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else if (homogeneous > 20) then
        begin
        s := Description[coordinate] + ': constant across periods';
        x := format('%s%d%s%s',[' (',j,')   ',s])+eol;
        ascii.Write(pchar(x)^,length(x));
        end;
      end;
    end;

//  st1 := stringofchar('_',2+length(inttostr(trunc(test_n)-1)));
  st2 := stringofchar(' ',2);
  x := '_________________________________________________'+eol;
  ascii.Write(pchar(x)^,length(x));
//  x:= st1+eol;
//  ascii.Write(pchar(x)^,length(x));
  x := '                          ';
  ascii.Write(pchar(x)^,length(x));
  x := st2+''+eol;
  ascii.Write(pchar(x)^,length(x));
  if (test_n > 1) then
    begin
    x := 'Joint test:'+eol;
    ascii.Write(pchar(x)^,length(x));
    x := '-----------'+eol;
    ascii.Write(pchar(x)^,length(x));
    end;
  x := format('%s%8.4f%s%d%s',['   c = ',testresult[0],'   d.f. = ',j,'   p-value ']); (*<j> is the number of restricted parameters.*)
  ascii.Write(pchar(x)^,length(x));
  pvalue := 1 - chi2(j,testresult[0]);
  if (pvalue < 0.0001) then begin
    x := '< 0.0001';
    ascii.Write(pchar(x)^,length(x));
  end
  else begin
    x := format('%s%8.4f',['= ',pvalue]);
    ascii.Write(pchar(x)^,length(x));
  end;
  if (test_n = 1) then
    begin
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    x := format('%s%8.4f',['   one-sided (normal variate): ',testresult_o[1]]);
    ascii.Write(pchar(x)^,length(x));
    end;
  if (test_n > 1) then (*Omnibus test.*)
    begin
    ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
    for k := 1 to j do
      begin
      x := format('%s%d%s',['(',k,') tested separately:'])+eol;
      ascii.Write(pchar(x)^,length(x));
      x := '----------------------'+eol;
      ascii.Write(pchar(x)^,length(x));
      x := ' - two-sided:'+eol;
      ascii.Write(pchar(x)^,length(x));
      x := format('%s%8.4f%s',['   c = ',testresult[k],'   d.f. = 1   p-value ']);
      ascii.Write(pchar(x)^,length(x));
      pvalue := 1 - chi2(1,testresult[k]);
      if (pvalue < 0.0001) then begin
        x := '< 0.0001';
        ascii.Write(pchar(x)^,length(x));
      end
      else begin
        x := format('%s%8.4f',['= ',pvalue]);
         ascii.Write(pchar(x)^,length(x));
      end;
      ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
      x := format('%s%8.4f',[' - one-sided (normal variate): ',testresult_o[k]]);
      ascii.Write(pchar(x)^,length(x));
      if (k < j) then
        begin
        ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
        end;
      end;
    end;
  x := st2+' '+eol;
  ascii.Write(pchar(x)^,length(x));
  x := '_________________________________________________'+eol;
  ascii.Write(pchar(x)^,length(x));
//  x := st1+''+eol;
//  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  (*One-step estimates:*)
  x := 'One-step estimates:'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  for i := 1 to dim do
    begin
    one_step_est[i] := one_step_est[i] + theta[i];
    s := Description[i];
    x := format('%s%8.4f',[Alignstring(s,50),one_step_est[i]])+eol;
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteSFiles(var run: integer; var cconditional: boolean; var observations: integer;
var nza: integer; var dim: byte; var fixed: pBVec; var theta: pvec);
(*Multiple sequences: write parameter estimates to file.*)
var
  j, k: byte;
  f: textfile;
begin
if (sequences > 1) then
  begin
  assignfile(f,paramstr(1)+'ThetaMultiple.m');
  if (run = 1) then rewrite(f)
  else append(f);
  (*CS: the following *could* be adjusted to conditioning on action variables.*)
  if cconditional then k := 1 (*Conditional moment estimation.*)
  else k := ((observations - 1) * (nza + 1)) + 1; (*Unconditional moment estimation; the rate estimates would otherwise swamp histograms;
  must be adjusted as soon as MLE is available.*)
  for j := k to dim do
    begin
    if not(fixed[j]) then write(f,theta[j]:20:10,' '); (*Boolean condition necessary because otherwise there is one very high box*)
    end;
  if (run < sequences) then write(f);
  writeln(f);
  closefile(f); (*Otherwise runtime errors may be raised.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteTFiles(const sequences: integer; var ascii: textfile);
(*Complete file containing the sampling distribution of test statistics.*)
begin
if (sequences > 1) then
  begin
  writeln(ascii);
  closefile(ascii); (*Otherwise runtime errors may be raised.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

end.
