Unit digraph;

{  Unit for defining the data structures for use with valued digraphs.         }
{  This unit was written by Tom A.B. Snijders, starting August 2005.           }
{  Types TDoublyLinkedDoubleInt and TDoublyLinkedIntList are based             }
{  on the unit cLinkedLists written by David Butler,                           }
{  which is part of Delphi Fundamentals.                                       }
{  This latter unit contained an omission; see below, noted TS.                }

interface

uses Forms, SysUtils, Math, classes, s_matrices, rangen, s_constants;

{                                                                              }
{ Linked items forming a list: TDoublyLinkedDoubleInt                          }
{                                                                              }

const bnodesize = 100;

const ConfigMax = 10;
(* Maximum number of configurations for which effects are defined and calculated
   in object TEffects. *)
      n_crit =  100;
(* Maximum number of nodes for which the TValuedDigraph type
   includes an adjacency matrix. *)      

type

  TConfigWeight0 = array[0..ConfigMax] of real;
  TConfigWeight1 = array[1..ConfigMax] of real;
  TConfigContr0 = array[0..ConfigMax] of Boolean;
  TConfigContr1 = array[1..ConfigMax] of Boolean;

type
  TDoublyLinkedDoubleInt = class
  protected
    FNext : TDoublyLinkedDoubleInt;
    FPrev : TDoublyLinkedDoubleInt;

  public
    Value1: Integer;
    Value2: Shortint;

    property  Next: TDoublyLinkedDoubleInt read FNext write FNext;
    property  Prev: TDoublyLinkedDoubleInt read FPrev write FPrev;

    function  HasNext: Boolean;
    function  HasPrev: Boolean;
    function  Count: Integer;
    function UseMem(const V1: Integer; const V2: Shortint) :TDoublyLinkedDoubleInt;
    procedure LibMem;

    private
    FElemIndex: integer;

    constructor Create(const V1: Integer; const V2: Shortint);
//abcd    procedure
    procedure DestroyList;  (* Changed by TS from destructor to procedure *)
    procedure Remove;
    procedure InsertAfter(const V1: Integer; const V2: Shortint);
    procedure InsertBefore(const V1: Integer; const V2: Shortint);
//    procedure Delete;
  end;

(* The delete procedures execute the corresponding remove procedure
   and also free the memory *)

{                                                                              }
{ Linked list with first and last element: TDoublyLinkedIntList                }
{                                                                              }

(* Although this is not implied by the type definition,
   the type TDoublyLinkedIntList is used here in the following way:
   values of Value1 are ordered after Ffirst, i.e.,
   if (it1.Next = it2) and (it1 <> Ffirst) then it1.Value1 < it2.Value1;
   and Ffirst.Value1 is the number of successors of First.
   Used in this way, TDoublyLinkedIntList is interpreted as a (sparse) vector,
   where element with index Value1 has the value Value2,
   all other elements have the value 0,
   and Ffirst.Value1 is the number of nonzero elements.
   In comments, for a variable defined as var vec: TDoublyLinkedIntList,
   element i of this sparse vector will be denoted by vec[i]:
   that is, if the list contains an element it with it.value1 = i,
   then vec[i] = it.value2.
    *)

type
  TDoublyLinkedIntList = class
  protected
    FFirst : TDoublyLinkedDoubleInt;
    FLast  : TDoublyLinkedDoubleInt;
    FCount : Integer;

  public

    property  First: TDoublyLinkedDoubleInt read FFirst write Ffirst;
    property  Last: TDoublyLinkedDoubleInt read FLast write FLast;
    function  IsEmpty: Boolean;
    property  Count: Integer read FCount;
    function RandomElement(const version: Byte): integer;
(* Returns a random draw from the Value1 of the vector, excluding First *)
    function VectorValueUp(const i: integer): integer;
    function VectorValueDown(const i: integer): integer;
(* The value of element i of the ShortInt vector represented by vec,
    computed by an upward or a downward search; both yield the same outcome. *)
    procedure CopyTo(it: TDoublyLinkedIntList);
(* Copies self to it; it must have been created earlier *)
    procedure WriteVector(var outf: textfile; const mode: byte);
    procedure StreamVector(var outf:  TFileStream; const mode: byte);

    private

    constructor Create;  (* added by TS *)
    destructor Destroy; override;
    procedure Remove(const Item: TDoublyLinkedDoubleInt);
    function  RemoveFirst: TDoublyLinkedDoubleInt;
    function  RemoveLast: TDoublyLinkedDoubleInt;  (* use deletelast! *)

    procedure Delete(const Item: TDoublyLinkedDoubleInt);
    procedure DeleteFirst;
    procedure DeleteLast;
    procedure DeleteList;
    procedure Append(const V1: Integer; const V2: Shortint);
    procedure InsertFront(const V1: Integer; const V2: Shortint);

  end; (* class  TDoublyLinkedIntList *)


function InnerProduct(const vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the ShortInt vectors represented by vec1 and vec2. *)
function CheckInnerProduct(vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the ShortInt vectors represented by vec1 and vec2. *)
function IndicatorInnerProduct(const vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the <> indicators of the
   ShortInt vectors represented by vec1 and vec2. *)
function IndicatorQuadProduct(const vec1,vec2,vec3,vec4: TDoublyLinkedIntList): integer;
(* The sum of the products of the <> indicators of the
   ShortInt vectors represented by vec1, vec2,v ec3, vec4. *)
function Subset(const vec1,vec2: TDoublyLinkedIntList): boolean;
(* indicates that vec1 is a subset of vec2 *)
function Equal(const vec1,vec2: TDoublyLinkedIntList): boolean;
(* indicates that vec1 is equal to vec2 *)
function ArrayInnerProduct(const vec1: TDoublyLinkedIntList; x: array of integer;
                           var ercode: integer): integer;
(* The inner product of the ShortInt vector represented by vec1 with x;
   low(x) must be 0. *)

type TProcedure = procedure(const i: integer; t1,t2: shortint);

procedure ForDifferencesDo(const vec1,vec2: TDoublyLinkedIntList; procd: TProcedure);
(* Executes procd(i,vec1[i],vec2[i]) for all i where vec1[i] <> vec2[i],
   where i goes from small to large. *)

{                                                                              }
{ Valued directed graph: TValuedDigraph                                        }
{                                                                              }

type

  TValuedDigraph = Class
    private
      NumberNodes: Integer;
      helplist:    array of boolean;
      AdjMatrix:   array of array of shortint;
      HasArray:    boolean;

(* The data structure is defined in the following way,
   with redundancy to give extra speed:
   if the tie from i to j has value v for v <> 0,
   then there is some next^p (for p >= 1) of Inties[j].Ffirst with value i, v,
        and some next^p (for p >= 1) of Inties[i].Ffirst with value j, v;
   (Item.Value2 =0) if and only if Item = Ffirst in some list;
   Inties[j].Ffirst.Value1  is the indegree of j:
   Number of i for which TieValue(i,j) <> 0 ;
   Outties[i].Ffirst.Value1 is the outdegree of i:
   Number of j for which TieValue(i,j) <> 0.
   By default, DiagonalPermitted = false;
   if (not DiagonalPermitted) then TieValue(i,i) = 0 for all i.
   If HasArray, then there is a redundant representation
   by the Adjacency matrix.
   *)

(* When error messages have been generated during execution time,
   only the last error message will be copied to outf by the Destroy method. *)

    protected
      ErMes: shortstring;
    public
      Inties: array of TDoublyLinkedIntList;
      Outties: array of TDoublyLinkedIntList;
      DiagonalPermitted: boolean;
(* if more variables are added to this class,
   change procedure CopyGraph accordingly! *)
      constructor Create(const n: Integer);  overload;
      destructor Destroy; overload;
      destructor Destroy(var outf: TFileStream; const s: string); overload;
      property Order: Integer read NumberNodes;
      property ErrorMessage: shortstring read ErMes;
      function IsCorrect: Boolean;
(* Checks consistency of data structure *)
      function TieValue(const i,j: integer): ShortInt;
      function TieValueDown(const i,j: integer): ShortInt;
      function SimpleInTieValue(const i,j: integer): ShortInt;
      function SimpleOutTieValue(const i,j: integer): ShortInt;
(* all give the same result; TieValue is computed more quickly *)
      function InDegree(const i: integer): Integer;
(* Number of i for which TieValue(i,j) <> 0 *)
      function OutDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) <> 0 *)
      function PosInDegree(const i: integer): Integer;
(* Number of i for which TieValue(i,j) > 0 *)
      function NegInDegree(const i: integer): Integer;
(* Number of i for which TieValue(i,j) < 0 *)
      function PosOutDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) > 0 *)
      function NegOutDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) < 0 *)
      function ReciDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) <> 0  and TieValue(j,i) <> 0 *)
      function TotalTies: integer;
(* Number of (i,j) for which TieValue(i,j) <> 0 *)
      function Mini: ShortInt;
      function Maxi: ShortInt;
(* Minimum and maximum of all tie values *)
      function SymmetricGraph: Boolean;
(* Indicates whether the graph is symmetric *)
      function TwoPaths(const i,j: integer): Integer;
(* Number of twopaths from i to j *)
      function TwoPathsTrunc2(const i,j: integer): Integer;
(* Number of twopaths from i to j truncated at two *)
      function NoTwoPath(const i,j,h: integer): Boolean;
(* Indicates that there is no twopath i => h' => j for any h' < h *)
      function ExistsTwoPath(const i,j: integer): Boolean;
(* Indicates that there is at least one twopath i => h => j for some h  *)
      function NoOrOneTwoPath(const i,j: integer; var tp: integer): Boolean;
(* Indicates that there is at most one twopath i => h' => j; tp = 0 or 1,
    depending on whether there is 0 or 1 found, if it is 0 or 1.  *)
      function NoOrOneTwoPathMax(const i,j,h: integer): Boolean;
(* Indicates that there is at most one twopath i => h' => j for any h' < h *)
      function OneTwoPath(const i,j,h: integer): Boolean;
(* Indicates that there is at most one twopath i => h' => j for any h' < h *)
      function OutTwoStars(const i,j: integer): Integer;
(* Number of outtwostars to i and j *)
      function InTwoStars(const i,j: integer): Integer;
(* Number of intwostars from i and j *)
      procedure ThirdCount(const i,j: integer; var n3, n4: Integer);
(* n3 (n4) is computed  as the number of third actors h,
   such that there are 3 (4) ties among the four possibilities
   (i,h), (h,i), (j,h), (h,j). *)
      function Correlation(var z: array of real;
                           var miz: array of boolean;
         var n0, nn: integer; var m0,s0,m1,s1,m2,s2,dd,cs,ds: real): real;
(* correlation between z[i] and z[j] for tied pairs (i,j),
   with miz[i]=miz[j] = false.
   m0 and s0 are regular mean and s.d. of z;
   m1, s1 and m2, s2 mean and s.d. of z in first/second in tied pairs;
   n0, nn numbers of individuals and ties;
   nn/n0 mean degree;
   dd standard deviation of outdegrees;
   cs standard deviation of personal network means;
   ds average of within-personal network variances.  *)

     function CliqueBound: integer;
(* Upper bound to the largest clique size, computed from the degrees. *)
      procedure Cliques(const report: Boolean; var outf: TFileStream;
                        const size,smin,smax: integer;
                        nodes, allnodes:  TDoublyLinkedIntList;  itp: TDoublyLinkedDoubleInt;
                        var nc: array of integer);
(* Finds cliques containing the set nodes, and contained in the set allnodes;
   nodes must itself be a clique;
   size must be number of nodes, used for efficiency;
   itp must be nodes.last, also used for efficiency.
   nc gives the clique count.
   if report, then all cliques found of size at least smin are reported to outf.
   *)

     procedure RedCliques(const report: Boolean; var outf: textfile;
                        const size,smin,smax,red: integer;
                        nodes, allnodes:  TDoublyLinkedIntList; itp: TDoublyLinkedDoubleInt;
                        var nc: IIArray);
     procedure Red1_Cliques(const report: Boolean; var outf: TFileStream;
                        const size,smin,smax,ii,jj: integer;
                        nodes, ijnodes, inodes, jnodes:  TDoublyLinkedIntList;
                        nodelinks: TValuedDigraph; itp: TDoublyLinkedDoubleInt;
                        var nc: array of integer);

     procedure AllCliques(const report: Boolean; var outf: TFileStream; const smin,smax: integer);
(* Finds all cliques of order  smin to smax. *)
     procedure AllCliques_ij(const report: Boolean; var outf: TFileStream;
                             smin: integer; const smax,ii,jj: integer; const t: shortint;
                             var ncl: array of integer);
(* Finds all cliques of order smin to smax, containing ii and jj. *)
     procedure All_red1_Cliques_ij(const report: Boolean; var outf: TFileStream;
            smin: integer; const smax,ii,jj: integer; const t: shortint;
            var ncl: array of integer);
     procedure ReducedCliques(const report: Boolean; var outf: textfile;
            const red, smin, smax: integer);
(* Finds all red-uced cliques of order at least smin. *)

      procedure SetZero;
(* Sets the graph to the empty graph without changing the order *)
      procedure PermitDiagonal;
(* Sets DiagonalPermitted to true *)
      procedure ChangeTieValue(const i,j: integer; const v: shortint);
(* If 1 <= i, j<= NumberNodes and i <> j,
   then this procedure changes the tie from i to j to value v *)
      procedure AddTie(const i,j: integer; const v: shortint;
                       const checkhi: boolean);
(* This procedure adds tie (i,j) with value v,
   assuming that in the adjacency matrix
   both in row i and in column j this is the largest element. *)
      function IncreaseTie(const i,j: integer; const v: shortint): ShortInt;
(* If 1 <= i, j<= NumberNodes and i <> j,
   then this function increases the tie value from i to j by v.
   The new value is the function result.  *)
      procedure SetRowTo(const i: integer; const vec: TDoublyLinkedIntList);
(* Sets row i to vec *)
      procedure SetColumnTo(const i: integer; const vec: TDoublyLinkedIntList);
(* Sets column i to vec *)
      function RandomOutTie(const version: Byte; const i: integer): integer;
(* Random draw from the outties of i;  returns i if outdegree(i) = 0 . *)
      procedure RandomTie(const version: Byte; const c: array of integer;
                                   var i,j: integer);
(* Random draw (i,j) from the outties;
   i with cumulative probabilities proportional to c[i]; (0,0) for empty graph *)
      procedure ListsToAdjacency(var x: array of ShortInt);
(* Writes data to adjacency matrix; x must be allocated before
   with High(x) >= sqr(NumberNodes)+1. *)
      procedure WriteAdjacency(var outf: textfile);
(* Writes data in adjacency matrix format to file. *)
      procedure WriteLists(var outf: TFileStream);
(* Writes data in double lists format to file. *)
      procedure WriteEdgeList(const uppertriangle: boolean; var outf: textfile);
(* Writes data in list of out-ties format to file. *)
      procedure WritePajekFile(const directed: boolean; const fn: shortstring);
(* writes data to file in Pajek format *)
      procedure ReadPajekFile(const fn: shortstring;
                                       var ertext: shortstring; var extra: boolean);
(* reads data to file in Pajek format;
   extra is a flag for seeing nodes larger than numbernodes. *)
      procedure WriteToFile(const fn: shortstring);
(* Writes data in order-dependent format to file with name fn. *)
      procedure ReadFromFile(const n: integer; const fn: shortstring);
(* reads data from file with name fn, in a format written by WriteToFile. *)
      procedure WriteToArray(var z: iiarray);
(* Writes adjacency matrix. *)
      procedure ComponentOne(var d,z: iarray);
(* output z = upper bounds to distances in max-symmetrized to
   node with highest symmetrized degree;
   infinite distances are equal to Numbernodes.            *)
      function Connected(var i: integer): boolean;
(* output = (from i there are paths to all nodes) *)
      function Path(var i,j: integer): boolean;
(* output = (there is a path from i to j) *)
      function OtherPath(var i,j: integer): boolean;
(* output = (there is a path from i to j that does not use the link i-j ) *)
end; (* Class TValuedDigraph *)


procedure CopyGraph(const a: TValuedDigraph; var b: TValuedDigraph); overload;
(* copies a to b; b must have been created earlier. *)

procedure Symmetrize(var a, b: TValuedDigraph; const how: integer);
(* defines b = symmetrized a;
   by min for how = 1, by max for how = 2;
   b must have been created earlier. *)

procedure BooleanProduct(var a, b, c: TValuedDigraph; const how: integer);
(* defines c = boolean product a*b;
   c must have been created earlier.
   if how = 2, c = a*(b+identity) *)

procedure Simmelian(var a, b: TValuedDigraph);
(* defines b = Simmelian ties in a;
   b must have been created earlier. *)

function JointIn(const a,b: TValuedDigraph; const k: integer): integer;
(* sum_i a(i,k)b(i,k) *)

{                                                                              }
{ Directed graph: TDigraph                                                     }
{ The restriction of TValuedDigraph to values 0 and 1.                         }
{                                                                              }

type TDigraph = Class(TValuedDigraph)
    public
      function IsCorrect: Boolean;
(* Checks consistency of data structure *)
      procedure AddTie(const i,j: integer);        // added 3.17s
      procedure ChangeTieValue(const i,j: integer; const v: shortint);  overload;
      function IncreaseTie(const i,j: integer; v: shortint): ShortInt;  overload;
end; (* Class TDigraph *)

procedure CopyGraph(const a: TDigraph; var b: TDigraph); overload;
(* copies a to b; b must have been created earlier. *)

function RestrictedSum(const a: TValuedDigraph; const b: TDigraph): integer;
(* sum_ij a[i,j]*b[i,j] *)

function EqualGraph(const a: TDigraph; var b: TDigraph): boolean; // added 3.17s
(* a = b *)

function NewArcs(a,b: TDigraph): boolean; // added 3.17s
(* b has arcs that are not in a *)


{                                                                              }
{ Directed TRedundantValuedDigraph                                                      }
{                                                                              }
{ This is a doubly redundant representation with linked lists as well as       }
{ an adjacency matrix. For checking; and for a quicker evaluation of TieValue. }

  type TRedundantValuedDigraph = Class(TValuedDigraph)
    private
     AdjMatrix: array of array of integer;
    public
      constructor Create(const n: Integer);  overload;
      function IsCorrect: Boolean;  overload;
      function TieValue(const i,j: integer): ShortInt;  overload;
      procedure ChangeTieValue(const i,j,v: integer);  overload;
      function IncreaseTie(const i,j,v: integer): ShortInt;  overload;
end; (* Class TRedundantValuedDigraph *)

{                                                                              }
{ Partial Order: TPOset                                                     }
{ A partial order with its transitive reduction.                         }
{                                                                              }

type TPOSet = Class
    private
      arrivals, intermediate: TValuedDigraph;
        (* arrivals = array N of La Poutre and Van Leeuwen,
           intermediate = M *)
    public
      ordered, hasse: TValuedDigraph;  (* ordered is the partial order,
                                    hasse is the transitive reduction *)
      constructor Create(const n: Integer);  overload;
      destructor Destroy;  overload;
      destructor Destroy(var outf: textfile);  overload;
      function Order: integer;  overload;  (* number of nodes *)
      function IsTransitive(var outf: textfile): Boolean; (* checks that ordered is transitive closure of hasse *)
      function IsReduced(var outf: textfile): Boolean; (* checks that ordered is transitive closure of hasse *)
      function IsTheClosure(var outf: textfile): Boolean; (* checks that ordered is transitive closure of hasse *)
      function CorrectArrivals(var outf: textfile): Boolean; (* checks definition of arrivals *)
      procedure InsertEdge(const i,j: integer);
      procedure DeleteEdge(const i,j: integer);
      procedure WritePOSet(var f: textfile); (* writes arrivals, ordered, hasse to f *)
      procedure CopyPOTo(var cpos: TPOSet); (* copies the TPOSet to cpos, which must have been defined  *)

      function lt(const i,j: integer): boolean;   (* i less than j in partial order *)
      function gt(const i,j: integer): boolean;   (* i greater than j in partial order *)
      function next(const i,j: integer): boolean;  (* hasse(i,j) *)
end; (* Class TPOSet *)



{                                                                              }
{ Effects for Valued directed graph: TEffect                                   }
{                                                                              }
{ Effect is a single component of the objective function.                      }
{ The SIENA manual indicates how its specification is used.                    }
{                                                                              }

type
  TRealFunction = function(const dg: TValuedDigraph;
                              const k,par: integer): Real of object;
type
  TRealFunctioni = function(const dg: TValuedDigraph;
                              const i,k,par: integer): Real of object;
type
  TRealFunctionij = function(const dg: TValuedDigraph;
                              const i,j,k,par: integer): Real of object;
type
  TRealFunctionijt = function(const dg: TValuedDigraph;
                              const i,j,k: integer; const t: shortint;
                              const par: integer): Real of object;

type
  TRealFunctionijr = function(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real of object;

(* The addition "of object" is required to pass use variables
   in assignment statements and as parameters. *)


type TEffect = class
  protected
    FEName, FFName: ShortString;  (* Effect name, Function Name *)
    FIndex: integer;           (* Index identifying the alpha parameter. *)
    FKind: byte;                (* Kind of effect:
                                   ergmcase:
                                   1 = symmetric covariate;
AA change                          2 = other covariate (not kind 1 or 3);
AA change                          3 = actor covariate which is constant within periods
AA change                              or within connected components;
                                   4 = interaction effect;
                                   0 = anything else;
                                   non ergmcase:
                                   1 = dyadic covariate or actor-variable based similarity,
                                       or symmetric data and some covariate;
                                   2 = alter-effect actor variable;
                                   3 = ego-effect actor variable;
                                   4 = interaction effect;
                                   0 = anything else *)
    FNextMod_f: TEffect;            (* For the model sequence used in TEffects *)
    FNextMod: TEffect;            (* For the model sequence used in TEffects *)
    FNextStat_f: TEffect;           (* For the statistics sequence used in TEffects *)
    FNextStat_g: TEffect;           (* For the statistics sequence used in TEffects *)
    ContributionWeight:  TConfigWeight0;
    HasContribution:  TConfigContr0;
    HasAnyContribution: boolean;
    ConfigWeight:  TConfigWeight1;
    (* Fixed parameters defining the effect:
    ContributionWeight defines contribution to objective function;
    HasContribution = (ContributionWeight <> 0.0);
    ConfigWeight defines contribution to statistic;
    For the observations = 1 (ERGM) case,
    only ContributionWeight and f1ijergm play a role.    *)
    alpha_f, alpha: Real;
    (* parameters used to weight f1ijc;
       DefineTotalContributions sets alpha := alpha_f + alpha_g *)
    hasweight_f, hasweight_g: Boolean;
    (* hasweight_f = (effect_f = true), hasweight_g = (effect_g = true) *)
    FParameter: Integer;       (* Changeable parameter *)
    FRParameter: real;       (* Function of Fparameter *)
    FparMin, FParMax: integer; (* Bounds for FParameter *)
    f1ijc: TRealFunctionijt; (* Contribution of tie (i,j)
                               to objective function of actor i *)
    f1ij: TRealFunctionij;  (* Contribution of tie (i,j) to statistic *)
    f1ijergm: TRealFunctionijr;  (* Contribution of tie (i,j) to change statistic for observations = 1 *)
    f1i: TRealFunctioni;      (* Overall contribution to statistic;
                                for observations = 1 this is the statistic for the empty graph *)
                              (* For effects including a non-nil f1i,
                                 there is no endowment effect possible in MoM. *)
    FIncludeEndowment: boolean; (* Inclusion of endowment effect in MoM. *)
    FDyadic: boolean;       (* The effect depends only on the dyad,
                               and can be used in the Reciprocity Model. *)
    function f1(const dg: TValuedDigraph; const k,par: integer): real;
  public
    property index: integer read FIndex;
    property NextMod_f: Teffect read FNextMod_f;
    property NextMod: Teffect read FNextMod;
    property NextStat_f: Teffect read FNextStat_f;
    property NextStat_g: Teffect read FNextStat_g;
    property Parameter: Integer read FParameter;
    property RParameter: real read FRParameter;
    property IncludeEndowment: boolean read FIncludeEndowment;
    property Dyadic: boolean read FDyadic;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const ConfigW:  TConfigWeight1;
                       const p, p0, p1: integer);  overload;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const ConfigW:  TConfigWeight1;
                       const p, p0, p1: integer;
                       fi: TRealFunctioni); overload;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fij: TRealFunctionij; dummy1, dummy2:boolean); overload;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fij: TRealFunctionij;
                       fi: TRealFunctioni; const DyadEf: boolean); overload;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fijr: TRealFunctionijr;
                       dummy3,dummy4,dummy5,dummy6:boolean); overload;
    constructor Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer;
                       fi: TRealFunctioni; fijr: TRealFunctionijr); overload;
(* The non-mentioned arguments in these Create constructors
   are set to 0 or nil.
   If desired, more such overloads could be defined.
   The AddEffect overloads then must be added correspondingly. *)
    function EName: shortstring;
    function FName: shortstring;
    function HasNextMod_f: Boolean;
    function HasNextMod: Boolean;
    function HasNextStat_f: Boolean;
    function HasNextStat_g: Boolean;
    function SetParameter(const p: integer): boolean;
(* AA start  *)
    function OneChangeStatistic(const i,j: integer;
         const t,tr: shortint; const dg: TValuedDigraph): real;
(* Calculates the change statistic for this effect for toggling tie variable (i,j);
   t is the current value of dg(i,j), tr is the current value of dg(j,i).
   Used for ERGM, i.e., observations = 1.
   Generate indicates that this is called for generating a MH step,
   (not generate) for calculating statistics.
   This is modeled after ChangeStatistics. It is used only in interactions.
   *)
(* AA end *)
  end; (* Class TEffect *)

{                                                                              }
{ Array of effects for Valued directed graph: TEffects                         }
{                                                                              }
{ Effects is an array of effects, which defines the whole objective function.  }
{ This is the main class of Digaph visibly used in other units in Siena.       }
{                                                                              }


type TEffects = class
  protected
    Effect: array of TEffect;
    FfirstMod_f: TEffect;
    FfirstMod: TEffect;
    FfirstStat_f: TEffect;
    FfirstStat_g: TEffect;
    FNumberStats_f: integer;
    FNumberStats_g: integer;
    FNum_f: integer;
    TotalContributionWeight_f: TConfigWeight1;
    DyadicContributionWeight_f: array[1..2] of real;
    TotalContributionNonZero_f: TConfigContr1;
    TotalConfigNonZero_f: TConfigContr1;
    AnyContributionNonZero_f: TConfigContr1;
    TotalContributionWeight: TConfigWeight0;
    DyadicContributionWeight: array[0..2] of real;
    TotalContributionNonZero: TConfigContr0;
    AnyContributionNonZero: TConfigContr0;
    TotalConfigNonZero_g: TConfigContr1;
    SkipNumber_g: array of integer;

(* The effects have numbers defined by the usual order in the array;
   and a sequence First - ...Next... used in the model specification,
   and following from the .MO file. *)
  public
    property FirstMod_f: TEffect read FfirstMod_f;  (* The first element of the sequence
                                             defined by TEffects.NextMod_f *)
    property FirstMod: TEffect read FfirstMod;  (* The first element of the sequence
                                             defined by TEffects.NextMod_g *)
    property FirstStat_f: TEffect read FfirstStat_f;  (* The first element of the sequence
                                             defined by TEffects.NextStat_f *)
    property FirstStat_g: TEffect read FfirstStat_g;  (* The first element of the sequence
                                             defined by TEffects.NextStat_g *)
    property num_f: integer read FNum_f; (* Number of activated evaluation effects *)
    property NumberStats_f: integer read FNumberStats_f;
    property NumberStats_g: integer read FNumberStats_g;
    constructor Create;
    destructor Destroy;  override;
    function Param(const i: integer): integer;
    function SetParam(const i: integer; var p: integer): boolean;
(* Param reports and SetParam changes the value of Effect[i-1].FParameter;
   SetParam reports whether p is a permitted value. *)
    function Endowed(const i: integer): Boolean;
    function EndowedIndex(const i: integer): Boolean;
    function NameIndex(const i: integer): shortstring;
    function Kind(const i: integer): byte;
    procedure AddEffect1(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const ConfigW:  TConfigWeight1;
                        const p, p0, p1: integer);
    procedure AddEffect2(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const ConfigW:  TConfigWeight1;
                        const p, p0, p1: integer;
                        fi: TRealFunctioni);
    procedure AddEffect3(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const p, p0, p1: integer; fijc: TRealFunctionijt;
                        fij: TRealFunctionij;
                        fi: TRealFunctioni);
    procedure AddEffect4(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fij: TRealFunctionij);
    procedure AddEffect5(const EN, FN: ShortString;
                        const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fij: TRealFunctionij; const DyadEf: boolean);
    procedure AddEffect6(const EN, FN: ShortString;
                       const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fi: TRealFunctioni);
    procedure AddEffect7(const EN, FN: ShortString;
                        const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fij: TRealFunctionij; const DyadEf: boolean; const KindEf: byte);
    procedure AddERGMEffect1(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const p, p0, p1: integer; fijr: TRealFunctionijr;
                        const KindEf: byte);
    procedure AddERGMEffect2(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const p, p0, p1: integer;
                        fi: TRealFunctioni; fijr: TRealFunctionijr;
                        const KindEf: byte);
(* The non-mentioned arguments in these AddEffect procedures
   are set to 0 or nil.
   If desired, more such overloads could be defined.
   The TEffect.Create Constructor overloads then must be added correspondingly. *)
    procedure SpecifyInteractionEffect(const i: integer; var pa, code: integer;
                                       var ert: shortstring);
 (* Defines effect i as an interaction effect defined by parameter value pa.
    code is an extra argument for interaction with other units. *)
    procedure AppendName(const i: integer; const st: shortstring);
 (* Appends st to Effects[i-1].FEName and FFName *)
    procedure DeleteName(const i: integer);
 (* Deletes Effects[i-1].FEName and FFName *)

(* The following procedures are for defining TEffects as follows from
   the model specification in S_Dat / S_Base. *)
    procedure DefineModSequences(var outf: TFileStream;
                                 const indf, indg: array of boolean;
                                 const IncludeNonendowed: boolean);
    procedure DefineStat_fSequence(const indicator: array of boolean);
    procedure DefineStat_gSequence(const indicator: array of boolean;
                                 const IncludeNonendowed: boolean);
    procedure DefineTotalContributions(const a_f, a_g: array of real);
(* This procedure calculates intermediate results for efficient calculations.
   MUST always be called after setting/changing alpa_f;
   therefore this is called in SetAlpha.
   The procedures calculates
   TotalContributionWeight_f[i] = sum_Ef alpa_f[Ef.index] Ef.ContributionWeight[i]
   and defines the alpha fields of the Effects in the sequence. *)
    function NumberEffects: integer;
    function EffectName(const i: integer): shortstring;
    function FunctionName(const i: integer): shortstring;
    procedure WriteMod_fSequenceParameters(const a: array of real;
                            var outf: textfile);
    procedure WriteModSequenceParameters(const a: array of real;
                            var outf: textfile);
    procedure WriteStatSequenceIndices(var outf: textfile);

    function  Interactiont_a
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
    function  Interaction_a
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
    function Interactiontot_a(const dg: TValuedDigraph;
                              const i,k,par: integer): Real;

    function  Interactiont_b
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
    function  Interaction_b
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;

    function Interaction_ergm(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction_ergm11(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction_ergm10(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction_ergm01(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction_ergm00(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function  Interaction3t_a
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
    function  Interaction3_a
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
    function Interaction3tot_a(const dg: TValuedDigraph;
                              const i,k,par: integer): Real;

    function  Interaction3t_b
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
    function  Interaction3_b
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
    function Interactiont_rec_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
    function Interaction_rec_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;

    function Interaction3ergm_0(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction3ergm_1(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction3ergm_2(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
    function Interaction3ergm_3(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;

(* The following procedures are used intensively in S_Base and S_ML
   for calculating objective functions, statistics, score functions, etc. *)
    procedure SetChasZero(const ch_first, nnb: integer; var chas: rrrarray); // added 3.17q
    function Contrib_n(const dg: TValuedDigraph; const i,j,k: integer;
                       const incoming, DyadicOnly: boolean): real;
    function Contrib_n_alpha(const dg: TValuedDigraph; const i,j,k: integer;
                    const alfa: array of real; const incoming: boolean;
                    const calcchangestat: boolean; var chas: rrrarray): real;
(* Contribution to evaluation function plus endowment function of actor i
   of changing tie variable (i,j) from 1 to 0, in period k.
   Used in SubtractTies. *)
    function Contrib_fn(const dg: TValuedDigraph; const i,j,k: integer;
                    const incoming, DyadicOnly: boolean): real;
    function Contrib_fn_alpha(const dg: TValuedDigraph; const i,j,k: integer;
                    const alfa: array of real; const incoming: boolean;
                    const calcchangestat: boolean; var chas: rrrarray): real;
(* Contribution to evaluation function of actor i
   of changing tie variable (i,j) from 0 to 1, in period k.
   Used in AddNoTies. *)
    procedure AddNoTies(const i,k: integer;
         var p: array of real; perm: array of boolean;
                          const dg: TValuedDigraph); overload;
    procedure AddNoTies(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const dg: TValuedDigraph; const calcchangestat: boolean;
         var chas: rrrarray); overload;
(* Adds Contrib_fn to p[j] over all j having no tie from i to j
                                             and satisfying perm[j] *)
    procedure SubtractTies(const i,k: integer;
         var p: array of real; perm: array of boolean;
                              const dg: TValuedDigraph); overload;
    procedure SubtractTies(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const dg: TValuedDigraph; const calcchangestat: boolean;
         var chas: rrrarray); overload;
(* Subtracts Contrib_n from p[j] over all j
   having a tie from i to j and satisfying perm[j]                         *)
    procedure CalcFunctions_f(var fun: array of real; var ordeg: array of integer;
               const k: integer;
               const dg: TValuedDigraph; const COD: Boolean);
    procedure CalcFunctions_g(var fun: array of real; const k: integer;
               const dg: TValuedDigraph);
(* Calculation of the statistics defined by the effects:
   CalcFunctions_f for the evaluation function,
   Calcfunctions_g for the endowment function. *)
    procedure ChangeStatistics(const i,j: integer; var s: pvec;
                              const t,tr: shortint; const generate: boolean;
                              const dg: TValuedDigraph);
(* Adds the change statistics (all effects h1) for toggling tie variable (i,j)
   to array s[j,h+h1];
   t is the current value of dg(i,j), tr is the current value of dg(j,i).
   Used for ERGM, i.e., observations = 1.
   For the symmetric case, this is the change when both (i,j) and (j,i)
   are toggled.
   For the a(nti)symmetric case, nothing special is done,
   so changestatistics(i,j) as well as (j,i) has to be called
   when toggling in an antisymmetric network.
   *)
   function ChangeLogOdds(const s: pvec): real;
(* The change log odds for a vector s of change statistics.
   Used for ERGM, i.e., observations = 1.  *)
    procedure CalcStatistics(var s: pvec; const symm: boolean;
                var od: array of integer; const generate: boolean;
                                  const dg: TValuedDigraph);
(* Calculate the statistics s by successively adding
   each of the ties in dg to an auxiliary digraph dg0.
   Used for ERGM, i.e., observations = 1.
   symm = symmetric network.
   Generate indicates that this is called for generating a MH step,
   (not generate) for calculating statistics. *)
    procedure UtilityComponents(const i,k,h,n: integer;
         var s: rrarray; perm: array of boolean;
         const dg: TValuedDigraph);
(* Stores the utility function contributions (h1) for changing tie variable (i,j)
   in array s[j,h+h1] , for all j with perm[j]; k is the period;
   first evaluation function, then endowment function;
   for ML estimation. *)
  end; (* Class TEffects *)

// is the destruction of TEffects OK?

procedure ChangeValues(const a,b: array of shortint; const k: byte;
                       var dg1: TValuedDigraph; var dg2: TDigraph);
(* For all (i,j), if dg1(i,j)=a[h] for some h in [0..k],
   changes dg1(i,j) into b[h] and sets dg2(i,j) to 1. *)


procedure YesNoNo(var dg: TValuedDigraph;
                  const d1: TValuedDigraph; const d2, d3: TDigraph);
(* Creates dg, with elements defined as
    dg(i,j) = d1(i,j)(1-d2(i,j))(1-d3(i,j)) *)

procedure YesNo(var dg: TDigraph; const d1, d2: TDigraph);  // added 3.17s
(* Fills dg with elements defined as
    dg(i,j) = d1(i,j)(1-d2(i,j)) *)

procedure YesYes(var dg: TValuedDigraph; const d1, d2: TValuedDigraph);
(* Fills dg with elements defined as
    dg(i,j) = (d1(i,j) <> 0) and (d2(i,j) <> 0) *)

{procedure ReduceYesNo(var dg1: TValuedDigraph; var dg2: TDigraph);
(* Changes existing dg1 to dg1(i,j) = dg1(i,j)(1-dg2(i,j)). *)
}

procedure DifNoNo(var dgdif: TValuedDigraph;
                  const da,db: TValuedDigraph; const d2,d3,d4,d5: TDigraph);
(* Creates dg, with elements defined as
    dg(i,j) = abso(da(i,j)-db(i,j))(1-d2(i,j))(1-d3(i,j)) *)

procedure CountDifNoNo(var s,sdown,sup: integer;
                  const da,db: TValuedDigraph; const d2, d3, d4, d5: TDigraph);
// d2,d3: missingness; CS: d4,d5: structural fixation
(* Calculates
    s = sum(i,j) abso(da(i,j)-db(i,j))(1-d2(i,j))(1-d3(i,j))
    st = sum(i,j) (1-d2(i,j))(1-d3(i,j))
    sup = sum(i,j) I(da(i,j)<db(i,j))(1-d2(i,j))(1-d3(i,j))
    sdown = sum(i,j) I(da(i,j)>db(i,j))(1-d2(i,j))(1-d3(i,j)) *)
    
procedure SplitTwo(var dg, dgs: TValuedDigraph; const d1, d2: TDigraph);
(* Calculates dgs
   with dgs(i,j) = max(dg(i,j), min(d1(i,j), d2(i,j))),
   and dg := dg - dgs. *)

procedure CopyMinTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Calculates dgs(i,j) = min(d1(i,j), d2(i,j)). *)

procedure AddMinTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Replaces dgs(i,j) for which d1(i,j) <> 0 and d2(i,j)=1, by d1(i,j). *)

procedure ReplaceTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Replaces dgs(i,j) for which d2(i,j)=1, by d1(i,j). *)

procedure MaxTwo(var dg, dgs: TValuedDigraph);
(* calculates dg := max(dg, dgs) *)


procedure OldLost(var dgoldlost: TValuedDigraph;
                  const da,db: TValuedDigraph; const d2: TDigraph);
(* da is old network with 0 values if missing;
   db is new network with missings d2 *)
(* OldLost Creates dg, with elements defined as
   da(i,j) if (db(i,j) >= da(i,j)) or d2(i,j);
     i.e., no evidence that tie (i,j) was lost;
   -da(i,j) if (db(i,j) < da(i,j)) and (not d2(i,j));
     i.e., evidence that tie (i,j) was lost; *)

function OrCount(const d1, d2: TValuedDigraph): integer;
(* Counts the number of (i,j) with d1(i,j) <> 0 or d2(i,j) <> 0 *)

(* AA start *)
function AndCount(const d1, d2: TValuedDigraph): integer;
(* Counts the number of (i,j) with d1(i,j) <> 0 and d2(i,j) <> 0 *)


procedure ValuesCount(const d1: TValuedDigraph; const d2: TDigraph;
                       const a1,a2: shortint; var a: array of integer);
(* a[h] for h = 0 to a2-a1 will be the number of (i,j)
   with  d1[i,j] = a1+h and d2(i,j) = 1. *)
(* AA end *)

function SameVar(const i,j,h: integer): boolean;

procedure Pointvar(var z: iiarray);



//function AbsDif(const d1, d2: TValuedDigraph): integer;
//(* Counts the sum of abs(d1(i,j) - d2(i,j)) *)

//procedure AbsDifExtra(var adif, ndif: integer;
//                      const d1, d2: TValuedDigraph; const da, db: TDigraph);
(* ndif is the number of (i,j) with not (da(i,j) <> 0 or db(i,j) <> 0);
   adif is the sum of abs(d1(i,j) - d2(i,j)) over all these (i,j). *)
//xxxxx dit moet nog geimplemteerd



var ergmcase: boolean;
    param_kinstars, param_koutstars, param_ktri, param_ktwop: integer;
    index_kinstars, index_koutstars, index_ktri, index_ktwop: integer;
    param1_clique, param2_clique, param3_clique, param4_clique,
      param1_rclique, param2_rclique: integer;
    param1_clique_s, param2_clique_s, param3_clique_s, param4_clique_s,
      param1_rclique_s, param2_rclique_s: integer;
    index_cliq1,  index_cliq2,  index_cliq3, index_cliq4,
      index_rcliq1, index_rcliq2,
      pclique_min, pclique_max, pclique_min_s, pclique_max_s: integer;
    extraclique, extrared1_clique: array of integer;
(* The parameters for the alternating weighted effects, and for the clique effects,
   used for communication with other units.
   They get values in S_Effects. *)

implementation

uses slib;

var nnn: integer;
  checkfile: textfile;
  CF_open: boolean;
  stst: shortstring;
  nelem: integer; (* number of tdoublylinkeddouebleints allocated *)
  firstfreeelem: integer; (* index of first non-used element *)
  elem: array of TDoublyLinkedDoubleInt; (* elements to be used as allocated memory *)
  nextfree: array of integer; (* pointers among used elements *)
  freeelem: array of boolean;
  varvar: ^iiarray;

function SameVar(const i,j,h: integer): boolean;
begin
  result := (varvar^[i,h] = varvar^[j,h]);
end;

procedure Pointvar(var z: iiarray);
begin
  varvar := @z;
end;

// abcd

procedure AllocateElem(k: integer);
var i: integer;
begin
//writeln(checkfile,'Allocate ',k:6,' to ',nelem+k:6);

  if (k <= 0) then exit;
  nelem := nelem+k;
  setlength(elem,nelem+1);
//  setlength(freeelem,nelem+1);
  setlength(nextfree,nelem+1);
  for i:=nelem-k+1 to nelem do
    begin
    elem[i] := TDoublyLinkedDoubleInt.Create(0,0);
    elem[i].FElemIndex := i;
//    freeelem[i] := true;
    nextfree[i] := succ(i);
    end;
  nextfree[nelem] := 0;
  if (firstfreeelem = 0) then firstfreeelem := nelem-k+1;
  nextfree[nelem-k] := nelem-k+1;

//abcdwriteln(checkfile,'alocated firstfree = ',firstfreeelem);
end; (* AllocateElem *)

{                                                                              }
{ TDoublyLinkedDoubleInt                                                            }
{                                                                              }

function TDoublyLinkedDoubleInt.HasNext: Boolean;
begin
  Result := Assigned(Next);
end;


function TDoublyLinkedDoubleInt.Count: Integer;
var N : TDoublyLinkedDoubleInt;
begin
  Result := 1;
  N := FNext;
  While Assigned(N) do
    begin
      Inc(Result);
      N := N.Next;
    end;
end;

function TDoublyLinkedDoubleInt.HasPrev: Boolean;
begin
  Result := Assigned(FPrev);
end;

constructor TDoublyLinkedDoubleInt.Create(const V1: Integer; const V2: Shortint);
begin
  inherited Create;
  Value1 := V1;
  Value2 := V2;
end;


function TDoublyLinkedDoubleInt.UseMem(const V1: Integer; const V2: Shortint):
         TDoublyLinkedDoubleInt;
var Item: TDoublyLinkedDoubleInt;
begin

{
item :=  TDoublyLinkedDoubleInt.create(V1,V2);
result := item;
exit;
}
// abcd

  if (firstfreeelem <= 0) then
    begin
    AllocateElem(1000);
    end
   else
    begin

    end;

//  if not freeelem[firstfreeelem] then writeln;  // rough check abcd
//abcd e writeln(checkfile,'used firstfree = ',firstfreeelem);

  item := elem[firstfreeelem];
  result := item;

//  freeelem[firstfreeelem] := false;

  firstfreeelem := nextfree[firstfreeelem];



  item.Value1 := V1;
  item.Value2 := V2;

end;


(*procedure TDoublyLinkedDoubleInt.Delete;
begin
  Remove;
 {
    FreeAndNil(Self);  // TS changed, was only Free previously
    exit;
 }

//abcd
//  freeelem[FElemIndex] := true;
  nextfree[FElemIndex] := firstfreeelem;
  firstfreeelem := FElemIndex;

writeln(checkfile,'liberated firstfree = ',firstfreeelem);


//abcd  FreeAndNil(Self);  // TS changed, was only Free previously
end;
*)

procedure TDoublyLinkedDoubleInt.Remove;
begin
  if Assigned(Next) then
    Next.Prev := FPrev;
  if Assigned(Prev) then
    Prev.Next := FNext;

  LibMem;


//  nextfree[FElemIndex] := firstfreeelem;
//  firstfreeelem := FElemIndex;

//abcd writeln(checkfile,'liberated firstfree = ',firstfreeelem);

end;

procedure TDoublyLinkedDoubleInt.LibMem;
begin

  nextfree[FElemIndex] := firstfreeelem;
  firstfreeelem := FElemIndex;

//abcd writeln(checkfile,'liberated mem firstfree = ',firstfreeelem);

end;


procedure TDoublyLinkedDoubleInt.InsertAfter(const V1: Integer; const V2: Shortint);
var Item: TDoublyLinkedDoubleInt;
begin
//abcd  Item := TDoublyLinkedDoubleInt.Create(V1, V2);
  Item := UseMem(V1,V2);
  Item.Next := FNext;
  Item.Prev := self;
  if Assigned(FNext) then FNext.Prev := Item;  (* Added by TS, this was not included in the original cLinkedLists *)
  FNext := Item;
end;

procedure TDoublyLinkedDoubleInt.InsertBefore(const V1: Integer; const V2: Shortint);
var Item: TDoublyLinkedDoubleInt;
begin
  Item := UseMem(V1,V2);
//abcd  Item := TDoublyLinkedDoubleInt.Create(V1, V2);
  Item.Next := self;
  Item.Prev := FPrev;
  if Assigned(FPrev) then FPrev.Next := Item;  (* Added by TS, this was not included in the original cLinkedLists *)
  FPrev := Item;
end;


//destructor TDoublyLinkedDoubleInt.DestroyList;
procedure TDoublyLinkedDoubleInt.DestroyList;
var N : TDoublyLinkedDoubleInt;
begin
  While Assigned(FNext) do
    begin
      N := FNext;
      FNext := N.Next;
      N.LibMem;
//bcde N.Free;
    end;
  inherited Destroy;
end;


procedure TDoublyLinkedIntList.CopyTo(it: TDoublyLinkedIntList);
var its: TDoublyLinkedDoubleInt;
begin
// @@@ does this work OK?
 it.FFirst.DestroyList;
 it.Insertfront(0,0);
 its := FFirst;
 while its.HasNext do
   begin
    its := its.FNext;
    it.Append(its.Value1,its.Value2);
   end;
end;

procedure TDoublyLinkedIntList.WriteVector(var outf: textfile; const mode: byte);
var it: TDoublyLinkedDoubleInt;
  kk: integer;
begin
  kk := 3;
  it := self.first;
  while it.HasNext do
    begin
    it := it.Next;
    case mode of
    0: write(outf,' ',it.value1:kk);
    1: write(outf,' (',it.value1:kk,', ',it.Value2:2,')');
    end;
    end;
  writeln(outf);
end;

procedure TDoublyLinkedIntList.StreamVector(var outf:  TFileStream; const mode: byte);
var it: TDoublyLinkedDoubleInt;
  kk: integer;
begin
  kk := 0;
  it := self.first;
  while it.HasNext and (kk <= 100) do
    begin
    inc(kk);
    it := it.Next;
    case mode of
    0: wwrite(outf,' '+inttostr(it.value1));
    1: wwrite(outf,' '+inttostr(it.value1)+', '+inttostr(it.Value2));
    end;
    end;
  if (kk >= 100) then wwriteln(outf,'!!!???!!!');    
  wwriteln(outf,' ');
end;



{                                                                              }
{ TDoublyLinkedIntList                                                         }
{                                                                              }
Constructor TDoublyLinkedIntList.Create;  (* added by TS *)
begin
  inherited Create;
  FCount := 0;
end;

Destructor TDoublyLinkedIntList.Destroy;
begin
  DeleteList;
  inherited Destroy;
end;

function TDoublyLinkedIntList.IsEmpty: Boolean;
begin
  Result := not Assigned(FFirst);
end;


procedure TDoublyLinkedIntList.Append(const V1: Integer; const V2: Shortint);
var Item: TDoublyLinkedDoubleInt;
begin
  if not Assigned(FLast) then
    begin
  Item := Item.UseMem(V1,V2);
//abcd      Item := TDoublyLinkedDoubleInt.Create(V1, V2);
      FFirst := Item;
      FLast := Item;
      Item.Prev := nil;
      Item.Next := nil;
    end else
    begin
      if ffirst.HasNext then
      stst :=inttostr(ffirst.FNext.Value1)+', '+inttostr(flast.Value1)+', '+inttostr(v1)
      else stst := '';      
      FLast.InsertAfter(V1, V2);
      FLast := Flast.Next;
      stst := stst+'='+inttostr(flast.Value1);
    end;
  Inc(FCount);
end;

procedure TDoublyLinkedIntList.InsertFront(const V1: Integer; const V2: Shortint);
var Item: TDoublyLinkedDoubleInt;
begin
  if not Assigned(FFirst) then
    begin
  Item := Item.UseMem(V1,V2);
if not assigned(Item) then writeln;
//abcd      Item := TDoublyLinkedDoubleInt.Create(V1, V2);
      FFirst := Item;
      FLast := Item;
      Item.Prev := nil;
      Item.Next := nil;
      end else
    begin
      FFirst.InsertBefore(V1, V2);
      FFirst := FFirst.Prev;
    end;
  Inc(FCount);
end;

procedure TDoublyLinkedIntList.Remove(const Item: TDoublyLinkedDoubleInt);
begin
  if not Assigned(Item) then
    exit;
  if FFirst = Item then
    FFirst := Item.Next;
  if FLast = Item then
    FLast := Item.Prev;


  if Assigned(Item.Next) then
    Item.Next.Prev := Item.FPrev;
  if Assigned(Item.Prev) then
    Item.Prev.Next := Item.FNext;

  Item.LibMem;

//abcd writeln(checkfile,'list liberated firstfree = ',firstfreeelem);

//  Item.Remove;
  Dec(FCount);
end;

function TDoublyLinkedIntList.RemoveFirst: TDoublyLinkedDoubleInt;
begin
  Result := FFirst;
  if not Assigned(Result) then
    exit;
  if Result = FLast then
    begin
      FFirst := nil;
      FLast := nil;
    end else
    begin
      Result.Remove;
      FFirst := Result.Next;
    end;
  Dec(FCount);
end;

function TDoublyLinkedIntList.RemoveLast: TDoublyLinkedDoubleInt;
begin
  Result := FLast;
  if not Assigned(Result) then
    exit;
  if Result = FFirst then
    begin
      Result.LibMem; // abcd
      Ffirst := nil;
      FLast := nil;
    end else
    begin
      Result.Remove;
      FLast := Result.Prev;
    end;
  Dec(FCount);
end;

procedure TDoublyLinkedIntList.Delete(const Item: TDoublyLinkedDoubleInt);
begin
//abcde
//Item.Delete;
  Remove(Item);
//abcd  Item.Free;
end;

procedure TDoublyLinkedIntList.DeleteFirst;
begin
//abcd  RemoveFirst.Free;
  RemoveFirst.LibMem;
end;

procedure TDoublyLinkedIntList.DeleteLast;
begin
//abcd   RemoveLast.Free;
  RemoveLast.LibMem;
end;

procedure TDoublyLinkedIntList.DeleteList;
begin
  if Assigned(FFirst) then begin
    FFirst.DestroyList;
    Ffirst.LibMem;
//abcd    Ffirst.Free;         (* Added by TS, this was not included in the original cLinkedLists *)
   end;
  FFirst := nil;
  FLast := nil;
end;

function TDoublyLinkedIntList.RandomElement(const version: Byte): integer;
var j,j0,od: integer;
    it:TDoublyLinkedDoubleInt;
begin
  result := 0;
  od := Count-1;
  if (od <= 0) then exit;
  j := 1 + randomd(version,od);
  it := First;
  j0 := 0;
  while (j0 < j) and it.HasNext do begin
    it := it.Next;
    inc(j0);
   end;
  result := it.Value1;
end; (* RandomElement *)



function TDoublyLinkedIntList.VectorValueUp(const i: integer): integer;
(* The value of element i of the ShortInt vector represented by vec. *)
var it: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it := Ffirst;
  while it.HasNext do begin
   it := it.Next;
   case sign(it.Value1 - i) of
    -1: ;
    0: begin Result := it.Value2; exit; end;
    1: exit;
    end; (* case *)
   end; (* while *)
end; (* VectorValueUp *)

function TDoublyLinkedIntList.VectorValueDown(const i: integer): integer;
(* The value of element i of the ShortInt vector represented by vec. *)
var it: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it := Flast;
  while it.HasPrev do begin
   case sign(it.Value1 - i) of
    -1: exit;
    0: begin Result := it.Value2; exit; end;
    1: ;
    end; (* case *)
   it := it.Prev;
   end; (* while *)
end; (* VectorValueDown *)

function InnerProduct(const vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the ShortInt vectors represented by vec1 and vec2. *)
var it1, it2: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it1 := vec1.First;
  it2 := vec2.First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: it1 := it1.next;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         Inc(Result,it1.Value2 * it2.Value2);
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
end; (* InnerProduct *)

function NegNumber(const vec: TDoublyLinkedIntList): integer;
(* The number of negative elements of the ShortInt vector represented by vec. *)
var it: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it := vec.First;
  while it.hasnext do
    begin
    it := it.next;
    if (it.Value2 < 0) then inc(Result);
    end;
end; (* NegNumber *)

function PosNumber(const vec: TDoublyLinkedIntList): integer;
(* The number of positive elements of the ShortInt vector represented by vec. *)
var it: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it := vec.First;
  while it.hasnext do
    begin
    it := it.next;
    if (it.Value2 > 0) then inc(Result);
    end;
end; (* PosNumber *)

function CheckInnerProduct(vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the ShortInt vectors represented by vec1 and vec2. *)
var i,i0: integer;
begin
  Result := 0;
  if vec1.First.HasNext then i0 := vec1.First.next.Value1 else exit;
  for i:= vec1.Last.Value1 downto i0 do Inc(Result,vec1.VectorValueUp(i)*vec2.VectorValueUp(i));
end; (* CheckInnerProduct *)

function IndicatorInnerProduct(const vec1,vec2: TDoublyLinkedIntList): integer;
(* The inner product of the <> indicators of the
   ShortInt vectors represented by vec1 and vec2. *)
var it1, it2: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it1 := vec1.First;
  it2 := vec2.First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: it1 := it1.next;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         Inc(Result);
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
end; (* IndicatorInnerProduct *)

function IndicatorQuadProduct(const vec1,vec2,vec3,vec4: TDoublyLinkedIntList): integer;
(* The sum of the products of the <> indicators of the
   ShortInt vectors represented by vec1, vec2,v ec3, vec4. *)
var it1, it2, it3, it4: TDoublyLinkedDoubleInt;
    j: integer;
begin
  Result := 0;
  it1 := vec1.First;
  it2 := vec2.First;
  it3 := vec3.First;
  it4 := vec4.First;
(* it1, it2 treated differently from it3, it4 just for convenience. *)
  if not (it3.HasNext and it4.HasNext) then exit;
  while (it1.hasnext and it2.hasnext) do
    begin
     case sign(it1.next.value1 - it2.next.value1) of
      -1: it1 := it1.next;
      0: begin
           it1 := it1.next;
           it2 := it2.next;
           j:= it1.value1;
           while ((it3.Value1 < j) and it3.HasNext) do it3 := it3.Next;
           while ((it4.Value1 < j) and it4.HasNext) do it4 := it4.Next;
           if ((it3.Value1 = j) and (it4.Value1 = j)) then Inc(Result);
(* Note that for it=it3 and it4, at this point,
   if it.value1 < j then this it.value1 is the highest existing it.value1;
   if it.value1 >= j then this it.value1 is the smallest existing it.value1
   larger than j-1 *)
         end; (* case 0 *)
      1: it2 := it2.next;
     end; (* case *)
     if ((it3.Value1 < j) or (it4.Value1 < j)) then exit;
   end;
end; (* IndicatorQuadProduct *)




function Subset(const vec1,vec2: TDoublyLinkedIntList): boolean;
(* indicates that vec1 is a subset of vec2 *)
var subs: boolean;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  subs := true;
  it1 := vec1.First;
  it2 := vec2.First;
  while (subs and it1.HasNext and it2.HasNext) do
    case sign(it1.Next.Value1 - it2.Next.Value1) of
    -1: subs := false; (* it1.next.value1 is in vec1 but not in vec2 *)
      0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          end;
       1: it2 := it2.Next;
      end; (* case *)
  if it1.HasNext then subs := false;
  subset := subs;
end;

function Equal(const vec1,vec2: TDoublyLinkedIntList): boolean;
(* indicates that vec1 is equal to vec2 *)
var equ: boolean;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  equ := true;
  it1 := vec1.First;
  it2 := vec2.First;
  while (equ and it1.HasNext and it2.HasNext) do
    case sign(it1.Next.Value1 - it2.Next.Value1) of
    -1: equ := false; (* it1.next.value1 is in vec1 but not in vec2 *)
      0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          if (it1.Value2 <> it2.Value2) then equ := false;
          end;
       1: equ := false;
      end; (* case *)
  if (it1.HasNext or it2.HasNext) then equ := false;
  equal := equ;
end;


function ArrayInnerProduct(const vec1: TDoublyLinkedIntList; x: array of integer;
                           var ercode: integer): integer;
(* The inner product of the ShortInt vector represented by vec1 with x;
   low(x) must be 0. *)
var it: TDoublyLinkedDoubleInt;
    n: integer;
begin
  Result := 0;
  ercode := 0;
  n := high(x);
  it := vec1.Ffirst;
  while it.HasNext do begin
   it := it.Next;
   if (it.Value1 <= 0) then
     begin
     ercode := -1;
     exit;
     end;
   if (it.Value1 > n) then
     begin
     ercode := 1;
     exit;
     end;
   Result := Result + it.Value2*x[it.Value1];
   end; (* while *)
end; (* ArrayInnerProduct *)


procedure ForDifferencesDo(const vec1,vec2: TDoublyLinkedIntList; procd: TProcedure);
(* Executes procd(i,vec1[i],vec2[i]) for all i where vec1[i] <> vec2[i],
   where i goes from small to large. *)
var it1, it2: TDoublyLinkedDoubleInt;
    v1,v2,n,nn: integer;
(* v1 is the current value of it1.value1, v2 of it2.value1;
   nn if there is no current value any more. *)
procedure continue1;
begin
  if (it1.HasNext) then
    begin
    it1 := it1.next;
    v1 := it1.Value1;
    end
   else v1 := nn;
end;
procedure continue2;
begin
  if (it2.HasNext) then
    begin
    it2 := it2.next;
    v2 := it2.Value1;
    end
   else v2 := nn;
end;
begin
  it1 := vec1.First;
  it2 := vec2.First;
  if (vec2.Last.Value1 > vec1.Last.Value1) then
   n := vec2.Last.Value1 else n := vec1.Last.Value1;
  nn := succ(n);
  continue1;
  continue2;
  while (v1 <= n) or (v2 <= n) do
   case sign(v1 - v2) of
    -1: begin
        procd(v1, it1.value2, 0);
        continue1;
        end;
    0:  begin
        if (it1.value2 <> it2.value2) then
          procd(v1, it1.value2, it2.value2);
        continue1;
        continue2;
        end; (* case 0 *)
    1:  begin
        procd(v2, 0, it2.value2);
        continue2;
        end;
   end; (* case *)
end;  (* ForDifferencesDo *)

{                                                                              }
{ TValuedDigraph                                                               }
{                                                                              }


constructor TValuedDigraph.Create(const n: Integer);
var i,j: integer;
begin
  if (n <= 1) then begin
    ErMes := 'call Create('+inttostr(n)+')';
    exit;
   end;
  inherited Create;
  DiagonalPermitted := false;
  ErMes := '';
  NumberNodes := n;
  SetLength(Inties,n+1);
  SetLength(Outties,n+1);
  SetLength(helplist,n+1);
  for i:=n downto 1 do begin
   Inties[i] := TDoublyLinkedIntList.Create;
   Inties[i].Insertfront(0,0);
   Outties[i] := TDoublyLinkedIntList.Create;
   Outties[i].Insertfront(0,0);
   helplist[i] := false;
  end;
  helplist[0] := false;
  if (n <= n_crit) then
    begin
    HasArray := true;
    SetLength(AdjMatrix,n+1,n+1);
    for i:=n downto 0 do for j:=n downto 0 do AdjMatrix[i,j] := 0;
    end
   else HasArray := false;
end; (* Create *)

destructor TValuedDigraph.Destroy;
var i: integer;
begin
  for i:=numbernodes downto 1 do begin
    Inties[i].Destroy;
    Outties[i].Destroy;
   end;
  inherited Destroy;
  if HasArray then SetLength(AdjMatrix,0,0);
  if CF_open then
    begin
    closefile(checkfile);
    CF_Open := false;
    end;
end;

destructor TValuedDigraph.Destroy(var outf: TFileStream; const s: string);
var i: integer;
var x: string;
begin
  if (ErMes > '') then begin
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'Warning: Unit Digraph gives the error message : '+eol;
    outf.Write(pchar(x)^,length(x));
    x := '<< '+ErMes+' >> for object '+s+'.'+eol;
    outf.Write(pchar(x)^,length(x));
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
   end;
//writeln(outf,'HasArray = ',HasArray);
  for i:=numbernodes downto 1 do begin
    Inties[i].Destroy;
    Outties[i].Destroy;
   end;
  inherited Destroy;
end;

function TValuedDigraph.SimpleInTieValue(const i,j: integer): ShortInt;
begin
(* Default : *)
  Result := 0;
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    ErMes := 'call SimpleInTieValue('+inttostr(i)+','+inttostr(j)+')';
    exit;
   end;
  if ((i = j) and (not DiagonalPermitted)) then exit;
  Result := InTies[j].VectorValueUp(i);
end; (* SimpleInTieValue *)

function TValuedDigraph.SimpleOutTieValue(const i,j: integer): ShortInt;
begin
(* Default : *)
  Result := 0;
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    ErMes := 'call SimpleOutTieValue('+inttostr(i)+','+inttostr(j)+')';
    exit;
   end;
  if ((i = j) and (not DiagonalPermitted)) then exit;
  Result := OutTies[i].VectorValueUp(j);
end; (* SimpleOutTieValue *)

function TValuedDigraph.IsCorrect: Boolean;
(* Checks consistency of data structure *)
var i,j,s,v1: integer;
    x1, x2, v2: ShortInt;
    it: TDoublyLinkedDoubleInt;
begin
  Result := true;
  if (Order <= 1) then begin
    Result := false;
    ErMes := 'Order = '+inttostr(Order);
    exit;
   end;
  for i:=1 to Order do begin
(* check redundant double representation *)
    s := 0;
    for j:=1 to Order do begin
      x1 := SimpleInTieValue(i,j);
      x2 := SimpleOutTieValue(i,j);
      if (x1 <> x2) then begin
        Result := false;
        ErMes := 'Tie('+inttostr(i)+','+inttostr(j)+') = '+inttostr(x1)+
                 '<>'+inttostr(x2);
        exit;
       end;
      if not (x1 = 0) then Inc(s);
     end; (* for j *)
(* check outdegrees *)
    if (s <> OutDegree(i)) then begin
      Result := false;
      ErMes := 'Outdegree('+inttostr(i)+') = '+inttostr(s)+
               '<>'+inttostr(OutDegree(i));
      exit;
     end;

(* check indegrees *)
    s := 0;
    for j:=1 to Order do begin
      x1 := SimpleInTieValue(j,i);
      if not (x1 = 0) then Inc(s);
     end; (* for j *)
    if (s <> InDegree(i)) then begin
      Result := false;
      ErMes := 'Indegree('+inttostr(i)+') = '+inttostr(s)+
               '<>'+inttostr(InDegree(i));
      exit;
     end;

(* Check values in linked lists *)
    for j:=1 to 2 do begin
      if (j <= 1) then it := OutTies[i].First else it := InTies[i].First;
      while it.HasNext do begin
        it := it.Next;
        v1 := it.Value1;
        v2 := it.Value2;
        if ((v1 <= 0) or (v1 > Order) or
            ((i = v1) and (not DiagonalPermitted)) or (v2 = 0))
         then begin
          Result := false;
          if (j=1) then ErMes := 'Out' else ErMes := 'In';
          ErMes := ErMes+'Ties('+inttostr(i)+') contains values '+inttostr(v1)+
               ', '+inttostr(v2);
          exit;
         end;
       end; (* while *)
      if (((j=1) and (it <> OutTies[i].Last)) or ((j=2) and (it <> InTies[i].Last)))
       then begin
          Result := false;
          if (j=1) then ErMes := 'Out' else ErMes := 'In';
          ErMes := 'Last element of '+ErMes+'Ties('+inttostr(i)+') is incorrect';
          exit;
         end;
     end; (* for j *)
   end; (* for i *)
end; (* IsCorrect *)

function TValuedDigraph.TieValue(const i,j: integer): ShortInt;
var it1, it2: TDoublyLinkedDoubleInt;
begin
(* Default : *)
  Result := 0;
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    ErMes := 'call TieValue('+inttostr(i)+','+inttostr(j)+')';
    exit;
   end;
  if ((i = j) and (not DiagonalPermitted)) then exit;
  if HasArray then
    begin
    Result := AdjMatrix[i,j];
    exit;
    end;
(* Search for deviations from default : *)
  it1 := InTies[j].First;
  it2 := OutTies[i].First;
(* Compare degrees : *)
  if ((it1.Value1 <= 0) or (it2.Value1 <= 0)) then exit
   else (* both degrees positive *) begin
(* The search for element (i,j) is divided in 4 cases to increase speed *)
     if (it1.Value1 <= it2.Value1) then begin
(* Indegree(j) <= Outdegree(i) *)
(* Search InTies[j] *)
       if (2*i <= Order) then Result := InTies[j].VectorValueUp(i)
        else Result := InTies[j].VectorValueDown(i);
    end else (* Indegree(j) > Outdegree(i)*) begin
(* Search OutTies[i] *)
       if (2*j <= Order) then Result := OutTies[i].VectorValueUp(j)
        else Result := OutTies[i].VectorValueDown(j);
    end; (* if Indegree(j) > Outdegree(i) *)
   end; (* if both degrees positive *)
end; (* TieValue *)

function TValuedDigraph.TieValueDown(const i,j: integer): ShortInt;
var it1, it2: TDoublyLinkedDoubleInt;
begin
(* Default : *)
  Result := 0;
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    ErMes := 'call TieValueDown('+inttostr(i)+','+inttostr(j)+')';
    exit;
   end;
  if ((i = j) and (not DiagonalPermitted)) then exit;
(* Search for deviations from default : *)
  it1 := InTies[j].First;
  it2 := OutTies[i].First;
(* Compare degrees : *)
  if ((it1.Value1 <= 0) or (it2.Value1 <= 0)) then exit
   else (* both degrees positive *) begin
(* The search for element (i,j) is divided in 4 cases to increase speed *)
     if (it1.Value1 <= it2.Value1) then
(* Indegree(j) <= Outdegree(i) *)
(* Search InTies[j] *)
       Result := InTies[j].VectorValueDown(i)
    else (* Indegree(j) > Outdegree(i)*)
(* Search OutTies[i] *)
       Result := OutTies[i].VectorValueDown(j);
   end; (* if both degrees positive *)
end; (* TieValueDown *)

function TValuedDigraph.Indegree(const i: integer): Integer;
begin
  if ((i <= 0) or (i > Order)) then begin
    Result := 0;
    ErMes := 'call InDegree('+inttostr(i)+')';
   end else
  Result := InTies[i].First.Value1;
end;

function TValuedDigraph.Outdegree(const i: integer): Integer;
begin
  if ((i <= 0) or (i > Order)) then begin
    Result := 0;
    ErMes := 'call OutDegree('+inttostr(i)+')';
   end else
  Result := OutTies[i].First.Value1;
end;

function TValuedDigraph.PosInDegree(const i: integer): Integer;
(* Number of i for which TieValue(i,j) > 0 *)
begin
  Result := PosNumber(InTies[i]);
end;

function TValuedDigraph.NegInDegree(const i: integer): Integer;
(* Number of i for which TieValue(i,j) < 0 *)
begin
  Result := NegNumber(InTies[i]);
end;

function TValuedDigraph.PosOutDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) > 0 *)
begin
  Result := PosNumber(OutTies[i]);
end;

function TValuedDigraph.NegOutDegree(const i: integer): Integer;
(* Number of j for which TieValue(i,j) < 0 *)
begin
  Result := NegNumber(OutTies[i]);
end;



function TValuedDigraph.Recidegree(const i: integer): Integer;
begin
  if ((i <= 0) or (i > Order)) then begin
    Result := 0;
    ErMes := 'call ReciDegree('+inttostr(i)+')';
   end else
  Result := IndicatorInnerProduct(OutTies[i],InTies[i]);
end;

function TValuedDigraph.TotalTies: integer;
(* Number of (i,j) for which TieValue(i,j) <> 0 *)
var s,i: integer;
begin
  s := 0;
  for i:= order downto 1 do inc(s,OutTies[i].First.Value1);
  result := s;
end;

function TValuedDigraph.Mini;
var i,d: integer;
    it:TDoublyLinkedDoubleInt;
    s: ShortInt;
begin
(* calculate d = lowest indegree; s = lowest value2 *)
  d := Order-1;
  s := High(ShortInt);
  for i:=1 to Order do begin
    if (InDegree(i) < d) then d := InDegree(i);
    it := InTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      if (it.Value2 < s) then s := it.Value2;
     end;
   end; (* i *)
  if (d >= Order-1) then begin
    Result := s
   end else begin
    if (s < 0) then Result := s else Result := 0;
   end;
end; (* Mini *)

function TValuedDigraph.Maxi;
var i,d: integer;
    it:TDoublyLinkedDoubleInt;
    s: ShortInt;
begin
(* calculate d = lowest indegree; s = highest value2 *)
  d := Order-1;
  s := Low(ShortInt);
  for i:=1 to Order do begin
    if (InDegree(i) < d) then d := InDegree(i);
    it := InTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      if (it.Value2 > s) then s := it.Value2;
     end;
   end; (* i *)
  if (d >= Order-1) then begin
    Result := s
   end else begin
    if (s > 0) then Result := s else Result := 0;
   end;
end; (* Maxi *)

function TValuedDigraph.SymmetricGraph;
var i: integer;
   sy: boolean;
begin
  sy := true;
  i := order;
  while (sy and (i >= 1)) do
    begin
    sy := Equal(outties[i],inties[i]);
    dec(i);
    end;
  SymmetricGraph := sy;
end; (* SymmetricGraph *)

function TValuedDigraph.TwoPaths(const i,j: integer): Integer;
begin
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    Result := 0;
    ErMes := 'call TwoPaths('+inttostr(i)+','+inttostr(j)+')';
   end else
  Result := IndicatorInnerProduct(OutTies[i],InTies[j]);
end;

function TValuedDigraph.TwoPathsTrunc2(const i,j: integer): Integer;
var it1, it2: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  while (it1.hasnext and it2.hasnext and (result <= 1)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         inc(result);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
end; (* TwoPathsTrunc2 *)


function TValuedDigraph.NoTwoPath(const i,j,h: integer): Boolean;
(* Indicates that there is no twopath i => h' => j for any h' < h *)
var it1, it2: TDoublyLinkedDoubleInt;
  Found: Boolean;
  nextvalue1: integer;
begin
  Found := false;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  nextvalue1 := 0;
  while (it1.hasnext and it2.hasnext and (nextvalue1 < h) and (not Found)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
          nextvalue1 := it1.Value1;
        end;
    0: begin
         nextvalue1 := it1.next.Value1;
         if (nextvalue1 < h) then Found := true;
           (* else the while condition will be false,
              so the loop terminates anyway*)
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
         nextvalue1 := it2.Value1;
        end;
   end; (* case *)
  Result := not Found;
end; (* NoTwoPath *)

function TValuedDigraph.ExistsTwoPath(const i,j: integer): Boolean;
(* Indicates that there is at least one twopath i => h => j for some h  *)
var it1, it2: TDoublyLinkedDoubleInt;
  Found: Boolean;
  nextvalue1: integer;
begin
  Found := false;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  nextvalue1 := 0;
  while (it1.hasnext and it2.hasnext and (not Found)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
          nextvalue1 := it1.Value1;
        end;
    0: begin
         nextvalue1 := it1.next.Value1;
         Found := true;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
         nextvalue1 := it2.Value1;
        end;
   end; (* case *)
  Result := Found;
end; (* ExistsTwoPath *)



function TValuedDigraph.NoOrOneTwoPath(const i,j: integer; var tp: integer): Boolean;
(* Indicates that there is at most one twopath i => h' => j; tp = 0 or 1,
    depending on whether there is 0 or 1 found, if it is 0 or 1.  *)
var it1, it2: TDoublyLinkedDoubleInt;
  Found: integer;
begin
  Found := 0;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  while (it1.hasnext and it2.hasnext and (Found <= 1)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         inc(Found);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  tp := Found;
  Result := (Found <= 1);
end; (* NoOrOneTwoPath *)

function TValuedDigraph.NoOrOneTwoPathMax(const i,j,h: integer): Boolean;
(* Indicates that there is at most one twopath i => h' => j for any h' < h *)
var it1, it2: TDoublyLinkedDoubleInt;
  Found: integer;
  nextvalue1: integer;
begin
  Found := 0;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  nextvalue1 := 0;
  while (it1.hasnext and it2.hasnext and (nextvalue1 < h) and (Found <= 1)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
          nextvalue1 := it1.value1;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         nextvalue1 := it1.value1;
         if (nextvalue1 < h) then inc(Found);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
         nextvalue1 := it2.value1;
        end;
   end; (* case *)
  Result := (Found <= 1);
end; (* NoOrOneTwoPathMax *)

function TValuedDigraph.OneTwoPath(const i,j,h: integer): Boolean;
(* Indicates that there is exactly one twopath i => h' => j for any h' < h *)
var it1, it2: TDoublyLinkedDoubleInt;
  Found: integer;
  nextvalue1: integer;
begin
  Found := 0;
  it1 := OutTies[i].First;
  it2 := InTies[j].First;
  nextvalue1 := 0;
  while (it1.hasnext and it2.hasnext and (nextvalue1 < h) and (Found <= 1)) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
          nextvalue1 := it1.value1;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         nextvalue1 := it1.value1;
         if (nextvalue1 < h) then inc(Found);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
         nextvalue1 := it2.value1;
        end;
   end; (* case *)
  Result := (Found = 1);
end; (* OneTwoPath *)


function TValuedDigraph.OutTwoStars(const i,j: integer): Integer;
begin
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    Result := 0;
    ErMes := 'call OutTwoStars('+inttostr(i)+','+inttostr(j)+')';
   end else
  Result := IndicatorInnerProduct(InTies[i],InTies[j]);
end;

function TValuedDigraph.InTwoStars(const i,j: integer): Integer;
begin
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)) then begin
    Result := 0;
    ErMes := 'call InTwoStars('+inttostr(i)+','+inttostr(j)+')';
   end else
  Result := IndicatorInnerProduct(OutTies[i],OutTies[j]);
end;

procedure TValuedDigraph.ThirdCount(const i,j: integer; var n3, n4: Integer);
(* n3 (n4) is computed  as the number of third actors h,
   such that there are 3 (4) ties among the four possibilities
   (i,h), (h,i), (j,h), (h,j). *)
var it1, it2, it3, it4: TDoublyLinkedDoubleInt;
  h1, h2, h3, h4, minh, s: integer;
  b1, b2, b3, b4, c1, c2, c3, c4: boolean;
function count4(c1,c2,c3,c4: boolean): byte;
begin
 result := 0;
 if c1 then inc(result);
 if c2 then inc(result);
 if c3 then inc(result);
 if c4 then inc(result);
end;
begin
  n3 := 0;
  n4 := 0;
  it1 := OutTies[i].First;
  it2 := InTies[i].First;
  it3 := OutTies[j].First;
  it4 := InTies[j].First;
  c1 := it1.HasNext;
  c2 := it2.HasNext;
  c3 := it3.HasNext;
  c4 := it4.HasNext;
  while (count4(c1,c2,c3,c4) >= 3) do begin
    if c1 then h1 := it1.next.value1 else h1 := high(integer);
    if c2 then h2 := it2.next.value1 else h2 := high(integer);
    if c3 then h3 := it3.next.value1 else h3 := high(integer);
    if c4 then h4 := it4.next.value1 else h4 := high(integer);
(* compute minh as the smallest h, i.e., the smallest node
   having some tie with i or j *)
    minh := h1;
    if (h2 < minh) then minh := h2;
    if (h3 < minh) then minh := h3;
    if (h4 < minh) then minh := h4;
(* compute b1 - b4 indicating whether these edge lists include an edge
   involving minh, and s as the number of these edges.
   note that bi implies ci (i=1,2,3,4) *)
    if (h1 <= minh) then
      begin b1 := true; s := 1; end
      else begin b1 := false; s := 0; end;
    if (h2 <= minh) then begin b2 := true; inc(s); end
      else b2 := false;
    if (h3 <= minh) then begin b3 := true; inc(s); end
      else b3 := false;
    if (h4 <= minh) then begin b4 := true; inc(s); end
      else b4 := false;
    if (s >= 4) then inc(n4) else if (s >= 3) then inc(n3);
(* increase all smallest edge indicators *)
    if b1 then begin it1 := it1.next; c1 := it1.HasNext; end;
    if b2 then begin it2 := it2.next; c2 := it2.HasNext; end;
    if b3 then begin it3 := it3.next; c3 := it3.HasNext; end;
    if b4 then begin it4 := it4.next; c4 := it4.HasNext; end;
   end; (* while *)
end;  (* ThirdCount*)

function TValuedDigraph.Correlation(var z: array of real;
                                    var miz: array of boolean;
         var n0, nn: integer; var m0,s0,m1,s1,m2,s2,dd,cs,ds: real): real;
(* correlation between z[i] and z[j] for tied pairs (i,j),
   with miz[i]=miz[j] = false.
   m0 and s0 are regular mean and s.d. of z;
   m1, s1 and m2, s2 mean and s.d. of z in first/second in tied pairs;
   n0, nn numbers of individuals and ties;
   nn/n0 mean degree;
   dd standard deviation of outdegrees;
   cs standard deviation of personal network means;
   ds average of within-personal network variances.  *)
var it: TDoublyLinkedDoubleInt;
    i,nd,nn0: integer;
    zi,zj,c,cs1,cs2,zii,zii2,szii2: real;
begin
  result := 0.0;
  if (high(z) < order) then
    begin
    ErMes := 'Error in Correlation: high(z) = '+inttostr(high(z));
    exit;
    end;
  if (high(miz) < order) then
    begin
    ErMes := 'Error in Correlation: high(miz) = '+inttostr(high(miz));
    exit;
    end;
  s0  := 0.0;
  m0  := 0.0;
  s1  := 0.0;
  m1  := 0.0;
  s2  := 0.0;
  m2  := 0.0;
  c   := 0.0;
  dd  := 0.0;
  cs  := 0.0;
  ds  := 0.0;
  cs1 := 0.0;
  cs2 := 0.0;
  szii2 := 0.0;
  n0  := 0;
  nn  := 0;
  nn0 := 0;
  for i:= order downto 1 do if not miz[i]then
    begin
    inc(n0);
    nd  := 0;
    zi  := z[i];
    zii := zi;
    zii2:= sqr(zi);
    m0  := m0+zi;
    s0  := s0+sqr(zi);
    it  := OutTies[i].First;
    while it.hasnext do
      begin
      it := it.next;
      if (not miz[it.Value1]) and ((i <> it.Value1) or DiagonalPermitted) then
        begin
        inc(nn);
        inc(nd);
        m1 := m1+zi;
        s1 := s1+sqr(zi);
        zj := z[it.value1];
        m2 := m2+zj;
        zii := zii + zj;
        zii2:= zii2 + sqr(zj);
        s2 := s2+sqr(zj);
        c  := c+zi*zj;
        end;
      end;
    dd := dd + sqr(nd);
    zii := zii/(nd+1);
    if (nd >= 1) then
      begin
      szii2:= szii2 + ((nd+1)/nd)*((zii2/(nd+1)) - sqr(zii));
      inc(nn0);
      end;
// weighted    szii2:= szii2 + zii2 - (nd+1)*sqr(zii);
    cs1 := cs1+zii;
    cs2 := cs2+sqr(zii);
    end; (* for i *)
  if (nn = 0) then exit;
  if (nn = 1) then
    begin
    s0 := 0.0;
    s1 := 0.0;
    s2 := 0.0;
    m0 := m1;
    n0 := 1;
    exit;
    end;
  m0 := m0/n0;
  m1 := m1/nn;
  m2 := m2/nn;
  s0 := root((s0/n0)-sqr(m0));
  s1 := root((s1/nn)-sqr(m1));
  s2 := root((s2/nn)-sqr(m2));
  dd := root((dd/n0)-sqr(nn/n0));
  if ((s1*s2) > 1e-20) then
    result := ((c/nn)-(m1*m2))/(s1*s2)
   else result := 0.0;
  cs := root((cs2/n0)-sqr(cs1/n0));
  ds := szii2/nn0;
// weighted    ds := szii2/(n0+nn);
end; (* Correlation *)



function TValuedDigraph.CliqueBound: integer;
(* Upper bound to the largest clique size, computed from the degrees. *)
var i,s,nn: integer;
 nd: array of integer;
begin
  nn := order;
  setlength(nd,nn+1);
  for i:=nn downto 0 do nd[i] := 0;
  for i:=nn downto 1 do inc(nd[outdegree(i)]);
  s := 0;
  i := nn;
  while (s < i+1) and (i >= 0) do
    begin
    dec(i);
    inc(s,nd[i]);
    end;
  result := i+1;
end; (* CliqueBound *)

procedure TValuedDigraph.Cliques(const report: Boolean; var outf: TFileStream;
                        const size,smin,smax: integer;
                        nodes, allnodes:  TDoublyLinkedIntList; itp: TDoublyLinkedDoubleInt;
                        var nc: array of integer);
(* Finds all graphs for which all possible edges between allnodes and nodes,
   and all possible edges within (allnodes \ nodes), are present.
   If nodes is a clique, this is the set of all cliques containing the set nodes, and contained in the set allnodes.
   nodes must correspond to an initial subset of allnodes.
   size must be number of nodes, used for efficiency;
   itp must be nodes.last, also used for efficiency.
   nc gives the clique count.
   if report, then all cliques found of size at least smin are reported to outf.
   *)
var nc0,i,nextsize,i0: integer;
    it,itw: TDoublyLinkedDoubleInt;
    x : string;
begin
  nc0 := low(nc);
  (* It is assumed this procedure is called only if nodes is a clique: *)
  inc(nc[nc0+size]);
  if (report and (size >= smin) and (size >= 3)) then
// if (size >= 1) then
    begin
    x := format('%3d%s',[size,': ']);
    outf.Write(pchar(x)^,length(x));
    itw := nodes.First.Next;
    x := format('%3d',[itw.value1]);
    outf.Write(pchar(x)^,length(x));
    while itw.HasNext do
      begin
      itw := itw.next;
      if (order <= 99) then begin
        x := format('%s%2d',[' - ',itw.value1]);
        outf.Write(pchar(x)^,length(x));
      end
       else begin
        x := format('%s%3d',[' - ',itw.value1]);
        outf.Write(pchar(x)^,length(x));
       end;
      end;
    wwriteln(outf,'');
    end;
//wwriteln(outf,'<<< cliques nodessize '+inttostr(nodes.count)+'; itp.value1 = '+inttostr(itp.Value1));
//if (size >= 2) then exit; // cde
//  if (size >= smax) then exit;
  (* go through all nodes it/i larger than itp *)
  it := itp;
  while it.HasNext do
    begin
    it := it.Next;
    i := it.value1;
(* check if it/i is linked to all nodes *)
    if (size <= 0) or subset(nodes,OutTies[i]) then
(* this means that nodes with i still is a clique *)
       begin
       nodes.Append(i,1);
       nextsize := succ(size);
       Cliques(report,outf,nextsize,smin,smax,nodes,allnodes,it,nc);
//       nodes.DeleteLast;
nodes.removelast; // abcd
       end;
    end;
end; (* Cliques *)


procedure TValuedDigraph.RedCliques(const report: Boolean; var outf: textfile;
                        const size,smin,smax,red: integer;
                        nodes, allnodes:  TDoublyLinkedIntList; itp: TDoublyLinkedDoubleInt;
                        var nc: IIArray);
(* Finds red-uced cliques containing the set nodes, and contained in the set allnodes;
   nodes must itself be a red-uced clique;
   a red-uced clique is a set of nodes all missing at most red edges wihin this set,
   so for red=0 this is a clique;
   for nodes it, (it.value2-1) is the number of missing edges in nodes.
   size must be number of nodes, used for efficiency;
   itp must be nodes.last, also used for efficiency.
   nc gives the red-clique count.
   if report, then all cliques found of size at least smin are reported to outf.
   *)
var i, missingi, maxmis, j: integer;
    it,it1,it2,itw: TDoublyLinkedDoubleInt;
    linked: boolean;
    missing: array of integer;
begin
    Application.ProcessMessages;
//    inc(nnn);
//    if (nnn > 10000) then exit;


  setlength(missing,order+1);

  if (size >= 1) then
    begin
    itw := nodes.First.Next;
    maxmis := itw.Value2;
    missingi := maxmis - 1;
    while itw.HasNext do
      begin
      itw := itw.next;
      inc(missingi,itw.Value2-1);
      if (itw.Value2 > maxmis) then maxmis := itw.Value2;
      end;
    maxmis := maxmis - 1;
    end
   else
    begin
    missingi := 0;
    maxmis := 0;
    end;

  if (report and (size >= smin) and (size >= 3)) then
    begin
    write(outf,size:3 ,' ( missing ',missingi:2,')' );
    itw := nodes.First.Next;
    write(outf,itw.value1:3,' (',itw.Value2-1,')');
    while itw.HasNext do
      begin
      itw := itw.next;
      if (order <= 99) then write(outf,' - ',itw.value1:2,' (',itw.Value2-1,')')
       else write(outf,' - ',itw.value1:3);
      end;
    writeln(outf);
    end;

  inc(nc[size,maxmis]);
  if (size >= smax) then exit;

  (* go with it through all nodes later than itp *)
  it := itp;
  while it.HasNext do
    begin
    it := it.Next;
    i := it.value1;
(* check that nodes with i still is a red-clique: that is, i is linked to all nodes possibly except for red nodes *)
    linked := true;
    it1 := nodes.First;
    it2 := OutTies[i].First;
    missingi := 0;    (* number of missings for i *)
    while (linked and it1.hasnext and it2.hasnext) do
      case sign(it1.Next.Value1 - it2.Next.Value1) of
      -1: begin
          (* no tie between it1.next.value1 and i *)
          it1 := it1.Next;
          inc(missingi);
          missing[missingi] := it1.value1;
          if (it1.Value2 > red) or (missingi > red)
             then linked := false;
          end;
      0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          end;
       1: it2 := it2.Next;
      end; (* case *)
    while (linked and it1.HasNext) do
(* there is no more edge between i and it1.next.value1 *)
      begin
      it1 := it1.Next;
      inc(missingi);
      missing[missingi] := it1.value1;
      if (it1.Value2 > red) or (missingi > red)
         then linked := false;
      end;
    if linked then
(* this means that nodes with i still is a red-clique *)
      begin
(* correct number of internal missings for nodes *)
      it1 := nodes.first;
      j := 1;
      while (it1.HasNext and (j <= missingi)) do
        begin
        it1 := it1.Next;
        if (it1.Value1 = missing[j]) then
          begin
          inc(it1.Value2);
          inc(j);
          end;
        end;
(* add node i to nodes *)
       nodes.Append(i,missingi+1);
       RedCliques(report,outf,succ(size),smin,smax,red,nodes,allnodes,it,nc);

(* after returning from redcliques, restore number of internal missings *)
       it1 := nodes.first;
       j := 1;
       while (it1.HasNext and (j <= missingi)) do
         begin
         it1 := it1.Next;
         if (it1.Value1 = missing[j]) then
           begin
           dec(it1.Value2);
           inc(j);
           end;
         end;

(* remove node i *)
       nodes.deleteLast;
       end;
    end;
end; (* RedCliques *)

procedure TValuedDigraph.Red1_Cliques(const report: Boolean; var outf: TFileStream;
                        const size,smin,smax,ii,jj: integer;
                        nodes, ijnodes, inodes, jnodes:  TDoublyLinkedIntList;
                        nodelinks: TValuedDigraph; itp: TDoublyLinkedDoubleInt;
                        var nc:  array of integer);
(* Finds red-uced cliques containing the set nodes, and contained in the set allnodes;
   nodes must itself be a red-uced clique;
   a red-uced clique is a set of nodes all missing at most red edges wihin this set,
   so for red=0 this is a clique;
   for nodes it, (it.value2-1) is the number of missing edges between it and the other nodes.
   size must be number of nodes, used for efficiency;
   itp must be nodes.last, also used for efficiency.
   nc gives the red-clique count.
   if report, then all cliques found of size at least smin are reported to outf.
   *)
var i, missingi, missingh, j, s1, s2, s3, s4, aa: integer;
    it,it1,it2: TDoublyLinkedDoubleInt;
    linked: boolean;
(* linedinodes, linedjnodes, indi, indj work like a list;
   I tried it with a  TDoublyLinkedDoubleIntList but for unknown reasons, perhaps
   the recursive call of the procedure, this failed. *)
    linedinodes, linedjnodes: array of integer;
    indi, indj: integer;
    x : string;
procedure extrai(const h: integer);
begin
  inc(indi);
  linedinodes[indi] := h;
end;
procedure extraj(const h: integer);
begin
  inc(indj);
  linedjnodes[indj] := h;
end;
begin
    Application.ProcessMessages;
//  inc(nnn);
// if (nnn > 10000) then
//   begin
//   writeln(outf,'stop!!!!');
//   exit;
//   end;


  setlength(linedinodes,order+1);
  setlength(linedjnodes,order+1);
  indi := 0;
  indj := 0;

//writeln(outf);
//write(outf,'inodes: ');
  s1 := 0;
(* s1 will be the number in inodes linked to all of nodes *)
(* of all inodes i, helplist[i] indicates if they are linke to all nodes *)
  it := inodes.FFirst;
  while it.HasNext do
    begin
    it := it.Next;
(* check that it is linked to all of nodes *)
    if subset(nodes,OutTies[it.value1]) then
      begin
      inc(s1);
      extrai(it.value1);
//write(outf,' ',it.value1);
      end;
    end;
//writeln(outf);

(* linkedjnodes will be the set of all jnodes linked to all nodes *)

//write(outf,'jnodes: ');
   s2 := 0;
(* s2 will be the number in jnodes linked to all of nodes *)
  it := jnodes.FFirst;
  while it.HasNext do
    begin
    it := it.Next;
(* check that it is linked to all of nodes *)
    if subset(nodes,OutTies[it.value1]) then
      begin
      inc(s2);
      extraj(it.value1);
//write(outf,' ',it.value1);
      end;
    end;
//writeln(outf);


(* s4 is the number of (i,j) where i in inodes, j in jnodes,
   i and j both are linked to all of nodes, and i is linked to j.*)
  s4 := 0;
  if (s2 > 0) and (size + 2 <= smax) then
    for i := 1 to indi do
      begin
      j := 0;
      it1 := outties[linedinodes[i]].First;
      while ((j < indj) and it1.HasNext) do
        case sign(linedjnodes[succ(j)] - it1.Next.value1) of
          -1: inc(j);
          0:  begin
              inc(j);
              it1 :=  it1.Next;
              inc(s4);
              end;
           1: it1 := it1.Next;
        end;
      end;

s3 := 0;

//  if (s1 + s2 + s3 >= 1) then
//    begin
//    if (size + 2 <= smax) then
    inc(nc[size+2],s4);
    inc(nc[size+1],s1+s2+s3);
//    end
//   else exit;

//writeln(outf,s4);
//writeln(outf,s1+s2+s3);

  if (report and (size >= smin) and (size >= 3)) then
    begin
    it := nodes.First.Next;
    missingi := it.value2 - 1;
    while it.HasNext do
      begin
      it := it.next;
      inc(missingi,it.Value2-1);
      end;

    x := format('%3d%s%2d%s',[size ,' ( missing ',missingi,')' ]);
    outf.Write(pchar(x)^,length(x));
    x := format('%s%4d%4d%s',[' (contributions ',s1+s2+s3,s1*s2,') ']);
    outf.Write(pchar(x)^,length(x));
    it := nodes.First.Next;
    x := format('%3d%s%d%s',[it.value1,' (',it.Value2-1,')']);
    outf.Write(pchar(x)^,length(x));
    while it.HasNext do
      begin
      it := it.next;
      if (order <= 99) then begin
       x := format('%s%2d%s%d%s',[' - ',it.value1,' (',it.Value2-1,')']);
       outf.Write(pchar(x)^,length(x));
      end
       else begin
        x := format('%s%3d',[' - ',it.value1]);
        outf.Write(pchar(x)^,length(x));
       end;
      end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    end;

//writeln(outf,'size = ',size,'smax = ',smax);

  if (size >= smax-1) then exit;

  (* go with it through all nodes later than itp *)
  it := itp;
  while it.HasNext do
    begin
    it := it.Next;
    i := it.value1;
(* check that nodes with i still is a 1-clique: that is, i is linked to all nodes possibly except for 1 node *)
    linked := true;
(* linked indicates that here is not more than one tie missing *)
    it1 := nodes.First;
    it2 := OutTies[i].First;
    missingi := 0;    (* number of missings for i *)
    missingh := 0;
    (* missingh will be the element in nodes to which i has no tie;
       there is at most one such element *)
    while (linked and it1.hasnext and it2.hasnext) do
      case sign(it1.Next.Value1 - it2.Next.Value1) of
      -1: begin
          (* no tie between it1.next.value1 and i *)
          it1 := it1.Next;
          inc(missingi);
          missingh := it1.value1;
          if (it1.Value2 > 1) or (missingi > 1) then linked := false;
          end;
      0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          end;
       1: it2 := it2.Next;
      end; (* case *)
    if (linked and it1.HasNext) then
(* then there is an extra element in nodes without an edge to i  *)
      begin
      if (missingi >= 1) or (it1.Next.HasNext) then linked := false
       else
        begin
        inc(missingi);
        missingh := it1.next.value1;
        end;
      end;

    if linked then
(* this means that nodes with i still is a 1-red-clique *)
      begin
(* correct the number of internal missings for node missingh; but first find it! *)
      if (missingi >= 1) then
        begin
        it1 := nodes.first;
        repeat
          it1 := it1.next;
         until (it1.value1 = missingh) or (not it1.HasNext);
(* the not hasnext condition should never occur *)
        if (it1.value1 = missingh) then inc(it1.Value2);
        end;
(* add node i to nodes *)
     nodes.Append(i,missingi+1);

      Red1_Cliques(report,outf,succ(size),smin,smax,ii,jj,nodes,ijnodes,inodes,jnodes,nodelinks,it,nc);
(* after returning from red1_cliques, restore number of internal missings *)
      if (missingi >= 1) then
        begin
        it1 := nodes.first;
        repeat
          it1 := it1.next;
         until (it1.value1 = missingh) or (not it1.HasNext);
(* the not hasnext condition should never occur *)
        if (it1.value1 = missingh) then dec(it1.Value2);
        end;

(* remove node i *)
      if (nodes.count > 1) then nodes.deleteLast;
      end; (* if linked *)
    end; (* while it.Hasnext *)
end; (* Red1_Cliques *)

procedure TValuedDigraph.AllCliques(const report: Boolean; var outf: TFileStream;
            const smin, smax: integer);
(* Finds all cliques of order smin to smax. *)
var size,i,j,s: integer;
    nodes, allnodes:  TDoublyLinkedIntList;
    it1, it2, it3, itp: TDoublyLinkedDoubleInt;
    nc,nd: array of integer;
    x : string;
begin
  setlength(nc,order+2);
  setlength(nd,order+1);
  nodes := TDoublyLinkedIntList.Create;
  nodes.Insertfront(0,0);
(* now nodes is the empty set *)
  allnodes := TDoublyLinkedIntList.Create;
  allnodes.Insertfront(0,0);
  for i := 1 to order do if (OutDegree(i) >= smin-1) then
    begin
    allnodes.Append(i,1);
    end;
(* code between <=== and ===> can be dropped; unknown degree of efficiency difference *)
(* <=== *)
{
  for i:=order downto 1 do nc[i] := 0;
(* determine degrees nc of allnodes in their own subgraph *)
  it1 := allnodes.First;
    while it1.HasNext do
      begin
      it1 := it1.Next;
      i := it1.Value1;
      it2 := OutTies[i].First;
      it3 := allnodes.first;
      j := 0;
      while (it2.HasNext and it3.HasNext and (j < i)) do
        case sign(it2.next.Value1 - it3.next.value1) of
        -1: it2 := it2.Next;
         0: begin
            it2 := it2.Next;
            it3 := it3.Next;
            j := it2.Value1;
            if (j < i) then
              begin
              inc(nc[i]);
              inc(nc[j]);
              end;
            end;
         1: it3 := it3.Next;
        end; (* case *)
      end;

(* determine degee counts nd[.] and highest degree i *)
  for i:=order downto 0 do nd[i] := 0;
  it1 := allnodes.First;
  i := 0;
  while it1.HasNext do
    begin
    it1 := it1.Next;
    inc(nd[nc[it1.value1]]);
    if (nc[it1.Value1] > i) then i := nc[it1.value1];
    end;
(* only nodes i with internal degree nc[i] >= smin  can be part of a clique *)

(* s is number of nodes with degree at least i *)
  s := nd[i];

  while (s < i+1) and (i >= 0) do
    begin
    dec(i);
    inc(s,nd[i]);
    end;
  s := i;
(* cliques cannot have more than s+1 elements *)

  allnodes.destroy;
  allnodes := TDoublyLinkedIntList.Create;
  allnodes.Insertfront(0,0);
  size := 0;
  for i := 1 to order do if (nc[i] >= smin) then
    begin
    allnodes.Append(i,1);
    inc(size);
    end;

(* now allnodes is the set of all nodes with induced degree smin or higher
   within  the subgraph induced by the nodes of degre smin or higher. *)

  if report then
    begin
    writeln(outf,'All cliques of order at least ',smin,'.');
    writeln(outf,'Upper bound to clique sizes is ', cliquebound,'.');
    writeln(outf,'This can be diminished to ',s,'.');
    writeln(outf,'Number of studied nodes is ',size,'.');
    end;

}
(* ==> *)
  for i := order+1 downto 0 do nc[i] := 0;
  itp := allnodes.First;
  size := 0;
  Cliques(report,outf,size,smin,smax,nodes,allnodes,itp,nc);
  nodes.Destroy;
  allnodes.Destroy;
//  if report then
    begin
    size := smin;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'Observed numbers of cliques:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'order   ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%4d',[size]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until nc[size] = 0;
    size := smin;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'number  ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%4d',[nc[size]]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until nc[size] = 0;
    //writeln(outf);
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    outf.Write(pchar(eol)^,length(eol));
    end;
end; (* AllCliques *)


procedure TValuedDigraph.AllCliques_ij(const report: Boolean; var outf: TFileStream;
            smin: integer; const smax,ii,jj: integer; const t: shortint;
            var ncl: array of integer);
(* Finds all cliques of order smin to smax, containing ii and jj.
   t = dg(ii,jj)  *)
var size,i,smaxx,sminn: integer;
    nodes, allnodes:  TDoublyLinkedIntList;
    it1, it2, itp: TDoublyLinkedDoubleInt;
    x : string;
begin
  if (smin < 2) then smin := 2;
//  t := tievalue(ii,jj); //must be input
  for i := low(ncl) to high(ncl) do ncl[i] := 0;
  if (ii = jj) or (outdegree(ii) - t < smin - 2) or (outdegree(jj) - t < smin - 2)
    then exit;
//writeln(outf,ii:4,jj:4,t:4);

  nodes := TDoublyLinkedIntList.Create;
  nodes.Insertfront(0,0);
(* now nodes is the empty set  *)
(* construct allnodes as the set consisting of all other nodes
   tied to ii and to jj and with outdegree at least smin-1.*)
  allnodes := TDoublyLinkedIntList.Create;
  allnodes.Insertfront(0,0);

  it1 := OutTies[ii].First;
  it2 := OutTies[jj].First;
  while (it1.hasnext and it2.HasNext) do
    case sign(it1.Next.Value1 - it2.Next.value1) of
      -1: it1 := it1.Next;
       0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          allnodes.Append(it1.value1,1);
//writeln(outf,' --- ',it1.value1:3);
          end;
       1: it2 := it2.Next;
       end; (* case *)

  itp := allnodes.First;
  size := 0;
  smaxx := smax-2;
  sminn := smin-2;
  if (sminn <= 1) then sminn := 2;
  if (smaxx <= sminn) then smaxx := sminn;
  Cliques(report,outf,size,sminn,smaxx,nodes,allnodes,itp,ncl);
  nodes.Destroy;
  allnodes.Destroy;

//for i:=low(ncl) to high(ncl) do write(outf,ncl[i]:4); writeln(outf,'>>>>>');

  for i:=smax downto 2 do ncl[i] := ncl[i-2];
  ncl[0] := 0;
  ncl[1] := 0;
  if report then
    begin
    size := smin;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := format('%s%d%s%d%s',['found cliques containing ',ii,'  and ',jj,' :'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'order   ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%4d',[size]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until ncl[size] = 0;
    size := smin;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'number  ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%4d',[ncl[size]]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until ncl[size] = 0;
   // writeln(outf);
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    outf.Write(pchar(eol)^,length(eol));
    end;
end; (* AllCliques_ij *)

procedure TValuedDigraph.All_red1_Cliques_ij(const report: Boolean; var outf: TFIleStream;
            smin: integer; const smax,ii,jj: integer; const t: shortint;
            var ncl: array of integer);
(* Finds all 1-reduced cliques of order smin to smax, containing ii and jj.
   t = dg(ii,jj)  *)
var size,i,smaxx,sminn,h1,h2: integer;
    nodes, ijnodes, inodes, jnodes:  TDoublyLinkedIntList;
    it1, it2, it3, itp: TDoublyLinkedDoubleInt;
    nodeslinks: TValuedDigraph;
    x : string;
begin
  if (smin < 2) then smin := 2;
//  t := tievalue(ii,jj);  must be input
  for i := low(ncl) to high(ncl) do ncl[i] := 0;
  if (ii = jj) or (outdegree(ii) - t < smin - 3) or   (outdegree(jj) - t < smin - 3) then exit;

  nodes := TDoublyLinkedIntList.Create;
  nodes.Insertfront(0,0);
(* now nodes is the empty set  *)
(* construct ijnodes as the set consisting of all nodes
   tied to ii and to jj and with outdegree at least smin-2;
   inodes the set of nodes connected to ii, not to jj;
   jnodes the set of nodes connected to jj, not ii.*)
  ijnodes := TDoublyLinkedIntList.Create;
  ijnodes.Insertfront(0,0);
  inodes := TDoublyLinkedIntList.Create;
  inodes.Insertfront(0,0);
  jnodes := TDoublyLinkedIntList.Create;
  jnodes.Insertfront(0,0);

(* in the following, h1 and h2 are the next smallest elements
   of outties[ii] ad outties[jj] *)
  it1 := OutTies[ii].First;
  if (it1.HasNext) then h1 := it1.Next.Value1 else h1 := succ(order);
  it2 := OutTies[jj].First;
  if (it2.HasNext) then h2 := it2.Next.value1 else h2 := succ(order);

//writeln(outf,ii:4,jj:4);
//writeln(outf,h1:4,h2:4,h3:4);
  while ((h1 <= order) or (h2 <= order)) do
    begin
      case sign(h1 - h2) of
      -1: begin
          it1 := it1.Next;
          if (outdegree(h1) >= smin-2) and
            (h1 <> jj) then inodes.Append(h1,1);
          if (it1.HasNext) then h1 := it1.Next.Value1 else h1 := succ(order);
          end;
       0: begin
          it1 := it1.Next;
          it2 := it2.Next;
          if (outdegree(h1) >= smin-2) then
          ijnodes.Append(h1,1);
          if (it1.HasNext) then h1 := it1.Next.Value1 else h1 := succ(order);
          if (it2.HasNext) then h2 := it2.Next.Value1 else h2 := succ(order);
          end;
       1: begin
          it2 := it2.Next;
          if (outdegree(h2) >= smin-2) and
            (h2 <> ii) then jnodes.Append(h2,1);
          if (it2.HasNext) then h2 := it2.Next.Value1 else h2 := succ(order);
          end;
      end; (* case *)
    end;

  nodeslinks := TDigraph.Create(self.order);
  it1 := inodes.First;
  while it1.HasNext do
    begin
    it1 := it1.next;
    it2 := jnodes.First;
    i := it1.Value1;
    it3 := OutTies[i].First;
(* add ties to nodeslinks for (i,j) with self(i,j) = 1 *)
    while it2.HasNext and it3.HasNext do
      case sign(it2.Next.Value1 - it3.Next.Value1) of
        -1: it2 := it2.Next;
         0: begin
            it2 := it2.Next;
            it3 := it3.Next;
            nodeslinks.AddTie(i,it2.Value1,1,true);
            end;
         1: it3 := it3.Next;
      end;
    end;
(* nodeslinks is a very sparse directed graph, with entries 1
   for (i,j) with i in inodes, j in jnodes, and linked in self *)
{
for i:=1 to order do write(outf,i:3);
writeln(outf);
for i:=1 to order do write(outf,tievalue(ii,i):3);
writeln(outf);
for i:=1 to order do write(outf,tievalue(jj,i):3);
writeln(outf);
writeln(outf,'nodes ',nodes.count);
nodes.writevector(outf,0);
writeln(outf,'ijnodes ',ijnodes.count);
ijnodes.writevector(outf,0);
writeln(outf,'inodes ',inodes.count);
inodes.writevector(outf,0);
writeln(outf,'jnodes ',jnodes.count);
jnodes.writevector(outf,0);
writeln(outf,'znodes ',znodes.count);
znodes.writevector(outf,0);
writeln(outf,'nodeslinks');
nodeslinks.writeadjacency(outf);
}
 // @@@

  itp := ijnodes.First;
  size := 0;
//writeln(outf,'<<<<');
  smaxx := smax-2;
  sminn := smin-2;
  if (sminn <= 1) then sminn := 2;
  Red1_Cliques(report,outf,size,sminn,smaxx,ii,jj,nodes,ijnodes,inodes,jnodes,nodeslinks,itp,ncl);
  nodes.Destroy;
  ijnodes.Destroy;
  inodes.Destroy;
  jnodes.Destroy;
  nodeslinks.Destroy;

//for i:=low(ncl) to high(ncl) do write(outf,ncl[i]:4); writeln(outf,'>>>>>');

  for i:=smax downto 2 do ncl[i] := ncl[i-2];
  ncl[0] := 0;
  ncl[1] := 0;
  if report then
    begin
    size := smin;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x:= format('%s%d%s%d%s',['contributions to change statistic for 1-reduced cliques containing ',
      ii,'  and ',jj,' :'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'order   ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%d',[size]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until ncl[size] = 0;
    size := smin;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'number  ';
    outf.Write(pchar(x)^,length(x));
    repeat
      x := format('%d',[ncl[size]]);
      outf.Write(pchar(x)^,length(x));
      inc(size);
     until ncl[size] = 0;
    //writeln(outf);
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    outf.Write(pchar(eol)^,length(eol));
    end;
end; (* All_red1_Cliques_ij *)


procedure TValuedDigraph.ReducedCliques(const report: Boolean; var outf: textfile;
            const red, smin, smax: integer);
(* Finds all red-uced cliques of order at least smin and at most smax. *)
var size,i,j: integer;
    nodes, allnodes:  TDoublyLinkedIntList;
    itp: TDoublyLinkedDoubleInt;
    nc: IIArray;
begin
  if (red >= smin-1) then
    begin
    writeln(outf,'Calculating the number of ',red,'-reduced cliques with at least ',
                   smin,' nodes is not meaningful.');
    exit;
    end;
  if (smin > smax) then
    begin
    writeln(outf,'Calculating the number of ',red,'-reduced cliques having between ',
                   smin,' and ',smax,' nodes is not meaningful.');
    exit;
    end;
(* nc[i,j] will contain number of subsets of order i nodewise missing at most j links
   and for some node missng exactly j links. *)
  setlength(nc,smax+2,red+1);
  nodes := TDoublyLinkedIntList.Create;
  nodes.Insertfront(0,0);
(* now nodes is the empty set *)
  allnodes := TDoublyLinkedIntList.Create;
  allnodes.Insertfront(0,0);
  (* let allnodes be the set of nodes of degre at least smin-red-1 *)
  for i := 1 to order do if (OutDegree(i) >= smin-1-red) then
    begin
    allnodes.Append(i,1);
    end;

  for j := red downto 0 do for i := smax+1 downto 0 do nc[i,j] := 0;
  itp := allnodes.First;
  size := 0;
  RedCliques(report,outf,size,smin,smax,red,nodes,allnodes,itp,nc);
  nodes.Destroy;
  allnodes.Destroy;

//  if report then
    begin
    for j:=1 to red do
      begin
      size := smin;
      writeln(outf);
      writeln(outf,'Observed numbers of ',j,'-reduced cliques:');
      write(outf,'order   ');
      repeat
        write(outf,size:4);
        inc(size);
       until nc[size,j] = 0;
      size := smin;
      writeln(outf);
      write(outf,'number  ');
      repeat
        write(outf,nc[size,j]:4);
        inc(size);
       until nc[size,j] = 0;
      writeln(outf);
      writeln(outf);
      end;
    end;
end; (* ReducedCliques *)

procedure TVAluedDigraph.SetZero;
var i: integer;
    it: TDoublyLinkedDoubleInt;
begin
  if HasArray then for i:=order downto 1 do
    begin
    it := OutTies[i].FFirst;
    while it.HasNext do
      begin
      it := it.Next;
      AdjMatrix[i,it.Value1] := 0;
      end;
    end;
  for i:=order downto 1 do
    begin
    Inties[i].Destroy;
    Inties[i] := TDoublyLinkedIntList.Create;
    Inties[i].Insertfront(0,0);
    Outties[i].Destroy;
    Outties[i] := TDoublyLinkedIntList.Create;
    Outties[i].Insertfront(0,0);
    end;
end; (* SetZero *)

procedure TValuedDigraph.PermitDiagonal;
begin
  DiagonalPermitted := true;
end;


procedure TValuedDigraph.ChangeTieValue(const i,j: integer; const v: shortint);
var it1, itstop: TDoublyLinkedDoubleInt;
    it1v1: Integer;
    Found: Boolean;
begin
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)
      or ((i = j) and (not DiagonalPermitted) and (v <> 0))) then begin
    ErMes := 'call ChangeTieValue('+inttostr(i)+','+inttostr(j)+','+inttostr(v)+')';
    exit;
   end;
  if ((i = j) and (not DiagonalPermitted) and (v = 0)) then exit;

(* Search InTies[j] *)
  it1 := InTies[j].First;
  if (it1.Value1 <= 0) then begin
    if (v <> 0) then begin
      InTies[j].Append(i,v);
      it1.Value1 := 1;
     end;
   end else begin
    if (2*i <= Order) then begin
(* case 1: Start searching from below *)
    repeat
      it1 := it1.Next;
      Found := (it1.Value1 = i);
    until ((it1.Value1 >= i) or (not it1.HasNext));
    if Found then begin
      if (v = 0) then begin
        InTies[j].Delete(it1);
        Dec(InTies[j].First.Value1);
       end else begin
        it1.Value2 := v;
       end;
     end else (* not Found *) begin
       if (v <> 0) then begin
         if (it1.Value1 >= i) then it1.InsertBefore(i,v)
           else InTies[j].Append(i,v);
         Inc(InTies[j].First.Value1);
        end;
       end; (* if not Found *)
  end (* case 1 *) else begin
(* case 2: Start searching from above *)
    it1 := Inties[j].Last;
    it1v1 := it1.Value1;
    Found := (it1v1 = i);
    itstop := InTies[j].First.next;
    while not ((it1v1 <= i) or (it1 = itstop)) do begin
      it1 := it1.Prev;
      it1v1 := it1.Value1;
      Found := (it1v1 = i);
     end;
    if Found then begin
      if (v = 0) then begin
        InTies[j].Delete(it1);
        Dec(InTies[j].First.Value1);
       end else begin
        it1.Value2 := v;
       end;
     end else (* not Found *) begin
       if (v <> 0) then begin
         if (it1v1 <= i) then begin
             if it1.HasNext then it1.InsertAfter(i,v) else InTies[j].Append(i,v);
          end else it1.InsertBefore(i,v);
         Inc(InTies[j].First.Value1);
        end;
     end; (* if not Found *)
    end; (* case 2 *)
  end; (* if it1.Value1 > 0 *)

(* Search OutTies[i] *)
  it1 := OutTies[i].First;
  if (it1.Value1 <= 0) then begin
    if (v <> 0) then begin
      OutTies[i].Append(j,v);
      it1.Value1 := 1;
     end;
   end else begin
    if (2*j <= Order) then begin
(* case 3: Start searching from below *)
    repeat
      it1 := it1.Next;
      Found := (it1.Value1 = j);
    until ((it1.Value1 >= j) or (not it1.HasNext));
    if Found then begin
      if (v = 0) then begin
        OutTies[i].Delete(it1);
        Dec(OutTies[i].First.Value1);
       end else begin
        it1.Value2 := v;
       end;
     end else (* not Found *) begin
       if (v <> 0) then begin
         if (it1.Value1 >= j) then it1.InsertBefore(j,v)
           else OutTies[i].Append(j,v);
         Inc(OutTies[i].First.Value1);
        end;
     end; (* if not Found *)
  end (* case 3 *) else begin
(* case 4: Start searching from above *)
    it1 := Outties[i].Last;
    it1v1 := it1.Value1;
    Found := (it1v1 = j);
    itstop := OutTies[i].First.next;
    while not ((it1v1 <= j) or (it1 = itstop)) do begin
      it1 := it1.Prev;
      it1v1 := it1.Value1;
      Found := (it1v1 = j);
     end;
    if Found then begin
      if (v = 0) then begin
        OutTies[i].Delete(it1);
        Dec(OutTies[i].First.Value1);
       end else begin
        it1.Value2 := v;
       end;
     end else (* not Found *) begin
       if (v <> 0) then begin
         if (it1v1 <= j) then begin
           if it1.HasNext then it1.InsertAfter(j,v) else OutTies[i].Append(j,v);
          end else it1.InsertBefore(j,v);
         Inc(OutTies[i].First.Value1);
        end;
     end; (* if not Found *)
    end; (* case 4 *)
  end; (* if it1.Value1 > 0 *)
  if HasArray then AdjMatrix[i,j] := v;
end; (* ChangeTieValue *)

procedure TValuedDigraph.AddTie(const i,j: integer; const v: shortint;
                       const checkhi: boolean);
(* This procedure adds tie (i,j) with value v,
   assuming that in the adjacency matrix
   both in row i and in column j this is the largest element. *)
var  OK: Boolean;
begin
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)
      or ((i = j) and (not DiagonalPermitted) and (v <> 0))) then begin
    ErMes := 'call AddTie('+inttostr(i)+','+inttostr(j)+','+inttostr(v)+')';
    exit;
   end;

  OK := (OutTies[i].Last.Value1 < j) and (InTies[j].Last.Value1 < i);
  if OK then
    begin
    if (v = 0) then exit;
    OutTies[i].Append(j,v);
    InTies[j].Append(i,v);
    Inc(OutTies[i].Ffirst.Value1);
    Inc(InTies[j].Ffirst.Value1);
    if HasArray then AdjMatrix[i,j] := v;
    end (* if OK *)
   else
    begin
    ChangeTieValue(i,j,v);
    if checkhi then ErMes := 'Call AddTie for non-maximal element';
    end;
end; (* AddTie *)

function TValuedDigraph.IncreaseTie(const i,j: integer; const v: shortint): ShortInt;
var it1, itstop: TDoublyLinkedDoubleInt;
    it1v1: Integer;
    Found: Boolean;
begin
(* This is pretty much like ChangeTieValue, except that
   in some places v is replaced by it1.Value2+v *)
(* Default: *)
  Result := 0;
(* Check : *)
  if ((i <= 0) or (i > Order) or (j <= 0) or (j > Order)
      or ((i = j) and (not DiagonalPermitted) and (v <> 0))) then begin
    ErMes := 'call IncreaseTie('+inttostr(i)+','+inttostr(j)+','+inttostr(v)+')';
    exit;
   end;
  if (v = 0) then begin
    Result := TieValue(i,j);
    exit;
   end;

(* Search InTies[j] *)
  it1 := InTies[j].First;
  if (it1.Value1 <= 0) then begin
    InTies[j].Append(i,v);
    Result := v;
    it1.Value1 := 1;
   end else begin
    if (2*i <= Order) then begin
(* case 1: Start searching from below *)
    repeat
      it1 := it1.Next;
      Found := (it1.Value1 = i);
    until ((it1.Value1 >= i) or (not it1.HasNext));
    if Found then begin
      if (it1.Value2+v = 0) then begin
        InTies[j].Delete(it1);
        Dec(InTies[j].First.Value1);
       end else begin
        it1.Value2 := it1.Value2+v;
        Result := it1.Value2;
       end;
     end else (* not Found *) begin
       if (it1.Value1 >= i) then it1.InsertBefore(i,v)
         else InTies[j].Append(i,v);
       Inc(InTies[j].First.Value1);
       Result := v;
      end; (* if not Found *)
  end (* case 1 *) else begin
(* case 2: Start searching from above *)
    it1 := Inties[j].Last;
    it1v1 := it1.Value1;
    Found := (it1v1 = i);
    itstop := InTies[j].First.next;
    while not ((it1v1 <= i) or (it1 = itstop)) do begin
      it1 := it1.Prev;
      it1v1 := it1.Value1;
      Found := (it1v1 = i);
     end;
    if Found then begin
      if (it1.Value2+v = 0) then begin
        InTies[j].Delete(it1);
        Dec(InTies[j].First.Value1);
       end else begin
        it1.Value2 := it1.Value2+v;
        Result := it1.Value2;
       end;
     end else (* not Found *) begin
       if (it1v1 <= i) then begin
           if it1.HasNext then it1.InsertAfter(i,v) else InTies[j].Append(i,v);
        end else it1.InsertBefore(i,v);
       Inc(InTies[j].First.Value1);
       Result := v;
     end; (* if not Found *)
    end; (* case 2 *)
  end; (* if it1.Value1 > 0 *)

(* Search OutTies[i] *)
  it1 := OutTies[i].First;
  if (it1.Value1 <= 0) then begin
    OutTies[i].Append(j,v);
    it1.Value1 := 1;
    Result := v;
   end else begin
    if (2*j <= Order) then begin
(* case 3: Start searching from below *)
    repeat
      it1 := it1.Next;
      Found := (it1.Value1 = j);
    until ((it1.Value1 >= j) or (not it1.HasNext));
    if Found then begin
      if (it1.Value2+v = 0) then begin
        OutTies[i].Delete(it1);
        Dec(OutTies[i].First.Value1);
       end else begin
        it1.Value2 := it1.Value2+v;
        Result := it1.Value2;
       end;
     end else (* not Found *) begin
       if (it1.Value1 >= j) then it1.InsertBefore(j,v)
        else OutTies[i].Append(j,v);
       Inc(OutTies[i].First.Value1);
       Result := v;
      end; (* if not Found *)
  end (* case 3 *) else begin
(* case 4: Start searching from above *)
    it1 := Outties[i].Last;
    it1v1 := it1.Value1;
    Found := (it1v1 = j);
    itstop := OutTies[i].First.next;
    while not ((it1v1 <= j) or (it1 = itstop)) do begin
      it1 := it1.Prev;
      it1v1 := it1.Value1;
      Found := (it1v1 = j);
     end;
    if Found then begin
      if (it1.Value2+v = 0) then begin
        OutTies[i].Delete(it1);
        Dec(OutTies[i].First.Value1);
       end else begin
        it1.Value2 := it1.Value2+v;
        Result := it1.Value2;
       end;
     end else (* not Found *) begin
       if (it1v1 <= j) then begin
         if it1.HasNext then it1.InsertAfter(j,v) else OutTies[i].Append(j,v);
        end else it1.InsertBefore(j,v);
       Inc(OutTies[i].First.Value1);
     end; (* if not Found *)
    end; (* case 4 *)
  end; (* if it1.Value1 > 0 *)
  if HasArray then Inc(AdjMatrix[i,j], v);
end; (* IncreaseTie *)

procedure TValuedDigraph.SetRowTo(const i: integer; const vec: TDoublyLinkedIntList);
(* Sets row i to vec *)
(* Basics of this procedure were copied from ForDifferencesDo *)
var it1, it2: TDoublyLinkedDoubleInt;
    v1,v2,n,nn: integer;
    vec1, vec2: TDoublyLinkedIntList;
procedure continue1;
begin
  if (it1.HasNext) then
    begin
    it1 := it1.next;
    v1 := it1.Value1;
    end
   else v1 := nn;
end;
procedure continue2;
begin
  if (it2.HasNext) then
    begin
    it2 := it2.next;
    v2 := it2.Value1;
    end
   else v2 := nn;
end;
begin
(* it is impossible in OutTies[i] to simultaneously check what must be changed,
   and change it; therefore the following procedure is used,
   with intermediate storage lists vec1 and vec2:
   vec1 will contain as value1 = j, value2 = 1 the entries in OutTies[i]
   that must be nullified;
   vec2 will contain as value1 = j, value2 the entries in OutTies[i]
   that must be changed into value2 <> 0. *)

  vec1 := TDoublyLinkedIntList.create;
  vec1.InsertFront(0,0);
  vec2 := TDoublyLinkedIntList.create;
  vec2.InsertFront(0,0);

  it1 := OutTies[i].First;
  it2 := vec.First;
  n := Order;
  nn := succ(n);
  continue1;
  continue2;

  while (v1 <= n) or (v2 <= n) do
   case sign(v1 - v2) of
    -1: begin
(* OutTies[i](v1) <> 0, must be changed into vec(v1) =  0 *)
        vec1.Append(v1,1);
        continue1;
        end;
    0:  begin
(* OutTies[i](v1) and vec(v2), where v1 = v2, both are nonzero *)
        if (it1.value2 <> it2.value2) then
          vec2.Append(v1,it2.value2);
        continue1;
        continue2;
        end; (* case 0 *)
    1:  begin
(* OutTies[i](v2) = 0, but vec(v2) <>  0 *)
        vec2.Append(v2,it2.value2);
        continue2;
        end;
   end; (* case *)

(* now carry out the changes *)
  it1 := vec1.Ffirst;
  while it1.HasNext do
    begin
    it1 := it1.next;
    ChangeTieValue(i,it1.Value1,0);
    end;
  it2 := vec2.Ffirst;
  while it2.HasNext do
    begin
    it2 := it2.next;
    ChangeTieValue(i,it2.Value1,it2.Value2);
    end;

  vec1.Destroy;
  vec2.Destroy;
end; (* SetRowTo *)

procedure TValuedDigraph.SetColumnTo(const i: integer; const vec: TDoublyLinkedIntList);
(* Sets column i to vec *)
(* Just like SetRowTo *)
var it1, it2: TDoublyLinkedDoubleInt;
    v1,v2,n,nn: integer;
    vec1, vec2: TDoublyLinkedIntList;
procedure continue1;
begin
  if (it1.HasNext) then
    begin
    it1 := it1.next;
    v1 := it1.Value1;
    end
   else v1 := nn;
end;
procedure continue2;
begin
  if (it2.HasNext) then
    begin
    it2 := it2.next;
    v2 := it2.Value1;
    end
   else v2 := nn;
end;
begin
  vec1 := TDoublyLinkedIntList.create;
  vec1.InsertFront(0,0);
  vec2 := TDoublyLinkedIntList.create;
  vec2.InsertFront(0,0);

  it1 := InTies[i].First;
  it2 := vec.First;
  n := Order;
  nn := succ(n);
  continue1;
  continue2;

  while (v1 <= n) or (v2 <= n) do
   case sign(v1 - v2) of
    -1: begin
(* InTies[i](v1) <> 0, must be changed into vec(v1) =  0 *)
        vec1.Append(v1,1);
        continue1;
        end;
    0:  begin
(* InTies[i](v1) and vec(v2), where v1 = v2, both are nonzero *)
        if (it1.value2 <> it2.value2) then
          vec2.Append(v1,it2.value2);
        continue1;
        continue2;
        end; (* case 0 *)
    1:  begin
(* InTies[i](v2) = 0, but vec(v2) <>  0 *)
        vec2.Append(v2,it2.value2);
        continue2;
        end;
   end; (* case *)

(* now carry out the changes *)
  it1 := vec1.Ffirst;
  while it1.HasNext do
    begin
    it1 := it1.next;
    ChangeTieValue(it1.Value1,i,0);
    end;
  it2 := vec2.Ffirst;
  while it2.HasNext do
    begin
    it2 := it2.next;
    ChangeTieValue(it2.Value1,i,it2.Value2);
    end;

  vec1.Destroy;
  vec2.Destroy;
end; (* SetColumnTo *)

function TValuedDigraph.RandomOutTie(const version: Byte; const i: integer):
            integer;
(* Random draw from the outties of i;
   returns i if outdegree(i) = 0 . *)
var j,j0,od: integer;
    it:TDoublyLinkedDoubleInt;
begin
  result := i;
  od := OutDegree(i);
  if (od <= 0) then exit;
  j := 1 + randomd(version,od);
  it := OutTies[i].First;
  j0 := 0;
  while (j0 < j) and it.HasNext do begin
    it := it.Next;
    inc(j0);
   end;
  result := it.Value1;
end; (* RandomOutTie *)

procedure TValuedDigraph.RandomTie(const version: Byte; const c: array of integer;
                                   var i,j: integer);
(* Random draw (i,j) from the outties;
   i with cumulative probabilities proportional to c[i]; (0,0) for empty graph *)
var i0: integer;
begin
  i0 := newranp(version,order,c);
  if (i0 >= 1) and (outdegree(i0) >= 1) then
    begin
    j := RandomOutTie(version,i0);
    i := i0;
    end
   else
    begin
    i := 0;
    j := 0;
    end;
end; (* RandomTie *)

procedure TValuedDigraph.ListsToAdjacency(var x: array of ShortInt);
(* Writes data in adjacency matrix format *)
var i,j: integer;
    it:TDoublyLinkedDoubleInt;
begin
  if (high(x) < sqr(Order)+1) then begin
   ErMes := 'ListToAdjacency called without a large enough array';
   exit;
  end;
  for i:=1 to Order do begin
    for j:=1 to Order do x[(i-1)*Order+j] := 0;
    it := OutTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      x[(i-1)*Order+it.Value1] := it.Value2;
     end;
   end;
end; (* ListsToAdjacency *)

procedure TValuedDigraph.WriteAdjacency(var outf: textfile);
var i,j,m,d: integer;
begin
  m := maxi;
  d := -mini;
  if (d > m) then m := d;
  if (d > 0) then m := 10*m; (* then an extra space is needed *)
  d := 2;
  if (m >= 10) then d := 3;
  if (m >= 100) then d := 4;
  for i:=1 to Order do begin
    for j:=1 to Order do
     if ((i = j) and (not DiagonalPermitted)) then
     case d of
      1: write(outf,'0');
      2: write(outf,' 0');
      3: write(outf,'  0');
      4: write(outf,'   0');
      end (* case *)
     else write(outf,TieValue(i,j):d);
    writeln(outf);
   end;
  writeln(outf);
end; (* WriteAdjacency *)

procedure TValuedDigraph.WriteLists(var outf: tFileStream);
var i,m,d1,d2: integer;
    it:TDoublyLinkedDoubleInt;
    graph: boolean;
    x: string;
begin
  m := maxi;
  d1 := -mini;
  graph := ((d1 = 0) and (m = 1));
  if (d1 > m) then m := d1;
  if (d1 > 0) then m := 10*m;
  d1 := 1;
  if (m >= 10) then d1 := 3;
  if (m >= 100) then d1 := 4;
  m := Order;
  case m of
    1..9: d2 := 1;
    10..99: d2 := 2;
    100..999: d2 := 3;
    1000..9999: d2 := 4;
    10000..99999: d2 := 5;
    else d2 := 6;
   end;
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  if graph then begin
    x := 'Out-ties:'+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x :='Out-ties. nodes:values'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  for i:=1 to Order do begin
    it := OutTies[i].First;
  //  write(outf,i:d2,'. Outdegree = ',it.Value1:d2,'; ');
    x:= format('%*d%s%*d%s',[d2,i,'. Outdegree = ',d2,it.Value1,'; ']);
    outf.Write(pchar(x)^,length(x));
    while it.HasNext do begin
      it := it.Next;
      if graph then begin
        x := format('%2d%s',[it.Value1,' ']);
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := format('%s%*d%s%*d%s',[d2,'(',it.Value1,':',d1,it.Value2,') ']);
        outf.Write(pchar(x)^,length(x));
      end;
     end;
x := format('%s%d%s%d%s',[' (Last ',OutTies[i].Last.Value1,':',
OutTies[i].Last.Value2,')'])+eol;
   outf.Write(pchar(x)^,length(x));
//    writeln(outf);
   end; (* for i *)
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  if graph then begin
    x := 'In-ties:'+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 'In-ties. nodes:values'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  for i:=1 to Order do begin
    it := InTies[i].First;
    x:= format('%*d%s%*d%s',[d2,i,'. Indegree = ',d2,it.Value1,'; ']);
    outf.Write(pchar(x)^,length(x));
    while it.HasNext do begin
      it := it.Next;
      if graph then begin
        x := format('%*d%s',[d2,it.Value1,' ']);
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := format('%s%*d%s%*d%s',[d2,'(',it.Value1,':',d1,it.Value2,') ']);
        outf.Write(pchar(x)^,length(x));
      end;
     end;
  x:= format('%s%d%s%d%s',[' (Last ',InTies[i].Last.Value1,':',
    InTies[i].Last.Value2,')'])+eol;
    outf.Write(pchar(x)^,length(x));
//    writeln(outf);
   end; (* for i *)
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
end; (* WriteLists *)

procedure TValuedDigraph.WriteEdgeList(const uppertriangle: boolean; var outf: textfile);
var i,m,d1,d2: integer;
    it:TDoublyLinkedDoubleInt;
begin
  m := maxi;
  d1 := -mini;
  if (d1 > m) then m := d1;
  if (d1 > 0) then m := 10*m;
  d1 := 1;
  if (m >= 10) then d1 := 3;
  if (m >= 100) then d1 := 4;
  m := Order;
  case m of
    1..9: d2 := 1;
    10..99: d2 := 2;
    100..999: d2 := 3;
    1000..9999: d2 := 4;
    10000..99999: d2 := 5;
    else d2 := 6;
   end;
  for i:=1 to Order do begin
    it := OutTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      if (not uppertriangle) or (i <= it.Value1) then
        writeln(outf,i:d2,' ',it.Value1:d2,' ',it.Value2:d1);
     end;
   end; (* for i *)
end; (* WriteEdgeList *)

procedure TValuedDigraph.WritePajekFile(const directed: boolean; const fn: shortstring);
(* writes data to file in Pajek format *)
var f: textfile;
  i: integer;
  it:TDoublyLinkedDoubleInt;
begin
  assignfile(f,fn);
  rewrite(f);
  writeln(f,'*Vertices ',order);
  for i:=1 to order do
    begin
    writeln(f,i:3,' ',char(39),i,char(39));
    end;
  if (maxi <= 1) and (mini >= 0) then
    begin
    if directed then writeln(f,'*Arcslist') else writeln(f,'*Edgeslist');
    for i:=1 to order do
      begin
      it := OutTies[i].First;
      if it.hasnext then
        begin
        write(f,i:3);
        while it.HasNext do
          begin
          it := it.Next;
          if (directed or (it.Value1 >= i)) then write(f,' ',it.Value1:3);
          end;
        writeln(f);
        end;
      end;
    end
   else
    begin
    if directed then writeln(f,'*Arcs') else writeln(f,'*Edges');
    for i:=1 to order do
      begin
      it := OutTies[i].First;
      if it.hasnext then
      while it.HasNext do
        begin
        it := it.Next;
        if (directed or (it.Value1 >= i)) then
          writeln(f,i:3,' ',it.Value1:3,' ',it.Value2:3);
        end;
      end;
    end;
  closefile(f);
end;


procedure TValuedDigraph.ReadPajekFile(const fn: shortstring;
                                       var ertext: shortstring; var extra: boolean);
(* reads data to file in Pajek format;
   extra is a flag for seeing nodes larger than numbernodes. *)
const maxkey = 6;
var keys : array[1..maxkey] of shortstring;
  signs: set of char;
  f: textfile;
  thiskey, nextkey, numb, i, j: integer;
  thisstring: shortstring;
  v: shortint;
  count: integer;
  function readnextstring: shortstring;
  (* Gives as result the next string read
     until end of line or any symbol not in the set <<signs>>,
     which contains all letters and digits and '*';
     letters are output as uppercase. *)
  var s: shortstring;
      a: char;
  begin
    s := '';
    repeat
      read(f,a);
      if (a in signs) then s := s+UpCase(a);
      until (eoln(f) or (not (a in signs)));
    readnextstring := s;
  end; (* readnextstring *)
  function readnextnumber: integer;
  (* Gives as result the next string read
     until end of line or any symbol not in the set <<signs>>,
     which contains all letters and digits and '*';
     letters are output as uppercase. *)
  var s: shortstring;
      a: char;
      num,h: integer;
  begin
    s := '';
    readnextnumber := -1;
    if eoln(f) then exit;
    repeat
      read(f,a);
      if (a in ['0'..'9']) then s := s+a;
      until (not (a in ['0'..'9'])) or eoln(f);
    if (s > '') then
      begin
      val(s,num,h);
      if (h = 0) then readnextnumber := num;
      end;
  end; (* readnextnumber *)
  procedure interpretnextstring(const currentkey: integer; var newkey: integer;
                                var num: integer; var s: shortstring);
  (* reads the next string;
     if there is no new string, newkey = 0 on output;
     if a new key is read, this is represented by output values 1-5 of newkey
       and output value -1 of num;
     if a new number is read, it is output as num.
     an error yields newkey = -1.
     processed as indicated by currentkey.
  *)
  var a: char;
    h: integer;
  begin
    num := -1;
    newkey := currentkey;
    repeat
      s := readnextstring;
      if eoln(f) then readln(f);
      until (s > '') or eof(f);
    if (s = '') then
      begin (* the file contains nothing any more *)
      newkey := 0;
      end
     else
      begin (* string s nonempty *)
      if Copy(s,1,1) = '*' then
        begin
        for h:=1 to maxkey do if (s = keys[h]) then newkey := h;
        end
       else
        begin (* string does not start with * *)
        val(s,num,h);
        if (h >= 1) then
          begin
          ertext := 'String <'+s+'> is read but not recognized.';
          newkey := -1;
          end;
        end;  (* string does not start with * *)
      end;  (* string s nonempty *)
  end; (* interpretnextstring *)
  procedure interpreterror(var tk,nk: integer);
  begin
  if (tk >= 1) and (nk >= 1) and (tk <= maxkey) and (nk <= maxkey) then
  ertext := 'incomplete '+keys[tk]+' line: terminated by '+keys[nk]
    else
  ertext := 'read '+thisstring+'; unknown keys '+inttostr(tk)+' and '+inttostr(nk);
  tk := -1;
  end;
begin  (* ReadPajekFile *)
  count := 0;
  extra := false;
  keys[1] := '*VERTICES';
  keys[2] := '*ARCS';
  keys[3] := '*EDGES';
  keys[4] := '*ARCSLIST';
  keys[5] := '*EDGESLIST';
  keys[6] := '*MATRIX';
  ertext := '';
  signs := ['*']+['0'..'9']+['A'..'Z']+['a'..'z'];
  assign(f,fn);
  reset(f);
  thiskey := 1;
  while (thiskey >= 1) do
    begin
(* start of a line *)
    repeat
      interpretnextstring(thiskey,nextkey,numb,thisstring);
      inc(count);
      if (nextkey = 0) then readln(f);
      until (nextkey <> 0) or eof(f);
    if eof(f) then nextkey := 0;
    thiskey := nextkey;
    case thiskey of
          1:  if not eoln(f) then readln(f); (* vertices: no action *)
          2,3: (* arcs / edges: read i, j, v and add this to the TValuedDigraph *)
              begin
              if (numb <= -1) then
                begin
(* Only the arcs command was read; now the first of the next line *)
                interpretnextstring(thiskey,nextkey,numb,thisstring);
                if (nextkey <> thiskey) then interpreterror(thiskey,nextkey);
                end;
(* Else numb is already the first of the next line. *)
              if (ertext <= '') then
                begin
                i := numb;
                interpretnextstring(thiskey,nextkey,numb,thisstring);
                if (nextkey <> thiskey) then interpreterror(thiskey,nextkey)
                else
                  begin
                  j := numb;
                  interpretnextstring(thiskey,nextkey,numb,thisstring);
                  if (nextkey <> thiskey) then interpreterror(thiskey,nextkey)
                  else
                    begin
                    if (numb >= -128) and (numb <= 127) then
                      begin
                      v := numb;
                      if (i <> j) then
                        begin
                        if (i > NumberNodes) or (j > NumberNodes) then extra := true
                         else
                          begin
                          ChangeTieValue(i,j,v);
                          if (thiskey = 3) then ChangeTieValue(j,i,v);
                               (* key is *Edges: nondirected *)
                          end;
                        end;
                      end
                     else
                      begin
                      ertext := 'value of tie('+inttostr(i)+','+inttostr(j)+
                                ') read as '+inttostr(v)+', which is not a shortint';
                      thiskey := -1;
                      end;
                    end;
                  end;
                end;
              end;
          4,5:  begin
(* arcslist *)
              if (numb <= -1) then
                begin
(* Only the arcslist command was read; now the first of the next line *)
                interpretnextstring(thiskey,nextkey,numb,thisstring);
                if (nextkey <> thiskey) then interpreterror(thiskey,nextkey);
                end;
(* Else numb is already the first of the next line. *)
              if (ertext <= '') then
                begin
                i := numb;

                while not eoln(f) do
                  begin
                  j := readnextnumber;
                  if (j >= 1) then
                    begin
                    if (i <> j) then
                      begin
                        if (i > NumberNodes) or (j > NumberNodes) then extra := true
                         else
                          begin
                          ChangeTieValue(i,j,1);
                          if (thiskey = 5) then ChangeTieValue(j,i,1);
                               (* key is *Edgeslist: nondirected *)
                          end;
                      end;
                    end;
                  end; (* while *)
                end;
              end ;
          6:  begin
(* *Matrix *)
              for i:=1 to order do
                begin
                for j:=1 to order do
                  begin
                  read(f,v);
                  if (i <> j) then
                    begin
                    ChangeTieValue(i,j,v);
                    end;
                  if eof(f) and (ertext <= '') then
                    ertext := 'in *Matrix option, the file ended when trying to read element ('+inttostr(i)+','+inttostr(j)+')';
                  end;
                readln(f);
                end;
              end;
       end; (* case *)
    end;
  closefile(f);
end; (* ReadPajekFile *)


procedure TValuedDigraph.WriteToFile(const fn: shortstring);
(* Writes data in order-dependent format to file with name fn. *)
var f: textfile;
begin
  assign(f,fn);
  rewrite(f);
  if (Order <= 40) then begin
    writeln(f,'Siena data file: adjacency matrix');
    writeAdjacency(f);
   end (* if order <= 40 *) else begin
    writeln(f,'Siena data file: edge list');
    WriteEdgeList(false, f);
   end; (* if order > 80 *)
  closefile(f);
end; (* WriteToFile *)

procedure TValuedDigraph.ReadFromFile(const n: integer; const fn: shortstring);
var f: textfile;
   fl: shortstring;
   h: shortint;
   i,j: integer;
begin
  if (n > Order) then begin
    ErMes := 'ReadFromFile called for n = '+inttostr(n)+' > '+inttostr(Order);
    exit;
   end;
  SetZero;
  assign(f,fn);
  reset(f);
  readln(f,fl);
  fl := trimright(fl);
  if (fl ='Siena data file: adjacency matrix') then begin
    for i:=1 to n do begin
      for j:=1 to n do begin
        read(f,h);
        if ((j <> i) or DiagonalPermitted) and (h <> 0) then AddTie(i,j,h,true);
       end; (* j *)
      readln(f);
     end; (* i *)
   end (* if adjacency matrix *) else
    if (fl = 'Siena data file: edge list') then begin
      while (not eof(f)) do begin
        readln(f,i,j,h);
        if (i >= 1) and (i <= n) and (j >= 1) and (j <= n) and
           (j <> i) and (h <> 0) then AddTie(i,j,h,true)
           else ErMes :=
           'incorrect data line '+inttostr(i)+' '+inttostr(j)+' '+inttostr(h)+' (n = '+inttostr(n)+')';
       end; (* while *)
   end; (* if edge list *)
  closefile(f);
end; (* ReadFromFile *)


procedure TValuedDigraph.WriteToArray(var z: iiarray);
(* Writes adjacency matrix. *)
var i,j,m,d1,d2: integer;
    it:TDoublyLinkedDoubleInt;
begin
  m := Order;
  SetLength(z,m+1,m+1);
  for i:= m downto 0 do for j := m downto 0 do z[i,j] := 0;
  for i:=1 to m do begin
    it := OutTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      z[i,it.Value1] := it.Value2;
     end;
   end; (* for i *)
end; (* WriteToArray *)


procedure TValuedDigraph.ComponentOne(var d,z: iarray);
(* output z = upper bounds to distances in max-symmetrized to
              node with highest symmetrized degree;
   infinite distances are Numbernodes.            *)
var depthmax,i,dmax,imax,nd: integer;
    it:TDoublyLinkedDoubleInt;
//    b: TValuedDigraph;
procedure recursive_newi(j: integer);
(* Adds to z the vertices to which j is tied and that were not yet in z,
   and recursively calls itself for the added vertices *)
var  itr:TDoublyLinkedDoubleInt;
   dist: integer;
begin
//  itr := b.OutTies[j].FFirst;
  itr := OutTies[j].FFirst;
  dist := z[j];
  while itr.HasNext do
    begin
    itr := itr.FNext;
    if (z[itr.Value1] >= NumberNodes) and (dist < depthmax) then
      begin
      z[itr.Value1] := succ(dist);
      recursive_newi(itr.Value1);
      end;
    end;
end; (* recursive_newi *)
begin (* ComponentOne *)
  depthmax := NumberNodes-1;
  setlength(d,NumberNodes+1);
  setlength(z,NumberNodes+1);
//  b := TValuedDigraph.Create(self.NumberNodes);
//  b.PermitDiagonal;
//  Symmetrize(self,b,2);
//  b.WritePajekFile(true,'b.dat');
  dmax := 0;
  imax := 1;
  for i := NumberNodes downto 1 do
    begin
//    if b.OutDegree(i) > dmax then
      if OutDegree(i) > dmax then
      begin
//      dmax := b.OutDegree(i);
      dmax := OutDegree(i);
      imax := i;
      end;
    end;
  for i:=NumberNodes downto 1 do
    begin
//    d[i] := b.OutDegree(i);
    d[i] := OutDegree(i);
    z[i] := NumberNodes;
    end;
  z[imax] := 0;
  recursive_newi(imax);
  nd := 0;
  for i:=1 to NumberNodes do
    if (z[i] < NumberNodes) then
      begin
      inc(nd);
      d[i] := nd;
      end;
   z[0] := nd;
end;

function TValuedDigraph.Connected(var i: integer): boolean;
(* output = {from i there are paths to all nodes } *)
var j: integer;
    it:TDoublyLinkedDoubleInt;
    z: boarray;
    conn: boolean;
procedure recursive_newzi(j: integer);
(* Adds to z the vertices to which j is tied and that were not yet in z,
   and recursively calls itself for the added vertices *)
var  itr:TDoublyLinkedDoubleInt;
begin
  itr := OutTies[j].FFirst;
  while itr.HasNext do
    begin
    itr := itr.FNext;
    if (z[itr.Value1]) then
      begin
      z[itr.Value1] := false;
      recursive_newzi(itr.Value1);
      end;
    end;
end; (* recursive_newzi *)
begin (* Connected *)
(* z[j] = false for all j that have been visited on a path from i *)
  setlength(z,NumberNodes+1);
  for j:= NumberNodes downto 1 do z[j] := true;
  z[i] := false;
  recursive_newzi(i);
  conn := true;
  j := NumberNodes;
  repeat
    conn := not z[j];
    dec(j);
   until (not conn) or (j <= 0);
  result := conn;
end;

function TValuedDigraph.Path(var i,j: integer): boolean;
(* output = (there is a path from i to j);
    note that for i=j and symmetric this is equivalent to tie(i,j). *)
var k: integer;
    it:TDoublyLinkedDoubleInt;
    z: boarray;
    pathfound: boolean;
procedure recursive_newpi(k: integer);
(* Adds to z the vertices to which k is tied and that were not yet in z,
   and recursively calls itself for the added vertices *)
var  itr:TDoublyLinkedDoubleInt;
begin
  if PathFound then exit;
  if (k = j) then
    begin
    PathFound := true;
    exit;
    end;
  itr := OutTies[k].FFirst;
  while itr.HasNext do
    begin
    itr := itr.FNext;
    if (z[itr.Value1]) then
      begin
      z[itr.Value1] := false;
      recursive_newpi(itr.Value1);
      end;
    end;
end; (* recursive_newpi *)
begin (* Path *)
(* z[k] = false for all k that have been visited on a path from i *)
  setlength(z,NumberNodes+1);
  for k:= NumberNodes downto 1 do z[k] := true;
  z[i] := (j=i);
  pathfound := false;
  recursive_newpi(i);
  Result := PathFound;
end;

function TValuedDigraph.OtherPath(var i,j: integer): boolean;
(* output = (there is a path from i to j that does not use the link i-j ) *)
var k: integer;
    it:TDoublyLinkedDoubleInt;
    z: boarray;
    pathfound: boolean;
procedure recursive_newpi(kk: integer);
(* Adds to z the vertices to which k is tied and that were not yet in z,
   and recursively calls itself for the added vertices *)
var  itrr:TDoublyLinkedDoubleInt;
begin
  if PathFound then exit;
  if (kk = j) then
    begin
    PathFound := true;
    exit;
    end;
  itrr := OutTies[kk].FFirst;
  while itrr.HasNext do
    begin
    itrr := itrr.FNext;
    if (z[itrr.Value1]) then
      begin
      z[itrr.Value1] := false;
      recursive_newpi(itrr.Value1);
      end;
    end;
end; (* recursive_newpi *)
procedure first_newpi;
(* Adds to z the vertices to which k is tied and that were not yet in z,
   and calls recursive+newpi for the added vertices *)
var  itr:TDoublyLinkedDoubleInt;
begin
  itr := OutTies[i].FFirst;
  while itr.HasNext do
    begin
    itr := itr.FNext;
    if (z[itr.Value1]) and (itr.Value1 <> j) then
      begin
      z[itr.Value1] := false;
      recursive_newpi(itr.Value1);
      end;
    end;
end; (* first_newpi *)
begin (* OtherPath *)
(* z[k] = false for all k that have been visited on a path from i *)
if (i <= 0) or (i > NumberNodes) then writeln;
  setlength(z,NumberNodes+1);
  for k:= NumberNodes downto 1 do z[k] := true;
  z[i] := false;
  pathfound := false;
  first_newpi;
  Result := PathFound;
end;

procedure CopyGraph(const a: TValuedDigraph; var b: TValuedDigraph);
(* copies a to b; b must have been created earlier. *)
var i: integer;
    it:TDoublyLinkedDoubleInt;
begin
  if b.NumberNodes < a.NumberNodes then begin
    b.ErMes := 'CopyGraph called with source graph with '+inttostr(a.numbernodes)+' nodes';
    exit;
   end;
  b.DiagonalPermitted := true;
  b.SetZero;
  for i:=1 to a.NumberNodes do begin
    it := a.OutTies[i].first;
    while it.HasNext do begin
      it := it.next;
      b.AddTie(i,it.Value1,it.Value2,true);
     end;
   end;
  b.DiagonalPermitted := a.DiagonalPermitted;
end; (* CopyGraph *)

procedure Symmetrize(var a, b: TValuedDigraph; const how: integer);
(* defines b = symmetrized a;
   by min for how = 1, by max for how = 2;
   b must have been created earlier. *)
var i,j: integer;
    it1, it2: TDoublyLinkedDoubleInt;
    v1, v2: shortint;
begin
  if b.NumberNodes < a.NumberNodes then begin
    b.ErMes := 'Symmetrize called with source graph with '+inttostr(a.numbernodes)+' nodes';
    exit;
   end;
  b.DiagonalPermitted := false;
  b.SetZero;
  for i:=1 to a.NumberNodes do begin
    it1 := a.OutTies[i].First;
    it2 := a.InTies[i].First;
    j := 0;
(* changed 3.181 *)
      while (it1.hasnext and it2.hasnext and (j < i)) do
       case sign(it1.next.value1 - it2.next.value1) of
       -1: begin
            it1 := it1.next;
            j := it1.Value1;
            if (how = 2) and (j < i) then
              begin
              b.ChangeTieValue(i,j,it1.value2);
              b.ChangeTieValue(j,i,it1.value2);
              end;
          end;
       0: begin
           it1 := it1.next;
           it2 := it2.next;
           v1 := it1.value2;
           v2 := it2.value2;
           j := it1.Value1;
           if (j < i) then
           case how of
             1: begin
                if (v1 > v2) then v1 := v2;
                b.ChangeTieValue(i,it1.Value1,v1);
                b.ChangeTieValue(it1.Value1,i,v1);
                end;
             2: begin
                if (v2 > v1) then v1 := v2;
                b.ChangeTieValue(i,it1.Value1,v1);
                b.ChangeTieValue(it1.Value1,i,v1);
                end;
              end; (* case how *)
          end; (* case 0 *)
       1: begin
           it2 := it2.next;
           j := it2.Value1;
            if (how = 2) and (j < i) then
              begin
              b.ChangeTieValue(i,j,it2.value2);
              b.ChangeTieValue(j,i,it2.value2);
              end;
           end;
      end; (* case *)

    if (how = 2) then
      begin
      while (it1.hasnext and (j < i)) do
        begin
        it1 := it1.next;
        j := it1.Value1;
        if (how = 2) and (j < i) then
          begin
          b.ChangeTieValue(i,j,it1.value2);
          b.ChangeTieValue(j,i,it1.value2);
          end;
        end; (* while *)
      while (it2.hasnext and (j < i)) do
       begin
       it2 := it2.next;
       j := it2.Value1;
        if (how = 2) and (j < i) then
          begin
          b.ChangeTieValue(i,j,it2.value2);
          b.ChangeTieValue(j,i,it2.value2);
          end;
       end;
     end; (* if how  = 2 *)

   end; (* i *)
 if (not b.SymmetricGraph) then
   b.ErMes := 'Symmetrization did not lead to symmetric graph.';
end; (* Symmetrize *)

procedure BooleanProduct(var a, b, c: TValuedDigraph; const how: integer);
(* defines c = boolean product a*b;
   c must have been created earlier.
   if how = 2, c = boolean a*(b+identity) *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  if (c.NumberNodes < a.NumberNodes) or (c.NumberNodes < b.NumberNodes) then begin
    c.ErMes := 'BooleanProduct called with source graphs with '
       +inttostr(a.numbernodes)+', '+inttostr(b.numbernodes)+' nodes';
    exit;
   end;
  c.DiagonalPermitted := false;
  c.SetZero;
  for i:=1 to a.NumberNodes do
    begin
    it1 := a.OutTies[i].First;
    while it1.HasNext do
      begin
      it1 := it1.next;
      if (how = 2) then
        if (it1.Value1 <> i) then c.ChangeTieValue(i,it1.Value1,1);
      it2 := b.OutTies[it1.Value1].First;
      while it2.HasNext do
        begin
        it2 := it2.Next;
        if (i <> it2.value1) then c.ChangeTieValue(i,it2.Value1,1);
        end;
      end;
    end;
end; (* BooleanProduct *)


procedure Simmelian(var a, b: TValuedDigraph);
(* defines b = Simmelian ties in a;
   b must have been created earlier. *)
var c,d: TValuedDigraph;
begin
  c := TValuedDigraph.Create(a.NumberNodes);
  d := TValuedDigraph.Create(a.NumberNodes);
  Symmetrize(a,c,1);
  BooleanProduct(c,c,d,1);
  YesYes(b,c,d);
  c.Destroy;
  d.Destroy;
end; (* Simmelian *)

procedure CopyGraph(const a: TDigraph; var b: TDigraph);
(* copies a to b; b must have been created earlier. *)
var i: integer;
    it:TDoublyLinkedDoubleInt;
begin
  if b.NumberNodes < a.NumberNodes then begin
    b.ErMes := 'CopyGraph called with source graph with '+inttostr(a.numbernodes)+' nodes';
    exit;
   end;
  b.SetZero;
  for i:=1 to a.NumberNodes do begin
    it := a.OutTies[i].first;
    while it.HasNext do begin
      it := it.next;
      b.AddTie(i,it.Value1);
     end;
   end;
  b.DiagonalPermitted := a.DiagonalPermitted;
end; (* CopyGraph *)

function RestrictedSum(const a: TValuedDigraph; const b: TDigraph): integer;
(* sum_ij a[i,j]*b[i,j] *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  result := 0;
  if b.NumberNodes < a.NumberNodes then
    begin
    b.ErMes := 'RestrictedSum called with source graph with different numbers of nodes';
    exit;
    end;
  for i:=a.NumberNodes downto 1 do
    begin
    it1 := a.OutTies[i].first;
    it2 := b.OutTies[i].first;
    while (it1.hasnext and it2.hasnext) do
     case sign(it1.next.value1 - it2.next.value1) of
      -1: it1 := it1.next;
      0: begin
           it1 := it1.next;
           it2 := it2.next;
           Inc(Result,it1.Value2);
         end; (* case 0 *)
      1: it2 := it2.next;
     end; (* case *)
    end;
end; (* RestrtictedSum *)


function EqualGraph(const a: TDigraph; var b: TDigraph): boolean;  // added 3.17s
(* a = b *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  if not (a.NumberNodes = b.NumberNodes) then
    begin
    Result := false;
    exit;
    end
   else Result := true;
  for i:= a.NumberNodes downto 1 do
    begin
    it1:= a.OutTies[i].Ffirst;
    it2:= b.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
             Result := false;
             exit;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
             Result := false;
             exit;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)

      if (it1.HasNext or it2.HasNext) then
        begin
        Result := false;
        exit;
        end;
    end; (* for i *)
end; (* EqualGraph *)


function NewArcs(a,b: TDigraph): boolean; // added 3.17s
(* b has arcs that are not in a *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  if not (a.NumberNodes = b.NumberNodes)  then
    begin
    Result := true;
    exit;
    end
   else Result := false;
  for i:= a.NumberNodes downto 1 do
    begin
    it1:= a.OutTies[i].Ffirst;
    it2:= b.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
             Result := true;   // changed 3.17x
             exit;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)

      if it2.HasNext then
        begin
        Result := true;
        exit;
        end;
    end; (* for i *)
end; (* NewArcs *)




function JointIn(const a,b: TValuedDigraph; const k: integer): integer;
(* sum_i a(i,k)b(i,k) *)
var ji, j: integer;
  ita, itb:TDoublyLinkedDoubleInt;
begin
  ita := a.InTies[k].first;
  itb := b.InTies[k].first;
  ji := 0;
  while (ita.hasnext and itb.HasNext) do
    begin
      case sign(ita.next.value1 - itb.next.value1) of
      -1: begin
           ita := ita.next;
          end; (* case -1 *)
      0: begin
           ita := ita.next;
           itb := itb.next;
           Inc(ji,ita.Value2*itb.Value2);
         end; (* case 0 *)
      1: begin
           itb := itb.next;
          end; (* case 1 *)
      end; (* case *)
    end; (* while (ita.hasnext and itb.hasnext) *)
  JointIn := ji;
end; (* JointIn *)

{                                                                              }
{ TDigraph                                                                     }
{                                                                              }

function TDigraph.IsCorrect: Boolean;
var i,j,v1: integer;
    v2: ShortInt;
    it:TDoublyLinkedDoubleInt;
begin
  Result := inherited IsCorrect;
(* Check values in linked lists *)
  for i:=1 to Order do for j:=1 to 2 do begin
    if (j <= 1) then it := OutTies[i].First else it := InTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      v1 := it.Value1;
      v2 := it.Value2;
      if (v2 <> 1) then begin
        Result := false;
        if (j=1) then ErMes := 'Out' else ErMes := 'In';
        ErMes := ErMes+'Ties('+inttostr(i)+') contains values '+inttostr(v1)+
             ', '+inttostr(v2);
        exit;
       end;
     end; (* while *)
   end; (* for i,j *)
end;

procedure TDigraph.AddTie(const i,j: integer);              // added 3.17s
(* This procedure adds tie (i,j),
   assuming that in the adjacency matrix
   both in row i and in column j this is the largest element. *)
var  OK: Boolean;
begin
(* Check : *)
  if ((i <= 0) or (i > NumberNodes) or (j <= 0) or (j > NumberNodes))
      or ((i = j) and (not DiagonalPermitted)) then begin
    ErMes := 'call AddTie('+inttostr(i)+','+inttostr(j)+')';
    exit;
   end;

  OK := (OutTies[i].Last.Value1 < j) and (InTies[j].Last.Value1 < i);
  if OK then
    begin
    OutTies[i].Append(j,1);
    InTies[j].Append(i,1);
    Inc(OutTies[i].Ffirst.Value1);
    Inc(InTies[j].Ffirst.Value1);
    if HasArray then AdjMatrix[i,j] := 1;
    end (* if OK *)
   else
    begin
    ChangeTieValue(i,j,1);
    end;
end; (* AddTie *)



procedure TDigraph.ChangeTieValue(const i,j: integer; const v: shortint);
begin
  if (v < 0) or (v > 1) then begin
    ErMes := 'call ChangeTieValue('+inttostr(i)+','+inttostr(j)+','+inttostr(v)+')';
    exit;
   end;
  inherited ChangeTieValue(i,j,v);
end;

function TDigraph.IncreaseTie(const i,j: integer; v: shortint): ShortInt;
var v1: Shortint;
begin
  v1 := inherited IncreaseTie(i,j,v);
  Result := v1;
  if (v1 < 0) or (v1 > 1) then begin
    ErMes := 'Digraph.IncreaseTie('+inttostr(i)+','+inttostr(j)+','+inttostr(v)+
             ') leads to new value '+inttostr(v1);
    exit;
   end;
end; (* IncreaseTie *)

procedure ChangeValues(const a,b: array of shortint; const k: byte;
                       var dg1: TValuedDigraph; var dg2: TDigraph);
(* For all (i,j), if dg1(i,j)=a[h] for some h in [0..k],
   changes dg1(i,j) into b[h] and sets dg2(i,j) to 1. *)
var it,it0: TDoublyLinkedDoubleInt;
 i,j,h: integer;
 v: shortint;
begin
  for i :=  dg1.NumberNodes downto 1 do
    begin
    it := dg1.OutTies[i].Ffirst;
    while it.HasNext do
      begin
      v := it.next.Value2;
      h := k;
      repeat
        if v = a[h] then
          begin
          j := it.next.Value1;
//writeln(checkfile,i:3,' ',j:3,' ',v:3,' =>',b[h]:2);
          if (b[h] = 0) then
            begin
            it0 := it.Next;
            if dg1.HasArray then dg1.AdjMatrix[i,it0.Value1] := 0;
            dg1.OutTies[i].Delete(it0);
            Dec(dg1.OutTies[i].Ffirst.Value1);
            end
           else
            begin
            it := it.Next;
            it.Value2 := b[h]; // 1234
            if dg1.HasArray then dg1.AdjMatrix[i,it.Value1] := b[h];
            end;
          h := -2;
          dg2.AddTie(i,j);
          end
         else
          dec(h);
      until (h < 0);
      if (h = -1) then it := it.Next;
      end;
    end;
  for i :=  dg1.NumberNodes downto 1 do
    begin
    it := dg1.InTies[i].Ffirst;
    while it.HasNext do
      begin
      v := it.next.Value2;
      h := k;
      repeat
        if v = a[h] then
          begin
          j := it.next.Value1;
          if (b[h] = 0) then
            begin
            it0 := it.Next;
            dg1.InTies[i].Delete(it0);
            Dec(dg1.InTies[i].Ffirst.Value1);
            end
            else
            begin
            it := it.Next;
            it.Value2 := b[h]; // 1234
            end;
          h := -2;
          end
         else
          dec(h);
      until (h < 0);
      if (h = -1) then it := it.Next;
      end;
    end; (* i *)
end; (* ChangeValues *)

procedure YesNoNo(var dg: TValuedDigraph;
                  const d1: TValuedDigraph; const d2, d3: TDigraph);
(* Creates dg, with elements defined as
    dg(i,j) = d1(i,j)(1-d2(i,j))(1-d3(i,j)) *)
var i, j, nn, j2, j3: integer;
    it1, it2, it3: TDoublyLinkedDoubleInt;
begin
  nn := d1.Order;
  dg := TValuedDigraph.Create(nn);
  for i:= 1 to nn do
    begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    it3:= d3.OutTies[i].Ffirst;
(* j2 will be the smallest Value1 in d2.OutTies[i]
   that is at least as large as the current it1.Value1 = j;
   and nn+1 if there is no such Value1;
   i.e., j2 is the smallest value with d2(i,j2) <> 0 larger than j,
   and nn+1 if this j2 does not exist.
   similar for j3 in d3.OutTies[i] *)
    if it2.HasNext then
      begin
      it2 := it2.Next;
      j2 := it2.Value1;
      end
     else j2 := succ(nn);
    if it3.HasNext then
      begin
      it3 := it3.Next;
      j3 := it3.Value1;
      end
     else j3 := succ(nn);
    while it1.HasNext do
      begin
      it1 := it1.Next;
      j := it1.Value1;
      if (j2 < j) then
        begin
        while ((it2.Value1 < j) and it2.HasNext) do it2 := it2.Next;
        j2 := it2.Value1;
        if (j2 < j) then j2 := succ(nn);
        end;
      if (j3 < j) then
        begin
        while ((it3.Value1 < j) and it3.HasNext) do it3 := it3.Next;
        j3 := it3.Value1;
        if (j3 < j) then j3 := succ(nn);
        end;
{
if (d2.TieValue(i,j) >= 1) and (j2 <> j) then
  begin
  writeln(checkfile,'graph d2 in YesNoNo:');
  writeln(checkfile,'i = ',i:3,' ','j = ',j:3,' ','j2 = ',j2:3,'. ');
  d2.OutTies[i].WriteVector(checkfile,0);
  end;
if (d3.TieValue(i,j) >= 1) and (j3 <> j) then
  begin
  writeln(checkfile,'graph d3 in YesNoNo:');
  writeln(checkfile,'i = ',i:3,' ','j = ',j:3,' ','j3 = ',j3:3,'. ');
  d3.OutTies[i].WriteVector(checkfile,0);
  end;
}
      if ((j2 > j) and (j3 > j)) then
//this is the same as:
//  if (not ((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1))) then
        dg.AddTie(i,j, it1.Value2,true);
     end; (* while *)
   end; (* for i *)
end; (* YesNoNo *)

procedure YesNo(var dg: TDigraph; const d1, d2: TDigraph);    // added 3.17s
(* Fills dg with elements defined as
    dg(i,j) = d1(i,j)(1-d2(i,j)) *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  dg.SetZero;
  for i:= d1.NumberNodes downto 1 do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
             dg.AddTie(i,it1.value1);
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)

    while it1.hasnext do
      begin
      it1 := it1.next;
      dg.AddTie(i,it1.value1);
      end; (*  while (it1.hasnext or it2.hasnext) *)
  end; (* for i *)
end; (* YesNo *)

procedure YesYes(var dg: TValuedDigraph; const d1, d2: TValuedDigraph);
(* Fills dg with elements defined as
    dg(i,j) = (d1(i,j) <> 0) and (d2(i,j) <> 0) *)
var i: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  dg.SetZero;
  for i:= 1 to d1.NumberNodes do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             dg.AddTie(i,it1.value1,1,true);
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (it1.hasnext and it2.hasnext) *)
  end; (* for i *)
end; (* YesYes *)


procedure DifNoNo(var dgdif: TValuedDigraph;
                  const da,db: TValuedDigraph; const d2, d3, d4, d5: TDigraph);
(* Creates dg, with elements defined as
    dg(i,j) = abso(da(i,j)-db(i,j))(1-d2(i,j))(1-d3(i,j)) *)
var i, j, j2, j3, j4, j5, v, nn: integer;                 //CS: j4, j5 added for structural considerations
    ita, itb, it2, it3, it4, it5: TDoublyLinkedDoubleInt; //CS: it4, it5 added for structural considerations

  procedure Advance_it2345; //CS: added the "45" to ensure non-counting of the structural values
  begin
        if (j2 < j) then
          begin
          while ((it2.Value1 < j) and it2.HasNext) do it2 := it2.Next;
          j2 := it2.Value1;
          if (j2 < j) then j2 := succ(nn);
          end;
        if (j3 < j) then
          begin
          while ((it3.Value1 < j) and it3.HasNext) do it3 := it3.Next;
          j3 := it3.Value1;
          if (j3 < j) then j3 := succ(nn);
          end;
        if (j4 < j) then
          begin
          while ((it4.Value1 < j) and it4.HasNext) do it4 := it4.Next;
          j4 := it4.Value1;
          if (j4 < j) then j4 := succ(nn);
          end;
        if (j5 < j) then
          begin
          while ((it5.Value1 < j) and it5.HasNext) do it5 := it5.Next;
          j5 := it5.Value1;
          if (j5 < j) then j5 := succ(nn);
          end;

  end; (* Advance_it2345 *)

begin
  nn := da.Order;
  dgdif := TValuedDigraph.Create(nn);
  for i:= 1 to nn do begin
    ita:= da.OutTies[i].Ffirst;
    itb:= db.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    it3:= d3.OutTies[i].Ffirst;
    it4:= d4.OutTies[i].Ffirst;
    it5:= d5.OutTies[i].Ffirst;
(* j2 will be the smallest Value1 in d2.OutTies[i]
   that is at least as large as the current it1.Value1 = j;
   and nn+1 if there is no such Value1;
   i.e., j2 is the smallest value with d2(i,j2) <> 0 larger than j,
   and nn+1 if this j2 does not exist.
   similar for j3 in d3.OutTies[i] *)
    if it2.HasNext then
      begin
      it2 := it2.Next;
      j2 := it2.Value1;
      end
     else j2 := succ(nn);
    if it3.HasNext then
      begin
      it3 := it3.Next;
      j3 := it3.Value1;
      end
     else j3 := succ(nn);
    if it4.HasNext then
      begin
      it4 := it4.Next;
      j4 := it4.Value1;
      end
     else j4 := succ(nn);
    if it5.HasNext then
      begin
      it5 := it5.Next;
      j5 := it5.Value1;
      end
     else j5 := succ(nn);
    while (ita.hasnext or itb.hasnext) do
      if (ita.hasnext and itb.hasnext) then
       case sign(ita.next.value1 - itb.next.value1) of
        -1: begin
             ita := ita.next;
             j := ita.value1;
             Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
              and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as             if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
              then dgdif.AddTie(i,j, abs(ita.value2),true);
            end; (* case -1 *)
        0: begin
             ita := ita.next;
             itb := itb.next;
             v := abs(ita.value2 - itb.value2);
             j := ita.value1;
             Advance_it2345;
             if ((v <> 0) and (j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
              and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as if (v <> 0) and not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
              then dgdif.AddTie(i,j,v,true);
           end; (* case 0 *)
        1: begin
             itb := itb.next;
             j := itb.value1;
             Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
              and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as  if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
              then dgdif.AddTie(i,j, abs(itb.value2),true);
            end; (* case 1 *)
         end (* case *)
        else (* not (ita.hasnext and itb.hasnext) *) begin
          if ita.hasnext then
            begin
            ita := ita.next;
            j := ita.value1;
            Advance_it2345;
            if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as  if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
             then dgdif.AddTie(i,j,abs(ita.value2),true)
            end
           else
            begin
            itb := itb.next;
            j := itb.value1;
            Advance_it2345;
            if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as  if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
             then dgdif.AddTie(i,j,abs(itb.value2),true)
            end;
         end; (* not (ita.hasnext and itb.hasnext) *)
   end; (* for i *)
end; (* DifNoNo *)

procedure CountDifNoNo(var s,sdown,sup: integer;
                  const da,db: TValuedDigraph; const d2, d3, d4, d5: TDigraph); // d2,d3: missingness; CS: d4,d5: structural fixation
(* Calculates
    s = sum(i,j) abso(da(i,j)-db(i,j))(1-d2(i,j))(1-d3(i,j))
    st = sum(i,j) (1-d2(i,j))(1-d3(i,j)) *)
(* Modeled after DifNoNo *)
var z, zdown, zup, i, j, j2, j3, j4, j5, v, nn: integer;  //CS: j4, j5 added for structural considerations
    ita, itb, it2, it3, it4, it5: TDoublyLinkedDoubleInt; //CS: it4, it5 added for structural considerations

  procedure Advance_it2345; //CS: added the "45" to ensure non-counting of the structural values
  begin
        if (j2 < j) then
          begin
          while ((it2.Value1 < j) and it2.HasNext) do it2 := it2.Next;
          j2 := it2.Value1;
          if (j2 < j) then j2 := succ(nn);
          end;
        if (j3 < j) then
          begin
          while ((it3.Value1 < j) and it3.HasNext) do it3 := it3.Next;
          j3 := it3.Value1;
          if (j3 < j) then j3 := succ(nn);
          end;
        if (j4 < j) then
          begin
          while ((it4.Value1 < j) and it4.HasNext) do it4 := it4.Next;
          j4 := it4.Value1;
          if (j4 < j) then j4 := succ(nn);
          end;
        if (j5 < j) then
          begin
          while ((it5.Value1 < j) and it5.HasNext) do it5 := it5.Next;
          j5 := it5.Value1;
          if (j5 < j) then j5 := succ(nn);
          end
  end; (* Advance_it2345 *)

begin
  z := 0; zdown := 0; zup := 0; // these values are returned below as s, sdown, sup; here initialised.
  nn := da.Order;
  for i:= 1 to nn
  do begin
    ita:= da.OutTies[i].Ffirst;
    itb:= db.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    it3:= d3.OutTies[i].Ffirst;
    it4:= d4.OutTies[i].Ffirst;
    it5:= d5.OutTies[i].Ffirst;
(* j2 will be the smallest Value1 in d2.OutTies[i]
   that is at least as large as the current it1.Value1 = j;              CS: I assume, 'it1' means 'ita' here?
   and nn+1 if there is no such Value1;
   i.e., j2 is the smallest value with d2(i,j2) <> 0 larger than j,
   and nn+1 if this j2 does not exist.
   similar for j3 in d3.OutTies[i], and so on also for j4 and j5 *)
    if it2.HasNext then
      begin
      it2 := it2.Next;
      j2 := it2.Value1;
      end
     else j2 := succ(nn);
    if it3.HasNext then
      begin
      it3 := it3.Next;
      j3 := it3.Value1;
      end
     else j3 := succ(nn);
    if it4.HasNext then
      begin
      it4 := it4.Next;
      j4 := it4.Value1;
      end
     else j4 := succ(nn);
    if it5.HasNext then
      begin
      it5 := it5.Next;
      j5 := it5.Value1;
      end
     else j5 := succ(nn);

    //CS: in the following, j is the next actor with nonzero tie from i to j in either d1 or in d2.
    //    j with d1[i,j]=0=d2[i,j] are jumped over by this procedure.
    while (ita.hasnext or itb.hasnext) do
      if (ita.hasnext and itb.hasnext) then
       case sign(ita.next.value1 - itb.next.value1) of
        -1: begin // next partner of i in d2 would have higher index than the next one in d1 ..
             ita := ita.next; // ..so only look for next partner in d1[i,.]
             j := ita.value1;
             Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as
//  if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
             then begin
               inc(z,abs(ita.value2)); // update distance by change between d1[i,j] and zero=d2[i,j]
               inc(zdown);             // and count this as a downward step, as d1 comes before d2.
             end;
            end; (* case -1 *)
        0: begin // next partner of i in d2 has same index as the next one in d1 ..
             ita := ita.next; // ..so look for next partners ..
             itb := itb.next; // ..in  both d1[i,.] and d2[i,.]
             v := ita.value2-itb.value2; //CS: this used to be "abs(ita.value2 - itb.value2);", which was inconsistent with the line below (see "HELA!"-remark).
             j := ita.value1;
             Advance_it2345;
             if ((v <> 0)
             and (j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as
// if (v <> 0) and not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
             then begin
               inc(z,abs(v)); //CS: this used to be "inc(z,v);", consistencised now with change above // update distance by change between d1[i,j] and d2[i,j]
               if (v > 0) then inc(zdown) else inc(zup); //CS: HELA!! v was prior to my changes defined as an absolute value above, leading to "wrong counting" here.
             end;
           end; (* case 0 *)
        1: begin // next partner of i in d2 would have lower index than the next one in d1 ..
             itb := itb.next; // ..so only look for next partner in d2[i,.]
             j := itb.value1;
             Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
             then begin
               inc(z,abs(itb.value2));
               inc(zup);
             end;
            end; (* case 1 *)
         end (* case *)
        else (* not (ita.hasnext and itb.hasnext) *) begin
          if ita.hasnext then // network d1 not yet fully counted through, meaning: some more ties are lost (see case -1 above)
            begin
            ita := ita.next;
            j := ita.value1;
            Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
            then begin
               inc(z,abs(ita.value2));
               inc(zdown);
               end;
            end
           else // 'else' implying here that itb.hasnext due to the "or" condition in the while..do loop
            begin // network d2 not yet fully counted through, meaning: some more new ties are added (see case 1 above)
            itb := itb.next;
            j := itb.value1;
            Advance_it2345;
             if ((j2 > j) and (j3 > j) // these conditions mean "this is not affected by missingness"
             and (j4 > j) and (j5 > j)) // CS: these added considerations mean "this is not structurally fixed"
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
            then begin
               inc(z,abs(itb.value2));
               inc(zup);
               end;
            end;
         end; (* not (ita.hasnext and itb.hasnext) *)
   end; (* for i *)
 s := z;
 sdown := zdown;
 sup := zup;
end; (* CountDifNoNo *)

{cdcd procedure CountDifNoNo(var s,sdown,sup: integer;
                  const da,db: TValuedDigraph; const d2, d3: TDigraph);
(* Calculates
    s = sum(i,j) abso(da(i,j)-db(i,j))(1-d2(i,j))(1-d3(i,j))
    st = sum(i,j) (1-d2(i,j))(1-d3(i,j)) *)
(* Modeled after DifNoNo *)
var z, zdown, zup, i, j, j2, j3, v, nn: integer;
    ita, itb, it2, it3: TDoublyLinkedDoubleInt;
procedure Advance_it23;
begin
      if (j2 < j) then
        begin
        while ((it2.Value1 < j) and it2.HasNext) do it2 := it2.Next;
        j2 := it2.Value1;
        if (j2 < j) then j2 := succ(nn);
        end;
      if (j3 < j) then
        begin
        while ((it3.Value1 < j) and it3.HasNext) do it3 := it3.Next;
        j3 := it3.Value1;
        if (j3 < j) then j3 := succ(nn);
        end;
end; (* Advance_it23 *)
begin
  z := 0; zdown := 0; zup := 0;
  nn := da.Order;
  for i:= 1 to nn do begin
    ita:= da.OutTies[i].Ffirst;
    itb:= db.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    it3:= d3.OutTies[i].Ffirst;
(* j2 will be the smallest Value1 in d2.OutTies[i]
   that is at least as large as the current it1.Value1 = j;
   and nn+1 if there is no such Value1;
   i.e., j2 is the smallest value with d2(i,j2) <> 0 larger than j,
   and nn+1 if this j2 does not exist.
   similar for j3 in d3.OutTies[i] *)
    if it2.HasNext then
      begin
      it2 := it2.Next;
      j2 := it2.Value1;
      end
     else j2 := succ(nn);
    if it3.HasNext then
      begin
      it3 := it3.Next;
      j3 := it3.Value1;
      end
     else j3 := succ(nn);
    while (ita.hasnext or itb.hasnext) do
      if (ita.hasnext and itb.hasnext) then
       case sign(ita.next.value1 - itb.next.value1) of
        -1: begin
             ita := ita.next;
             j := ita.value1;
             Advance_it23;
             if ((j2 > j) and (j3 > j)) then
// this is the same as
//  if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
               begin
               inc(z,abs(ita.value2));
               inc(zdown);
               end;
            end; (* case -1 *)
        0: begin
             ita := ita.next;
             itb := itb.next;
             v := abs(ita.value2 - itb.value2);
             j := ita.value1;
             Advance_it23;
             if ((v <> 0) and (j2 > j) and (j3 > j)) then
// this is the same as
// if (v <> 0) and not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
               begin
               inc(z,v);
               if (v > 0) then inc(zdown) else inc(zup);
               end;
           end; (* case 0 *)
        1: begin
             itb := itb.next;
             j := itb.value1;
             Advance_it23;
             if ((j2 > j) and (j3 > j)) then
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
               begin
               inc(z,abs(itb.value2));
               inc(zup);
               end;
            end; (* case 1 *)
         end (* case *)
        else (* not (ita.hasnext and itb.hasnext) *) begin
          if ita.hasnext then
            begin
            ita := ita.next;
            j := ita.value1;
            Advance_it23;
            if ((j2 > j) and (j3 > j)) then
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
               begin
               inc(z,abs(ita.value2));
               inc(zdown);
               end;
            end
           else
            begin
            itb := itb.next;
            j := itb.value1;
            Advance_it23;
            if ((j2 > j) and (j3 > j)) then
// this is the same as
// if not((d2.TieValue(i,j) >= 1) or (d3.TieValue(i,j) >= 1)) then
               begin
               inc(z,abs(itb.value2));
               inc(zup);
               end;
            end;
         end; (* not (ita.hasnext and itb.hasnext) *)
   end; (* for i *)
 s := z;
 sdown := zdown;
 sup := zup;
end; (* CountDifNoNo *)
}

procedure SplitTwo(var dg, dgs: TValuedDigraph; const d1, d2: TDigraph);
(* Calculates dgs
   with dgs(i,j) = max(dg(i,j), min(d1(i,j), d2(i,j))),
   and dg := dg - dgs. *)
var i, j, nn: integer;
    v: shortint;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  nn := dg.Order;
  dgs.SetZero;
  for i:= 1 to nn do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
             j := it1.value1;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             j := it1.value1; (* = it2.value1 *)
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
             j := it2.value1;
            end; (* case 1 *)
        end; (* case *)
      v := dg.TieValue(i,j);
      if (v <> 0) then
        begin
        dgs.AddTie(i,j,v,true);
        dg.ChangeTieValue(i,j,0);
        end;
      end; (* while (ita.hasnext and itb.hasnext) *)

    while (it1.hasnext or it2.hasnext) do
      begin
      if it1.hasnext then
        begin
        it1 := it1.next;
        j := it1.value1;
        end
       else
        begin
        it2 := it2.next;
        j := it2.value1;
        end;
      v := dg.TieValue(i,j);
      if (v <> 0) then
        begin
        dgs.AddTie(i,j,v,true);
        dg.ChangeTieValue(i,j,0);
        end;
      end; (*  while (it1.hasnext or it2.hasnext) *)
  end; (* for i *)
end; (* SplitTwo *)

procedure CopyMinTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Calculates dgs(i,j) = min(d1(i,j), d2(i,j)). *)
var i, j, nn: integer;
    v: shortint;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  nn := d1.Order;
  dgs.SetZero;
  for i:= 1 to nn do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             j := it1.value1; (* = it2.value1 *)
             v := it1.Value2;
             dgs.AddTie(i,j,v,true);
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)
  end; (* for i *)
end; (* CopyMinTwo *)

procedure AddMinTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Replaces dgs(i,j) for which d1(i,j) <> 0 and d2(i,j)=1, by d1(i,j). *)
var i, j, nn: integer;
    v: shortint;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  nn := d1.Order;
  for i:= 1 to nn do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             j := it1.value1; (* = it2.value1 *)
             v := it1.Value2;
             dgs.ChangeTieValue(i,j,v);
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)
  end; (* for i *)
end; (* AddMinTwo *)

procedure ReplaceTwo(var dgs: TValuedDigraph; const d1: TvaluedDigraph;
                 const d2: TDigraph);
(* Replaces dgs(i,j) for which d2(i,j)=1, by d1(i,j). *)
var i, j, nn: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  nn := d1.Order;
  for i:= 1 to nn do begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin (* d2.OutTies[i] has a tie *)
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             j := it1.value1; (* = it2.value1 *)
             dgs.ChangeTieValue(i,j,it1.Value2);
           end; (* case 0 *)
        1: begin (* d1.OutTies[i] has a tie but d2 does not *)
             it2 := it2.next;
             dgs.ChangeTieValue(i,it2.value1,0);
            end; (* case 1 *)
        end; (* case *)
      end; (* while (ita.hasnext and itb.hasnext) *)
    while it2.HasNext do
      begin
      it2 := it2.next;
      dgs.ChangeTieValue(i,it2.value1,0);
      end;
  end; (* for i *)
end; (* ReplaceTwo *)

procedure MaxTwo(var dg, dgs: TValuedDigraph);
(* calculates dg := max(dg, dgs) *)
var i, j: integer;
    v: shortint;
    it: TDoublyLinkedDoubleInt;
begin
  dgs.SetZero;
  for i:=dg.Order downto 1 do
    begin
    it := dgs.OutTies[i].FFirst;
    while it.Hasnext do
      begin
      it := it.Next;
      j := it.Value1;
      v := it.Value2;
      dg.ChangeTieValue(i,j,v);
      end;
    end;
end; (* MaxTwo *)


procedure OldLost(var dgoldlost: TValuedDigraph;
                  const da,db: TValuedDigraph; const d2: TDigraph);
(* da is old network with 0 values if missing;
   db is new network with missings d2 *)
(* OldLost Creates dg, with elements defined as
   da(i,j) if (db(i,j) >= da(i,j)) or d2(i,j);
     i.e., no evidence that tie (i,j) was lost;
   -da(i,j) if (db(i,j) < da(i,j)) and (not d2(i,j));
     i.e., evidence that tie (i,j) was lost; *)
var i, j, nn: integer;
    ita, itb: TDoublyLinkedDoubleInt;
begin
  nn := da.Order;
  dgoldlost := TValuedDigraph.Create(nn);
  for i:= 1 to nn do begin
    ita:= da.OutTies[i].Ffirst;
    itb:= db.OutTies[i].Ffirst;
    while ita.hasnext do
      if itb.hasnext then
(* increase j=value1 to the next smallest of da and db: *)
       case sign(ita.next.value1 - itb.next.value1) of
        -1: begin
             ita := ita.next;
             j := ita.value1; (* tie a(i,j) was lost *)
             if (d2.TieValue(i,j) >= 1) then
               (* seemingly lost: new value is missing; tie is retained *)
               dgoldlost.AddTie(i,j, ita.value2,true) else
               (* really lost *) dgoldlost.AddTie(i,j, -ita.value2,true);
            end; (* case -1 *)
        0: begin
             ita := ita.next;
             itb := itb.next;
(* there are ties a(i,j) and b(i,j), so there cannot be missings *)
             j := ita.value1; (* tie a(i,j) was kept or changed *)
             if (itb.value2 >= ita.value2) then (* tie a(i,j) was kept or increased *)
               dgoldlost.AddTie(i,j, ita.value2,true) else (* tie a(i,j) was decreased *)
               dgoldlost.AddTie(i,j,-ita.value2,true);
           end; (* case 0 *)
        1: begin
             itb := itb.next; (* tie b(i,j) was newly created *)
            end; (* case 1 *)
         end (* case *)
        else (* not itb.hasnext; only lost da ties remain *) begin
          ita := ita.next;
          j := ita.value1; (* tie a(i,j) was lost; like above: *)
          if (d2.TieValue(i,j) >= 1) then dgoldlost.AddTie(i,j, ita.value2,true)
           else dgoldlost.AddTie(i,j, -ita.value2,true);
         end; (* not itb.hasnext *)
   end; (* for i *)
end; (* OldLost *)


{procedure ReduceYesNo(var dg1: TValuedDigraph; var dg2: TDigraph);
(* Changes existing dg1 so that all ties for which dg2 is larger than 0 are deleted. *)
var i, j, nn: integer;
    it: TDoublyLinkedDoubleInt;
begin
  nn := dg1.Order;
  for i:= 1 to nn do begin
    it:= dg1.OutTies[i].Ffirst;
    while it.HasNext do begin
      it := it.Next;
      j := it.Value1;
      if (dg2.TieValue(i,j) >= 1) then
(* @@@ perhaps this can be made more efficient *)
        dg1.ChangeTieValue(i,j, 0);
     end; (* while *)
   end; (* for i *)
end; (* ReduceYesNo *)
}


(* AA start *)
function OrCount(const d1, d2: TValuedDigraph): integer;
(* Counts the number of (i,j) with d1(i,j) <> 0 or d2(i,j) <> 0 *)
begin
  OrCount := d1.TotalTies + d2.TotalTies - AndCount(d1,d2);
end;

function AndCount(const d1, d2: TValuedDigraph): integer;
(* Counts the number of (i,j) with d1(i,j) <> 0 and d2(i,j) <> 0 *)
var i, ac: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  ac := 0;
  for i:= d1.order downto 1 do
    begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             inc(ac);
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (it1.hasnext and it2.hasnext) *)
  end; (* for i *)
  AndCount := ac;
end; (* AndCount *)

procedure ValuesCount(const d1: TValuedDigraph; const d2: TDigraph;
                       const a1,a2: shortint; var a: array of integer);
(* a[h] for h = 0 to a2-a1 will be the number of (i,j)
   with  d1[i,j] = a1+h <> 0 and d2(i,j) = 1. *)
(* AA end *)

var i, ind: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  if (length(a) < a2-a1) then
    begin
    exit;
    end;
  for ind := 0 to a2-a1 do a[ind] := 0;
  for i:= d1.order downto 1 do
    begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             inc(a[it1.Value2-a1]);
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
        end; (* case *)
      end; (* while (it1.hasnext and it2.hasnext) *)
  end; (* for i *)
end;




{function AbsDif(const d1, d2: TValuedDigraph): integer;
(* Counts the sum of abs(d1(i,j) - d2(i,j)) *)
var i, ad: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
  ad := 0;
  for i:= d1.order downto 1 do
    begin
    it1:= d1.OutTies[i].Ffirst;
    it2:= d2.OutTies[i].Ffirst;
    while (it1.hasnext and it2.hasnext) do
      begin
        case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             it1 := it1.next;
             inc(ad,abs(it1.Value2));
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
             inc(ad,abs(it1.Value2 - it2.Value2));
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
             inc(ad,abs(it2.Value2));
            end; (* case 1 *)
        end; (* case *)
      end; (* while (it1.hasnext and it2.hasnext) *)
  end; (* for i *)
  AbsDif := ad;
end;
}

{                                                                              }
{ TRedundantValuedDigraph                                                      }
{                                                                              }

constructor TRedundantValuedDigraph.Create(const n: Integer);
var i,j: integer;
begin
  inherited Create(n);
  SetLength(AdjMatrix,n+1,n+1);
  for i:=n downto 0 do for j:=n downto 0 do AdjMatrix[i,j] := 0;
end;


function TRedundantValuedDigraph.IsCorrect: Boolean;
var i,h,v1,v10: integer;
    v2: ShortInt;
    it:TDoublyLinkedDoubleInt;
begin
  Result := inherited IsCorrect;
(* Check Adjacency Matrix *)
  for i:=1 to Order do begin
   it := OutTies[i].First;
   v10 := 0;
   while it.HasNext do begin
     it := it.Next;
     v1 := it.Value1;
     v2 := it.Value2;
     for h:= v10 + 1 to v1 - 1 do if (AdjMatrix[i,h] <> 0)
      then begin
        Result := false;
        ErMes := 'AdjMatrix('+inttostr(i)+','+inttostr(h)+') = '
                 +inttostr(AdjMatrix[i,h])+'; should be 0';
          exit;
       end;
     if (AdjMatrix[i,v1] <> v2)
      then begin
        Result := false;
        ErMes := 'AdjMatrix('+inttostr(i)+','+inttostr(v1)+') = '
                 +inttostr(AdjMatrix[i,v1])+'; should be '+inttostr(v2);
          exit;
       end;
      v10 := v1;
     end; (* while *)
   end; (* for i *)
end;

function TRedundantValuedDigraph.TieValue(const i,j: integer): ShortInt;
begin
  Result := AdjMatrix[i,j];
end;

procedure TRedundantValuedDigraph.ChangeTieValue(const i,j,v: integer);
begin
  inherited ChangeTieValue(i,j,v);
  AdjMatrix[i,j] := v;
end;

function TRedundantValuedDigraph.IncreaseTie(const i,j,v: integer): ShortInt;
var r: shortint;
begin
  r := inherited IncreaseTie(i,j,v);
  AdjMatrix[i,j] := r;
end;

{                                                                              }
{ Partial Order: TPOset                                                     }
{ A partial order with its transitive reduction.                         }
{                                                                              }

constructor TPOset.Create(const n: Integer);
var i: integer;
begin
  inherited Create;
  arrivals      := TValuedDigraph.Create(n);
  ordered       := TValuedDigraph.Create(n);
  intermediate  := TValuedDigraph.Create(n);
  ordered.PermitDiagonal;
  for i:=n downto 1 do ordered.ChangeTieValue(i,i,1);
  hasse    := TDigraph.Create(n);
end;

destructor TPOSet.Destroy;
begin
  arrivals.Destroy;
  ordered.Destroy;
  hasse.Destroy;
  intermediate.Destroy;
  inherited Destroy;
end;

destructor TPOSet.Destroy(var outf: textfile);
begin
  if (arrivals.ermes > '') then writeln(outf,'arrivals error message: ',arrivals.ermes);
  if (ordered.ermes > '') then writeln(outf,'ordered error message: ',ordered.ermes);
  if (hasse.ermes > '') then writeln(outf,'hasse error message: ',hasse.ermes);
  if (intermediate.ermes > '') then
    writeln(outf,'intermediate error message: ',intermediate.ermes);
  arrivals.Destroy;
  ordered.Destroy;
  hasse.Destroy;
  intermediate.Destroy;
  inherited Destroy;
end;

function TPOSet.Order: Integer;
begin
  result := ordered.NumberNodes;
end;

function TPOSet.IsTransitive(var outf: textfile): Boolean;
var it1, it2: TDoublyLinkedDoubleInt;
    i,j,h: integer;
    OK: boolean;
begin
  OK := true;
  for i:=order downto 1 do
      begin
      it1 := ordered.OutTies[i].FFirst;
      while it1.HasNext do
        begin
        it1 := it1.Next;
        h := it1.value1;
        it2 := ordered.OutTies[h].FFirst;
        while it2.HasNext do
          begin
          it2 := it2.Next;
          j := it2.value1;
          if (ordered.TieValue(i,j) <= 0) then
            begin
            OK := false;
            writeln(outf,i:2,' => ',h:2,' => ',j:2,'; not ', i:2,' => ',j:2);
            end;
          end;
        end;
      end;
  result := OK;
end; (* IsTransitive *)


function TPOSet.IsReduced(var outf: textfile): Boolean;
var it1, it2: TDoublyLinkedDoubleInt;
    i,j,h: integer;
    OK: boolean;
begin
  OK := true;
  for i:=order downto 1 do
      begin
      it1 := hasse.OutTies[i].FFirst;
      while it1.HasNext do
        begin
        it1 := it1.Next;
        h := it1.value1;
        it2 := hasse.OutTies[h].FFirst;
        while it2.HasNext do
          begin
          it2 := it2.Next;
          j := it2.value1;
          if (hasse.TieValue(i,j) >= 1) then
            begin
            OK := false;
            writeln(outf,'Hasse: ',i:2,' => ',h:2,' => ',j:2,'; and ', i:2,' => ',j:2);
            end;
          end;
        end;
      end;
  result := OK;
end; (* IsReduced *)


function TPOSet.IsTheClosure(var outf: textfile): Boolean;
var a,b: TValuedDigraph;
    it1, it2: TDoublyLinkedDoubleInt;
    i,j,h,k: integer;
    changed,OK: boolean;
begin
  i := Order;
  a := TValuedDigraph.Create(i);
  OK := true;
  for k:=1 to 2 do
  begin
  if k = 1 then b := hasse else b := intermediate;
  if (k = 2) then a.SetZero;
  CopyGraph(b,a);             
  a.PermitDiagonal;
  for i:=1 to order do a.ChangeTieValue(i,i,1);
  repeat
  (* compute a := a * b, Boolean multiplication *)
    changed := false;
    for i:=order downto 1 do
      begin
      it1 := a.OutTies[i].FFirst;
      while it1.HasNext do
        begin
        it1 := it1.Next;
        h := it1.value1;
        it2 := b.OutTies[h].FFirst;
        while it2.HasNext do
          begin
          it2 := it2.Next;
          j := it2.value1;
          if (a.TieValue(i,j) <= 0) then
            begin
            changed := true;
            a.ChangeTieValue(i,j,1);
            end;
          end;
        end;
      end;
  until (not changed);
  i := Order;
  if OK then repeat
    if (not equal(a.OutTies[i],ordered.OutTies[i])) then
      begin
      OK := false;
      if (k = 1) then writeln(outf,'Ordered is not the closure of hasse; ',i)
       else  writeln(outf,'Ordered is not the closure of intermediate; ',i);
      writeln(outf,'Row ',i,':');
      write(outf,'Ordered ');
      for j:=1 to order do write(outf,ordered.TieValue(i,j):2);
      writeln(outf);
      write(outf,'Closure ');
      for j:=1 to order do write(outf,a.TieValue(i,j):2);
      writeln(outf);
      end;
    dec(i);
  until (not OK) or (i <= 0);
  end;
  result := OK;
  a.Destroy;
end; (* IsTheClosure *)

function TPOSet.CorrectArrivals(var outf: textfile): Boolean;
var a: TDigraph;
    i,j: integer;
    OK: boolean;
begin
  OK := true;
  i := Order;
  repeat
    for j := Order downto 1 do
    begin
    if (InnerProduct(ordered.OutTies[i],intermediate.InTies[j]) <> arrivals.TieValue(i,j))
      then
      begin
      writeln(outf,'For (',i:2,', ',j:2,'): ordered*intermediate = ',
                   InnerProduct(ordered.OutTies[i],intermediate.InTies[j]),
                   ', arrivals = ',arrivals.TieValue(i,j),'.');
      OK := false;
      end;
    end;
    dec(i);
  until (not OK) or (i <= 0);
  result := OK;
  if not OK then
    begin
    writeln(outf);
    writeln(outf,'Ordered:');
    ordered.WriteAdjacency(outf);
    writeln(outf,'Intermediate:');
    intermediate.WriteAdjacency(outf);
    writeln(outf,'Arrivals:');
    arrivals.WriteAdjacency(outf);
    writeln(outf);
    end;
end; (* CorrectArrivals *)

procedure TPOSet.InsertEdge(const i,j: integer);
(* Page 110 of La Poutre and Van Leeuwen 1988 *)
var it1,it2: TDoublyLinkedDoubleInt;
    h,k,n,r,dif: integer;
    newarrivals: shortint;
    color: array of byte; (* 0 = neutral, 1 = red, 2 = blue *)
    difm: TValuedDigraph;
procedure adapt(kk: integer);
var h,m: integer;
    itc: TDoublyLinkedDoubleInt;
    red: boolean;
    c: char;
begin (* adapt *)
  repeat
  (* search for a red node *)
    h := order+1;
    red := false;
    repeat
      dec(h);
      if (color[h] = 1) then red := true;
    until (red or (h <= 1));
    if red then
    (* h is a red node *)
      begin
      if (ordered.TieValue(kk,h) <= 0) then
        begin
        ordered.ChangeTieValue(kk,h,1);
        difm.IncreaseTie(kk,h,1);
        end
       else
         begin
writeln('   **** ');
readln(c);
         end;
      color[h] := 2;
      itc := intermediate.OutTies[h].Ffirst;
      while itc.HasNext do
        begin
        itc := itc.Next;
        m := itc.Value1;
        arrivals.IncreaseTie(kk,m,1);
        hasse.ChangeTieValue(kk,m,0);
        if (ordered.TieValue(kk,m) <= 0) then color[m] := 1;
        end;
      end; (* if red *)
   until (not red);
end; (* adapt *)
begin (* InsertEdge *)
  if (i = j) then exit;
  if (ordered.TieValue(i,j) >= 1) then exit;   (* superfluous to do anything *)
  if (ordered.TieValue(j,i) >= 1) then exit;  (* forbidden change *)
  n := Order;
  CopyGraph(ordered,intermediate);
  for h:=n downto 1 do intermediate.ChangeTieValue(h,h,0);
  intermediate.DiagonalPermitted := false;
  difm := TValuedDigraph.Create(n);
  intermediate.ChangeTieValue(i,j,1);
  difm.ChangeTieValue(i,j,-1);
  it1 := ordered.InTies[i].FFirst;
  it2 := intermediate.InTies[j].FFirst;

  while (it1.HasNext) do
    if (it1.HasNext and it2.HasNext) then
       case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             (* in this case, intermediate(k,j) = 0, ordered(k,i) = 1 *)
             it1 := it1.next;
             k := it1.Value1;
             newarrivals := arrivals.IncreaseTie(k,j,1);
             if (newarrivals = 127) then arrivals.ErMes := '127 arrivals';
             hasse.ChangeTieValue(k,j,0);
            end; (* case -1 *)
        0: begin
             (* in this case, intermediate(k,j) = ordered(k,i) = 1 *)
             it1 := it1.next;
             it2 := it2.next;
             k := it1.Value1;
             newarrivals := arrivals.IncreaseTie(k,j,1);
             if (newarrivals = 127) then arrivals.ErMes := '127 arrivals';
             if (newarrivals = 1) then hasse.ChangeTieValue(k,j,1)
              else
                begin
                hasse.ChangeTieValue(k,j,0);
                end;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
         end (* case *)
        else (* not (it1.hasnext and it2.hasnext) *)
          begin
          if it1.hasnext then
           (* in this case, intermediate(k,j) = 0, ordered(k,i) = 1 *)
            begin
            it1 := it1.next;
            k := it1.Value1;
            newarrivals := arrivals.IncreaseTie(k,j,1);
            if (newarrivals = 127) then arrivals.ErMes := '127 arrivals';
            hasse.ChangeTieValue(k,j,0);
            end;
         end; (* not (it1.hasnext and it2.hasnext) *)

  n :=  order;
  setlength(color,n+1);
  for k := n downto 0 do color[k] := 0;
(* colors: 0 = neutral, 1 = red, 2 = blue *)
  it1 := ordered.InTies[i].FFirst;
  it2 := ordered.InTies[j].FFirst;

  while (it1.HasNext) do
    if (it1.HasNext and it2.HasNext) then
      begin
       case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             (* in this case, ordered(k,j) = 0, ordered(k,i) = 1 *)
             it1 := it1.next;
             k := it1.Value1;
             color[j] := 1;
             adapt(k);
             for r:=n downto 1 do if (color[r] = 2) then color[r] := 0;
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
         end; (* case *)
       end
     else (* not (it1.hasnext and it2.hasnext) *)
      begin
      if it1.hasnext then
        begin
         (* in this case, hasse(k,j) = 0, hasse(k,i) = 1 *)
        it1 := it1.next;
        k := it1.Value1;
        color[j] := 1;
        adapt(k);
        for r:=n downto 1 do if (color[r] = 2) then color[r] := 0;
        end;
      end; (* not (it1.hasnext and it2.hasnext) *)

  for h := 1 to n do
    begin
    it1 := difm.OutTies[h].FFirst;
    while it1.HasNext do
      begin
      it1 := it1.Next;
      it2 := ordered.Inties[h].first;
      while it2.HasNext do
        begin
        it2 := it2.Next;
        arrivals.IncreaseTie(it2.Value1,it1.Value1,it1.Value2);
        end;
      end;
    end;
end; (* InsertEdge *)

procedure TPOSet.DeleteEdge(const i,j: integer);
(* Page 116 of La Poutre and Van Leeuwen 1988 *)
var it1,it2: TDoublyLinkedDoubleInt;
    h,k,n,r,dif: integer;
    newarrivals: shortint;
    marked: array of boolean;
    difm: TValuedDigraph;
procedure adjust(kk: integer);
var h,m: integer;
    itc: TDoublyLinkedDoubleInt;
    mark: boolean;
    c: char;
begin (* adjust *)
  repeat
(* search for a marked node *)
    mark := false;
    h := order+1;
    repeat
      dec(h);
      if marked[h] then mark := true;
    until (mark or (h <= 1));
    if mark then
(* h is a marked node *)
      begin
      if (ordered.TieValue(kk,h) >= 1) then
        begin
        ordered.ChangeTieValue(kk,h,0);
        difm.IncreaseTie(kk,h,1);
        end
       else begin
writeln('   **** ');
readln(c);
         end;
      marked[h] := false;
      itc := intermediate.OutTies[h].Ffirst;
      while itc.HasNext do
        begin
        itc := itc.Next;
        m := itc.Value1;
        newarrivals := arrivals.IncreaseTie(kk,m,-1);
        if (newarrivals <= -1) then arrivals.ErMes := 'negative arrivals';
        if (newarrivals <= 0) then marked[m] := true
         else if (newarrivals = 1) and (intermediate.TieValue(kk,m) = 1)
         then hasse.ChangeTieValue(kk,m,1);
        end;
      end; (* if mark *)
   until (not mark);
end; (* adjust *)
begin (* DeleteEdge *)
  if (i = j) then exit;
  if (hasse.TieValue(i,j) <= 0) then exit;  (* nothing needs to be done *)
  n := order;

  CopyGraph(ordered,intermediate);
  for h:=n downto 1 do intermediate.ChangeTieValue(h,h,0);
  intermediate.DiagonalPermitted := false;
  difm := TValuedDigraph.Create(n);
  intermediate.ChangeTieValue(i,j,0);
  difm.ChangeTieValue(i,j,-1);
  it1 := ordered.InTies[i].FFirst;
  it2 := intermediate.InTies[j].FFirst;

  while (it1.HasNext) do
    if (it1.HasNext and it2.HasNext) then
       case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             (* in this case, intermediate(k,j) = 0, ordered(k,i) = 1 *)
             it1 := it1.next;
             k := it1.Value1;
             newarrivals := arrivals.IncreaseTie(k,j,-1);
             if (newarrivals <= -1) then arrivals.ErMes := 'negative arrivals';
             hasse.ChangeTieValue(k,j,0);
            end; (* case -1 *)
        0: begin
             (* in this case, intermediate(k,j) = ordered(k,i) = 1 *)
             it1 := it1.next;
             it2 := it2.next;
             k := it1.Value1;
             newarrivals := arrivals.IncreaseTie(k,j,-1);
             if (newarrivals <= -1) then arrivals.ErMes := 'negative arrivals';
             if (newarrivals = 1) then hasse.ChangeTieValue(k,j,1)
              else
                begin
                hasse.ChangeTieValue(k,j,0);
                end;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
         end (* case *)
        else (* not (it1.hasnext and it2.hasnext) *)
          begin
          if it1.hasnext then
           (* in this case, intermediate(k,j) = 0, ordered(k,i) = 1 *)
            begin
            it1 := it1.next;
            k := it1.Value1;
            newarrivals := arrivals.IncreaseTie(k,j,-1);
            if (newarrivals <= -1) then arrivals.ErMes := 'negative arrivals';
            hasse.ChangeTieValue(k,j,0);
            end;
         end; (* not (it1.hasnext and it2.hasnext) *)

  n :=  order;
  setlength(marked,n+1);
  for k := n downto 0 do marked[k] := false;
  it1 := ordered.InTies[i].FFirst;
  it2 := arrivals.InTies[j].FFirst;
  while (it1.HasNext) do
    if (it1.HasNext and it2.HasNext) then
       case sign(it1.next.value1 - it2.next.value1) of
        -1: begin
             (* in this case, arrivals(k,j) = 0, ordered(k,i) = 1 *)
             it1 := it1.next;
             k := it1.Value1;
             marked[j] := true;
             adjust(k);
            end; (* case -1 *)
        0: begin
             it1 := it1.next;
             it2 := it2.next;
           end; (* case 0 *)
        1: begin
             it2 := it2.next;
            end; (* case 1 *)
         end (* case *)
        else (* not (it1.hasnext and it2.hasnext) *)
          begin
          if it1.hasnext then
            begin
            (* in this case, arrivals(k,j) = 0, ordered(k,i) = 1 *)
            it1 := it1.next;
            k := it1.Value1;
            marked[j] := true;
            adjust(k);
            end;
         end; (* not (it1.hasnext and it2.hasnext) *)

  for h:=1 to n do
    begin
    it1 := difm.OutTies[h].FFirst;
    while it1.HasNext do
      begin
      it1 := it1.Next;
      it2 := ordered.Inties[h].first;
      while it2.HasNext do
        begin
        it2 := it2.Next;
        arrivals.IncreaseTie(it2.Value1,it1.Value1,-it1.Value2);
        end;
      end;
    end;
end; (* DeleteEdge *)


procedure TPOSet.WritePOSet(var f: textfile); (* writes arrivals, ordered, hasse to f *)
var i,j: integer;
begin
  writeln(f,'Adjacency matrix of partial order;');
  writeln(f,'underlined elements are irreducible.');
  writeln(f);
  write(f,'     ');
  for j:=1 to order do write(f,j mod 10:2);
  writeln(f);
  for i:=1 to Order do
    begin
    write(f,i:3,'. ');
    for j:=1 to Order do write(f,ordered.TieValue(i,j):2);
    writeln(f);
    write(f,'     ');
    for j:=1 to Order do if (hasse.TieValue(i,j) >= 1) then
     write(f,' =') else write(f,'  ');
    writeln(f);
    end;
  writeln(f);

end; (* WritePOSet *)


function TPOSet.lt(const i,j: integer): boolean;   (* i less than j in partial order *)
begin
  lt := (i = j) or (ordered.TieValue(i,j) >= 1);
end;

function TPOSet.gt(const i,j: integer): boolean;   (* i greater than j in partial order *)
begin
  gt := (i = j) or (ordered.TieValue(j,i) >= 1);
end;

function TPOSet.next(const i,j: integer): boolean;
begin
  next := (hasse.TieValue(i,j) >= 1);
end;

procedure TPOSet.CopyPOTo(var cpos: TPOSet); (* copies the TPOSet to cpos, which must have been defined  *)
begin
  CopyGraph(arrivals,cpos.arrivals);
  CopyGraph(intermediate,cpos.intermediate);
  CopyGraph(hasse,cpos.hasse);
  CopyGraph(ordered,cpos.ordered);
end; (* CopyPOTo *)



{                                                                              }
{ Effects for Valued directed graph: TEffect                                   }
{                                                                              }

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                           const ContriW:  TConfigWeight0;
                           const ConfigW:  TConfigWeight1;
                           const p, p0, p1: integer);
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1i := nil;
  f1ij := nil;
  f1ijc := nil;
  f1ijergm := nil;
  FDyadic := true;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
    if ((i >= 3) and (ContributionWeight[i] <> 0.0)) then Fdyadic := false;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := ConfigW[i];
  FIncludeEndowment := true;
end;

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                           const ContriW:  TConfigWeight0;
                           const p, p0, p1: integer; fij: TRealFunctionij;
                           dummy1, dummy2: boolean);
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1ij := fij;
  f1i := nil;
  f1ijc := nil;
  f1ijergm := nil;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := 0;
  FDyadic := false;
  FIncludeEndowment := true;
end;

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fij: TRealFunctionij;
                       fi: TRealFunctioni; const DyadEf: boolean);
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1ij := fij;
  f1i := fi;
  f1ijc := fijc;
  f1ijergm := nil;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := 0;
  FDyadic := DyadEf;
  FIncludeEndowment := false;
end;

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const ConfigW:  TConfigWeight1;
                       const p, p0, p1: integer;
                       fi: TRealFunctioni);
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1ij := nil;
  f1i := fi;
  f1ijc := nil;
  f1ijergm := nil;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := ConfigW[i];
  FDyadic := false;
  FIncludeEndowment := false;
end;

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer; fijr: TRealFunctionijr;
                       dummy3, dummy4, dummy5, dummy6: boolean);
(* This is one of the overloads for creating ergm effects.*)
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1ij := nil;
  f1i := nil;
  f1ijc := nil;
  f1ijergm := fijr;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := 0;
  FDyadic := false;
  FIncludeEndowment := false;
end;

constructor TEffect.Create(const EN, FN: ShortString; const ind: integer;
                       const ContriW:  TConfigWeight0;
                       const p, p0, p1: integer;
                       fi: TRealFunctioni; fijr: TRealFunctionijr);
(* This is one of the overloads for creating ergm effects.*)
var i: integer;
begin
  Inherited Create;
  FEName := EN;
  FFName := FN;
  FIndex := ind;
  FKind := 0;
  FParameter := p;
  FParMin := p0;
  FParMax := p1;
  f1ij := nil;
  f1i := fi;
  f1ijc := nil;
  f1ijergm := fijr;
  HasAnyContribution := false;
  for i:=0 to ConfigMax do begin
    ContributionWeight[i] := ContriW[i];
    HasContribution[i] := (ContriW[i] <> 0.0);
    if HasContribution[i] then HasAnyContribution := true;
   end;
  for i:=1 to ConfigMax do ConfigWeight[i] := 0;
  FDyadic := false;
  FIncludeEndowment := false;
end;

function TEffect.HasNextMod_f: Boolean;
begin
  Result := Assigned(FNextMod_f);
end;

function TEffect.HasNextMod: Boolean;
begin
  Result := Assigned(FNextMod);
end;

function TEffect.HasNextStat_f: Boolean;
begin
  Result := Assigned(FNextStat_f);
end;

function TEffect.HasNextStat_g: Boolean;
begin
  Result := Assigned(FNextStat_g);
end;

function TEffect.SetParameter(const p: integer): Boolean;
begin
 if (p >= FParMin) and (p <= FParMax) then
   begin
   FParameter := p; Result := true;
//   if (p in [0..1]) then FRparameter := 0 else FRparameter := log(1.0 - (1.0/p));
// not in use now; rparameter may be used later for something else.
   end
  else Result := false;
end;

function TEffect.EName: shortstring;
begin
  Result := StringReplace (FEName, '#', inttostr(Parameter), [rfReplaceAll]);
end;

function TEffect.FName: shortstring;
begin
  Result := StringReplace (FFName, '#', inttostr(Parameter), [rfReplaceAll]);
end;



{                                                                              }
{ Array of Effects for Valued directed graph: TEffects                         }
{                                                                              }

constructor TEffects.Create;
var i: integer;
begin
 inherited Create;
 FNumberStats_f := 0;
 FNumberStats_g := 0;
 for i:=0 to 2 do DyadicContributionWeight[i] := 0.0;
 for i:=0 to ConfigMax do begin
   TotalContributionWeight[i] := 0.0;
   TotalContributionNonZero[i] := false;
  end;
 for i:=1 to 2 do DyadicContributionWeight_f[i] := 0.0;
 for i:=1 to ConfigMax do begin
   TotalContributionWeight_f[i] := 0.0;
   TotalContributionNonZero_f[i] := false;
   TotalConfigNonZero_f[i] := false;
   TotalConfigNonZero_g[i] := false;
  end;
end;

destructor TEffects.Destroy;
var i: integer;
begin
  for i:= High(Effect) downto 0 do Effect[i].Destroy;
  SetLength(Effect,0);
  SetLength(skipnumber_g,0);
  inherited Destroy;
end;

function TEffects.NumberEffects: integer;
begin
  if assigned(self) then Result := High(Effect)+1 else Result :=0;
end;


function TEffects.Param(const i: integer): integer;
begin
  if assigned (Effect[i-1]) then Result := Effect[i-1].FParameter
   else Result := 0;
end;

function TEffects.SetParam(const i: integer; var p: integer): boolean;
(* Param reports and SetParam changes the value of Effect[i-1].FParameter;
   SetParam reports whether p is a permitted value. *)
var pp: integer;
begin
  if assigned (Effect[i-1]) then begin
    pp := Effect[i-1].Parameter;
    Result := Effect[i-1].SetParameter(p);
    if (not Result) then p := pp;
   end else Result := false;
end;

function TEffects.Endowed(const i: integer): Boolean;
begin
  if assigned (Effect[i-1]) then Result := Effect[i-1].IncludeEndowment
   else Result := false;
end;

function TEffects.EndowedIndex(const i: integer): Boolean;
var h: integer;
begin
  Result := false;
  if (NumberEffects <= 0) then
    exit
   else begin
    h := 0;
    repeat
     inc(h);
     if (Effect[h-1].index = i) then Result := Effect[h-1].IncludeEndowment;
    until (Effect[h-1].index = i) or (h >= NumberEffects);
   end;
end;

function TEffects.NameIndex(const i: integer): shortstring;
var h: integer;
begin
  Result := '';
  if (NumberEffects <= 0) then
    exit
   else begin
    h := 0;
    repeat
     inc(h);
     if (Effect[h-1].index = i) then Result := Effect[h-1].FEName;
    until (Effect[h-1].index = i) or (h >= NumberEffects);
   end;
end;

function TEffects.Kind(const i: integer): byte;
begin
  if assigned (Effect[i-1]) then Result := Effect[i-1].FKind
   else Result := 255;
end;


procedure TEffects.AddEffect1(const EN, FN: ShortString;
                             const ContriW:  TConfigWeight0;
                             const ConfigW:  TConfigWeight1;
                             const p, p0, p1: integer);
begin
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, ConfigW, p, p0, p1);
end; (* AddEffect1 *)

procedure TEffects.AddEffect2(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const ConfigW:  TConfigWeight1;
                        const p, p0, p1: integer; fi: TRealFunctioni);
var fij: TRealfunctionij;
    fijc: TRealFunctionijt;
begin
  fij := nil;
  fijc := nil;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, ConfigW, p, p0, p1,fi);
end; (* AddEffect2 *)


procedure TEffects.AddEffect4(const EN, FN: ShortString;
                             const ContriW:  TConfigWeight0;
                             const p, p0, p1: integer; fij: TRealFunctionij);
begin
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1,fij, false, false );
end; (* AddEffect4 *)

procedure TEffects.AddEffect3(const EN, FN: ShortString;
                             const ContriW:  TConfigWeight0;
                             const p, p0, p1: integer; fijc: TRealFunctionijt;
                             fij: TRealFunctionij; fi: TRealFunctioni);
begin
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1,fijc,fij,fi,false);
end; (* AddEffect3 *)

procedure TEffects.AddEffect6(const EN, FN: ShortString;
                             const p, p0, p1: integer; fijc: TRealFunctionijt;
                             fi: TRealFunctioni);
var fij: TRealfunctionij;
    ContriW:  TConfigWeight0;
    i: integer;
begin
  for i:= 1 to ConfigMax do ContriW[i] := 0;
  fij := nil;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1,fijc,fij,fi,false);
end; (* AddEffect6 *)

procedure TEffects.AddEffect5(const EN, FN: ShortString;
                             const p, p0, p1: integer; fijc: TRealFunctionijt;
                             fij: TRealFunctionij; const DyadEf: boolean);
var fi: TRealFunctioni;
    ContriW:  TConfigWeight0;
    i: integer;
begin
  for i:= 1 to ConfigMax do ContriW[i] := 0;
  fi := nil;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1,fijc,fij,fi,DyadEf);
  Effect[NumberEffects-1].FIncludeEndowment := true;
  (* overriding the false assignment in Create because still f1i = nil *)
end; (* AddEffect5 *)


procedure TEffects.AddEffect7(const EN, FN: ShortString;
                        const p, p0, p1: integer; fijc: TRealFunctionijt;
                       fij: TRealFunctionij; const DyadEf: boolean; const KindEf: byte);
var fi: TRealFunctioni;
    ContriW:  TConfigWeight0;
    i: integer;
begin
  for i:= 1 to ConfigMax do ContriW[i] := 0;
  fi := nil;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1,fijc,fij,fi,DyadEf);
  Effect[NumberEffects-1].FIncludeEndowment := true;
  (* overriding the false assignment in Create because still f1i = nil *)
  Effect[NumberEffects-1].FKind := KindEf;
  (* overriding the 0 assignment in Create *)
end; (* AddEffect7 *)


procedure TEffects.AddERGMEffect1(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const p, p0, p1: integer; fijr: TRealFunctionijr; const KindEf: byte);
(* This defines ergm effects where the statistic for the empty graph is 0. *)
var ddyadic: boolean;
    i: integer;
begin
  ergmcase := true;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1, fijr, false, false, false, false);
  Effect[NumberEffects-1].FKind := KindEf;
  (* overriding the 0 assignment in Create *)
  if (KindEf in [1,2,3]) then Effect[NumberEffects-1].FDyadic := true;
  ddyadic := true;
  for i:=3 to ConfigMax do begin
    if ((i >= 3) and (ContriW[i] <> 0.0)) then ddyadic := false;
   end;
  if (not assigned(fijr)) then Effect[NumberEffects-1].FDyadic := ddyadic;
end; (* AddERGMEffect1 *)



procedure TEffects.AddERGMEffect2(const EN, FN: ShortString;
                        const ContriW:  TConfigWeight0;
                        const p, p0, p1: integer;
                        fi: TRealFunctioni; fijr: TRealFunctionijr; const KindEf: byte);
(* This defines ergm effects where the statistic for the empty graph
   is not 0, but defined by fi. *)
var ddyadic: boolean;
    i: integer;
begin
  ergmcase := true;
  SetLength(Effect,NumberEffects+1);
  (* Note that this increases NumberEffects by 1 *)
  SetLength(SkipNumber_g,NumberEffects+1);
  Effect[NumberEffects-1] :=
    TEffect.Create(EN, FN, NumberEffects, ContriW, p, p0, p1, fi, fijr);
  Effect[NumberEffects-1].FKind := KindEf;
  (* overriding the 0 assignment in Create *)
  if (KindEf in [1,2,3]) then Effect[NumberEffects-1].FDyadic := true;
  ddyadic := true;
  for i:=3 to ConfigMax do begin
    if ((i >= 3) and (ContriW[i] <> 0.0)) then ddyadic := false;
   end;
  if (not assigned(fijr)) then Effect[NumberEffects-1].FDyadic := ddyadic;
end; (* AddERGMEffect2 *)


procedure TEffects.SpecifyInteractionEffect(const i: integer;
                        var pa, code: integer; var ert: shortstring);
(* This procedure is called in S_Base.ReviseDefinition.
   There the names are modified, as indicated there by 'hack'.
*)
var p1, p2, p3, p4: integer;
    av1, av2, dya: boolean;
    k1, k2, k3, k4, numego, numdyad, numstruc: byte;
begin
  ert := '';
  code := 0;
  if (Effect[i-1].FKind <> 4) then exit; (* no interaction effect *)
  if (pa = 0) then exit;                 (* no definition of interaction effect *)
  if (pa <= 999999) then
    begin
(* two-way interaction *)
    p2 := pa mod 1000;
    p1 := (pa div 1000) mod 1000;
    if (p1 <= 1) or (p1 > NumberEffects) or (p2 <= 1) or (p2 > Numbereffects) then
      begin
      ert := 'An interaction effect with code '+inttostr(pa)+', decomposed into ('
             +inttostr(p1)+', '+inttostr(p2)+'), is not implemented.';
      pa := 0;
      exit;
      end;
(* AA start *)
   (*  Check if FKind = 3, i.e.,
       either is an ego actor variable effect for non-ergm,
       or an actor variable constant within connected components. *)
    k1 := Effect[p1-1].FKind;
    k2 := Effect[p2-1].FKind;
    if (k1 = 4) or (k2 = 4) then
      begin
      ert := 'An interaction effect of interaction effects is not implemented.';
      pa := 0;
      exit;
      end;
    if (k1 <> 3) and (k2 = 3) then (* swap *)
      begin
      p3 := p1;
      p1 := p2;
      p2 := p3;
      k3 := k1;
      k1 := k2;
      k2 := k3;
      end;

    if not ((k1 = 3)
            or (Effect[p1-1].FDyadic and Effect[p2-1].FDyadic)) then
      begin
      ert := 'An interaction effect with code '+inttostr(pa)+', decomposed into ('
             +inttostr(p1)+', '+inttostr(p2)+') is not implemented. See the SIENA manual.';
      pa := 0;
      exit;
      end;

    pa := p2+ 1000*p1;
    if (length(Effect[p1-1].FEName)+length(Effect[p2-1].FEName) > 35) then
      begin
      Effect[i-1].FEName := Effect[p1-1].FEName+'  '+Effect[p2-1].FEName;
      Effect[i-1].FFName := Effect[p1-1].FEName+'  '+Effect[p2-1].FEName;
      end
     else
      begin
      Effect[i-1].FEName := 'int. '+Effect[p1-1].FEName+'  '+Effect[p2-1].FEName;
      Effect[i-1].FFName := 'int. '+Effect[p1-1].FEName+'  '+Effect[p2-1].FEName;
      end;
  (* The FFNames do not combine so nicely. *)

    if ergmcase then
      begin
(* The following selection uses the fact that
   for ergm effects, either Contriw is <> 0, or f1ijergm is defined,
   but not both. *)
//if assigned(Effect[p1-1].f1ijergm) then ert := 'aa' else ert := 'bb';
//if assigned(Effect[p2-1].f1ijergm) then ert := ert+ ' aa' else ert := ert + ' bb';

      if assigned(Effect[p1-1].f1ijergm) then
        begin
        if assigned(Effect[p2-1].f1ijergm) then
          Effect[i-1].f1ijergm := Interaction_ergm11
         else
          Effect[i-1].f1ijergm := Interaction_ergm10;
        end
       else
        begin
        if assigned(Effect[p2-1].f1ijergm) then
          Effect[i-1].f1ijergm := Interaction_ergm01
         else
           Effect[i-1].f1ijergm := Interaction_ergm00;
        end;
      end
     else (* not ergmcase *)
      begin
      if (k1 = 3) then
        begin
        Effect[i-1].f1ijc := Interactiont_a;
        Effect[i-1].f1ij := Interaction_a;
        Effect[i-1].f1i := Interactiontot_a;
        Effect[i-1].Fdyadic := Effect[p2-1].Fdyadic;
        end
       else
        begin
        Effect[i-1].f1ijc := Interactiont_b;
        Effect[i-1].f1ij := Interaction_b;
        Effect[i-1].f1i := nil;
        Effect[i-1].Fdyadic := true;
        end;
      end;
    end (* pa <= 999999*)
   else
    begin (* pa >= 1000000 *)
(* three-way interaction *)
    p3 := pa mod 1000;
    p2 := (pa div 1000) mod 1000;
    p1 := (pa div 1000000);
(* first the special cases *)
(*
1003def, 2003def, 3003def, 4003def:
  (similarity on actor variable # def) x transitive triplets
8003def, 8005def, 8006def:
  transitive triplets, 3-cycles, transitive ties
  only within same values of actor variable # def
11003def, 12003def, 13003def, 14003def:
  (similarity on dyadic variable # def) x transitive triplets
18003def, 18005def, 18006def:
  transitive triplets, 3-cycles, transitive ties
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.
*)

    if (not ergmcase) and (p1 in [1..2]) and (p2 = 2) and (p3 = 3) then
      begin
(* interactions reciprocity by transitive triplets *)
      Effect[i-1].FEName := 'reciprocity  transitive triplets '
                             +inttostr(p1);
      Effect[i-1].FFName := 'number of reciprocal and transitive triplets type '
                             +inttostr(p1);
      Effect[i-1].f1ijc := Interactiont_rec_trans;
      Effect[i-1].f1ij := Interaction_rec_trans;
      Effect[i-1].f1i := nil;
      Effect[i-1].Fdyadic := false;
      exit;
      end;

    if (not ergmcase) and ((p1 in [1..4]) or (p1 in [11..14])) and (p2 = 3) then
      begin
(* interactions transitive triplets by covariate *)
      dya := (p1 >= 11);
      if dya then
        begin
        p1 := p1-10;
        code := 1000 + p3;
        end
       else
        begin
        code := p3;
        end;

      if dya then
        begin
        Effect[i-1].FEName := 'trans. triplets  ';
        Effect[i-1].FFName := 'number trans. trip. weighted by ';
        end
       else
        begin
        Effect[i-1].FEName := 'trans. triplets  similarity ';
        Effect[i-1].FFName := 'number trans. trip. weighted by similarity ';
        end;
//      Effect[i-1].f1ijc := Interactiont_var_trans;  is already defined in S_Effects!
//      Effect[i-1].f1ij := Interaction_var_trans;    is already defined!
//      These two are the default choices in the statement in S_Effects
//        Effs.Addeffect7('unspecified interaction effect',
//                      'unspecified interaction statistic',
//                    nextpar(0),low(integer),high(integer),Interactiont_var_trans,Interaction_var_trans,true,4);
//      This tric was used to have access to the variables, which are not available in Digraph.
      Effect[i-1].f1i := nil;
      Effect[i-1].Fdyadic := false;
      exit;
      end;

    if (not ergmcase) and ((p1 = 8) or (p1 = 18)) and (p2 in [3,5,6]) then
      begin
(* interactions transitive triplets by covariate *)
      dya := (p1 >= 11);
      if dya then
        begin
        p1 := p1-10;
        code := 1000 + p3;
        end
       else
        begin
        code := p3;
        end;
      case p2 of
        3: begin
           if dya then
             begin
             Effect[i-1].FEName := 'trans. trip. for ';
             Effect[i-1].FFName := 'number tr. trip. for ';
             end
            else
             begin
             Effect[i-1].FEName := 'trans. trip. within same ';
             Effect[i-1].FFName := 'number tr. trip. within same ';
             end;
           end; (* case 3 *)
        5: begin
           if dya then
             begin
             Effect[i-1].FEName := '3-cycles for ';
             Effect[i-1].FFName := 'number 3-cycles for ';
             end
            else
             begin
             Effect[i-1].FEName := '3-cycles within same ';
             Effect[i-1].FFName := 'number 3-cycles within same ';
             end;
           end; (* case 5 *)
        6: begin
           if dya then
             begin
             Effect[i-1].FEName := 'trans. ties for ';
             Effect[i-1].FFName := 'number tr. ties for ';
             end
            else
             begin
             Effect[i-1].FEName := 'trans. ties within same ';
             Effect[i-1].FFName := 'number ties within same ';
             end;
           end; (* case 6 *)
        end; (* case *)
//      Effect[i-1].f1ijc := Interactiont_var_trans;  is already defined in S_Effects!
//      Effect[i-1].f1ij := Interaction_var_trans;    is already defined!
//      These two are the default choices in the statement in S_Effects
//        Effs.Addeffect7('unspecified interaction effect',
//                      'unspecified interaction statistic',
//                    nextpar(0),low(integer),high(integer),Interactiont_var_trans,Interaction_var_trans,true,4);
//      This tric was used to have access to the variables, which are not available in Digraph.
      Effect[i-1].f1i := nil;
      Effect[i-1].Fdyadic := false;
      exit;
      end;

(* now the standard case *)
    if (p1 <= 1) or (p1 > NumberEffects) or (p2 <= 1) or (p2 > Numbereffects)
       or (p3 <= 1) or (p3 > Numbereffects) then
      begin
      ert := 'An interaction effect with code '+inttostr(pa)+', decomposed into ('
             +inttostr(p1)+', '+inttostr(p2)+', '+inttostr(p3)+'), is not implemented.';
      pa := 0;
      exit;
      end;
    k1 := Effect[p1-1].FKind;
    k2 := Effect[p2-1].FKind;
    k3 := Effect[p3-1].FKind;
    if (k1 = 4) or (k2 = 4) or (k3 = 4) then
      begin
      ert := 'An interaction effect of interaction effects is not implemented.';
      pa := 0;
      exit;
      end;

    if ergmcase then
      begin
      numstruc := 0;
      if (k1 = 0) then inc(numstruc);
      if (k2 = 0) then inc(numstruc);
      if (k3 = 0) then inc(numstruc);
(* number of structural effects *)
      if (numstruc >= 1) then
        begin
(* reorder so that the structural effects come first *)
        if (k1 <> 0) then
          begin
          if (k3 = 0) then
            begin
            p4 := p3;
            p3 := p1;
            p1 := p4;
            k4 := k3;
            k3 := k1;
            k1 := k4;
            end
           else if (k2 = 0) then
            begin
            p4 := p2;
            p2 := p1;
            p1 := p4;
            k4 := k2;
            k2 := k1;
            k1 := k4;
            end;
          end;
        if (k2 <> 0) and (k3 = 0) then
          begin
          p4 := p2;
          p2 := p3;
          p3 := p4;
          end;
        end;
      end
     else
      begin (* not ergmcase *)
   (*  Check if either is an ego actor variable effect, i.e., FKind = 3: *)
      numego := 0;
      if (k1 = 3) then inc(numego);
      if (k2 = 3) then inc(numego);
      if (k3 = 3) then inc(numego);
      numdyad := 0;
      if Effect[p1-1].FDyadic and (Effect[p1-1].FKind <> 4) then inc(numdyad);
      if Effect[p2-1].FDyadic and (Effect[p2-1].FKind <> 4) then inc(numdyad);
      if Effect[p3-1].FDyadic and (Effect[p3-1].FKind <> 4) then inc(numdyad);
      (* note that (fkind = 3) implies fdyadic *)
      if (numdyad <= 2)  and (numego <= 1) then
        begin
        ert := 'An interaction effect with code '+inttostr(pa)+', decomposed into ('
               +inttostr(p1)+', '+inttostr(p2)+', '+inttostr(p3)+
               '), is not implemented. See the SIENA manual.';
        pa := 0;
        exit;
        end;
      if (numego <= 2) then
        begin
(* reorder so that the ego effects come first *)
        if (k1 <> 3) then
          begin
          if (k3 = 3) then
            begin
            p4 := p3;
            p3 := p1;
            p1 := p4;
            k4 := k3;
            k3 := k1;
            k1 := k4;
            end
           else if (k2 = 3) then
            begin
            p4 := p2;
            p2 := p1;
            p1 := p4;
            k4 := k2;
            k2 := k1;
            k1 := k4;
            end;
          end;
        if (k2 <> 3) and (k3 = 3) then
          begin
          p4 := p2;
          p2 := p3;
          p3 := p4;
          end;
        end;
      end;

    pa := p3 + 1000*p2+ 1000000*p1;
//    ert := inttostr(pa);
    Effect[i-1].FEName := 'i3 '+Effect[p1-1].FEName+'  '+Effect[p2-1].FEName
                          +'  '+Effect[p3-1].FEName;
    Effect[i-1].FFName := 'i3 '+Effect[p1-1].FEName+'  '+Effect[p2-1].FEName
                          +'  '+Effect[p3-1].FEName;
  (* The FFNames do not combine so nicely. *)

    if ergmcase then
      begin
      case numstruc of
        0: Effect[i-1].f1ijergm := Interaction3ergm_0;
        1: Effect[i-1].f1ijergm := Interaction3ergm_1;
        2: Effect[i-1].f1ijergm := Interaction3ergm_2;
        3: Effect[i-1].f1ijergm := Interaction3ergm_3;
        end; (* case *)
      end
     else (* not ergmcase *)
      begin
      if (numdyad <= 2) then
(* there are two dyadic effects, which must be ego effects,
   and which must be the first two *)
        begin
        Effect[i-1].f1ijc := Interaction3t_a;
        Effect[i-1].f1ij := Interaction3_a;
        Effect[i-1].f1i := Interaction3tot_a;
        Effect[i-1].Fdyadic := false;
        end
       else
        begin
        Effect[i-1].f1ijc := Interaction3t_b;
        Effect[i-1].f1ij := Interaction3_b;
        Effect[i-1].f1i := nil;
        Effect[i-1].Fdyadic := true;
        end;
      end;
    end;  (* pa >= 1000000 *)
end; (* SpecifyInteractionEffect *)

procedure TEffects.AppendName(const i: integer; const st: shortstring);
 (* Appends st to Effects[i-1].FEName and FFName *)
begin
  Effect[i-1].FEName := Effect[i-1].FEName+st;
  Effect[i-1].FFName := Effect[i-1].FFName+st;
end;

procedure TEffects.DeleteName(const i: integer);
 (* Appends st to Effects[i-1].FEName and FFName *)
begin
  Effect[i-1].FEName := 'unspecified effect';
  Effect[i-1].FFName := 'unspecified effect';
end;

procedure TEffects.DefineModSequences(var outf: TFileStream;
                                 const indf, indg: array of boolean;
                                 const IncludeNonendowed: boolean);
var h: integer;
    Effct1, Effct2: TEffect;
procedure nexth_f;
(* Find smallest next h for which indf[h] = true *)
begin
  repeat
    inc(h)
   until (indf[h] or (h > NumberEffects));
   if (h > High(indf)) then h:= NumberEffects+1;
end; (* nexth_f *)
procedure nexth_fg;
(* Find smallest next h for which indf[h] = true or indg[h] = true *)
begin
  repeat
    inc(h)
   until (indf[h] or indg[h] or (h > NumberEffects));
   if (h > High(indf)) then h:= NumberEffects+1;
end; (* nexth_fg *)
begin (* DefineModSequences *)
(* Low(indf) is defined as 0, even though externally
   the array bounds of indf may be defined differently.
   Therefore, Effect[h] must be used below rather than Effect[h-1]. *)

(* First define the model_f sequence *)
  h := Low(indf);
(* note that effects_f is a BBarray starting with the non-used index 0 *)
  FNum_f := 0;

  nexth_f;                                            
  if (h <= NumberEffects) then
    begin
    Effct2 := Effect[h-1];
    FfirstMod_f := Effct2;
    inc(FNum_f);
    nexth_f;
    while (h <= NumberEffects) do begin
      Effct1 := Effct2;
      Effct2 := Effect[h-1];
      Effct1.FnextMod_f := Effct2;
      inc(FNum_f);
      nexth_f;
     end; (* while *)
    end; (* if *)

(* Next define the model sequence ( _f or _g ) *)
  h := Low(indf);
  nexth_fg;
  if (h > NumberEffects) then exit;
  Effct2 := Effect[h-1];
  if (indf[h] or (indg[h] and (Effct2.IncludeEndowment or IncludeNonendowed))) then begin
    FfirstMod := Effct2;
    Effct2.hasweight_f := indf[h];
    Effct2.hasweight_g := (indg[h] and (Effct2.IncludeEndowment or IncludeNonendowed));
   end;
  nexth_fg;
  while (h <= NumberEffects) do begin
    Effct1 := Effct2;
    Effct2 := Effect[h-1];
    if (indf[h] or (indg[h] and (Effct2.IncludeEndowment or IncludeNonendowed))) then begin
      if assigned(FfirstMod) then Effct1.FnextMod := Effct2
       else FfirstMod := Effct2;
      Effct2.hasweight_f := indf[h];
      Effct2.hasweight_g := (indg[h] and (Effct2.IncludeEndowment or IncludeNonendowed));
     end (* if *) else Effct2 := Effct1;
    nexth_fg;
   end;
end; (* DefineModSequences *)

procedure TEffects.DefineTotalContributions(const a_f, a_g: array of real);
(* This defines intermediate results for efficient calculations.
   MUST always be called after setting/changing alpa_f;
   therefore this is called in SetAlpha. *)
var Ef: TEffect;
    alo, i: integer;
begin
  (* First the evaluation effects
     used when a tie variable is increased from 0 to 1 . *)
  for i:=1 to ConfigMax do AnyContributionNonZero_f[i] := false;
  alo := Low(a_f)-1;
(* a_f and a_g must have the same array bounds *)
  if Assigned(FirstMod_f) then begin
    Ef := FirstMod_f;
    Ef.alpha_f := a_f[Ef.index+alo];
    for i:=1 to 2 do if Ef.Dyadic then
     DyadicContributionWeight_f[i] := Ef.alpha_f*Ef.ContributionWeight[i]
      else DyadicContributionWeight_f[i] := 0.0;
(* @@@  check:  (Ef.ContributionWeight[i] <> 0) = Ef.HasContribution[i] *)
    for i:=1 to ConfigMax do if (Ef.ContributionWeight[i] <> 0) then begin
        AnyContributionNonZero_f[i] := true;
        TotalContributionWeight_f[i] := Ef.alpha_f*Ef.ContributionWeight[i];
       end else TotalContributionWeight_f[i] := 0.0;
    while Ef.HasNextMod_f do begin
      Ef := Ef.NextMod_f;
      Ef.alpha_f := a_f[Ef.index+alo];
      for i:=1 to 2 do if Ef.Dyadic then
       DyadicContributionWeight_f[i] :=
         DyadicContributionWeight_f[i] + Ef.alpha_f*Ef.ContributionWeight[i];
(* @@@  check:  (Ef.ContributionWeight[i] <> 0) = Ef.HasContribution[i] *)
      for i:=1 to ConfigMax do if (Ef.ContributionWeight[i] <> 0) then begin
          AnyContributionNonZero_f[i] := true;
          TotalContributionWeight_f[i] :=
           TotalContributionWeight_f[i] + Ef.alpha_f*Ef.ContributionWeight[i];
         end;
      end; (* while *)
   end (* if assigned *) else begin
    for i:=1 to ConfigMax do TotalContributionWeight_f[i] := 0.0;
    for i:=1 to 2 do DyadicContributionWeight_f[i] := 0.0;
   end;

  for i:=1 to ConfigMax do
    TotalContributionNonZero_f[i] :=
                          (TotalContributionWeight_f[i] <> 0.0);

  (* Now the evaluation and endowment effects
     used when a tie variable is decreased from 1 to 0 . *)

  for i:=0 to ConfigMax do AnyContributionNonZero[i] := false;

(* a_f and a_g must have the same array bounds *)
  if Assigned(FirstMod) then begin
    Ef := FirstMod;
    Ef.alpha := a_f[Ef.index+alo]+a_g[Ef.index+alo];   (* @@@ was -a_g in 3bet2 *)
    for i:=0 to 2 do if Ef.Dyadic then
     DyadicContributionWeight[i] := Ef.alpha*Ef.ContributionWeight[i] else
     DyadicContributionWeight[i] := 0.0;
(* @@@  check:  (Ef.ContributionWeight[i] <> 0) = Ef.HasContribution[i] *)
    for i:=0 to ConfigMax do
      if (Ef.ContributionWeight[i] <> 0) then begin
        TotalContributionWeight[i] := Ef.alpha*Ef.ContributionWeight[i];
        AnyContributionNonZero[i] := true;
       end else TotalContributionWeight[i] := 0.0;
    while Ef.HasNextMod do begin
      Ef := Ef.NextMod;
      Ef.alpha := a_f[Ef.index+alo]+a_g[Ef.index+alo];    (* @@@ was -a_g in 3bet2 *)
      for i:=0 to 2 do if Ef.Dyadic then
       DyadicContributionWeight[i] :=
         DyadicContributionWeight[i] + Ef.alpha*Ef.ContributionWeight[i];
(* @@@  check:  (Ef.ContributionWeight[i] <> 0) = Ef.HasContribution[i] *)
      for i:=0 to ConfigMax do
        if (Ef.ContributionWeight[i] <> 0) then begin
          TotalContributionWeight[i] :=
            TotalContributionWeight[i] + Ef.alpha*Ef.ContributionWeight[i];
          AnyContributionNonZero[i] := true;
         end;
      end; (* while *)
   end (* if assigned *) else begin
    for i:=0 to ConfigMax do TotalContributionWeight[i] := 0.0;
    for i:=0 to 2 do DyadicContributionWeight[i] := 0.0;
   end;

  for i:=0 to ConfigMax do TotalContributionNonZero[i] :=
                          (TotalContributionWeight[i] <> 0.0);
end; (* DefineTotalContributions *)

procedure TEffects.DefineStat_fSequence(const indicator: array of boolean);
var h, i, nef: integer;
    Effct1, Effct2: TEffect;
procedure nexth;
(* Find smallest next h for which indicator[h] = true *)
begin
  repeat
    inc(h)
   until (indicator[h] or (h > NumberEffects));
   if (h > High(indicator)) then h:= NumberEffects+1;
end; (* nexth *)
begin (* DefineStat_fSequence *)
  for i:=1 to ConfigMax do begin
    TotalConfigNonZero_f[i] := false;
   end;
   FNumberStats_f := 0;

(* Low(indicator) is defined as 0, even though externally
   the array bounds of indicator may be defined differently.
   Therefore, Effect[h] must be used below rather than Effect[h-1]. *)
  h := Low(indicator);
(* note that effects_f is a BBarray starting with the non-used index 0 *)

  nexth;
  if (h > NumberEffects) then exit;
  Effct2 := Effect[h-1];
  FfirstStat_f := Effct2;
  for i:=1 to ConfigMax do begin
    if (Effct2.ConfigWeight[i] <> 0.0) then TotalConfigNonZero_f[i] := true;
   end;
  nef := 1;
  nexth;
  while (h <= NumberEffects) do begin
    Effct1 := Effct2;
    Effct2 := Effect[h-1];
    Effct1.FnextStat_f := Effct2;
    for i:=1 to ConfigMax do begin
      if (Effct2.ConfigWeight[i] <> 0.0) then TotalConfigNonZero_f[i] := true;
     end;
    inc(nef);
    nexth;
   end;

  FNumberStats_f := nef;
end; (* DefineStat_fSequence *)


procedure TEffects.DefineStat_gSequence(const indicator: array of boolean;
                                 const IncludeNonendowed: boolean);
(* indicator gives the statistics to be included;
   SkipNumber_g[Effect.index] gives the number of effects skipped before Effect,
   skipping being necessary for other effects not defining an endowment function. *)
var h, i, nef, skips: integer;
    Effct1, Effct2: TEffect;
procedure nexth;
(* Find smallest next h for which indicator[h] = true *)
begin
  repeat
    inc(h)
   until (indicator[h] or (h > NumberEffects));
   if (h > High(indicator)) then h:= NumberEffects+1;
end; (* nexth *)
begin (* DefineStat_gSequence *)
  for i:=1 to ConfigMax do begin
    TotalConfigNonZero_g[i] := false;
   end;
   FNumberStats_g := 0;
   for i:= NumberEffects-1 downto 0 do SkipNumber_g[i] := 0;

(* Low(indicator) is defined as 0, even though externally
   the array bounds of indicator may be defined differently.
   Therefore, Effect[h] must be used below rather than Effect[h-1]. *)
  h := Low(indicator);
(* note that effects_f is a BBarray starting with the non-used index 0 *)

  nexth;
  if (h > NumberEffects) then exit;
  Effct2 := Effect[h-1];
  skips := 0;

  if (Effct2.IncludeEndowment or IncludeNonendowed) then begin
    FfirstStat_g := Effct2;
    for i:=1 to ConfigMax do
      if (Effct2.ConfigWeight[i] <> 0.0) then TotalConfigNonZero_g[i] := true;
   end else skips := 1;
  nexth;
  nef := 1;
  while (h <= NumberEffects) do begin
    Effct1 := Effct2;
    Effct2 := Effect[h-1];
    if (Effct2.IncludeEndowment or IncludeNonendowed) then begin
      if assigned(FfirstStat_g) then Effct1.FnextStat_g := Effct2
       else FfirstStat_g := Effct2;
      SkipNumber_g[Effct2.index] := skips;
      skips := 0;
      for i:=1 to ConfigMax do
        if (Effct2.ConfigWeight[i] <> 0.0) then TotalConfigNonZero_g[i] := true;
     end else begin (* not includeendowment *)
      Effct2 := Effct1;
      inc(skips);
     end;
    inc(nef);
    nexth;
   end; (* while *)

  FNumberStats_g := nef;
end; (* DefineStat_gsequence *)


procedure TEffects.WriteMod_fSequenceParameters(const a: array of real;
                            var outf: textfile);
var Ef: TEffect;
    alo,h,hh: integer;
begin
  alo := Low(a)-1;
  h := 1;
  writeln(outf,'Parameters in network effect_f sequence are ');
  if Assigned(FirstMod_f) then begin
    Ef := FirstMod_f;
    writeln(outf,h:3,'. ',a[Ef.index+alo]:10:6,'; ',Ef.alpha_f:10:6);
    while Ef.HasNextMod_f do
      begin
      Ef := Ef.NextMod_f;
      inc(h);
      writeln(outf,h:3,'. ',a[Ef.index+alo]:10:6,'; ',Ef.alpha_f:10:6);
      for hh:=0 to ConfigMax do writeln(outf,'  ',hh:3,'. ',Ef.ContributionWeight[hh]:6:1);
      writeln(outf);
      end;
   end else writeln(outf,' ... none ... ');

   writeln(outf,'TotalContributionWeight_f:');
   for h:=1 to ConfigMax do writeln(outf,h:3,'. ',TotalContributionWeight_f[h]:12:8);
end; (* WriteMod_fSequenceParameters *)

procedure TEffects.WriteModSequenceParameters(const a: array of real;
                            var outf: textfile);
var Ef: TEffect;
    alo,h: integer;
begin
  alo := Low(a)-1;
  h := 1;
  writeln(outf,'Parameters in network effect sequence are ');
  if Assigned(FirstMod) then begin
    Ef := FirstMod;
    writeln(outf,h:3,'. ',a[Ef.index+alo]:10:6,'; ',Ef.alpha:10:6);
    while Ef.HasNextMod do begin
      Ef := Ef.NextMod;
      inc(h);
      writeln(outf,h:3,'. ',a[Ef.index+alo]:10:6,'; ',Ef.alpha:10:6);
      end;
   end else writeln(outf,' ... none ... ');

   writeln(outf,'TotalContributionWeight:');
   for h:=0 to ConfigMax do writeln(outf,h:3,'. ',TotalContributionWeight[h]:12:8);
end; (* WriteModSequenceParameters *)

procedure TEffects.WriteStatSequenceIndices(var outf: textfile);
var Ef: TEffect;
    h: integer;
begin
  h := 1;
  writeln(outf,'Parameters in network statistics sequence are ');
  if Assigned(FirstStat_f) then begin
    Ef := FirstStat_f;
    writeln(outf,h:3,'. ',Ef.index:4);
    while Ef.HasNextStat_f do begin
      Ef := Ef.NextStat_f;
      inc(h);
      writeln(outf,h:3,'. ',Ef.index:4);
      end;
   end else writeln(outf,' ... none ... ');
  writeln(outf,'Count is given as ',NumberStats_f,'.');
end; (* WriteStatSequenceIndices *)

function TEffects.EffectName(const i: integer): shortstring;
begin
  if (not (assigned(self))) then begin
    Result := ' ??? ';
    exit;
   end;
  if (high(Effect) < 0) then begin
    Result := ' ???? ';
    exit;
   end;
  if (i <= 0) or (i > NumberEffects) then Result := ' (unknown effect) '
   else Result := Effect[i-1].EName;
end;

function TEffects.FunctionName(const i: integer): shortstring;
begin
  if (not (assigned(self))) then begin
    Result := ' ??? ';
    exit;
   end;
  if (high(Effect) < 0) then begin
    Result := ' ???? ';
    exit;
   end;
  if (i <= 0) or (i > NumberEffects) then Result := ' (unknown statistic) '
   else Result := Effect[i-1].FName;
end;

function TEffects.Contrib_n(const dg: TValuedDigraph; const i,j,k: integer;
                    const incoming, DyadicOnly: boolean): real;
(* Minus contribution to evaluation plus endowment function
   of changing tie variable (i,j) from 1 to 0.
   Used in Utilitygain, AddNoTies and SubtractTies.
   k is the period.
   incoming is the tie indicator for (j,i), used for looking this up only once. *)
var Ef: TEffect;
begin
  if (not Assigned(FirstMod)) then begin
    Result := 0.0;
    exit;
   end else begin
  if DyadicOnly then begin
    Result := DyadicContributionWeight[0] + DyadicContributionWeight[1];
    if incoming then Result := Result + DyadicContributionWeight[2];
   end else (* not DyadicOnly *) begin
    Result := TotalContributionWeight[0] + TotalContributionWeight[1];
    if incoming then Result := Result + TotalContributionWeight[2];
    if TotalContributionNonZero[3] then
      Result := Result +  TotalContributionWeight[3]*dg.OutDegree(i);
    if TotalContributionNonZero[4] then
      Result := Result +  TotalContributionWeight[4]*dg.InDegree(i);
    if TotalContributionNonZero[5] then
      Result := Result +  TotalContributionWeight[5]*dg.OutDegree(j);
    if TotalContributionNonZero[6] then
      Result := Result +  TotalContributionWeight[6]*dg.InDegree(j);
    if TotalContributionNonZero[7] then
      Result := Result + TotalContributionWeight[7]*dg.TwoPaths(i,j);
    if TotalContributionNonZero[8] then
      Result := Result + TotalContributionWeight[8]*dg.OutTwoStars(i,j);
    if TotalContributionNonZero[9] then
      Result := Result + TotalContributionWeight[9]*dg.InTwoStars(i,j);
    if TotalContributionNonZero[10] then
      Result := Result + TotalContributionWeight[10]*dg.TwoPaths(j,i);
   end; (* not DyadicOnly *)

  Ef := FirstMod;
  if (Assigned(Ef.f1ijc) and (Ef.Dyadic or (not DyadicOnly))) then
     Result := Result + Ef.alpha*Ef.f1ijc(dg,i,j,k,1,Ef.parameter);
  while Ef.HasNextMod do begin
    Ef := Ef.NextMod;
    if (Assigned(Ef.f1ijc) and (Ef.Dyadic or (not DyadicOnly))) then
      Result := Result + Ef.alpha*Ef.f1ijc(dg,i,j,k,1,Ef.parameter);
   end; (* while *)
  end;
end; (* Contrib_n *)

function TEffects.Contrib_n_alpha(const dg: TValuedDigraph; const i,j,k: integer;
                    const alfa: array of real; const incoming: boolean;
                    const calcchangestat: boolean; var chas: rrrarray): real;
(* Minus contribution to evaluation plus endowment function
   of changing tie variable (i,j) from 1 to 0.
   Used in SubtractTies.
   Does not use the total contributions calculated in DefineTotalContributions.
   k is the period.
   incoming is the tie indicator for (j,i), used for looking this up only once. *)
var Ef: TEffect;
  h,alo: integer;
  co: TConfigWeight0;
  sum: real;
  hasnexteffect: boolean;
begin
  Result := 0.0;
  if (not Assigned(FirstMod)) then exit;

  for h:=2 to ConfigMax do co[h] := 0.0;
  co[0] := 1.0;
  co[1] := 1.0;
  if incoming then co[2] := 1.0;
  if AnyContributionNonZero[3] then co[3] := dg.OutDegree(i);
  if AnyContributionNonZero[4] then co[4] := dg.InDegree(i);
  if AnyContributionNonZero[5] then co[5] := dg.OutDegree(j);
  if AnyContributionNonZero[6] then co[6] := dg.InDegree(j);
  if AnyContributionNonZero[7] then co[7] := dg.TwoPaths(i,j);
  if AnyContributionNonZero[8] then co[8] := dg.OutTwoStars(i,j);
  if AnyContributionNonZero[9] then co[9] := dg.InTwoStars(i,j);
  if AnyContributionNonZero[10] then co[10] := dg.TwoPaths(j,i);

  alo := low(alfa)-1; (* Note that Ef.index starts from 1 *)
  Ef := FirstMod;
  repeat
    sum := 0.0;
    for h:=0 to ConfigMax do
      sum := sum + Ef.ContributionWeight[h]*co[h];
    if Assigned(Ef.f1ijc) then
      sum := sum + Ef.f1ijc(dg,i,j,k,1,Ef.parameter);
(* now sum = s_{i, Ef}(x(i \leadsto j)) - s_{i, Ef}(x) abcd *)
//      if calcchangestat then begin   @@@@ gaf foutmelding, uitgeschakeld
//      chas[2,Ef.index,j] := -sum;
//      chas[3,Ef.index,j] := sum;
//     end;
    Result := Result + alfa[Ef.index+alo]*sum;
    hasnexteffect := Ef.HasNextMod;
    if hasnexteffect then Ef := Ef.NextMod;
  until (not hasnexteffect);
end; (* Contrib_n_alpha *)

function TEffects.Contrib_fn(const dg: TValuedDigraph; const i,j,k: integer;
                    const incoming, DyadicOnly: boolean): real;
(* Contribution to evaluation function of actor i
   of changing tie variable (i,j) from 0 to 1, in period k.
   Used in AddNoTies.
   k is the period.
   incoming is the tie indicator for (j,i), used for looking this up only once. *)
var Ef: TEffect;
begin
  if (not Assigned(FirstMod_f)) then begin
    Result := 0.0;
    exit;
   end;

  if DyadicOnly then begin
    Result := DyadicContributionWeight_f[1];
    if incoming then Result := Result + DyadicContributionWeight_f[2];
   end else (* not DyadicOnly *) begin
    Result := TotalContributionWeight_f[1];
    if incoming then Result := Result + TotalContributionWeight_f[2];
    if TotalContributionNonZero_f[3] then
      Result := Result +  TotalContributionWeight_f[3]*dg.OutDegree(i);
    if TotalContributionNonZero_f[4] then
      Result := Result +  TotalContributionWeight_f[4]*dg.InDegree(i);
    if TotalContributionNonZero_f[5] then
      Result := Result +  TotalContributionWeight_f[5]*dg.OutDegree(j);
    if TotalContributionNonZero_f[6] then
      Result := Result +  TotalContributionWeight_f[6]*dg.InDegree(j);
    if TotalContributionNonZero_f[7] then
      Result := Result + TotalContributionWeight_f[7]*dg.TwoPaths(i,j);
    if TotalContributionNonZero_f[8] then
      Result := Result + TotalContributionWeight_f[8]*dg.OutTwoStars(i,j);
    if TotalContributionNonZero_f[9] then
      Result := Result + TotalContributionWeight_f[9]*dg.InTwoStars(i,j);
    if TotalContributionNonZero_f[10] then
      Result := Result + TotalContributionWeight_f[10]*dg.TwoPaths(j,i);
   end; (* not DyadicOnly *)

  Ef := FirstMod_f;
  if (Assigned(Ef.f1ijc) and (Ef.Dyadic or (not DyadicOnly))) then
      Result := Result + Ef.alpha_f*Ef.f1ijc(dg,i,j,k,0,Ef.parameter);
  while Ef.HasNextMod_f do begin
    Ef := Ef.NextMod_f;
    if (Assigned(Ef.f1ijc) and (Ef.Dyadic or (not DyadicOnly))) then
      Result := Result + Ef.alpha_f*Ef.f1ijc(dg,i,j,k,0,Ef.parameter);
   end; (* while *)
end; (* Contrib_fn *)

function TEffects.Contrib_fn_alpha(const dg: TValuedDigraph; const i,j,k: integer;
                    const alfa: array of real; const incoming: boolean;
                    const calcchangestat: boolean; var chas: rrrarray): real;
(* Contribution to objective function of actor i
   of changing tie variable (i,j) from 0 to 1,
   and minus contribution of changing tie variable (i,j) from 1 to 0.
   Used in AddNoTies.
   Does not use the total contributions calculated in DefineTotalContributions.
   k is the period.
   incoming is the tie indicator for (j,i), used for looking this up only once. *)
var Ef: TEffect;
  h,alo: integer;
  co: TConfigWeight0;
  sum: real;
  hasnexteffect: boolean;
begin
  Result := 0.0;
  if (not Assigned(FirstMod_f)) then exit;

  for h:=0 to ConfigMax do co[h] := 0.0;
(* co[0] is 0 here always; not used. *)
  co[1] := 1.0;
  if incoming then co[2] := 1.0;
  if AnyContributionNonZero_f[3] then co[3] := dg.OutDegree(i);
  if AnyContributionNonZero_f[4] then co[4] := dg.InDegree(i);
  if AnyContributionNonZero_f[5] then co[5] := dg.OutDegree(j);
  if AnyContributionNonZero_f[6] then co[6] := dg.InDegree(j);
  if AnyContributionNonZero_f[7] then co[7] := dg.TwoPaths(i,j);
  if AnyContributionNonZero_f[8] then co[8] := dg.OutTwoStars(i,j);
  if AnyContributionNonZero_f[9] then co[9] := dg.InTwoStars(i,j);
  if AnyContributionNonZero_f[10] then co[10] := dg.TwoPaths(j,i);

  alo := low(alfa)-1; (* Note that Ef.index starts from 1 *)
  Ef := FirstMod_f;
  repeat
    sum := 0.0;
    for h:=1 to ConfigMax do
      sum := sum + Ef.ContributionWeight[h]*co[h];
    if Assigned(Ef.f1ijc) then
      sum := sum + Ef.f1ijc(dg,i,j,k,0,Ef.parameter);
(* now sum = s_{i, Ef}(x(i \leadsto j)) - s_{i, Ef}(x) *)
//    if calcchangestat then chas[2,Ef.index+1,j] := sum;  @@@@ gaf foutmelding, uitgeschakeld
    Result := Result + alfa[Ef.index + alo]*sum;
    hasnexteffect := Ef.HasNextMod_f;
    if hasnexteffect then Ef := Ef.NextMod_f;
  until (not hasnexteffect);
end; (* Contrib_fn_alpha *)

procedure TEffects.AddNoTies(const i,k: integer;
         var p: array of real; perm: array of boolean;
          const dg: TValuedDigraph);
(* Adds Contrib_fn(Effs,a,i,j) to p[j] over all j having no tie from i to j
                                             and satisfying perm[j] *)
var it,itr: TDoublyLinkedDoubleInt;
    j, pt, ptr: integer;
function prevouttie: integer;
(* next element with tie from i, going from high to low *)
begin
  Result := it.Value1;
  if it.hasprev then it := it.prev
   else Result := 0;
end; (* prevouttie *)
function previntie: integer;
(* next element with tie to i, going from high to low *)
begin
  Result := itr.Value1;
  if itr.hasprev then itr := itr.prev
   else Result := 0;
end; (* previntie *)
begin (* AddNoTies *)
  it := dg.OutTies[i].Last;
  itr := dg.InTies[i].Last;
  pt := prevouttie;
  ptr := previntie;
  for j:=dg.Order downto 1 do begin
    if (j > pt) then begin
        if perm[j] then
            (* (ptr >= j) = there is an incoming tie(j,i) *)
            p[j] := p[j] + Contrib_fn(dg,i,j,k,(ptr >= j),false);
     end (*if j > pt *) else pt := prevouttie;
    if (j <= ptr) then ptr := previntie;
   end; (* for j *)
end; (* AddNoTies *)


procedure TEffects.AddNoTies(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const dg: TValuedDigraph; const calcchangestat: boolean; var chas: rrrarray);
(* Adds Contrib_fn(Effs,a,i,j) to p[j] over all j having no tie from i to j
                                             and satisfying perm[j] *)
var it,itr: TDoublyLinkedDoubleInt;
    j, pt, ptr: integer;
function prevouttie: integer;
(* next element with tie from i, going from high to low *)
begin
  Result := it.Value1;
  if it.hasprev then it := it.prev
   else Result := 0;
end; (* prevouttie *)
function previntie: integer;
(* next element with tie to i, going from high to low *)
begin
  Result := itr.Value1;
  if itr.hasprev then itr := itr.prev
   else Result := 0;
end; (* previntie *)
begin (* AddNoTies *)
  it := dg.OutTies[i].Last;
  itr := dg.InTies[i].Last;
  pt := prevouttie;
  ptr := previntie;
  for j:=dg.Order downto 1 do begin
    if (j > pt) then begin
        if perm[j] then
            (* (ptr >= j) = there is an incoming tie(j,i) *)
            p[j] := p[j] + Contrib_fn_alpha(dg,i,j,k,a,(ptr >= j),calcchangestat,chas);
     end (*if j > pt *) else pt := prevouttie;
    if (j <= ptr) then ptr := previntie;
   end; (* for j *)
end; (* AddNoTies *)

(* following procedure added 3.17q *)
procedure TEffects.SetChasZero(const ch_first, nnb: integer; var chas: rrrarray);
var j: integer;
    Ef: TEffect;
    hasnexteffect: boolean;
begin
for j:=nnb downto 1 do
  begin
  Ef := FirstMod;
  repeat
    chas[1,ch_first+Ef.index,j] := 0.0;
    chas[2,ch_first+Ef.index,j] := 0.0;
    hasnexteffect := Ef.HasNextMod;
    if hasnexteffect then Ef := Ef.NextMod;
  until (not hasnexteffect);
  end; (* for j *)
end; (* SetChasZero *)

procedure TEffects.SubtractTies(const i,k: integer;
         var p: array of real; perm: array of boolean;
         const dg: TValuedDigraph);
(* Subtracts Contrib_n(Effs,a,self,i,j)
   from p[j] over all j
   having a tie from i to j and satisfying perm[j] *)
var it: TDoublyLinkedDoubleInt;
    j: integer;
    recipr: boolean;
begin (* SubtractTies *)
  it := dg.OutTies[i].First;
  while it.HasNext do begin
   it := it.next;
   j := it.Value1;
   if perm[j] then begin
     recipr := (dg.TieValue(j,i) >= 1);
     p[j] := p[j] - Contrib_n(dg,i,j,k,recipr,false);
    end;
  end;
end; (* SubtractTies *)

procedure TEffects.SubtractTies(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const dg: TValuedDigraph; const calcchangestat: boolean;
         var chas: rrrarray);
(* Subtracts Contrib_n(Effs,a,self,i,j)
   from p[j] over all j
   having a tie from i to j and satisfying perm[j] *)
var it: TDoublyLinkedDoubleInt;
    j: integer;
    recipr: boolean;
begin (* SubtractTies *)
  it := dg.OutTies[i].First;
  while it.HasNext do begin
   it := it.next;
   j := it.Value1;
   if perm[j] then begin
     recipr := (dg.TieValue(j,i) >= 1);
     p[j] := p[j] - Contrib_n_alpha(dg,i,j,k,a,recipr,calcchangestat,chas);
    end;
  end;
end; (* SubtractTies *)

{
procedure TEffects.ChangeStatistics(const i,j: integer;
         var s: array of real; var eu: real; const t,tr: shortint; const dg: TValuedDigraph);
(* Stores the change statistics (all effects h) for changing tie variable (i,j)
   in array element s[h];
   t is the current value of dg(i,j); tr is the current value of dg(j,i).
   Used for ERGM, i.e., observations = 1. *)
(* This is very much like the j-step in UtilityComponents;
   it is programmed separately because there is no endowment function here. *)
var Ef: TEffect;
    h1: integer;
    hasnextmod: boolean;
    Ef_f1, eeu: real;
procedure AddCo_f(const hh: integer; const co: real);
var coo: real;
begin
  coo := co;
  Ef := FirstMod_f;
  h1 := 0;
  eeu := 0.0;
  repeat
    (* evaluation function only: *)
    inc(h1);
    if Ef.HasContribution[hh] then
      s[h1] := s[h1] + Ef.ContributionWeight[hh]*coo;
    hasnextmod := Ef.HasNextMod_f;
    if  hasnextmod then Ef := Ef.NextMod_f;
   until (not hasnextmod);
end; (* AddCo_f *)
begin
  if (i = j) then exit;
  if (t <= 0) then
    begin
    AddCo_f(1,1.0);
    if (tr >= 1) then AddCo_f(2,1.0);
    if AnyContributionNonZero_f[3] then AddCo_f(3,dg.OutDegree(i));
    if AnyContributionNonZero_f[4] then AddCo_f(4,dg.InDegree(i));
    if AnyContributionNonZero_f[5] then AddCo_f(5,dg.OutDegree(j));
    if AnyContributionNonZero_f[6] then AddCo_f(6,dg.InDegree(j));
    if AnyContributionNonZero_f[7] then AddCo_f(7,dg.TwoPaths(i,j));
    if AnyContributionNonZero_f[8] then AddCo_f(8,dg.OutTwoStars(i,j));
    if AnyContributionNonZero_f[9] then AddCo_f(9,dg.InTwoStars(i,j));
    if AnyContributionNonZero_f[10] then AddCo_f(10,dg.TwoPaths(j,i));
(* loop along all effects with a nonzero evaluation function parameter *)
    Ef := FirstMod_f;
    h1 := 0;
     repeat
        (* evaluation function: *)
        if Ef.hasweight_f then
          begin
          inc(h1);
          if (Assigned(Ef.f1ijc)) then
            s[h1] := s[h1] + Ef.f1ijc(dg,i,j,1,0,Ef.parameter);
          end;
        hasnextmod := Ef.HasNextMod_f;
        if  hasnextmod then Ef := Ef.NextMod_f;
     until (not hasnextmod);
   end (* if t <= 0 *)
  else
   begin
    AddCo_f(1,-1.0);
    if (tr >= 1) then AddCo_f(2,-1.0);
    if AnyContributionNonZero_f[3] then AddCo_f(3,-dg.OutDegree(i));
    if AnyContributionNonZero_f[4] then AddCo_f(4,-dg.InDegree(i));
    if AnyContributionNonZero_f[5] then AddCo_f(5,-dg.OutDegree(j));
    if AnyContributionNonZero_f[6] then AddCo_f(6,-dg.InDegree(j));
    if AnyContributionNonZero_f[7] then AddCo_f(7,-dg.TwoPaths(i,j));
    if AnyContributionNonZero_f[8] then AddCo_f(8,-dg.OutTwoStars(i,j));
    if AnyContributionNonZero_f[9] then AddCo_f(9,-dg.InTwoStars(i,j));
    if AnyContributionNonZero_f[10] then AddCo_f(10,-dg.TwoPaths(j,i));
(* loop along all effects with a nonzero evaluation function parameter *)
    Ef := FirstMod_f;
    h1 := 0;
     repeat
        (* evaluation function: *)
        if Ef.hasweight_f then
          begin
          inc(h1);
          if (Assigned(Ef.f1ijc)) then
            s[h1] := s[h1] - Ef.f1ijc(dg,i,j,1,1,Ef.parameter);
          end;
        hasnextmod := Ef.HasNextMod_f;
        if  hasnextmod then Ef := Ef.NextMod_f;
     until (not hasnextmod);
   end; (* if t > 0 *)
end; (* ChangeStatistics *)
}

procedure TEffects.ChangeStatistics(const i,j: integer;
         var s: pvec; const t,tr: shortint;
         const generate: boolean; const dg: TValuedDigraph);
(* Adds the change statistics (effect h1) for toggling tie variable (i,j)
   to array s[j,h+h1];
   t is the current value of dg(i,j), tr is the current value of dg(j,i).
   Used for ERGM, i.e., observations = 1.
   Generate indicates that this is called for generating a MH step,
   (not generate) for calculating statistics. *)
var Ef: TEffect;
  h,sh: integer;
  co: TConfigWeight0;
  hasnexteffect: boolean;
  off: TFileStream;
  ncl: array of integer;
begin
//assignfile(off,'off.out');
//append(off);
  if generate then
    begin
    if (not Assigned(FirstMod_f)) then exit;

    if (t >= 1) then co[0] := 1.0 else co[0] := 0.0;
    co[1] := 1.0;
    if (tr >= 1) then co[2] := 1.0 else co[2] := 0.0;
    if AnyContributionNonZero_f[3] then co[3] := dg.OutDegree(i)
       else co[3] := 0.0;
    if AnyContributionNonZero_f[4] then co[4] := dg.InDegree(i)
       else co[4] := 0.0;
    if AnyContributionNonZero_f[5] then co[5] := dg.OutDegree(j)
       else co[5] := 0.0;
    if AnyContributionNonZero_f[6] then co[6] := dg.InDegree(j)
       else co[6] := 0.0;
    if AnyContributionNonZero_f[7] then co[7] := dg.TwoPaths(i,j)
       else co[7] := 0.0;
    if AnyContributionNonZero_f[8] then co[8] := dg.OutTwoStars(i,j)
       else co[8] := 0.0;
    if AnyContributionNonZero_f[9] then co[9] := dg.InTwoStars(i,j)
       else co[9] := 0.0;
    if AnyContributionNonZero_f[10] then co[10] := dg.TwoPaths(j,i)
       else co[10] := 0.0;
    end
   else
    begin
    if (t >= 1) then co[0] := 1.0 else co[0] := 0.0;
    co[1] := 1.0;
    if (tr >= 1) then co[2] := 1.0 else co[2] := 0.0;
    co[3] := dg.OutDegree(i);
    co[4] := dg.InDegree(i);
    co[5] := dg.OutDegree(j);
    co[6] := dg.InDegree(j);
    co[7] := dg.TwoPaths(i,j);
    co[8] := dg.OutTwoStars(i,j);
    co[9] := dg.InTwoStars(i,j);
    co[10] := dg.TwoPaths(j,i);
    end;

//setlength(ncl,11);


//writeln(off,i:4,j:4,t:4);
//writeln(off,'==========');

  if generate then
    begin
    if (pclique_max > 2) then
      dg.AllCliques_ij(false,off,pclique_min,pclique_max,i,j,t,extraclique);

//    if (param1_rclique_s > 2) or (param2_rclique > 2) then
//    begin
//writeln(off,'param  ',param1_rclique:4,' ',param2_rclique:4);
//      dg.All_red1_Cliques_ij(false,off,param1_rclique,param2_rclique,i,j,t,extrared1_clique);
//    dg.All_red1_Cliques_ij(true,off,2,5,i,j,t,extrared1_clique);
//    end;
    end
     else
    begin
    if (pclique_max_s > 2) then
      dg.AllCliques_ij(false,off,pclique_min_s,pclique_max_s,i,j,t,extraclique);

//    if (param1_rclique_s > 2) or (param2_rclique > 2) then  (* <<<<<>>>>> *)
//      dg.All_red1_Cliques_ij(false,off,param1_rclique_s,param2_rclique_s,i,j,t,extrared1_clique);
    end;

//dg.AllCliques_ij(true,off,3,5,i,j,t,ncl);
//for h:=2 to 5 do write(off,ncl[h]:2); writeln(off);


//closefile(off);
  sh := low(s);
  if generate then Ef := FirstMod_f else Ef := FirstStat_f;
  repeat
    for h:=0 to ConfigMax do (* @@@ try if this saves time : *)
// @@@@      if (h <= 0) or AnyContributionNonZero_f[h] then
        begin
        if (t <= 0) then s[sh] := s[sh] + Ef.ContributionWeight[h]*co[h]
         else s[sh] := s[sh] - Ef.ContributionWeight[h]*co[h];
        end;
    if Assigned(Ef.f1ijergm) then
        begin
        if (t <= 0) then s[sh] := s[sh] + Ef.f1ijergm(dg,i,j,0,Ef.parameter,Ef.RParameter)
          else  s[sh] := s[sh] - Ef.f1ijergm(dg,i,j,1,Ef.parameter,Ef.RParameter);
        end;
    inc(sh);
    if generate then
      begin
      hasnexteffect := Ef.HasNextMod_f;
      if hasnexteffect then Ef := Ef.NextMod_f;
      end
     else
      begin
      hasnexteffect := Ef.HasNextStat_f;
      if hasnexteffect then Ef := Ef.NextStat_f;
      end;
  until (not hasnexteffect);
end; (* ChangeStatistics *)

(* AA start  *)
function TEffect.OneChangeStatistic(const i,j: integer;
         const t,tr: shortint; const dg: TValuedDigraph): real;
(* Calculates the change statistic for this effect for toggling tie variable (i,j);
   t is the current value of dg(i,j), tr is the current value of dg(j,i).
   Used for ERGM, i.e., observations = 1.
   Generate indicates that this is called for generating a MH step,
   (not generate) for calculating statistics.
   This is modeled after ChangeStatistics. It is used only in interactions.
   *)
var co: TConfigWeight0;
  h: integer;
  s: real;
begin
    if (t >= 1) then co[0] := 1.0 else co[0] := 0.0;
    co[1] := 1.0;
    if (tr >= 1) then co[2] := 1.0 else co[2] := 0.0;
    if HasContribution[3] then co[3] := dg.OutDegree(i)
       else co[3] := 0.0;
    if HasContribution[4] then co[4] := dg.InDegree(i)
       else co[4] := 0.0;
    if HasContribution[5] then co[5] := dg.OutDegree(j)
       else co[5] := 0.0;
    if HasContribution[6] then co[6] := dg.InDegree(j)
       else co[6] := 0.0;
    if HasContribution[7] then co[7] := dg.TwoPaths(i,j)
       else co[7] := 0.0;
    if HasContribution[8] then co[8] := dg.OutTwoStars(i,j)
       else co[8] := 0.0;
    if HasContribution[9] then co[9] := dg.InTwoStars(i,j)
       else co[9] := 0.0;
    if HasContribution[10] then co[10] := dg.TwoPaths(j,i)
       else co[10] := 0.0;

{  if generate then
    begin
    if (param2_clique > 2) then
      dg.AllCliques_ij(false,off,param1_clique,param2_clique,i,j,t,extraclique);

    if (param2_rclique > 2) then
      dg.All_red1_Cliques_ij(false,off,param1_rclique,param2_rclique,i,j,t,extrared1_clique);
    end
     else
    begin
    if (param2_clique_s > 2) then
      dg.AllCliques_ij(false,off,param1_clique_s,param2_clique_s,i,j,t,extraclique);

    if (param2_rclique > 2) then  (* <<<<<>>>>> *)
      dg.All_red1_Cliques_ij(false,off,param1_rclique_s,param2_rclique_s,i,j,t,extrared1_clique);
    end;
}
  s := 0.0;
  for h:=0 to ConfigMax do s := s + ContributionWeight[h]*co[h];
  if Assigned(f1ijergm) then
    s := s + f1ijergm(dg,i,j,0,parameter,RParameter);
  OneChangeStatistic := s;
end; (* OneChangeStatistic *)
(* AA end *)

function TEffects.ChangeLogOdds(const s: pvec): real;
(* The change log odds for a vector s of change statistics *)
var Ef: TEffect;
  sh: integer;
  hasnexteffect: boolean;
begin
  Result := 0.0;
  if (not Assigned(FirstMod_f)) then exit;
  sh := low(s);
  Ef := FirstMod_f;
  repeat
    Result := Result + Ef.alpha_f*s[sh];
    inc(sh);
    hasnexteffect := Ef.HasNextMod_f;
    if hasnexteffect then Ef := Ef.NextMod_f;
  until (not hasnexteffect);
end; (* ChangeLogOdds *)


procedure TEffects.CalcStatistics(var s: pvec; const symm: boolean;
                    var od: array of integer; const generate: boolean;
                                  const dg: TValuedDigraph);
(* Calculate the statistics s by successively adding
   each of the ties in dg to an auxiliary digraph dg0.
   Used for ERGM, i.e., observations = 1.
   Generate indicates that this is called for generating a MH step,
   (not generate) for calculating statistics.
   if (not generate), then the ordered out-degrees are stored in od. *)
var it,itr: TDoublyLinkedDoubleInt;
  i,j,ptr,ptrn,n,lo: integer;
  s0: pvec;
  od0: array of integer;
  dg0: TValuedDigraph;
  Ef: TEffect;
  sh: integer;
  hasnexteffect: boolean;
begin
  n := dg.Order;
  dg0 := TValuedDigraph.Create(n);

(* First set s[.] to the values for the empty graph: *)
  if generate then
    begin
    if (not Assigned(FirstMod_f)) then exit;
    sh := low(s);
    Ef := FirstMod_f;
    repeat
      if Assigned(Ef.f1i) then s0[sh] := Ef.f1(dg,1,Ef.Parameter) else s0[sh] := 0.0;
      inc(sh);
      hasnexteffect := Ef.HasNextMod_f;
      if hasnexteffect then Ef := Ef.NextMod_f;
    until (not hasnexteffect);
    end
   else
    begin
    if (not Assigned(FirstStat_f)) then exit;
    sh := low(s);
    Ef := FirstStat_f;
    repeat
      if Assigned(Ef.f1i) then s0[sh] := Ef.f1(dg,1,Ef.Parameter) else s0[sh] := 0.0;
      inc(sh);
      hasnexteffect := Ef.HasNextStat_f;
      if hasnexteffect then Ef := Ef.NextStat_f;
    until (not hasnexteffect);
    end;
(* Then sequentially add all ties to dg0 and add the change statistics to s0 *)
  if symm then
    for i:=2 to n do
     begin
     it := dg.OutTies[i].Ffirst;
     j := 0;
     while (it.HasNext and (j < i)) do
       begin
       it := it.Next;
       j := it.Value1;
(* dg has a (i,j) tie *)
       if (j < i) then
         begin
         ChangeStatistics(i,j,s0,0,1,generate,dg0);
         dg0.AddTie(i,j,1,true);
         dg0.AddTie(j,i,1,true);
         end;
       end;
     end (* for i *)
   else (* not symm *)
   for i:=1 to n do
    begin
    it := dg.OutTies[i].Ffirst;
(* itr is used to calculate ptr. *)
    itr := dg.InTies[i].Ffirst;
    if itr.HasNext then
      begin
      ptr := 0;
      itr := itr.next;
      ptrn := itr.Value1;
      if (ptrn > i) then ptrn := n+1; (* there is no dg0(j,i) tie at all *)
      end
     else
      begin
      ptr := n+1;
      ptrn := ptr;
      end;
    while it.HasNext do
      begin
      it := it.Next;
      j := it.Value1;
(* dg has a (i,j) tie *)
      while (ptr < j) do if itr.HasNext then
        begin
        ptr := ptrn;
        itr := itr.next;
        ptrn := itr.Value1;
        if (ptrn > i) then ptrn := n+1; (* then dg(j,i) is a tie, but dg0(j,i) is not *)
        end
       else
        begin
        ptr := ptrn;
        ptrn := n+1;
        end;
(* ptr is the smallest j' >= j for which dg0 has a (j', i) tie;
   and n+1 if there is no such tie. *)
      if (ptr <= j) then ChangeStatistics(i,j,s0,0,1,generate,dg0)
       else ChangeStatistics(i,j,s0,0,0,generate,dg0);
      dg0.AddTie(i,j,1,true);
      end; (* while it.Hasnext *)
    end; (* for i *)

   for i := NumberStats_f downto 1 do s[i] := s0[i]; // this was in error until version 152 ! before it said NumberEffects.
//   for i := NumberEffects downto 1 do s[i] := s0[i];
   dg0.Destroy;

   if (not generate) then
     begin
     SetLength(od0,n+1);
     for i:= n downto 1 do
       begin
       od0[i] := dg.OutDegree(i);
       end;
     quicksort(od0,1,n);
     lo := low(od)-1;
     for i := n downto 1 do od[lo+i] := od0[i];
     end;
end; (* CalcStatistics *)


procedure TEFfects.UtilityComponents(const i,k,h,n: integer;
         var s: rrarray; perm: array of boolean;
         const dg: TValuedDigraph);
(* Stores the utility function contributions (h1) for changing tie variable (i,j)
   in array s[j,h+h1] , for all j with perm[j]; k is the period;
   note that j=i is a formal way of representing no change, hence contribution 0;
   first evaluation function, then endowment function *)
var Ef: TEffect;
    h1,h2,j,pt,ptr: integer;
    it,itr: TDoublyLinkedDoubleInt;
    hasnextmod: boolean;
    Ef_f1: real;
function prevouttie: integer;
(* next element with tie from i, going from high to low *)
begin
  Result := it.Value1;
  if it.hasprev then it := it.prev
   else Result := 0;
end; (* prevouttie *)
function previntie: integer;
(* next element with tie to i, going from high to low *)
begin
  Result := itr.Value1;
  if itr.hasprev then itr := itr.prev
   else Result := 0;
end; (* previntie *)
procedure AddCo_f(const hh: integer; const co: real);
var coo: real;
begin
  coo := co;
  Ef := FirstMod_f;
  h1 := h;
  repeat
    (* evaluation function only: *)
    inc(h1);
    if Ef.HasContribution[hh] then
      s[j,h1] := s[j,h1] + Ef.ContributionWeight[hh]*coo;
    hasnextmod := Ef.HasNextMod_f;
    if  hasnextmod then Ef := Ef.NextMod_f;
   until (not hasnextmod);
end; (* AddCo_f *)
procedure SubtractCo(const hh: integer; const co: real);
var Ef_co,coo: real;
begin
  Ef := FirstMod;
  h1 := h;
  h2 := h1 + num_f;
  coo := co;
  repeat
    if Ef.HasContribution[hh] then Ef_co := Ef.ContributionWeight[hh]*coo
     else Ef_co := 0.0;
    (* evaluation function: *)
    if Ef.hasweight_f then begin
      inc(h1);
      if Ef.HasContribution[hh] then s[j,h1] := s[j,h1] - Ef_co;
     end;
    (* endowment function: *)
    if Ef.hasweight_g then begin
      inc(h2);
      if Ef.HasContribution[hh] then s[j,h2] := s[j,h2] - Ef_co;
     end;
    hasnextmod := Ef.HasNextMod;
    if hasnextmod then Ef := Ef.NextMod;
   until (not hasnextmod);
end; (* SubtractCo *)
begin (* UtilityComponents *)
  if (not Assigned(FirstMod)) then exit;
(* First the set of j with a tie (i,j), and then the set of j without a tie,
   will be visited - both in the order from high to low. *)
(* First the values of j with a tie (i,j);
   evaluation and endowment functions: *)
  it := dg.OutTies[i].Last;
  itr := dg.InTies[i].Last;
  ptr := previntie;
  while it.HasPrev do
   begin
   j := it.Value1;
(* j is the next largest value for which there is a tie (i,j) *)
   while (ptr > j) do ptr := previntie;
(* ptr is the largest value <= j, for which there is a tie (ptr,i) *)
   if (perm[j] and (j <> i)) then
     begin
     SubtractCo(0,1.0);
     SubtractCo(1,1.0);
     if (ptr >= j) (* this means there is an incoming tie (j,i) *)
       then SubtractCo(2,1.0);
     if AnyContributionNonZero[3] then SubtractCo(3,dg.OutDegree(i));
     if AnyContributionNonZero[4] then SubtractCo(4,dg.InDegree(i));
     if AnyContributionNonZero[5] then SubtractCo(5,dg.OutDegree(j));
     if AnyContributionNonZero[6] then SubtractCo(6,dg.InDegree(j));
     if AnyContributionNonZero[7] then SubtractCo(7,dg.TwoPaths(i,j));
     if AnyContributionNonZero[8] then SubtractCo(8,dg.OutTwoStars(i,j));
     if AnyContributionNonZero[9] then SubtractCo(9,dg.InTwoStars(i,j));
     if AnyContributionNonZero[10] then SubtractCo(10,dg.TwoPaths(j,i));
(* loop along all effects with a nonzero evaluation or endowment function parameter *)
     Ef := FirstMod;
     h1 := h;
     h2 := h1 + num_f;
     repeat
       if (Assigned(Ef.f1ijc)) then Ef_f1 := Ef.f1ijc(dg,i,j,k,1,Ef.parameter)
        else Ef_f1 := 0.0;
(* Because of the construction of NextMod,
  at least one of the next if statements will be executed: *)
         (* evaluation function: *)
       if Ef.hasweight_f then
         begin
         inc(h1);
         s[j,h1] := s[j,h1] - Ef_f1;
         end;
         (* endowment function: *)
       if Ef.hasweight_g then
         begin
         inc(h2);
         s[j,h2] := s[j,h2] - Ef_f1;
         end;
       hasnextmod := Ef.HasNextMod;
       if hasnextmod then Ef := Ef.NextMod;
      until (not hasnextmod);
     end; (* if perm[j] *)
   it := it.Prev;
   end; (* while *)  (* not(it.HasPrev) <=> it = outties[i].First *)

(* Now the values of j without a tie (i,j); evaluation function only: *)
  it := dg.OutTies[i].Last;
  itr := dg.InTies[i].Last;
  pt := prevouttie;
  ptr := previntie;
//   for j:=dg.Order downto 1 do begin In view of multi-group projects, the extra parameter n is required.
  for j:=n downto 1 do begin
    if (j > pt) then  (* this condition means there is no tie (i,j) *)
      begin
      if (perm[j] and (j <> i)) then
        begin
(* Next part used in ChangeStatistics; but calling ChangeStatistics here
   seems to slow down operation. *)
        AddCo_f(1,1.0);
        if (ptr >= j) then AddCo_f(2,1.0);
           (* (ptr >= j) = there is an incoming tie (j,i) *)
        if AnyContributionNonZero_f[3] then AddCo_f(3,dg.OutDegree(i));
        if AnyContributionNonZero_f[4] then AddCo_f(4,dg.InDegree(i));
        if AnyContributionNonZero_f[5] then AddCo_f(5,dg.OutDegree(j));
        if AnyContributionNonZero_f[6] then AddCo_f(6,dg.InDegree(j));
        if AnyContributionNonZero_f[7] then AddCo_f(7,dg.TwoPaths(i,j));
        if AnyContributionNonZero_f[8] then AddCo_f(8,dg.OutTwoStars(i,j));
        if AnyContributionNonZero_f[9] then AddCo_f(9,dg.InTwoStars(i,j));
        if AnyContributionNonZero_f[10] then AddCo_f(10,dg.TwoPaths(j,i));
(* loop along all effects with a nonzero evaluation function parameter *)
        Ef := FirstMod_f;
        h1 := h;
         repeat
            (* evaluation function: *)
            if Ef.hasweight_f then
              begin
              inc(h1);
              if (Assigned(Ef.f1ijc)) then
                s[j,h1] := s[j,h1] + Ef.f1ijc(dg,i,j,k,0,Ef.parameter);
              end;
            hasnextmod := Ef.HasNextMod_f;
            if  hasnextmod then Ef := Ef.NextMod_f;
         until (not hasnextmod);
        end; (* if perm[j] *)
      end (* if j > pt *) else pt := prevouttie;
(* now pt is the largest value <= j, for which there is a tie (i,pt) *)
    if (j <= ptr) then ptr := previntie;
(* now ptr is the largest value <= j, for which there is a tie (ptr,i) *)
    end; (* for j *)
end; (* UtilityComponents *)


procedure quicksort(var a: array of integer; Lo,Hi: integer);
(* copied from Slib *)
procedure sort(l,r: integer);
var
  i,j,x,y: integer;
begin
  i:=l; j:=r; x:=a[(l+r) DIV 2];
  repeat
    while a[i]<x do i:=i+1;
    while x<a[j] do j:=j-1;
    if i<=j then
    begin
      y:=a[i]; a[i]:=a[j]; a[j]:=y;
      i:=i+1; j:=j-1;
    end;
  until i>j;
  if l<j then sort(l,j);
  if i<r then sort(i,r);
end;

begin (* quicksort *);
  sort(Lo,Hi);
end;

procedure TEffects.CalcFunctions_f
   (var fun: array of real; var ordeg: array of integer; const k: integer;
    const dg: TValuedDigraph; const COD: Boolean);
(* Calculation of the statistics for evaluation effects;
   must be called with k = thisperiod *)
var flo, odlo, odhi, i, j, h, jj: integer;
    reciproc: boolean;
    it,itr: TDoublyLinkedDoubleInt;
    Ef: Teffect;
    ConfigCount: array[1..ConfigMax] of Integer;
begin
  flo := Low(fun);
  for i := flo + NumberStats_f - 1 downto flo do fun[i] := 0.0;
  if (not Assigned(FirstStat_f)) then exit;

(* First calculate the sum of the statistics
   for the ConfigMax configurations: *)
  for i:=1 to ConfigMax do ConfigCount[i] := 0;
  for i:=1 to dg.Order do begin
    it := dg.OutTies[i].First;
    itr:= dg.InTies[i].First;
    jj := 0;
    while it.Hasnext do begin
(* find next j with an (i,j) tie *)
      it := it.Next;
      j := it.Value1;
      if (jj < j) then
        begin
        while (itr.HasNext) and (itr.Next.Value1 < j) do
          begin
          itr := itr.Next;
          end;
        if itr.HasNext then jj := itr.Next.Value1 else jj := succ(dg.Order);
        end;
(* jj is the lowest h >= j for which there is a tie h -> i *)
//      reciproc :=  (dg.TieValuedown(j,i) >= 1);
//     without jj:  if itr.HasNext then reciproc := (itr.Next.Value1 <= j) else reciproc := false;
      reciproc := (jj <= j);

      if TotalConfigNonZero_f[1] then Inc(ConfigCount[1]);
      if (reciproc and TotalConfigNonZero_f[2]) then Inc(ConfigCount[2]);
      if TotalConfigNonZero_f[3] then Inc(ConfigCount[3], dg.OutDegree(i));
      if TotalConfigNonZero_f[4] then Inc(ConfigCount[4], dg.InDegree(i));
      if TotalConfigNonZero_f[5] then Inc(ConfigCount[5], dg.OutDegree(j));
      if TotalConfigNonZero_f[6] then Inc(ConfigCount[6], dg.InDegree(j));
      if TotalConfigNonZero_f[7] then Inc(ConfigCount[7], dg.TwoPaths(i,j));
      if TotalConfigNonZero_f[8] then Inc(ConfigCount[8], dg.OutTwoStars(i,j));
      if TotalConfigNonZero_f[9] then Inc(ConfigCount[9], dg.InTwoStars(i,j));
      if TotalConfigNonZero_f[10] then Inc(ConfigCount[10], dg.TwoPaths(j,i));

      Ef := FirstStat_f;
      if Assigned(Ef.f1ij) then fun[flo] := fun[flo] + Ef.f1ij(dg,i,j,k,Ef.parameter);
      h := flo;
      while Ef.HasNextStat_f do begin
        Ef := Ef.NextStat_f;
        inc(h);
        if Assigned(Ef.f1ij) then fun[h] := fun[h] + Ef.f1ij(dg,i,j,k,Ef.parameter);
       end; (* while *)
     end; (* while it.HasNext *)
   end; (* for i *)

(* Now calculate the statistics as the weighted sums of these counts,
   and add the contributions from f1 : *)
    Ef := FirstStat_f;
    for i:=1 to ConfigMax do
      fun[flo] := fun[flo] + Ef.ConfigWeight[i]*ConfigCount[i];
    if Assigned (Ef.f1i) then fun[flo] := fun[flo] + Ef.f1(dg,k,Ef.parameter);
    h := flo;
    while Ef.HasNextStat_f do begin
      Ef := Ef.NextStat_f;
      inc(h);
      for i:=1 to ConfigMax do
        fun[h] := fun[h] + Ef.ConfigWeight[i]*ConfigCount[i];
      if Assigned(Ef.f1i) then fun[h] := fun[h] + Ef.f1(dg,k,Ef.parameter);
     end; (* while *)

(* Finally calculate the ordered outdegrees: *)
  odlo := Low(ordeg)-1;
  odhi :=  High(ordeg);
  for i:= odlo+1 to odhi do ordeg[i] := 0;
  if COD then begin
    for i:= dg.Order downto 1 do ordeg[odlo+i] := dg.Outdegree(i);
    quicksort(ordeg,1,dg.Order);
   end;
end; (* CalcFunctions_f *)


procedure TEffects.CalcFunctions_g
   (var fun: array of real; const k: integer; const dg: TValuedDigraph);
(* It will be more efficient to integrate CalcFunctions_g with CalcFunctions_f;
   if this can be handled well with the missings.
   But this function is not called extremely often. *)
(* Calculation of the statistics for endowment effects;
   must be called with k = thisperiod.
   dg must be a valued digraph with values +1 for ties existing at time k
   and retained in the current digraph
   (or with missing information at time k+1),
   values -1 for ties existing at time k and lost in the current digraph,
   and no other ties. *)
var flo, i, j, h, jj: integer;
    v: ShortInt;
    reciproc: boolean;
    it, itr: TDoublyLinkedDoubleInt;
    Ef: Teffect;
    ConfigCount: array[1..ConfigMax] of Integer;
begin
  flo := Low(fun);
  for i := flo + NumberStats_f + NumberStats_g - 1 downto flo + NumberStats_f
    do fun[i] := 0.0;
  if (not Assigned(FirstStat_g)) then exit;

(* First calculate the sum of the statistics
   for the ConfigMax configurations: *)
  for i:=1 to ConfigMax do ConfigCount[i] := 0;
//  itr:= dg.InTies[i].First;
//  jj := 0;
  for i:=1 to dg.Order do
    begin
    it := dg.OutTies[i].First;
    while it.Hasnext do
      begin
(* find next j with an (i,j) tie *)
      it := it.Next;
      j := it.Value1;
      v := it.Value2;
      if (v <= -1) then
        begin  (* the tie was lost *)
// hierin zit een fout:
//        if (jj < j) then
//          begin
//          while (itr.HasNext) and (itr.Next.Value1 < j) do
//            begin
//            itr := itr.Next;
//            end;
//          if itr.HasNext then jj := itr.Next.Value1 else jj := succ(dg.Order);
//          end;
(* jj is the lowest h >= j for which there is a tie h -> i *)
//       reciproc := (jj <= j);
       reciproc := (dg.TieValuedown(j,i) <> 0);
(* indicating that there was a reciprocating tie at the beginning of the period *)

        if TotalConfigNonZero_g[1] then Inc(ConfigCount[1]);
        if (reciproc and TotalConfigNonZero_g[2]) then Inc(ConfigCount[2]);
(* The dg.OutDegree(i) etc. are counted over all non-zero ties dg(i,j),
   irrespective of their value2 (being -1 or +1).
   This refers exactly to the network at the start of the period. *)
        if TotalConfigNonZero_g[3] then Inc(ConfigCount[3], dg.OutDegree(i));
        if TotalConfigNonZero_g[4] then Inc(ConfigCount[4], dg.InDegree(i));
        if TotalConfigNonZero_g[5] then Inc(ConfigCount[5], dg.OutDegree(j));
        if TotalConfigNonZero_g[6] then Inc(ConfigCount[6], dg.InDegree(j));
        if TotalConfigNonZero_g[7] then Inc(ConfigCount[7], dg.TwoPaths(i,j));
        if TotalConfigNonZero_g[8] then Inc(ConfigCount[8], dg.OutTwoStars(i,j));
        if TotalConfigNonZero_g[9] then Inc(ConfigCount[9], dg.InTwoStars(i,j));
        if TotalConfigNonZero_g[10] then Inc(ConfigCount[10], dg.TwoPaths(j,i));

        Ef := FirstStat_g;
        h := flo + NumberStats_f + SkipNumber_g[Ef.index];
(* h is the number for outputting the corresponding statistic *)
        if Assigned(Ef.f1ij) then fun[h] := fun[h] + Ef.f1ij(dg,i,j,k,Ef.parameter);
        while Ef.HasNextStat_g do
          begin
          Ef := Ef.NextStat_g;
          inc(h, 1 + SkipNumber_g[Ef.index]);
          if Assigned(Ef.f1ij) then fun[h] := fun[h] + Ef.f1ij(dg,i,j,k,Ef.parameter);
          end; (* while *)
        end; (* if v <= -1 *)
      end; (* while it.HasNext *)
    end; (* for i *)
(* Now calculate the statistics as the weighted sums of these counts : *)
    Ef := FirstStat_g;
    h := flo + NumberStats_f + SkipNumber_g[Ef.index];
    for i:=1 to ConfigMax do
      fun[h] := fun[h] + Ef.ConfigWeight[i]*ConfigCount[i];
    while Ef.HasNextStat_g do begin
      Ef := Ef.NextStat_g;
      inc(h, 1 + SkipNumber_g[Ef.index]);
      for i:=1 to ConfigMax do
        fun[h] := fun[h] + Ef.ConfigWeight[i]*ConfigCount[i];
     end; (* while *)
// strange error - a line had to be omitted, was done 3.17w.      
end; (* CalcFunctions_g *)

function TEffect.f1(const dg: TValuedDigraph;
                                     const k,par: integer): real;
var i: integer;
    rs: real;
begin
  result := 0.0;
  for i:= dg.Order downto 1 do result := result + f1i(dg,i,k,par);
end;

function TEffects.Interactiont_a
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Interaction effect where the first effect p1 corresponds to the
   ego effect of actor variable fp1, which has no contributionweight *)
var p1, p2, fp1, fp2: integer;
  x2: real;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  x2 := 0.0;
  if Effect[p2].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p2].HasContribution[0] then x2 := x2 + Effect[p2].ContributionWeight[0];
    if Effect[p2].HasContribution[1] then x2 := x2 + Effect[p2].ContributionWeight[1];
    if Effect[p2].HasContribution[2] and (dg.TieValue(j,i) >= 1)
       then x2 := x2 + Effect[p2].ContributionWeight[2];
    if Effect[p2].HasContribution[3] then
      x2 := x2 + Effect[p2].ContributionWeight[3]*dg.OutDegree(i);
    if Effect[p2].HasContribution[4] then
      x2 := x2 + Effect[p2].ContributionWeight[4]*dg.InDegree(i);
    if Effect[p2].HasContribution[5] then
      x2 := x2 + Effect[p2].ContributionWeight[5]*dg.OutDegree(j);
    if Effect[p2].HasContribution[6] then
      x2 := x2 + Effect[p2].ContributionWeight[6]*dg.InDegree(j);
    if Effect[p2].HasContribution[7] then
      x2 := x2 + Effect[p2].ContributionWeight[7]*dg.TwoPaths(i,j);
    if Effect[p2].HasContribution[8] then
      x2 := x2 + Effect[p2].ContributionWeight[8]*dg.OutTwoStars(i,j);
    if Effect[p2].HasContribution[9] then
      x2 := x2 + Effect[p2].ContributionWeight[9]*dg.InTwoStars(i,j);
    if Effect[p2].HasContribution[10] then
      x2 := x2 + Effect[p2].ContributionWeight[10]*dg.TwoPaths(j,i);
    end;

    if assigned(Effect[p2].f1ijc) then
      Result := Effect[p1].f1ijc(dg,i,j,k,tv,fp1)*(x2+Effect[p2].f1ijc(dg,i,j,k,tv,fp2))
     else
      Result := Effect[p1].f1ijc(dg,i,j,k,tv,fp1)*x2;
end;


function TEffects.Interaction_a
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
(* Interaction effect where the first effect p1 corresponds to the
   ego effect of actor variable fp1, which has no contributionweight *)
var p1, p2, fp1, fp2: integer;
  x2: real;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  x2 := 0.0;
  if Effect[p2].HasAnyContribution then
    begin
    if Effect[p2].HasContribution[1] then x2 := x2 + Effect[p2].Configweight[1];
    if Effect[p2].HasContribution[2] and (dg.TieValue(j,i) >= 1)
       then x2 := x2 + Effect[p2].Configweight[2];
    if Effect[p2].HasContribution[3] then
      x2 := x2 + Effect[p2].Configweight[3]*dg.OutDegree(i);
    if Effect[p2].HasContribution[4] then
      x2 := x2 + Effect[p2].Configweight[4]*dg.InDegree(i);
    if Effect[p2].HasContribution[5] then
      x2 := x2 + Effect[p2].Configweight[5]*dg.OutDegree(j);
    if Effect[p2].HasContribution[6] then
      x2 := x2 + Effect[p2].Configweight[6]*dg.InDegree(j);
    if Effect[p2].HasContribution[7] then
      x2 := x2 + Effect[p2].Configweight[7]*dg.TwoPaths(i,j);
    if Effect[p2].HasContribution[8] then
      x2 := x2 + Effect[p2].Configweight[8]*dg.OutTwoStars(i,j);
    if Effect[p2].HasContribution[9] then
      x2 := x2 + Effect[p2].Configweight[9]*dg.InTwoStars(i,j);
    if Effect[p2].HasContribution[10] then
      x2 := x2 + Effect[p2].Configweight[10]*dg.TwoPaths(j,i);
    end;

  if assigned(Effect[p2].f1ij) then
    Result := Effect[p1].f1ij(dg,i,j,k,fp1)*
               (x2+Effect[p2].f1ij(dg,i,j,k,fp2))
   else
    Result := Effect[p1].f1ij(dg,i,j,k,fp1)*x2;
end;

function TEffects.Interactiontot_a
           (const dg: TValuedDigraph; const i,k: integer; const par: integer): real;
(* Interaction effect where the first effect p1 corresponds to the
   ego effect of actor variable fp1, which has no contributionweight *)
var p1, p2, fp1, fp2: integer;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  if (not assigned(Effect[p2].f1i)) then
    begin
    result := 0.0;
    exit;
    end;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  Result := Effect[p1].f1ij(dg,i,1,k,fp1)*Effect[p2].f1i(dg,i,k,fp2);
(* This utilizes the fact that for this type of Effect[p1],
   f1ij is the covariate value of actor i, and does not depend on j *)
end;


function TEffects.Interactiont_b
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Interaction effect where both effects are dyadic. *)
var p1, p2, fp1, fp2: integer;
  x1, x2: real;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  x1 := 0.0;
  if Effect[p1].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p1].HasContribution[0] then x1 := x1 + Effect[p1].ContributionWeight[0];
    if Effect[p1].HasContribution[1] then x1 := x1 + Effect[p1].ContributionWeight[1];
    end;
(* It is not allowed that effect[p1] has non-dyadic weights. *)
  x2 := 0.0;
  if Effect[p2].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p2].HasContribution[0] then x2 := x2 + Effect[p2].ContributionWeight[0];
    if Effect[p2].HasContribution[1] then x2 := x2 + Effect[p2].ContributionWeight[1];
    end;
  if (Effect[p1].HasContribution[2] or Effect[p2].HasContribution[2]) then
    begin
    if (dg.TieValue(j,i) >= 1) then
      begin
      if Effect[p1].HasContribution[2] then
        x1 := x1 + Effect[p1].ContributionWeight[2];
      if Effect[p2].HasContribution[2] then
        x2 := x2 + Effect[p2].ContributionWeight[2];
      end;
    end;

(* changed 3.305 : *)
  if assigned(Effect[p1].f1ij) then
    begin
    if assigned(Effect[p2].f1ij) then
      Result := (x1+Effect[p1].f1ijc(dg,i,j,k,tv,fp1))*(x2+Effect[p2].f1ijc(dg,i,j,k,tv,fp2))
     else
      Result := (x1+Effect[p1].f1ijc(dg,i,j,k,tv,fp1))*(x2);
    end
   else
    begin
    if assigned(Effect[p2].f1ij) then
      Result := (x1)*(x2+Effect[p2].f1ijc(dg,i,j,k,tv,fp2))
     else
      Result := (x1)*(x2);
    end;

// earlier:  Result := (x1+Effect[p1].f1ijc(dg,i,j,k,tv,fp1))*
//             (x2+Effect[p2].f1ijc(dg,i,j,k,tv,fp2));
end;

function TEffects.Interaction_b
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
(* Interaction effect where both effects are dyadic. *)
var p1, p2, fp1, fp2: integer;
  x1, x2: real;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
   else x1 := 0.0;
(* It is not allowed that effect[p1] has non-dyadic weights. *)
  if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
   else x2 := 0.0;
  if (Effect[p1].HasContribution[2] or Effect[p2].HasContribution[2]) then
    begin
    if (dg.TieValue(j,i) >= 1) then
      begin
      if Effect[p1].HasContribution[2] then
        x1 := x1 + Effect[p1].ContributionWeight[2];
      if Effect[p2].HasContribution[2] then
        x2 := x2 + Effect[p2].ContributionWeight[2];
      end;
    end;

(* changed 3.17n : *)
  if assigned(Effect[p1].f1ij) then
    begin
    if assigned(Effect[p2].f1ij) then
      Result := (x1+Effect[p1].f1ij(dg,i,j,k,fp1))*(x2+Effect[p2].f1ij(dg,i,j,k,fp2))
     else
      Result := (x1+Effect[p1].f1ij(dg,i,j,k,fp1))*(x2);
    end
   else
    begin
    if assigned(Effect[p2].f1ij) then
      Result := (x1)*(x2+Effect[p2].f1ij(dg,i,j,k,fp2))
     else
      Result := (x1)*(x2);
    end;

// earlier:  Result := (x1+Effect[p1].f1ij(dg,i,j,k,fp1))*(x2+Effect[p2].f1ij(dg,i,j,k,fp2));
end;

function TEffects.Interaction_ergm(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
(* ERGM interaction effect. *)
var p1, p2, fp1, fp2: integer;
  x1, x2: real;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
(* The following selection uses the fact that
   for ergm effects, either Contriw is <> 0, or f1ijergm is defined,
   but not both. *)
(* It is not allowed that effect[p1,p2] has non-dyadic Contriw weights. *)
  if assigned(Effect[p1].f1ijergm) then
    begin
    if assigned(Effect[p2].f1ijergm) then
      Result := Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar)*
                 Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar)
     else
      begin
      if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
        else x2 := 0.0;
      if Effect[p2].HasContribution[2] then
        begin
          if ((dg.TieValue(j,i) >= 1) and Effect[p2].HasContribution[2]) then
           x2 := x2 + Effect[p2].ContributionWeight[2];
        end;
      Result := x2*Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar);
      end;
    end
   else
    begin
    if assigned(Effect[p2].f1ijergm) then
      begin
      if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
        else x1 := 0.0;
      if Effect[p1].HasContribution[2] then
        begin
          if ((dg.TieValue(j,i) >= 1) and Effect[p1].HasContribution[2]) then
           x1 := x1 + Effect[p1].ContributionWeight[2];
        end;
      Result := x1*Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar);
      end
     else
      begin
      if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
       else x1 := 0.0;
      if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
       else x2 := 0.0;
      if (Effect[p1].HasContribution[2] or Effect[p2].HasContribution[2]) then
        begin
        if (dg.TieValue(j,i) >= 1) then
          begin
          if Effect[p1].HasContribution[2] then
            x1 := x1 + Effect[p1].ContributionWeight[2];
          if Effect[p2].HasContribution[2] then
            x2 := x2 + Effect[p2].ContributionWeight[2];
          end;
        end;
      Result := x1*x2;
      end;
    end;


(*  if assigned(Effect[p1].f1ijergm) then
    begin
    if assigned(Effect[p2].f1ijergm) then
      Result := (x1+Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar))*
            (x2+Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar))
     else Result := (x1+Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar))*x2;
    end
   else
    begin
    if assigned(Effect[p2].f1ijergm) then
      Result := x1*(x2+Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar))
     else Result := x1*x2;
    end;
    *)
end; (* Interaction_ergm *)

function TEffects.Interaction_ergm11(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
(* ERGM interaction effect. *)
var p1, p2, fp1, fp2: integer;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
(* The following selection uses the fact that
   for ergm effects, either Contriw is <> 0, or f1ijergm is defined,
   but not both. *)
  Result := Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar)*
             Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar)
end; (* Interaction_ergm11 *)

(* AA begin *)
function TEffects.Interaction_ergm10(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
(* ERGM interaction effect. *)
var p1, p2, fp1: integer;
  tr: shortint;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
(* Note that for ergm effects, either Contriw is <> 0, or f1ijergm is defined,
   but not both. *)
  tr := dg.TieValue(j,i);
  Result := Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar)
               *Effect[p2].OneChangeStatistic(i,j,t,tr,dg);
end; (* Interaction_ergm10 *)

function TEffects.Interaction_ergm01(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
(* ERGM interaction effect. *)
var p1, p2, fp2: integer;
  tr: shortint;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  fp2 := Effect[p2].FParameter;
(* Note that for ergm effects, either Contriw is <> 0, or f1ijergm is defined,
   but not both. *)
  tr := dg.TieValue(j,i);
  Result := Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar)
               *Effect[p1].OneChangeStatistic(i,j,t,tr,dg);
end; (* Interaction_ergm01 *)

function TEffects.Interaction_ergm00(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
(* ERGM interaction effect. *)
var p1, p2: integer;
  tr: shortint;
begin
  p2 := (par mod 1000) - 1;
  p1 := ((par div 1000) mod 1000)-1;
  tr := dg.TieValue(j,i);
  Result := Effect[p1].OneChangeStatistic(i,j,t,tr,dg)
               *Effect[p2].OneChangeStatistic(i,j,t,tr,dg);
end; (* Interaction_ergm00 *)
(* AA end *)

function TEffects.Interaction3t_a
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Interaction effect where the first and second effects p1, p2 corresponds to the
   ego effects of actor variables fp1, fp2, which have no contributionweight *)
var p1, p2, p3, fp1, fp2, fp3: integer;
  x3: real;
begin
  p3 := (par mod 1000) - 1;
  p2 := ((par div 1000) mod 1000)-1;
  p1 := ((par div 1000000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  x3 := 0.0;
  if Effect[p3].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p3].HasContribution[0] then x3 := x3 + Effect[p3].ContributionWeight[0];
    if Effect[p3].HasContribution[1] then x3 := x3 + Effect[p3].ContributionWeight[1];
    if Effect[p3].HasContribution[2] and (dg.TieValue(j,i) >= 1)
       then x3 := x3 + Effect[p3].ContributionWeight[2];
    if Effect[p3].HasContribution[3] then
      x3 := x3 + Effect[p3].ContributionWeight[3]*dg.OutDegree(i);
    if Effect[p3].HasContribution[4] then
      x3 := x3 + Effect[p3].ContributionWeight[4]*dg.InDegree(i);
    if Effect[p3].HasContribution[5] then
      x3 := x3 + Effect[p3].ContributionWeight[5]*dg.OutDegree(j);
    if Effect[p3].HasContribution[6] then
      x3 := x3 + Effect[p3].ContributionWeight[6]*dg.InDegree(j);
    if Effect[p3].HasContribution[7] then
      x3 := x3 + Effect[p3].ContributionWeight[7]*dg.TwoPaths(i,j);
    if Effect[p3].HasContribution[8] then
      x3 := x3 + Effect[p3].ContributionWeight[8]*dg.OutTwoStars(i,j);
    if Effect[p3].HasContribution[9] then
      x3 := x3 + Effect[p3].ContributionWeight[9]*dg.InTwoStars(i,j);
    if Effect[p3].HasContribution[10] then
      x3 := x3 + Effect[p3].ContributionWeight[10]*dg.TwoPaths(j,i);
    end;

    if assigned(Effect[p3].f1ijc) then
      Result := Effect[p1].f1ijc(dg,i,j,k,tv,fp1)*
                Effect[p2].f1ijc(dg,i,j,k,tv,fp2)*(x3+Effect[p3].f1ijc(dg,i,j,k,tv,fp3))
     else
      Result := Effect[p1].f1ijc(dg,i,j,k,tv,fp1)*
                Effect[p2].f1ijc(dg,i,j,k,tv,fp2)*x3;
end;



function TEffects.Interaction3_a
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
(* Interaction effect where the first and second effects p1, p2 corresponds to the
   ego effects of actor variables fp1, fp2, which have no contributionweight *)
var p1, p2, p3, fp1, fp2, fp3: integer;
  x3: real;
begin
  p3 := (par mod 1000) - 1;
  p2 := ((par div 1000) mod 1000)-1;
  p1 := ((par div 1000000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  x3 := 0.0;
  if Effect[p3].HasAnyContribution then
    begin
    if Effect[p3].HasContribution[1] then x3 := x3 + Effect[p3].Configweight[1];
    if Effect[p3].HasContribution[2] and (dg.TieValue(j,i) >= 1)
       then x3 := x3 + Effect[p3].Configweight[2];
    if Effect[p3].HasContribution[3] then
      x3 := x3 + Effect[p3].Configweight[3]*dg.OutDegree(i);
    if Effect[p3].HasContribution[4] then
      x3 := x3 + Effect[p3].Configweight[4]*dg.InDegree(i);
    if Effect[p3].HasContribution[5] then
      x3 := x3 + Effect[p3].Configweight[5]*dg.OutDegree(j);
    if Effect[p3].HasContribution[6] then
      x3 := x3 + Effect[p3].Configweight[6]*dg.InDegree(j);
    if Effect[p3].HasContribution[7] then
      x3 := x3 + Effect[p3].Configweight[7]*dg.TwoPaths(i,j);
    if Effect[p3].HasContribution[8] then
      x3 := x3 + Effect[p3].Configweight[8]*dg.OutTwoStars(i,j);
    if Effect[p3].HasContribution[9] then
      x3 := x3 + Effect[p3].Configweight[9]*dg.InTwoStars(i,j);
    if Effect[p3].HasContribution[10] then
      x3 := x3 + Effect[p3].Configweight[10]*dg.TwoPaths(j,i);
    end;

  if assigned(Effect[p3].f1ij) then
    Result := Effect[p1].f1ij(dg,i,j,k,fp1)*Effect[p2].f1ij(dg,i,j,k,fp2)*
               (x3+Effect[p3].f1ij(dg,i,j,k,fp3))
   else
    Result := Effect[p1].f1ij(dg,i,j,k,fp1)*Effect[p2].f1ij(dg,i,j,k,fp2)*x3;
end;

function TEffects.Interaction3tot_a
           (const dg: TValuedDigraph; const i,k: integer; const par: integer): real;
(* Interaction effect where the first and second effects p1, p2 corresponds to the
   ego effects of actor variables fp1, fp2, which have no contributionweight *)
var p1, p2, p3, fp1, fp2, fp3: integer;
  x3: real;
begin
  p3 := (par mod 1000) - 1;
  p2 := ((par div 1000) mod 1000)-1;
  p1 := ((par div 1000000) mod 1000)-1;
  if (not assigned(Effect[p3].f1i)) then
    begin
    result := 0.0;
    exit;
    end;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  Result := Effect[p1].f1ij(dg,i,1,k,fp1)*Effect[p2].f1ij(dg,i,1,k,fp2)*
                 Effect[p3].f1i(dg,i,k,fp3);
(* This utilizes the fact that for this type of Effect[p1] and Effect[p2],
   f1ij is the covariate value of actor i, and does not depend on j *)
end;


function TEffects.Interaction3t_b
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Three-way interaction effect where all three effects are dyadic. *)
var p1, p2, p3, fp1, fp2, fp3: integer;
  x1, x2, x3, y1, y2, y3: real;
begin
  p3 := (par mod 1000) - 1;
  p2 := ((par div 1000) mod 1000)-1;
  p1 := ((par div 1000000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  x1 := 0.0;
  if Effect[p1].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p1].HasContribution[0] then x1 := x1 + Effect[p1].ContributionWeight[0];
    if Effect[p1].HasContribution[1] then x1 := x1 + Effect[p1].ContributionWeight[1];
    end;
(* It is not allowed that effect[p1] has non-dyadic weights. *)
  x2 := 0.0;
  if Effect[p2].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p2].HasContribution[0] then x2 := x2 + Effect[p2].ContributionWeight[0];
    if Effect[p2].HasContribution[1] then x2 := x2 + Effect[p2].ContributionWeight[1];
    end;
  x3 := 0.0;
  if Effect[p3].HasAnyContribution then
    begin
    if (tv >= 1) and Effect[p3].HasContribution[0] then x3 := x3 + Effect[p3].ContributionWeight[0];
    if Effect[p3].HasContribution[1] then x3 := x3 + Effect[p3].ContributionWeight[1];
    end;
  if (Effect[p1].HasContribution[2] or Effect[p2].HasContribution[2]
   or Effect[p3].HasContribution[2]) then
    begin
    if (dg.TieValue(j,i) >= 1) then
      begin
      if Effect[p1].HasContribution[2] then
        x1 := x1 + Effect[p1].ContributionWeight[2];
      if Effect[p2].HasContribution[2] then
        x2 := x2 + Effect[p2].ContributionWeight[2];
      if Effect[p3].HasContribution[2] then
        x3 := x3 + Effect[p3].ContributionWeight[2];
      end;
    end;
  if assigned(Effect[p1].f1ijc) then y1 := Effect[p1].f1ijc(dg,i,j,k,tv,fp1) else y1 := 0.0;
  if assigned(Effect[p2].f1ijc) then y2 := Effect[p2].f1ijc(dg,i,j,k,tv,fp2) else y2 := 0.0;
  if assigned(Effect[p3].f1ijc) then y3 := Effect[p3].f1ijc(dg,i,j,k,tv,fp3) else y3 := 0.0;
  Result := (x1+y1)*(x2+y2)*(x3+y3);
end;

function TEffects.Interaction3_b
           (const dg: TValuedDigraph; const i,j,k: integer; const par: integer): real;
(* Three-way interaction effect where all three effects are dyadic. *)
var p1, p2, p3, fp1, fp2, fp3: integer;
  x1, x2, x3, y1, y2, y3: real;
begin
  p3 := (par mod 1000) - 1;
  p2 := ((par div 1000) mod 1000)-1;
  p1 := ((par div 1000000) mod 1000)-1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
   else x1 := 0.0;
(* It is not allowed that effect[p1] has non-dyadic weights. *)
  if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
   else x2 := 0.0;
  if Effect[p3].HasContribution[1] then x3 := Effect[p3].ContributionWeight[1]
   else x3 := 0.0;
  if (Effect[p1].HasContribution[2] or Effect[p2].HasContribution[2]
     or Effect[p3].HasContribution[2]) then
    begin
    if (dg.TieValue(j,i) >= 1) then
      begin
      if Effect[p1].HasContribution[2] then
        x1 := x1 + Effect[p1].ContributionWeight[2];
      if Effect[p2].HasContribution[2] then
        x2 := x2 + Effect[p2].ContributionWeight[2];
      if Effect[p3].HasContribution[2] then
        x3 := x3 + Effect[p3].ContributionWeight[2];
      end;
    end;
  if assigned(Effect[p1].f1ij) then y1 := Effect[p1].f1ij(dg,i,j,k,fp1) else y1 := 0.0;
  if assigned(Effect[p2].f1ij) then y2 := Effect[p2].f1ij(dg,i,j,k,fp2) else y2 := 0.0;
  if assigned(Effect[p3].f1ij) then y3 := Effect[p3].f1ij(dg,i,j,k,fp3) else y3 := 0.0;
  Result := (x1+y1)*(x2+y2)*(x3+y3);
end;

function TEffects.Interactiont_rec_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var p1,resu: integer;
    it1, it2: TDoublyLinkedDoubleInt;
begin
(*
See Siena manual, "additional interaction effects".
1002003, 2002003: reciprocity x transitive triplets
The following interactions are defined in Interactiont_var_trans in S_Effects:
1003def, 2003def, 3003def:
  (similarity on actor variable # def) x transitive triplets
11003def, 12003def, 13003def:
  (similarity on dyadic variable # def) x transitive triplets
*)
  Resu := 0;
  p1 := (par div 1000000);
  case p1 of
  1: begin
     if (dg.TieValue(j,i) >= 1) then
       begin
       it1 := dg.OutTies[i].Ffirst;
       it2 := dg.Inties[j].FFirst;
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       end; (* if *)
     it1 := dg.OutTies[i].Ffirst;
     it2 := dg.OutTies[j].Ffirst;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if (dg.TieValue(it1.Value1,i) >= 1) then Inc(resu);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 1 *)
  2: begin
     if (dg.TieValue(j,i) >= 1) then
       begin
       it1 := dg.OutTies[i].Ffirst;
       it2 := dg.Inties[j].FFirst;
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (dg.TieValue(j,it1.Value1) >= 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       end; (* if *)
     it1 := dg.OutTies[i].Ffirst;
     it2 := dg.OutTies[j].Ffirst;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if ((dg.TieValue(it1.Value1,i) >= 1)
             and (dg.TieValue(it1.Value1,j) >= 1)) then Inc(resu);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end;
  end; (* case *)
  Result := resu;
end; (* Interactiont_rec_trans *)

function TEffects.Interaction_rec_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;
begin
  Result := Interactiont_rec_trans(dg,i,j,k,1,par);
end; (* Interaction_rec_trans *)


function TEffects.Interaction3ergm_0(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var p1, p2, p3, fp1, fp2, fp3: integer;
begin
  p3 := (par mod 1000) -1;
  p2 := ((par div 1000) mod 1000) - 1;
  p1 := ((par div 1000000) mod 1000) - 1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;
  Result := Effect[p1].f1ijergm(dg,i,j,t,fp1,rpar)*
     Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar)* Effect[p3].f1ijergm(dg,i,j,t,fp3,rpar);
end; (* Interaction3ergm_0*)


function TEffects.Interaction3ergm_1(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var p1, p2, p3, fp1, fp2, fp3: integer;
    x1: real;
begin
  p3 := (par mod 1000) -1;
  p2 := ((par div 1000) mod 1000) - 1;
  p1 := ((par div 1000000) mod 1000) - 1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;

  if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
    else x1 := 0.0;
  if (Effect[p1].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x1 := x1 + Effect[p1].ContributionWeight[2];

  Result := x1*Effect[p2].f1ijergm(dg,i,j,t,fp2,rpar)
              *Effect[p3].f1ijergm(dg,i,j,t,fp3,rpar);
end; (* Interaction3ergm_1*)

function TEffects.Interaction3ergm_2(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var p1, p2, p3, fp1, fp2, fp3: integer;
    x1, x2: real;
begin
  p3 := (par mod 1000) -1;
  p2 := ((par div 1000) mod 1000) - 1;
  p1 := ((par div 1000000) mod 1000) - 1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;

  if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
    else x1 := 0.0;
  if (Effect[p1].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x1 := x1 + Effect[p1].ContributionWeight[2];

  if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
    else x2 := 0.0;
  if (Effect[p2].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x2 := x2 + Effect[p2].ContributionWeight[2];

  Result := x1*x2*Effect[p3].f1ijergm(dg,i,j,t,fp3,rpar);
end; (* Interaction3ergm_2*)

function TEffects.Interaction3ergm_3(const dg: TValuedDigraph;
                               const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var p1, p2, p3, fp1, fp2, fp3: integer;
    x1, x2, x3: real;
begin
  p3 := (par mod 1000) -1;
  p2 := ((par div 1000) mod 1000) - 1;
  p1 := ((par div 1000000) mod 1000) - 1;
  fp1 := Effect[p1].FParameter;
  fp2 := Effect[p2].FParameter;
  fp3 := Effect[p3].FParameter;

  if Effect[p1].HasContribution[1] then x1 := Effect[p1].ContributionWeight[1]
    else x1 := 0.0;
  if (Effect[p1].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x1 := x1 + Effect[p1].ContributionWeight[2];

  if Effect[p2].HasContribution[1] then x2 := Effect[p2].ContributionWeight[1]
    else x2 := 0.0;
  if (Effect[p2].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x2 := x2 + Effect[p2].ContributionWeight[2];

  if Effect[p3].HasContribution[1] then x3 := Effect[p3].ContributionWeight[1]
    else x3 := 0.0;
  if (Effect[p3].HasContribution[2] and (dg.TieValue(j,i) >= 1)) then
     x3 := x3 + Effect[p3].ContributionWeight[2];

  Result := x1*x2*x3;
end; (* Interaction3ergm_3 *)

begin
  index_cliq1 := -1;
  index_cliq2 := -1;
  index_cliq3 := -1;
  index_cliq4 := -1;
  index_rcliq1 := -1;
  index_rcliq2 := -1;

nelem := 0;
firstfreeelem := 0;
ergmcase := false;
nnn := 0;
setlength(extraclique,40);
setlength(extrared1_clique,40);
CF_Open := false;
assignfile(checkfile,'checkfile.txt');
rewrite(checkfile);
CF_Open := true;
end.


