{$if not defined(lazarus)}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q+,R+,S-,T-,U-,V+,W-,X+,Y+,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00010000}
{$IMAGEBASE $00400000}
{$ifend}

                              (* S_Dat.PAS *)

unit s_dat;

(* This unit contains the object DataSpecification for SIENA.
   The fundamental data included are the basic variables for the model
   that can be written to, and recovered from, the project files. *)

interface

uses SysUtils, s_constants, 
     eight, slib, digraph, Math, Dialogs, classes;

(* Unit Eight contains variable storage for adjacency matrices.
   ynm(.,.,k) for k = 1 to observations are the observed digraphs. *)

(* constants nzmax, nzzmax are defined in Eight; other constants are defined in S_Constants. *)

const epsi = 1e-6; (* used to compare integers in S_Effects *)

type b_array = array of boolean;
     resmatrix = array of array of byte;

var SienaVersion  : string[30];
    hit           : boolean; (*MS: keeps track of whether or not output writing procedures have been called in the past.*)
    simulate      : boolean; (*MS: whether or not to simulate data according to specifiction in <pname.MO> file.*)
    initialFEestimate, initialFEestimate_input: boolean; (*MS: random effects model: obtain initial estimates of fixed effects by unconditional moment estimation.*)
    randomeffects : boolean; (*MS: indicates whether or not actor-dependent random effects are included in the network objective and / or gratification function.*)
    homogeneous   : byte; (*MS: what test (actors vs. periods) to be performed, and on what parameter coordinate.*)
    alternative   : real; (*MS: alternative value of the parameter coordinate on which outlier test is to be performed.*)
    derivatives   : boolean; (*MS: whether or not the algorithm is in phase three.*)
    altern        : boolean; (*MS: indicates whether or not data are to be generated according to a model where some homogeneity assumption is violated: useful in Monte Carlo simulation to evaluate test statistic.*)
    outlier       : b_array; (*MS: specifies the actors / period of which a parameter coordinate is restricted by the null hypothesis, involved in testing homogeneity assumptions.*)
    period        : integer; (*MS: homogeneous periods test: specifies period.*)
    coordinate    : byte; (*MS: needed for tests of homogeneity assumptions.*)
    nre           : byte; (*MS: number of random coefficients.*)
    res           : resmatrix; (*MS: links the fixed effects alpa-arrays to the random effects array.*)
    re_position   : array of byte; (*MS: links the fixed effects alpa-arrays to the random effects array; used in <S_ml>.*)
    test          : BBarray; (*MS: indicates the restrictions on the parameter vector which are to be tested.*)
    d_estAB, d_estA, d_estB, FinDiff_method: boolean;
                     (*MS: These are set in S_Phase3.
                     <d_estAB = true> if the model is
                     (1) estimated (not: simulated),
                     (2) by the MoM,
                     (3) and the LR method is used to estimate derivative;
                     TS: Differentiated with d_estAB = previous d_est;
                      d_estA = (d_est and (d_estimator in [2..4])),
                      d_estB = (d_est and (d_estimator = 1)),
                      d_estAB = d_estA or d_estB;
                      since d_estimator = 1 is a new option,
                      and d_estimator = 4 replaces the old d_estimator = 1,
                      the old d_est is functionally equivalent to the new d_estAB
                      as well as to the new d_estA.
                      Note: d_estimator >= 2 discontinued in 3.17.
                     <FinDiff_method = true> when the Crude MC method (based on finite differences) is used.*)
    d_estimator: byte; (*MS: moment estimation: method of derivative estimation (three in total, see paper).*)
    ppc: boolean; (*MS: Posterior Predictive Checks for random effects models.*)
    phaseI_III    : integer; (*MS: phase 1 or phase 3 iteration number.*)
    cdensity1: byte; (*MS: Bayesian estimation of parameter: candidate-generating density, corresponding to three classes Metropolis-Hastings algorithms.*)
    cdensity2: byte; (*MS: Actor-dependent random effects: candidate-generating density, corresponding to three classes Metropolis-Hastings algorithms.*)
    scalebeta: real; (*MS: Bayesian estimation of objective function parameter: scale factor: 0 calibrated during burn-in, other number: user input, constant scale factor.*)
    scalere: real; (*MS: Bayesian estimation of objective function parameter: scale factor: 0 calibrated during burn-in, other number: user input, constant scale factor.*)
    CurrentProjectName : shortString; (* CS: for compatibility between SIENA08 and S_RE *)
    thetarandom: pbvec; (*MS: Boolean array, indicating (1) whether some coefficients are modeled as random coefficients, (2) in what functions, (3) what coefficients are modeled as random coefficients.*)
    CallingProgram : byte;  (* indicates which SIENA program is using S_Dat; 1, 2, 4, 5, 7;
                                this is set in Siena_7, etc. *)

type
 DMatrix = array[1..pmax,1..3*ccmax+1] of real;


(* Order of statistic types is as follows.
   this fixed order may eventually supersede the arrays "contribution".
   (observations - 1)                    for network base rate of change;
   networkeffects_l - (observations - 1) for network change, in addition to base rates;
   networkeffects_f                      for network evaluation;
   networkeffects_f                      for network endowment;
   nza*(observations - 1)                for action base rate of change;
   nza*(actioneffects_l - (observations - 1)) for action change,
                                         in addition to base rates.
   nza*actioneffects_f                   for action evaluation;
   nza*actioneffects_f                   for action endowment;
   Correspondence between order of effects in parameter estimation
   and order of effect types in model definition is determined in
   procedures SetTheta and SetAlpha in S_Base.
   For parameters in theta, the order is:
   first ActiveNetworkEffects_l network rate effects;
   then ActiveNetworkEffects_f  network evaluation effects;
   then ActiveNetworkEffects_g  network endowment effects;
   then ActiveActionEffects_l   action  rate effects;
   then ActiveActionEffects_f   action  evaluation effects;
   then the action endowment effects. *)

 DataSpecification = object

   pname         : shortstring;   (* project name                                            *)

   (* variables defining the data format: *)
   observations,  (* number of observations                                  *)
   ActivePeriods, (* number of observations k for which the transition from k to k+1 is simulated;
                     if (not multi) this equals observations-1. *)
   n, nmax,                  (* number of actors                                        *)
   nz,                       (* number of constant individual covariates                *)
   nzc,                      (* number of changing individual variables                 *)
   nza,                      (* number of dependent actor variables                     *)
   nxa,                      (* CS: number of dependent network variables;
                                    THE CASE nxa > 1 IS NOT YET SUPPORTED BY SIENA.     *)
   nzz,                      (* number of constant dyadic covariates                    *)
   nzzc,                     (* CS: number of changing dyadic covariates.    *)
   nzx: integer;             (* MH: Composition change as exogenous events:
                                    0=no change,
                                    1=change: joiners 0, leaver last observed value,
                                    2=change: joiners 0, leavers missing,
                                    3=change, treat as missing
                                    4=change, treat as structural zeros                 *)
   exo: boolean;             (* MH: Indicates whether there is an exogenous events file *)

   // some flags that indicate basic characteristics of the data (values are set at data input):
   anynetworkmissings,
   symmetric,                (* indicates that the network(s) is(are) undirected        *)
   asymmetric,               (* ERGM case: asymmetric = all dyads are asymmetric        *)
   signed,                   (* indicates whether there is a negative tie code in the
                                data (and hence whether these define a signed graph)    *)
   valued : boolean;         (* CS: indicates whether the network contains valued tie
                                    variables, i.e., the data set contains values not
                                    declared as "one" in the <pname>.IN file.           *)
   uponly, downonly: array of boolean; (* indicates per period that
                                          only changes up/down were observed *)

   // two more data characteristics calculated at data input:
   valmin,                   (* minimum and ...                                         *)
   valplus: shortint;        (* ...maximum values of tie codes;
                                if signed=TRUE, these are -1 and +1, respectively;
                                if valued=TRUE, these are 0 and k, where
                                  1<k<min(set of positive missing-codes,carryforward,struct0,struct1),
                                  with the set of missing-codes being defined in the .IN-file
                                  and (currently) carryforward=7,struct0=10 and struct1=11;
                                if signed=FALSE & valued=FALSE, these are 0 and 1.      *)

   balmean: real;
   (* sum_(i,j,h, k = 1..,observations-1; i<>j, i<>h, j<>h)|ynm(i,h,k) - ynm(j,h,k)| /
               ((n*(n-1)*(n-2)*(observations-1)));
      used as subtraction term in balance calculations *)

      (* signed indicates if there are any negative values in the data matrix;
   valmin and valplus indicate the minimum and maximum codes after processing the data;
   this is a precursor to working with valued graphs,
   used now to allow signed graphs *)

   (* variable labels: *)
   ZZNames  : Snzzarray;     (* names of pairwise covariates                            *)
   labels_l,
   labels_f : SSArray;       (* names of rate function and objective function effects   *)
   (* statistics labels *)
   FDescriptions  : SSarray; (* for storing names of all statistics used in the model   *)

   (* parameter vectors: *)

   (* arrays calculated in procedure CALCULATIONS: *)
   zcmax,            (* maximum of the variable, and ...                *)
   zcmin  : array[1..nzmax] of real;  (* ... minimum of the variable.   *)
   zzmean : array[1..nzzmax] of real; (* grand mean of zz               *)
   pzamean : array of array of real;   (* observationwise average of dependent actor variables *)
   czamean : array[1..nzmax] of real;   (* current average of dependent actor variables; is dynamically updated *)
(* AA start *)
   poszvar : array[1..nzmax] of boolean;  (* positive variance of z or zc in at least one period *)
   compzvar : array[1..nzmax] of boolean; (* z or zc is constant within components
                                             for all observations *)
   (* poszvar and compzvar are used in S_Effects for the definition of effects. *)                                             
(* AA end *)
   nni     : array of integer;           (* period-dependent n *)
   skipi   : array of boolean;           (* indicates that period i is skipped *)
   ActiveObservationRank: array of integer;
     (* ActiveObservationRank[h] = if skipi[h], then 999;
                                   else number of non-skipped[h'] with h' <= h. *)

   NewNetworkEffects: TEffects;
(* For unpacked behavior interaction effects: *)
   ni,pv0,pv1,pv2,pv3: array of array of integer;

   ObservedNetworkDistance : array of array of integer;
   ObservedActionDistance : array of array of integer;
   ObservedNetworkData : array of array of integer;
   ObservedActionData : array of array of integer;
     (* The observed distances, and the number of data points on which
        these are based, between observation times t[k] and t[k+1];
        dynamic length is defined
        in procedure SetNumberOfPeriods.                               *)
   MoreComponents: boolean;
   ComponentIndicator: IntVec;
   ComponentNumber: array of array of integer;
   (* ComponentIndicator and ComponentNumber are used only if ExistStruct.
      ComponentIndicator[i] is the component of vertex i.
      Then ComponentNumber[0,0] is the number of components;
      ComponentNumber[h,k] is the k'th vertex in the h'th component;
      ComponentNumber[h,0] is the number of vertices in the h'th component.
      MoreComponents is calculated in initialise_y as ComponentNumber[0,0] >= 2.*)


   (* starting values for estimation, calculated at data input: *)
   densty: real;              (* Density for ERGM case                                  *)

   alpa_f,
   alpa_g,
   alpa_l   : CRarray;       (* parameters of statistical distribution                  *)
   vertex_a,                 (* incidental vertex parameter for outdegrees,             *)
   vertex_b : Realnvec;      (* incidental vertex parameter for indegrees,              *)
                             (* both used only in the Vertexpar (observations = 1) case *)

   param_effect      : SSIarray;
   (* If an effect includes a parameter, param_effect is its value.
      Note that rate function effects do not contain a parameter,
      and that the parameters for evaluation effects and endowment effects
      are identical. *)
   Effects_l,Effects_f, Effects_g, Effects_s, Effects_sg: BBarray;
   (* These Boolean arrays indicate whether the indicated effects
      are present or absent.
      _l indicates rate effects, _f evaluation effects, _g endowment effects.
      In each case, the order is:
      first the network effects, then for each of the nza dependent behavior variables
      the effects for this variable.
      Effects_x[0] = true iff no effect of type x is present.
      if estimation then
        Effects_s[j] = Effects_f[j] or Effects_g[j]
        (or j is a main effect corresponding to an interaction included in the model;
         this is done at the end of DefineFunctions)
        else Effects_s[j] indicates the effects
        corresponding to the simulated statistics,
        and is calculated in DefineSimFunctions. *)
   FirstActionRateParameter: byte;
   (* The index number of the first action rate parameter *)
   thetarandom_f, thetarandom_g: BBarray;
   (*MS: these Boolean arrays indicate whether or not some coefficients in the network or behavior
   objective or gratification function are assumed to be fixed coefficients
   (= parameters) or random coefficients (= random variables):
   choosing the last option implies specifying a random effects model.*)
   variable_rc: b_array;
   (*MS: this Boolean array indicates for each of the dependent variables whether or not
   there are random coefficients in the corresponding objective or gratification function.*)
   fix_l, fix_f, fix_g: BBarray;
   (* These Boolean arrays indicate whether the indicated effects
      are to be fixed in the estimation *)
   test_l, test_f, test_g: BBarray;
   (*MS: these Boolean arrays indicate whether the indicated effects
      are to be tested.  *)
   test_CovRE: BBarray;
   (*MS: indicates whether and what restrictions on the variance-covariance matrix of the
   actor-dependent random effects are to be tested.*)
   Functions: pWordVec;
   (* Indicate the numbers of the calculated statistics.
      From the first 0 onward, no more statistics are calculated.  *)
   CConditional, MaxLike, Bayesian: Boolean;
   (* This indicates whether simulations go on until a specified
      distance is reached (CConditional = true),
      or with a given change rate during a specified
      time period (CConditional = false);
      and, for observations >= 2, whether Maximum Likelihood
      rather than Method of Moments is used for estimation. *)
   ConditioningNetworkVariable, ConditioningActionVariable: integer;
   (* If Cconditional, this indicates which variable
      is the conditioning variable:
      only one of these two variables can be active, the other must be 0! *)
   InitialStandard: Boolean;
   (* This indicates whether the initial estimates in S_Est
      are to be calculated in a standard way (True),
      or are the current parameter values (False). *)
   Continuous: Boolean;
   (* This indicates whether, for observations = 1,
      the Gibbs steps for each newly generated network start
      from the existing network or from an independently
      generated random network. *)
   Gibbs: byte;
   (* This indicates the type of steps for the observations = 1 case;
      explanation is given in Gibbsstring. *)
   ModelType, Gibbscode, OutputType, StaySameType: Integer;
   (* This indicates the type of model and output options for the observations >= 2 case.
   StaySameType = 1 is the older default, now close to obsolete version,
   where actors have to make a network change if they are allowed to;
   the current default version is StaySameType = 2. *)
{   SymmetricGibbs: set of byte;
   (* The set of Gibbs codes for symmetric = nondirected graphs *)
   DefaultSymmetricGibbs: byte;
   (* The default Gibbs code for symmetric = nondirected graphs *)
 }  ReciprocityModel, ExtendedReciprocityModel, GibbsConnected: Boolean;
   (* This indicates that the rate function is defined as in the reciprocity model
      by equations (25), (30) in Snijders & van Duijn 1997.
      The ExtendedReciprocityModel is the implementation with covariates.
      But without covariates it is faster than the standard
      (not extended) reciprocity model - because the latter requires
      many calculations of exponentials.
      For checking purposes, the implementation of the standard reciprocity model
      is retained, but the extended model is carried out in practice. *)
   MakeStep3: boolean;
   (* This indicates that after phase 3 in the Robbins-Monro algorithm in PolRup,
      a quasi-Newton Raphson step is carried out. *)
   VertexPar, VertexParNew: boolean;
   (* This indicates that incidental vertex parameters are used
      for the observations = 1 case. *)
   arraykoutstars, arraykinstars, arrayktri, arrayktwop: array of real;
         (* used for storage of expensive functions used in weighted parallel
             k-stars, k-triangles, and k-parallel twopaths *)
   maxdegree: integer;
   (* All networks are constrained to outdegrees not larger than than maxdegree *)
   MaxRuns: integer;
   (* Upper bound to number of single steps ("runstep"),
      only valid for observations >= 2 *)
   rstep0, rstep: real;
   (* (Number of steps) / (n*n) in Runstep for observations = 1 *)
   nsubphase: integer;
   (* number of subphases in POLRUP *)
   nextra: integer;
   (* added 3.181: maximum number of repetitions in s_est *)
   nrun3: integer;
   (* default number of phase 3 iteration runs in POLRUP *)
   niter: integer;
   (* default number of iterations for straight simulation *)
   firstgain: real;
   (* initial value for gain parameter in the estimation algorithm *)
   nrunMH: integer;
   (* number of MCMC iterations per RM step *)
   nrunMH_batches: integer;
   (*MS: number of batches of MCMC iterations for random effects or Bayesian models.*)
   estimation: boolean;
   (* indicates whether this unit is used for estimation
      - true when it has passed through S_est => Estimate. *)
   (* The following variables are calculated from the data or project input,
      and are not stored in the project files : *)
   NumberFunctions: integer;
   NumberParameters: integer;

   modelseed : integer; (* CS: random seed as read from the .MO file *)
   seedsown  : boolean; (* CS: indicator whether this seed was already used for initialisation *)
    SoniaGraphics,            (* indicates whether graphic output of SoNIA type shall be generated *)
    PajekGraphics : boolean;  (* indicates whether graphic output of PAJEK type shall be generated *)
    nSoniaVars,               (* number of variables to be visualised with SoNIA *)
    nPajekVars    : integer;  (* number of variables to be visualised with Pajek *)
    SoniaVar,                 (* id-numbers of the variables to be visualised with SoNIA *)
    PajekVar,                 (* id-numbers of the variables to be visualised with Pajek *)
    SoniaType,                (* visualisation type for SoNIA *)
    PajekType,                (* ... and for Pajek *)
    nSoniaCat,                (* number of categories for recoding the variables to visual *)
    nPajekCat     : array of integer; (* dimensions, for SoNIA and Pajek, respectively. *)
    SoniaCatBoundary,         (* Category definitions (left & right boundaries) *)
    PajekCatBoundary : array of array of array[1..2] of integer;
    SoniaCatLabel,
    PajekCatLabel : array of array of string;

  procedure Component(var ncomp: integer;
            const determine: boolean; var compo: IntVec; k: integer);
      (* Indicates components in the case of structurally determined data
         in period k.
         If determine, then also the number of components is determined
         and the dimensions of ComponentNumber are set. *)
(* Note TS: Component is very inefficient because it is called often
         and it uses (n+1)^2 arrays. *)
  procedure Spatialfunctions(stra: array of boolean;   // changed 3.17u
                             h,k: integer; var Mi, Gc: real);
(* Spatial autocorrelation statistics using current data
   as calculated for dependent variable h, period k. *)
  function NetworkEffects_f: integer;
  function ActionEffects_f: integer;
  function MaxEffects_f: integer;
  function NetworkEffects_g: integer;
  function ActionEffects_g: integer;
  function MaxEffects_g: integer;
  function NetworkEffects_l: integer;
  function ActionEffects_l: integer;
  function MaxEffects_l: integer;
  function NetworkFunctions: integer;
  function ActionFunctions: integer;
  function MaxFunctions: integer;
(* These maxima define the .mo* files, and therefore
   are not allowed to be changed by changing the model specification.
   The "Network..." refer to effects and statistics for the network evolution part,
   and are the model part that as already present in the Siena 1.xx versions.
   The "Action..." refer to effects and statistics for the behavior evolution part,
   and are new to the Siena 2.0 version. *)

  function Minimalya(i,h: integer): boolean;
  function Maximalya(i,h: integer): boolean;
(* Indicates that the action value ya is minimal or maximal, respectively. *)

(* Functions used for defining effects in S_Effects: *)
  function BalanceSimilarity(const dg: TValuedDigraph;
                                             const i,j,k,par: integer): real;
  (* Contribution of tie variable (i,j) in period k to balance similarity *)
  function Dist2Contr(const dg: TValuedDigraph; const i,j,k,par: integer): Real;
(* Contribution of tie (i,j) to number of distances 2 *)
  function HalfDist2Contr(const dg: TValuedDigraph;
                                      const i,j,k,par: integer): real;
  function Dist2ActorContr(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): Real;
(* Distance 2 contribution to actor evaluation function. *)
  function CorNDist2(const dg: TValuedDigraph; const i,k,par: integer): real;
  function HalfCorNDist2(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;

  function TwoDist2ActorContr(const dg: TValuedDigraph;
                          const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function TwoDist2Contr(const dg: TValuedDigraph; const i,j,k,par: integer): Real;
  function HalfTwoDist2Contr(const dg: TValuedDigraph; const i,j,k,par: integer): Real;
  function TwoCorNDist2(const dg: TValuedDigraph; const i,k,par: integer): real;
  function HalfTwoCorNDist2(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;

  function CorBetween(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifMax(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function MaxTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifRoot(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function RootPlus(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifPow15(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Pow15Plus(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifPow2(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Pow2Plus(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifOutSqrtInt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function OutSqrtIn(const dg: TValuedDigraph; const i,k,par: integer): real;
  function InSqrtOut(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifOutInt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function OutIn(const dg: TValuedDigraph; const i,k,par: integer): real;

  function Difddt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Difododt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Prododod(const dg: TValuedDigraph; const i,j,k,par: integer): real;
  function Difodidt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Prododid(const dg: TValuedDigraph; const i,j,k,par: integer): real;
  function Difidodt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Prodidod(const dg: TValuedDigraph; const i,j,k,par: integer): real;
  function Difididt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Prodidid(const dg: TValuedDigraph; const i,j,k,par: integer): real;
  function DifSquare(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function SquarePlus(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifRec(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function RecTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifRec2(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Rec2Tot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifDyadicCovt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function DifDyadicCov(const dg: TValuedDigraph; const i,j,k: integer;
                        const par: integer): real;
  function DyadicCovTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifDyadicCovRect(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function DifDyadicCovRec(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DyadicCovRecTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifChangeDyadicCov(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DifChangeDyadicCovt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function DifChangeDyadicCovRect(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function DifChangeDyadicCovRec(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DyadicCovWWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DyadicCovWWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function ChangeDyadicCovWWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function ChangeDyadicCovWWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function DyadicCovWXX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DyadicCovWXXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function ChangeDyadicCovWXX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function ChangeDyadicCovWXXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function DyadicCovXWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function DyadicCovXWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
  function ChangeDyadicCovXWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
  function ChangeDyadicCovXWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;

  function DifCovAltert(const dg: TValuedDigraph; const i,j,k: integer;
                       const tv: shortint; const par: integer): real;
  function DifCovAlter(const dg: TValuedDigraph; const i,j,k: integer;
                       const par: integer): real;
  function DifSquareCovAltert(const dg: TValuedDigraph; const i,j,k: integer;
                       const tv: shortint; const par: integer): real;
  function DifSquareCovAlter(const dg: TValuedDigraph; const i,j,k: integer;
                       const par: integer): real;
  function CovBotht(const dg: TValuedDigraph; const i,j,k: integer;
                    const tv: shortint; const par: integer): real;
  function CovBoth(const dg: TValuedDigraph; const i,j,k: integer;
                   const par: integer): real;
  function CovBothRect(const dg: TValuedDigraph; const i,j,k: integer;
                    const tv: shortint; const par: integer): real;
  function CovBothRec(const dg: TValuedDigraph; const i,j,k: integer;
                   const par: integer): real;
// AA start
  function CovIndirectt(const dg: TValuedDigraph; const i,j,k: integer;
                    const tv: shortint; const par: integer): real;
  function CovIndirect(const dg: TValuedDigraph; const i,j,k: integer;
                   const par: integer): real;
  function CorCovIndirect(const dg: TValuedDigraph; const i,k: integer;
                   const par: integer): real;
// AA end
  function DifCovEgot(const dg: TValuedDigraph; const i,j,k: integer;
                     const tv: shortint; const par: integer): real;
  function DifCovEgo(const dg: TValuedDigraph; const i,j,k: integer;
                     const par: integer): real;

  function DifCovSimt(const dg: TValuedDigraph; const i,j,k: integer;
                     const tv: shortint; const par: integer): real;
  function DifCovSim(const dg: TValuedDigraph; const i,j,k: integer;
                     const par: integer): real;
  function DifCovSimRect(const dg: TValuedDigraph; const i,j,k: integer;
                        const tv: shortint; const par: integer): real;
  function DifCovSimRec(const dg: TValuedDigraph; const i,j,k: integer;
                        const par: integer): real;
  function DifCovSamet(const dg: TValuedDigraph; const i,j,k: integer;
                     const tv: shortint; const par: integer): real;
  function DifCovSame(const dg: TValuedDigraph; const i,j,k: integer;
                     const par: integer): real;
  function DifCovSameRect(const dg: TValuedDigraph; const i,j,k: integer;
                        const tv: shortint; const par: integer): real;
  function DifCovSameRec(const dg: TValuedDigraph; const i,j,k: integer;
                        const par: integer): real;
  function CovSimTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function CovSimRecTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifDense(const dg: TValuedDigraph; const i,j,k: integer;
                    const tv: shortint; const par: integer): real;
  function DenseTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifSqrtInt(const dg: TValuedDigraph; const i,j,k: integer;
                     const tv: shortint; const par: integer): real;
  function DifSqrtIn(const dg: TValuedDigraph; const i,j,k: integer;
                     const par: integer): real;
  function SqrtInTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifSqrtOutt(const dg: TValuedDigraph; const i,j,k: integer;
                      const tv: shortint; const par: integer): real;
  function DifSqrtOut(const dg: TValuedDigraph; const i,j,k: integer;
                      const par: integer): real;
  function SqrtOutTot(const dg: TValuedDigraph; const i,k,par: integer): real;
  function DifIndirectt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function DifIndirect(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;


  function Interactiont_var_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Interaction_var_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;


  function ChooseOut(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ChooseIn(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ChooseBoth(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function InvDEmpty(const dg: TValuedDigraph; const i,k,par: integer): Real;
  function InvDEmpty2(const dg: TValuedDigraph; const i,k,par: integer): Real;
  function InvDegrees(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function InvDegree(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function InvDegrees2(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function InvDegree2(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function DirectIndirect(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Indirect(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function AltStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function AltOutStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function AltInStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function AltTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function TwoTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ThreeTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ThreePaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ThreeTwoPaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function AltParTwoPaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function CoatHanger(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function ThreeLine(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Covari(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Covarj(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Simcovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Samecovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function DyCovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function CliqueChange(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
  function Red1_CliqueChange(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
{  function Interactiont(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
  function Interaction(const dg: TValuedDigraph; const i,k,par: integer): real;
 }




  procedure CheckMaxDegree(var outf: TFileStream);
  (* checks whether the largest degree of the current network is
     not higher than MaxDegree *)
  procedure CheckMaxima(var halterr: Boolean; var outf, lf: TFileStream);
  (* checks whether the functions MaxEffects_l, MaxEffects_f, MaxEffects_g
     and MaxFunctions yield a value less than ccmax *)
  procedure initialise;
   (* basic initialisation *)
  procedure AttachProjectName(var proname: shortstring; var outf, lf: TFileStream;
                              var apf: boolean; writedate: boolean);
  (* If pname > '', then this sets pname to proname,
      otherwise to the default name siena. *)
  procedure AttachProject(var proname, moname: shortstring;
                          var halterr: Boolean; var outf, lf: TFileStream;
                          var apf: boolean; writedate: boolean);
  (* This procedure first executes AttachProjectName,
     and then reads the internal data and definition files.
     The .mo file is read from moname.mo *)
  procedure InitialAlphas(var alp,alpf: CRarray; show: byte; var outf: tFileStream);
  (* Calculates initial standard estimates *)
  procedure SetNumberOfPeriods(nn,hzc,hzz,hzzc,k: integer);
  procedure FunctionsToBooleans(var FunctInd: BBarray);
  (* Makes an array: FunctInd[i] = true if Function number i
     is calculated. This is determined from array Functions. *)
  procedure IndicateFixedThetaAndRC(var fix, thetarandom: pBVec);
  (*Indicates in array fix which parameters are fixed in estimation, and in array thetarandom
   which coefficients are regarded as random coefficients: random effects model.*)
  procedure IndicateFixedAlpa(fix: pBVec);
  (* Changes arrays fix_l, fix_f, fix_g
    according to what is defined in array fix;
    fixing values for parameters not in the current model is unaffected. *)
  procedure writestart(var wf: textfile; h: integer);
  (* writes first lines of definition file *)
  (* CS: the following two procedures generate SIENA 2 format files: *)
  procedure WriteDefinitionModel(var lf: TFileStream);
  (* writes model definition to file .MO *)
  procedure WriteDefinitionStats(var lf: TFileStream);
  (* writes definition for simulation statistics to file .SI *)
  procedure SienaToPajek(var lf: TFileStream);

  procedure DefineSignedNetworkEffects(var Effs: TEffects);
  procedure DefineNetworkEffects(var Effs: TEffects);

  procedure DefineModel_lNames
            (var modelnames: SSarray; var Correspondence: SBYarray;
             var param: SSIarray);
  (* labels rate function effects *)
  procedure DefineModel_fNames
            (var modelnames: SSarray; var Correspondence: SBYarray;
             var VertPar: BBarray; var param: SSIArray);
  (* labels objective function effects *)
  procedure DefineModel_gNames
            (var modelnames: SSarray; var Correspondence: SBYarray;
             var param: SSIarray);
  (* labels endowment function effects; transferred from S_Base to S_Dat by TS,
    13-jun-2005, for reasons of consistency *)
  procedure DefineFunctionNames(var outf: TFileStream);
  (* labels statistics in array FDescriptions *)
  procedure ReadBehavior(const dim: byte);
  (*MS: used when simulating data: read only first behavioral outcome variable.*)
  procedure WritePData(var halterr: Boolean; var outf,lf:TFileStream);
  (* Writes data to project data files *)
  procedure ReadDefinition1(var halterr: Boolean; var outf, lf: TFileStream);
  procedure ReadDefinition2(var moname: shortstring;
                            var halterr: Boolean; var outf, lf: TFileStream);
  (* Reads data definition from files .MO and .SI.
     Defines effects according to S_Effects. *)
  procedure ReadPData(var halterr: boolean; var outf, lf: TFileStream);
  (* Reads data from project data files *)
  procedure ReadGraphicOptions(var logfile,
                                 outfile : TFileStream;
                             var halterr : boolean);
(* CS 7-apr-2005: This procedure reads - if it exists - a graphic options file <pname.vis>
   from the current directory, and sets the graphical output options of SIENA accordingly. *)
  procedure Calculations(writeoutput: boolean; var outf: TFileStream);
  (* calculates averages for covariates and balance dissimilarity,
     observeddistances, etc. *)
  procedure AdaptModel(var outf: TFileStream);
    (* Adapts model; only for use
       when running special versions of Siena04. *)
  procedure HaltError(s: string; var f, lf: TFileStream);
  procedure FinishDat(var outfile: TFileStream);
end; (* DataSpecification *)

var errorstring: string;

procedure unpack(const pa: integer; var nia, p1a, p2a, p3a: integer);

implementation

uses StrUtils;

const facmax = 10;

var
 MadeError: boolean;
 iii: integer;
 (* used for storing change statistics for osrvations = 1 *)
 oldmofile: boolean; (* old .mo file, compatible with Stocnet 1.7 *)
 factorial: array[1..facmax] of real; // storage of factorials


{$I s_effects} (* TS: Input file <S_Effects>.*)


procedure DataSpecification.Component(var ncomp: integer;
            const determine: boolean; var compo: IntVec; k: integer);
(* Indicates components in the case of structurally determined data,
   for period k.
   If determine, then also the number of components is determined
   and the dimensions of ComponentNumber are set.
   Modeled on function Components in S_desc.  *)
(* Note TS: This is very inefficient because it is called often. *)
var i,j,h,reached,pow,comps,LocalNumber: integer;
 m, m2: array of array of Boolean;
 reach: array of Boolean;
 mm: Boolean;
begin
  n := nni[k];
  SetLength(m,n+1,n+1);
  SetLength(m2,n+1,n+1);
  SetLength(reach,n+1);
  for i:=n downto 1 do for j:=n downto 1 do
(* symmetrized places for random=non-determined values & "true" diagonal *)
    m2[i,j] := ((not struct(i,j,k)) or (not struct(j,i,k)) or (i=j));
  pow := 1;
  repeat
    m := m2;
    for i:=n downto 1 do for j:=n downto 1 do begin
      mm := m[i,j];
      if (not mm) then begin
        h := n;
        repeat
         if (m[i,h] and m[j,h]) then mm := true;
         dec(h);
        until (mm or (h <= 0));
       end; (* if not mm *)
      m2[i,j] := mm;
     end; (* i, j *)
    pow := pow*2;
(* Now m2 is the Boolean power ynm^pow *)
   until (pow >= n);
(* Now m2[i,j] indicates whether there is a path between i and j *)

 if determine then
   begin
   for i:=1 to n do reach[i] := false;
   i := 1;
   reached := 0;
   comps := 0;
   repeat
     inc(comps);
     for j:=1 to n do if (m2[i,j] and (not reach[j])) then
       begin
       reach[j] := true;
       inc(reached);
       end;
     if (reached < n) then
      repeat inc(i) until ((not reach[i]) or (i > n));
     until ((reached >= n) or (i > n));
    ncomp := comps;
    end (* if determine *)
   else
    begin
    for i:=1 to n do reach[i] := false;
    for i:=1 to n do Compo[i] := 0;
    i := 1;
    reached := 0;
    comps := 0;
    repeat
      inc(comps);
      LocalNumber := 0;
      for j:=1 to n do if (m2[i,j] and (not reach[j])) then begin
        inc(LocalNumber);
        reach[j] := true;
        inc(reached);
        Compo[j] := comps;
        ComponentNumber[comps,LocalNumber] := j;
       end;
      ComponentNumber[comps,0] := LocalNumber; (* number of vertices in this component *)
      if (reached < n) then
       repeat inc(i) until ((not reach[i]) or (i > n));
     until ((reached >= n) or (i > n));
     ncomp := comps;
     ComponentNumber[0,0] := comps;
    end; (* if not determine *)
  finalize(m);
  finalize(m2);
end; (* Component *)

procedure DataSpecification.Spatialfunctions(stra: array of boolean;  // changed 3.17u
                                             h,k: integer; var Mi, Gc: real);
(* Spatial autocorrelation statistics using current data
   as calculated for dependent variable h, period k;
  from S_desc.spatials *)
var i,j,nn: integer;
  s1,s2,s3,s4,s5,s6,s5i,s6ij,ym: real;
begin
(* Computation of Moran's I and Geary's c *)
(* Actors i with missing zc[i,h,k] are disregarded;
   missings tie variables are set to 0;
   tie variables are weights. These are not necessarily symmetric.
   For the formulae, see e.g.
   B. Ripley, "Spatial Statistics", Wiley 1981, p. 98-99. *)
      s1 := 0.0; s2 := 0.0; ym := 0.0;
      s3 := 0.0; s4:= 0.0; nn := 0;
      s5 := 0.0; s6 := 0.0;
      for i:=1 to n do
        if stra[i] and (not miszc(i,h,k)) then // changed 3.17u
          begin
          ym := ym + ya(i,h);
          nn := nn+1;
          end;
      if (nn >= 1) then ym := ym/nn
       else ym := 0.0; (* mean zc *)
(* Correspondence to Ripley's symbols:
   s2 = Omega, s5 = S2, s6 = S1.*)

      for i:=1 to n do if stra[i] then //  stra[i] added 3.17u
        begin
        s5i := 0.0;
        for j:=1 to n do if stra[j] then
 //  stra[j] added 3.17u
          begin
          s6ij := 0;
          if (yn(i,j) and (not (mis(i,j,k) or miszc(i,h,k) or miszc(j,h,k)))) then begin
            s1 := s1 + sqr(ya(i,h)-ya(j,h));
            s2 := s2 + 1;
            s3 := s3 + (ya(i,h) - ym)*(ya(j,h) - ym);
            s5i := s5i + 1;
            s6ij := s6ij + 1;
           end; (* if  *)
          if (yn(j,i) and (not (mis(j,i,k) or miszc(i,h,k) or miszc(j,h,k)))) then begin
            s5i := s5i + 1;
            s6ij := s6ij + 1;
           end;
          s6 := s6 + sqr(s6ij);
         end; (* for j *)
        s5 := s5 + sqr(s5i);
        if (not miszc(i,h,k))
          then s4 := s4 + sqr(ya(i,h) - ym);
       end; (* for i *)
      s6 := 0.5*s6;
      if (nn >= 2) and (s2*s4 > 1e-50) then begin
        Gc := (nn-1)*s1/(2*s2*s4);
        Mi := nn*s3/(s2*s4);
       end else begin
         Gc := 1.0;
         Mi := 0.0;
        end;
end; (* Spatialfunctions *)




(* The following maxima must correspond with the numbers of effects and functions in S_Model *)

function DataSpecification.NetworkEffects_l: integer;
(* number of rate function effects per dependent network variable *)
begin
  if (observations = 1)
  then NetworkEffects_l := 0
  else NetworkEffects_l := (observations-1) + 4 + 2*(nz+nzc); (* The 2* is for ModelType = 2 *)
end;

function DataSpecification.ActionEffects_l: integer;
(* number of rate function effects per dependent behavioural variable *)
begin
  if (observations = 1)
  then ActionEffects_l := 0
  else ActionEffects_l := (observations-1) + 3 + (nz+nzc);
end;

function DataSpecification.MaxEffects_l: integer;
(* total number of rate function effects *)
begin
  MaxEffects_l := nxa*NetworkEffects_l + nza*ActionEffects_l;
end;

function DataSpecification.NetworkEffects_f: integer;
(* number of objective function effects per dependent network variable *)
begin
 NetworkEffects_f := NewNetworkEffects.NumberEffects;
end;

function DataSpecification.ActionEffects_f: integer;
(* number of objective function effects per dependent behavioural variable *)
begin
  ActionEffects_f := 19 + 2*(nz+nzc-1) + ninta + 1;
  (* 1 Added 27/10/8: *)
end;

function DataSpecification.MaxEffects_f: integer;
(* total number of objective function effects *)
begin
  MaxEffects_f := nxa*NetworkEffects_f + nza*ActionEffects_f;
end;

function DataSpecification.NetworkEffects_g: integer;
(* number of gratification function effects per dependent network variable *)
begin
  if ((observations = 1) or signed) then Networkeffects_g := 0
    else NetworkEffects_g := NewNetworkEffects.NumberEffects;
end;

function DataSpecification.ActionEffects_g: integer;
(* number of gratification function effects per dependent behavioural variable *)
begin
  if ((observations = 1) or signed)
  then ActionEffects_g := 0
  else ActionEffects_g := ActionEffects_f;
end;

function DataSpecification.MaxEffects_g: integer;
(* total number of gratification function effects *)
begin
  if (observations = 1)
  then MaxEffects_g := 0
  else  MaxEffects_g := nxa*NetworkEffects_g + nza*ActionEffects_g
end;

function DataSpecification.NetworkFunctions: integer;
(* total number of statistics in MM-estimation for approximating network dynamics *)
begin
 if (observations = 1)
 then NetworkFunctions := nxa*NetworkEffects_f
  else NetworkFunctions := nxa*(Networkeffects_l + NetworkEffects_f + NetworkEffects_g);
end;

function DataSpecification.ActionFunctions: integer;
(* total number of statistics in MM-estimation for approximating behavioural dynamics *)
begin
 if (observations = 1)
 then ActionFunctions := 0
 else ActionFunctions := nza*(Actioneffects_l + ActionEffects_f + ActionEffects_g)
end;

function DataSpecification.MaxFunctions: integer;
(* total number of statistics for MM-estimation *)
begin
  MaxFunctions := NetworkFunctions + ActionFunctions;
end;


function DataSpecification.Minimalya(i,h: integer): boolean;
(* Indicates that the action value ya is minimal. *)
begin
  Result := (ya(i,h) <= round(zcmin[h]));
end;

function DataSpecification.Maximalya(i,h: integer): boolean;
(* Indicates that the action value ya is maximal. *)
begin
  Result := (ya(i,h) >= round(zcmax[h]));
end;



procedure DataSpecification.CheckMaxDegree(var outf: TFileStream);
  (* checks whether the largest degree of the current network is
     not higher than MaxDegree *)
var i: integer; wrong: boolean; x: string;
begin
  wrong := false;
  for i:=1 to n do
    if outd_yn(i) > MaxDegree then begin
//    writeln(outf,'Outdegree of vertex ',i,' = ',outd_yn(i),' > ',MaxDegree,',');
    x :=format('%s %d %s %d %s %d%s',
      ['Outdegree of vertex',i,'=',outd_yn(i),'>',MaxDegree,','])+eol;
    outf.Write(pchar(x)^,length(x));
    wrong := true;
   end;
  if wrong then
  begin
  x :='contrasting with the model specification.'+eol;
  outf.write(pchar(x)^,length(x));
  end;
end;


procedure DataSpecification.CheckMaxima(var halterr: Boolean; var outf, lf: TFileStream);
procedure videlicet;
var x: string;
begin
  x := 'Number of variables:'+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%d%s',[nz,' non-changing individal covariates;'])+eol;
  outf.Write(pchar(x)^,length(x));
  if (nza > 0) then begin
  //writeln(outf,'      of which ',nza,' are dependent variables;');
  x := format('%s%d%s',['      of which ',nza,' are dependent variables;'])+eol;
  outf.write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[nzz,' dyadic covariates.'])+eol;
  outf.Write(pchar(x)^,length(x));
 // writeln(outf, nzz,' dyadic covariates.');
end;
begin
  halterr := false;
  if (MaxFunctions > ccmax) then begin
    videlicet;
(* AA change *)    Halterror('Too many statistics. Increase ccmax or use fewer statistics.',outf,lf);
    halterr := true;
   end;
  if (MaxEffects_l > ccmax) then begin
    videlicet;
    Halterror('Too many effects for the rate function. Increase cmax.',outf,lf);
    halterr := true;
   end;
  if (NetworkEffects_f > ccmax) then begin
    videlicet;
    Halterror('Too many effects for the evaluation (n) function. Increase ccmax.',outf,lf);
    halterr := true;
   end;
  if (nza*ActionEffects_f > ccmax) then begin
    videlicet;
    Halterror('Too many effects for the evaluation (a) function. Increase ccmax.',outf,lf);
    halterr := true;
   end;
  if (MaxEffects_g > ccmax) then begin
    videlicet;
    Halterror('Too many effects for the gratification function. Increase ccmax.',outf,lf);
    halterr := true;
   end;
end;

procedure DataSpecification.initialise;
(* This initialisation must not uses any variables as input,
   only constants. *)
var i : integer;
begin
  for i:=1 to nzmax do ZNames[i] := 'ind. cov. '+st(i);
  for i:=1 to nzzmax do ZZNames[i] :='dya. cov. '+st(i);
  for i:=1 to pmax do functions[i] := 0;
  Functions[1] := 1; Functions[2] := 2;
  for i:=1 to ccmax do begin
    alpa_f[i] := 0.0;
    alpa_g[i] := 0.0;
    alpa_l[i] := 0.0;
    effects_f[i] := false;
    effects_l[i] := false;
    effects_g[i] := false;
    fix_l[i] := false;
    fix_f[i] := false;
    fix_g[i] := false;
    test_l[i] := false;
    test_f[i] := false;
    test_g[i] := false;
    param_effect[i] := 0;
  end;
  symmetric := false;
  asymmetric := false;
  signed := false;
  VertexPar := false;
  VertexParNew := false;

  setlength(vertex_a,n+1);
  setlength(vertex_b,n+1);
  for i:=1 to n
  do begin
    vertex_a[i] := 0.0;
    vertex_b[i] := 0.0;
  end; (* for i *)
  for i:=1 to nzzmax do zzmean[i] := 0.0;
(* AA start *)
  for i:=1 to nzmax do
    begin
    zsmean[i] := 0.0;
    poszvar[i] := true;
    compzvar[i] := false;
    end;
(* AA end *)
  CConditional := true;
  ConditioningNetworkVariable := 1;
  ConditioningActionVariable := 0;
  ReciprocityModel := false;
  ExtendedReciprocityModel := false;
  MaxLike := false;
  Bayesian := false;
  Gibbs := 14;
  d_estAB := true;
  d_estA := false;
  d_estB := true;
  d_estimator := 1; // (* Score function method I.*)
  cdensity1 := 1; (*MS: Bayesian estimation of parameter: candidate-generating density, corresponding to three classes Metropolis-Hastings algorithms.*)
  cdensity2 := 3; (*MS: Actor-dependent random effects: candidate-generating density, corresponding to three classes Metropolis-Hastings algorithms.*)
  scalebeta := 0; (*MS: Bayesian estimation of objective function parameter: scale factor: 0 calibrated during burn-in, other number: user input, constant scale factor.*)
  scalere := 0; (*MS: Bayesian estimation of objective function parameter: scale factor: 0 calibrated during burn-in, other number: user input, constant scale factor.*)
  ModelType := 1;
  OutputType := 1;
  StaySameType := 2;
  GibbsCode := 14;
  GibbsConnected := false;
  InitialStandard := false;
  Continuous := false;
  effects_l[0] := true;
  effects_f[0] := true;
  effects_g[0] := false;

  rstep0 := 2.0;
  rstep := 2.0;
  nsubphase := 4;
  nextra := 0;
  nrun3 := 1000;
  niter := 1000;
  firstgain := 0.2;
  factorial[1] := 1.0;
  for i:=2 to facmax do factorial[i] := factorial[i-1]*i;
end;


function va(b: boolean): integer;
begin
 if b then va := 1 else va := 0;
end;

procedure DataSpecification.AttachProjectName
          (var proname: shortstring; var outf, lf: TFileStream;
           var apf: boolean; writedate: boolean);
var x: string;
begin
  if (proname = '')
  then pname := 'siena'
  else pname := proname;
  if not(hit)  (*MS: keeps track of whether or not the procedure was <hit> by the program in the past.*)
  then
  outputfiles(apf,writedate,pname,outf,lf); (*MS: please do not disable the condition again!*)
  hit := true;
//  writeln(lf, 'Project ',pname,' attached.');
  x := 'Project '+pname+' attached.'+eol;
  lf.Write(pchar(x)^,length(x));
  if (proname = '')
  then begin
    x := 'The project name for attaching a SIENA project was empty.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Therefore, the default name <<'+pname+'>> was chosen.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'If there is no basic information file <<'+pname+'.IN>>,'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'this will lead to an error.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Empty project name; default name <<'+pname+'>> used.'+eol;
    lf.Write(pchar(x)^,length(x));
  end;
end;

procedure DataSpecification.AttachProject
       (var proname, moname: shortstring;
        var halterr: Boolean;
        var outf, lf: TFileStream; var apf: boolean; writedate: boolean);
(* This is not excuted in Siena01.
   There it is replaced in Siena_1 by its constituents. *)
begin
  halterr := false;
  AttachProjectName(proname,outf,lf,apf,writedate);
  initialise;
  ReadDefinition1(halterr, outf, lf);
  if halterr then exit;
  ReadPData(halterr,outf,lf);
  if halterr then exit;
  ReadDefinition2(moname, halterr, outf, lf);
  if halterr then exit;
(* This procedure MUST be followed by ReviseDefinition. *)
end;  (* AttachProject *)

procedure DataSpecification.InitialAlphas(var alp, alpf: CRarray; show: byte; var outf: TFileStream);
  (* Calculates initial standard estimates .
     If show=1, also shows output; not for show = 0.
     TS 16-jun-05: changes in view of signed digraphs, and canceled duplication
     in S_Start. *)
var n0,n1,kmin,kmax,
    ModelCode : integer;
    aten,
    dif,dif2   : real;
    n00, n10,
    n01, n11   : array of integer;
    i,j,h,h1,k  : integer;
    ntot, ndtot: array of integer;
    nval, ndval: array of array of integer;
    nvalval: array of array of array of integer;
    ndyad: array of array of integer;
    alpha,
    prec       : array of real;
      nmisd, ndy    : integer;
      p00,p11,
      p01,p10,
      sum,
      deviance   : real;
      alphaf1, alphaf2, alphaconditionalreciprocity: real;
      pm,pa,pn   : real;
      stru, difficult  : boolean;
      sigg,siggi : pmatrix;
      erte       : shortstring;
      nplus,splus,nmin,smin,nnoskip: integer;
      x : string;

 procedure truncate(var x: real);
 begin
   if (x < 0.02) then x:= 0.02;
   if (x > 0.98) then x:= 0.98;
 end; (* truncate *)


    function MLEconditionalreciprocity(ndy,mut,asym,null: integer): real;
    var alpha2,
        alpha2old,
        mu, sig2   : real;

      procedure firstmoments(alp: real; var mm,ss: real);
      var h, s   : integer;
          pr, s1,
          s2, s0  : extended;
          (* these numbers can become awfully large, therefore extended *)
      begin
        s := 2*mut + asym;
        s1 := 0.0; s2 := 0.0; s0 := 0.0;
        pr := 1e-20;
        if (ndy >= s)
        then h := 0
        else h := s - ndy;
        while (2*h <= s)
        do begin
          s0 := s0 + pr;
          s1 := s1 + pr*h;
          s2 := s2 + pr*sqr(h);
          pr := pr*expo(alp)*(s-2*h)*(s-(2*h)-1)/(4.0*(h+1)*(ndy-s+h+1.0));
          (* number of possibilities of h mutuals, s-2h asymmetrics, and ndy-s+h nulls
             in ndy dyads is trinomial(ndy; h, s-2h, ndy-s+h)*(2^(s-2h));
             the factor (s-2*h)*(s-(2*h)-1)/(4.0*(h+1)*(ndy-s+h+1.0))
             is the ratio of this value for h+1 to its value for h *)
          inc(h);
        end;
        mm := s1/s0;
        ss := (s2/s0) - sqr(s1/s0);
      end;  (* firstmoments *)

    begin  (* MLEconditionalreciprocity *)
      if (asym <= 0)
      then begin
        MLEconditionalreciprocity := 50.0;
        exit
      end;
      if ((null <= 0) or (mut <= 0))
      then begin
        MLEconditionalreciprocity := -50.0;
        exit;
      end;
      alpha2 := alphaf2;
      repeat
        alpha2old := alpha2;
        (* Fisher scoring step *)
        firstmoments(alpha2,mu,sig2);
        alpha2 := alpha2 + (mut - mu)/sig2;
      until (abs(alpha2 - alpha2old) < 1e-10);
      MLEconditionalreciprocity := alpha2;
    end;  (* MLEconditionalreciprocity *)

  begin (* InitialAlphas *)
   stru := false;
   for h:=1 to observations do if ExistStruct[h] then stru := true;
   h := valplus - valmin + 1;
   if (h <= 1) then h := 2;
   setlength(nval,observations+2,h+2);
   setlength(ndval,observations+1,h+1);
   setlength(nvalval,observations+1,h+1,h+1);
   setlength(ndyad,h+1,h+1);
   setlength(ntot,observations+1);
   setlength(ndtot,observations+1);
   difficult := false;
   if (show = 1) then
     begin
     x := 'For the following statistics, missing values (if any) are not counted.'+eol;
     outf.Write(pchar(x)^,length(x));
     if stru then begin
     x := 'Further, structurally determined entries are treated as observed entries.'+eol;
     outf.Write(pchar(x)^,length(x));
     end;
     end;

  if symmetric
  then begin
    if (ModelType <= 2)
    then ModelCode := 1
    else ModelCode := 4;
    if (ModelType < 0) then ModelCode := - ModelCode;
  end
  else ModelCode := ModelType;


   (*  note that missing cells have the value 0=false;
       this is no more true due to the carry forward option! *)
   if (observations = 1)
   then begin
     if signed then begin
       x := 'Signed graphs not implemented for ERGM model.'+eol;
       outf.Write(pchar(x)^,length(x));
       exit;
      end;
     nval[1,1] := 0;
     ntot[1] := 0;
     ndyad[0,0] := 0; ndyad[0,1] := 0; ndyad[1,1] := 0; ndy := 0;
     for i:=1 to n
     do for j:=1 to n
        do begin
          if ynm(i,j,1) then inc(nval[1,1]);
          if not((i = j) or mis(i,j,1)) then inc(ntot[1]);
          if ((i < j) and (not mis(i,j,1)) and (not mis(j,i,1)))
          then begin
            k := 0;
            if ynm(i,j,1) then inc(k);
            if ynm(j,i,1) then inc(k);
            case k of
              0 : inc(ndyad[0,0]);
              1 : inc(ndyad[0,1]);
              2 : inc(ndyad[1,1]);
            end; (* case *)
            inc(ndy);
          end; (* if *)
        end; (* for i, j *)
     if (show = 1) then begin
        x := format('%s%d%s%d%s%8.4f%s',['Network density  = ',nval[1,1],'/',ntot[1],' = ',
                   nval[1,1]/(ntot[1]+1e-20),' .'])+eol;
       outf.Write(pchar(x)^,length(x));
       x := format('%s%8.4f%s',['Average degree   = ', (n-1)*nval[1,1]/(ntot[1]+1e-20),' .'])+eol;
       outf.Write(pchar(x)^,length(x));
       if ((nval[1,1] = 0) or (nval[1,1] = ntot[1]))
       then begin
        x := 'WARNING: meaningless data set for this model!!'+eol;
        outf.Write(pchar(x)^,length(x));
       end;
      end;
   end
   else begin (* observations >=2 *)
     setlength(alpha,observations+1);
     setlength(prec,observations+1);
     if symmetric then begin
       if (show = 1) then
         begin
         if (observations = 2) then begin
           x := 'Both observed networks are symmetric.'+eol;
           outf.Write(pchar(x)^,length(x));
         end
          else begin
          x := format('%s%d%s',['All ',observations ,' observed networks are symmetric.'])+eol;
          outf.Write(pchar(x)^,length(x));
          end;
         x := 'Therefore, it is assumed that this is an analysis of an non-directed relation.'+eol;
         outf.Write(pchar(x)^,length(x));
         outf.Write(pchar(eol)^,length(eol));
         end;
// waarom stond dit hier???
//gevaarlijk, want modeltype wordt voor symmetric niet teruggezet.
// if signed then ModelType := -2 else ModelType := 2; (* default at initialisation *)
      end;
     if (show = 1) then begin
      x := 'Network density indicators:'+eol;
      outf.Write(pchar(x)^,length(x));
     end;

     for k := 1 to observations do
       begin
       n := nni[k];
       for h := 0 to valplus-valmin do
         begin
         nval[k,h] := 0;
         ndval[k,h] := 0;
         for h1 := 0 to valplus-valmin do nvalval[k,h,h1] := 0;
         end;
       ntot[k] := 0;
       ndtot[k] := 0;
       for i:=1 to n do begin
         for j:=1 to n do begin
            if not(mis(i,j,k) or (i=j)) then begin
              inc(ntot[k]);
              inc(nval[k,ynvm(i,j,k)-valmin]);
             end;
            if ((k >= 2) and (not (mis(i,j,k-1) or (i=j))))
              then inc(ndval[k,ynvm(i,j,k)-valmin]);
            if (k <observations)
            then begin
              if not(mis(i,j,k) or mis(i,j,k+1) or (i=j)) then begin
                inc(ndtot[k]);
                inc(nvalval[k,ynvm(i,j,k)-valmin,ynvm(i,j,k+1)-valmin]);
               end;
            end;
          end;
          end; (* for i,j *)
       if ((valmin = 0) and (valplus = 1) and ((nval[k,1] >= ntot[k]) or (nval[k,1] = 0)))
         then difficult := true;
       if symmetric then begin
         for h := 0 to valplus-valmin do begin
           nval[k,h] := nval[k,h] div 2;
           ndval[k,h] := ndval[k,h] div 2;
           for h1 := 0 to valplus-valmin do nvalval[k,h,h1] := nvalval[k,h,h1] div 2;
          end; (* h *)
         ntot[k] := ntot[k] div 2;
         ndtot[k] := ndtot[k] div 2;
        end;
     end; (* for k *)

     if (show = 1) then
       begin
       kmax := 0;
       repeat
       kmin := kmax + 1;
       kmax := kmax + 7;
       if (kmax > observations) then kmax := observations;
       x := 'observation time         ';
       outf.Write(pchar(x)^,length(x));
       for k:=kmin to kmax do outf.write(pchar(format('%5d%s',[k,'   ']))^,8);
       outf.write(pchar(eol)^,length(eol));
       if signed then begin
         x := '=========================';
         outf.Write(pchar(x)^,length(x));
         for k:=kmin to kmax do outf.Write(pchar(string('========'))^,8);
         outf.write(pchar(eol)^,length(eol));
         for h := valmin to valplus do begin
           x := format('%s%2d%s',['density value ',h,'        ']);
           outf.Write(pchar(x)^,length(x));
           for k:=kmin to kmax
           do if (ntot[k] = 0)
              then begin
              x := '     --';
              outf.Write(pchar(x)^,length(x));
              end
              else begin
              x := format('%s%7.3f',[' ',nval[k,h-valmin]/ntot[k]]);
              outf.Write(pchar(x)^,length(x));
              end;
           outf.write(pchar(eol)^,length(eol));
           x := format('%s%2d%s',['average degree value ',h,' ']);
           outf.Write(pchar(x)^,length(x));
          for k:=kmin to kmax
           do if (ntot[k] = 0)
              then begin
              x := '     --';
              outf.Write(pchar(x)^,length(x));
              end
              else begin
              n := nni[k];
              x := format('%s%7.3f%s',[' ',(n-1)*nval[k,h-valmin]/ntot[k]]);  // corrected 3.17m
              outf.Write(pchar(x)^,length(x));
              end;
           outf.write(pchar(eol)^,length(eol));
           x := format('%s%2d%s',['number pairs value ',h,'   ']);
           outf.Write(pchar(x)^,length(x));
           for k:=kmin to kmax
           do if (ntot[k] = 0)
              then outf.Write(pchar(string('     --'))^, 7)
              else begin
              x := format('%s%7d',[' ',nval[k,h-valmin]]);
              outf.Write(pchar(x)^,length(x));
              end;
           outf.write(pchar(eol)^,length(eol));
           outf.write(pchar(eol)^,length(eol));
         end; (* for h *)
        end else begin (* not signed *)
         x := 'density                ';
         outf.Write(pchar(x)^,length(x));
         for k:=kmin to kmax
         do if (ntot[k] = 0)
            then outf.Write(pchar(string('     --'))^, 7)
            else begin
              x := format('%s%7.3f',[' ',nval[k,1]/ntot[k]]);
              outf.Write(pchar(x)^,length(x));
            end;
         outf.write(pchar(eol)^,length(eol));
         x := 'average degree         ';
         outf.Write(pchar(x)^,length(x));
         for k:=kmin to kmax
         do if (ntot[k] = 0)
            then outf.Write(pchar(string('     --'))^, 7) //write(outf,'     --')
            else begin
            n := nni[k];
            x := format('%s%7.3f',[' ',(n-1)*nval[k,1]/ntot[k]]);   // corrected 3.17m
            outf.Write(pchar(x)^,length(x));
            end;
            outf.write(pchar(eol)^,length(eol));
            x := 'number of ties         ';
            outf.Write(pchar(x)^,length(x));
         for k:=kmin to kmax
         do if (ntot[k] = 0)
            then outf.Write(pchar(string('     --'))^, 7)//write(outf,'     --')
            else begin
              x := format('%s%7d',[' ',nval[k,1]]);
              outf.Write(pchar(x)^,length(x));
            end;
         outf.write(pchar(eol)^,length(eol));
        end; (* if not signed *)
       x := 'missing fraction       ';
       outf.Write(pchar(x)^,length(x));
       for k:=kmin to kmax
       do if (ntot[k] = 0)
          then outf.write(pchar(string('   1.0  '))^,8)
          else begin
            n := nni[k];
            if symmetric then begin
              x := format('%s%7.3f',[' ',(1.0 - 2*ntot[k]/(n*(n-1)))]);
              outf.Write(pchar(x)^,length(x));
            end
             else begin
             x := format('%s%7.3f',[' ',(1.0 - ntot[k]/(n*(n-1)))]);
             outf.Write(pchar(x)^,length(x));
             end;
         end;
       if (kmax < observations) then
         begin
         outf.write(pchar(eol)^,length(eol));
         outf.write(pchar(eol)^,length(eol));
         end;
       until (kmax >= observations);

       outf.write(pchar(eol)^,length(eol));
       outf.write(pchar(eol)^,length(eol));

       if symmetric then outf.write(pchar(string('Edge '))^,5) else outf.write(pchar(string('Tie '))^,4);
       x := 'changes between subsequent observations:'+eol;
       outf.Write(pchar(x)^,length(x));
       x := ' periods       ';
       outf.Write(pchar(x)^,length(x));
       for h:=valmin to valplus do for h1:=valmin to valplus do begin
       x := format('%2d%s%2d%s',[h,' => ',h1,'  ']);
        outf.Write(pchar(x)^,length(x));
       end;
       if (valmin = 0) and (valplus = 1) then
        x := 'Distance  Jaccard   Missing'+eol
           else x := 'Distance     Missing'+eol;
        outf.Write(pchar(x)^,length(x));
      end; (* if show = 1 *)

     for k:=1 to observations-1 do
       begin
       n := nni[k];
       if symmetric then nmisd := (n*(n-1) div 2) -ndtot[k]
         else nmisd := n*(n-1)-ndtot[k];
       if ((show = 1) and (not skipi[k])) then
         begin
         x := format('%3d%s%3d',[k,' ==>',k+1]);
         outf.Write(pchar(x)^,length(x));
         if symmetric then
           begin
           for h:= valmin to valplus do for h1 := valmin to valplus do
             begin
             x := format('%10d',[nvalval[k,h-valmin,h1-valmin]]);
             outf.Write(pchar(x)^,length(x));
             end;
(* Version 3.17n: extra space in output *)
           if (valmin = 0) and (valplus = 1) then
             x := format('%10d%s%9.3f%s%6d%s%2.0f%s',[ ObservedNetworkDistance[1,k] div 2,
                         ' ',d(nvalval[k,1,1],(nvalval[k,1,1]+nvalval[k,1,0]+nvalval[k,0,1])),' ',
                         nmisd,' (',(200.0*nmisd/(n*(n-1))),'%)'])
            else
             x := format('%10d%s%10d%s%2.0f%s',[ ObservedNetworkDistance[1,k] div 2,' ', nmisd,
                          ' (',(200.0*nmisd/(n*(n-1))),'%)']);
           outf.Write(pchar(x)^,length(x));
           end
          else  (* if not symmetric *)
           begin
           for h:= valmin to valplus do for h1 := valmin to valplus do
             begin
             x := format('%10d',[nvalval[k,h-valmin,h1-valmin]]);
             outf.Write(pchar(x)^,length(x));
             end;
(* Version 3.17n: extra space in output *)
           if (valmin = 0) and (valplus = 1) then
             x := format('%10d%s%9.3f%s%6d%s%2.0f%s',[ ObservedNetworkDistance[1,k],
                         ' ',d(nvalval[k,1,1],(nvalval[k,1,1]+nvalval[k,1,0]+nvalval[k,0,1])),' ',
                         nmisd,' (',(100.0*nmisd/(n*(n-1))),'%)'])
            else
             x := format('%10d%s%10d%s%2.0f%s',[ ObservedNetworkDistance[1,k],' ', nmisd,
                          ' (',(100.0*nmisd/(n*(n-1))),'%)']);
           outf.Write(pchar(x)^,length(x));
           end;
         end; (* if show = 1 *)
       if ((valmin = 0) and (valplus = 1)) then
         begin
         if ((show = 1) and (not skipi[k])) then
           begin
//           if ((nvalval[k,1,1]*(nvalval[k,0,0]+nvalval[k,0,1]) < nvalval[k,0,1]*(nvalval[k,1,0]+nvalval[k,1,1]))
//            and (nvalval[k,0,0]+nvalval[k,0,1] > 0))
// changed to division to prevent integer overflow
           if ((d(nvalval[k,1,1], nvalval[k,1,0]+nvalval[k,1,1]) <
               d(nvalval[k,0,1], nvalval[k,0,0]+nvalval[k,0,1]))
              and (nvalval[k,0,0]+nvalval[k,0,1] > 0))
           then
             begin
             wwriteln(outf,'');
             wwriteln(outf,'');
             x := format('%s%d%s',['This means that in period ',k,', proportionately less 1-ties stayed 1,'])+eol;
             wwrite(outf,x);
             x := 'than 0-ties became 1. A great reversal of the network pattern!'+eol;
             wwrite(outf,x);
             wwriteln(outf,'WARNING: ');
             x := 'For some model specifications this may lead to problems in estimation.'+eol;
             wwrite(outf,x);
             end; (* if *)
           end; (* if show = 1 *)

         if ((nvalval[k,0,0]+nvalval[k,0,1]) >= 1)
         then p01 := nvalval[k,0,1]/(nvalval[k,0,0]+nvalval[k,0,1])
         else p01 := 0.5;
         if ((nvalval[k,1,0]+nvalval[k,1,1]) >= 1)
          then p10 := nvalval[k,1,0]/(nvalval[k,1,0]+nvalval[k,1,1])
          else p10 := 0.5;
         truncate(p01);
         truncate(p10);
         alpha[k] := 0.5*log(p01/p10);
         if ((nvalval[k,0,0]+nvalval[k,0,1]) >= 1) then p00 := nvalval[k,0,0]/(nvalval[k,0,0]+nvalval[k,0,1]);
         if ((nvalval[k,1,0]+nvalval[k,1,1]) >= 1) then p11 := nvalval[k,1,1]/(nvalval[k,1,0]+nvalval[k,1,1]);
         truncate(p00);
         truncate(p11);
         (* truncation in order not to give one of the observations too much weight *)
         if ((nvalval[k,0,1]*nvalval[k,1,0]) >= 1)
         then prec[k] := 4.0/((p00/nvalval[k,0,1]) + (p11/nvalval[k,1,0]))
         else prec[k] := 1e-6;
         (* prec is 1/(estimation variance) *)
         end (* if valmin = 0 and valplus = 1 *)
       else
         begin
(* the following assumes we are in the signed digraph case, and gives a rough approximation;
   for positive tie tendency, neglect negative ties;
   for negative tie tendency, neglect positive ties  *)
         if ((nvalval[k,1,2]+nvalval[k,1,2]) >= 1)
         then p01 := nvalval[k,1,2]/(nvalval[k,1,1]+nvalval[k,1,2])
         else p01 := 0.5;
         if ((nvalval[k,2,1]+nvalval[k,2,2]) >= 1)
          then p10 := nvalval[k,2,1]/(nvalval[k,2,1]+nvalval[k,2,2])
          else p10 := 0.5;
         truncate(p01);
         truncate(p10);
         alpha[k] := 0.5*log(p01/p10);

         if ((nvalval[k,1,1]+nvalval[k,1,0]) >= 1)
         then p01 := nvalval[k,1,0]/(nvalval[k,1,1]+nvalval[k,1,0])
         else p01 := 0.5;
         if ((nvalval[k,0,1]+nvalval[k,0,0]) >= 1)
          then p10 := nvalval[k,0,1]/(nvalval[k,0,1]+nvalval[k,0,0])
          else p10 := 0.5;
         truncate(p01);
         truncate(p10);
         prec[k] := 0.5*log(p01/p10);
    (* prec in this case is the analogue to alpha, but for negative ties;
       it is not the precision! *)
         end; (* if not (valmin = 0 and valplus = 1) *)
       if (show = 1) then outf.Write(pchar(eol)^,length(eol));
       end; (* for k *)

     if stru and (show=1)
     then begin
       wwriteln(outf,'');
       x := 'Distances do not consider structurally fixed network ties.'+eol;
       outf.Write(pchar(x)^,length(x));
       x := 'This may imply that the columns 0 => 1 and 1 => 0 do not add up to the distance column.'+eol;
       outf.Write(pchar(x)^,length(x));
     end;

     if (symmetric and (show = 1)) then begin
       x := '(The distances reported in the output file for conditional estimation'+eol;
       outf.Write(pchar(x)^,length(x));
       x := 'for the network variable refer to the total symmetric adjacency matrix,'+eol;
       outf.Write(pchar(x)^,length(x));
       x :='and therefore are double the distance reported above.)'+eol;
       outf.Write(pchar(x)^,length(x));
       outf.Write(pchar(eol)^,length(eol));
      end; (* for k *)

     if (show = 1) then begin
       outf.Write(pchar(eol)^,length(eol));
       x := 'Dyad counts:'+eol;
       outf.Write(pchar(x)^,length(x));
       if ((valmin=0) and (valplus=1)) then  begin
         x := ' observation    total    mutual    asymm.     null'+eol;
         outf.Write(pchar(x)^,length(x));
       end
        else begin
         x := ' observation    total  ';
         outf.Write(pchar(x)^,length(x));
         for h:=valmin to valplus do for h1:= h to valplus do begin
         x := format('%s%2d%s%2d%s',['(',h,',',h1,') '] );
         outf.Write(pchar(x)^,length(x));
         end;
         outf.Write(pchar(eol)^,length(eol));
        end;
      end; (* if show = 1 *)

     for k:=1 to observations do begin
       n := nni[k];
       ndy := 0;
       for h:=0 to valplus-valmin do for h1 := 0 to valplus-valmin do ndyad[h,h1] := 0;
       for i:=1 to n do for j:=i+1 to n do begin
         if ((not mis(i,j,k)) and (not mis(j,i,k)))
          then begin
            inc(ndyad[ynvm(i,j,k)-valmin,ynvm(j,i,k)-valmin]);
            if (ynvm(i,j,k) <> ynvm(j,i,k)) then
              inc(ndyad[ynvm(j,i,k)-valmin,ynvm(i,j,k)-valmin]);
            inc(ndy);
          end; (* if *)
        end; (* for i, j *)
       if (show = 1) then begin
         if ((valmin = 0) and (valplus = 1)) then begin
           x := format('%6d%s%6d%s%6d%s%6d%s%6d',[k,'.        ',ndy,'   ',ndyad[1,1],'    ',ndyad[0,1],'    ',ndyad[0,0]])+eol;
           outf.Write(pchar(x)^,length(x));
           end
           else begin
             x := format('%6d%s%6d',[k,'.        ',ndy]);
             outf.Write(pchar(x)^,length(x));
             for h:=valmin to valplus do for h1 := h to valplus do begin
             x := format( '%s%6d%s',[' ',ndyad[h-valmin,h1-valmin],' ']);
             outf.Write(pchar(x)^,length(x));
             end;
             //writeln(outf);
             outf.Write(pchar(eol)^,length(eol));
            end;
         end; (* if show = 1 *)
      end; (* for k *)
   end; (* if observations >= 2 *)

   if (difficult and (show = 1))
   then begin
     x := 'There is a density equal to 0.0 or 1.0.'+eol;
     outf.Write(pchar(x)^,length(x));
     x := 'This may lead to difficulties.'+eol;
     outf.Write(pchar(x)^,length(x));
     //writeln(outf)
     outf.Write(pchar(eol)^,length(eol));
   end;

   alphaf1 := 0.0;
   alphaf2 := 0.0;
   alphaconditionalreciprocity := 0.0;

   if (observations = 1)
   then begin
     n0 := ntot[1] - nval[1,1];  (*[s]*)
     if (n0 >= 1) then alphaf1 := log(nval[1,1]/n0)   (*[s]*)
       else begin
         if (show = 1) then begin
           x := 'A density of 1.0 leads to trivial difficulties.'+eol;
           outf.Write(pchar(x)^,length(x));
           x := 'Do not go on with this data set!'+eol;
           outf.Write(pchar(x)^,length(x));
           alphaf1 := 10.0;
          end;
       end;
     if (show = 1) then begin
       x := 'Parameter estimate in the independent ';
       outf.Write(pchar(x)^,length(x));
       if symmetric then outf.write(pchar(string('edges '))^,6) else outf.write(pchar(string('ties '))^,5);
       x := format('%s%8.4f%s',['model =', alphaf1,' ,'])+eol;
       outf.Write(pchar(x)^,length(x));
       if ((n0 >= 1) and (nval[1,1] >= 1))      (*[s]*)
       then begin
         pm := nval[1,1]/ntot[1];                      (*[s]*)
         if symmetric then sigg[1,1] :=1/sqrt(0.5*pm*(1.0-pm)*ntot[1])
          else sigg[1,1] :=1/sqrt(pm*(1.0-pm)*ntot[1]);
         x := format('%s%8.4f%s',['standard error   = ',sigg[1,1],' .'])+eol;
          outf.Write(pchar(x)^,length(x));
       end;
      // writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       x := 'Dyad count:    total    mutual    asymm.     null'+eol;
       outf.Write(pchar(x)^,length(x));
       x := format('%s%6d%s%6d%s%6d%s%6d',['             ',ndy,'    ',ndyad[1,1],'    ',ndyad[0,1],'    ',ndyad[0,0]])+eol;
       outf.Write(pchar(x)^,length(x));
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));

       if ExistStruct[1] then begin
         x := 'For the parameter estimates first reported here,'+eol;
         outf.Write(pchar(x)^,length(x));
         x := 'structurally determined values are treated as outcomes.'+eol;
         outf.Write(pchar(x)^,length(x));
        end;
      end; (* if show = 1 *)

     if ((ndyad[0,0] >= 1) and (ndyad[0,1] >= 1) and (ndyad[1,1] >= 1))
     then begin
       alphaf1 := log(ndyad[0,1]/(2.0*ndyad[0,0]));
       alphaf2 := log((4.0*ndyad[1,1]*ndyad[0,0])/sqr(ndyad[0,1]));
       if (show = 1) then begin
         x := 'For the reciprocity model, the ML parameter estimates are'+eol;
         outf.Write(pchar(x)^,length(x));
         x := format('%s%8.4f',['density:     ',alphaf1])+eol;
         outf.Write(pchar(x)^,length(x));
         x := format('%s%8.4f',['reciprocity: ',alphaf2])+eol;
         outf.Write(pchar(x)^,length(x));
         pm := ndyad[1,1]/ndy;
         pa := ndyad[0,1]/ndy;
         pn := ndyad[0,0]/ndy;
         (* covariance matrix of sufficient statistic (2*ndyad[1,1] + ndyad[0,1], ndyad[1,1]) : *)
         sigg[1,1] := 4*pm*(1.0-pm) + pa*(1.0-pa) - 4*pm*pa;
         sigg[2,2] := pm*(1.0-pm);
         sigg[1,2] := 2*pm*(1.0-pm) - pm*pa;
         sigg[2,1] := sigg[1,2];
         for i:=1 to 2 do for j:=1 to 2 do sigg[i,j] := ndy*sigg[i,j];
         x := 'Covariance matrix of sufficient statistics (2M+A, M) (correlation below diagonal):'+eol;
         outf.Write(pchar(x)^,length(x));
         x := format('%10.4f%s%10.4f',[sigg[1,1],' ',sigg[1,2]])+eol;
         outf.Write(pchar(x)^,length(x));
         x := format('%10.4f%s%10.4f',[(sigg[1,2]/sqrt(sigg[1,1]*sigg[2,2])),' ',sigg[2,2]])+eol;
         outf.Write(pchar(x)^,length(x));
         pinverse(2,sigg,siggi,erte);
         if (erte = '')
         then begin
           x := 'Covariance matrix of ML estimates (correlation below diagonal'+eol;
           outf.Write(pchar(x)^,length(x));
           x := format('%10.4f%s%10.4f',[siggi[1,1],' ',siggi[1,2]])+eol;
           outf.Write(pchar(x)^,length(x));
           x := format('%10.4f%s%10.4f',[(siggi[1,2]/sqrt(siggi[1,1]*siggi[2,2])),' ',
                         siggi[2,2]])+eol;
           outf.Write(pchar(x)^,length(x));
           x := 'Standard errors:'+eol;
           outf.Write(pchar(x)^,length(x));
           x := format('%s%8.4f',['density:     ',sqrt(siggi[1,1])])+eol;
           outf.Write(pchar(x)^,length(x));
           x := format('%s%8.4f',['reciprocity: ',sqrt(siggi[2,2])])+eol;
           outf.Write(pchar(x)^,length(x));
           //writeln(outf);
           outf.Write(pchar(eol)^,length(eol));
         end;
         deviance := ndyad[1,1]*log(pm) + ndyad[0,1]*log(pa) + ndyad[0,0]*log(pn);
         (* no trinomial coefficient: deviance of non-reduced data! *)
          x := format('%s%10.4f',['Deviance:  ', -2.0*deviance])+eol;
         outf.Write(pchar(x)^,length(x));
        end; (* if show = 1 *)
      end (* if ((ndyad[0,0] >= 1) and (ndyad[0,1] >= 1) and (ndyad[1,1] >= 1)) *)

     else begin
(* Note that this procedure InitialStatistics is always preceded by Calculations,
   in which the variable symmetric has been calculated; it implies ndyad[0,1] = 0. *)
       if (symmetric and (show = 1))
       then begin
         x := 'This data set is treated as a non-directed graph.'+eol;
         outf.Write(pchar(x)^,length(x));
         SetDigraph(1);
//         ConnectedCheck(outf);
         wwrite(outf,'The observed network is ');
         if (not ConnectedNetwork) then wwrite(outf,'not ');
         wwriteln(outf,'connected.');
         Gibbs := 6;
         pm := ndyad[1,1]/ndy;
         deviance := ndyad[1,1]*log(pm) + ndyad[0,0]*log(1.0-pm);
         x := format('%s%10.4f%s',['Deviance under the independent edges model =', -2.0*deviance,' .']);
         outf.Write(pchar(x)^,length(x));
         (* alphaf1 was defined already *)
       end;
       if ((ndyad[1,1] = 0) and (ndyad[0,0] = 0) and (show = 1)) then begin
         x := 'This data set contains only asymmetric dyads.'+eol;
         outf.Write(pchar(x)^,length(x));
         x := 'By retaining the Gibbs code as 7 or 17, it is treated as a network'+eol;
         outf.Write(pchar(x)^,length(x));
         x := 'that is forced to be asymmetric (a tournament).'+eol;
         outf.Write(pchar(x)^,length(x));
         Gibbs := 7;
        end; (* if ndyad[1,1] = ndyad[0,0] = 0 *)
       alphaf2 := 0.0;
     end; (* if not (ndyad[1,1] > 0 and ndyad[0,1] > 0 and ndyad[0,0] > 0) *)


     alpf[1] := alphaf1;
     alpf[2] := alphaf2;
     (* Last part for observations = 1 case: computation of alphaconditionalreciprocity,
        that is the MLE for the reciprocity parameter in the reciprocity model,
        conditional on the number of edges. *)
     if ((ndyad[0,1] > 0) and (show = 1))
     then begin
       alphaconditionalreciprocity := MLEconditionalreciprocity(ndy,ndyad[1,1],ndyad[0,1],ndyad[0,0]);
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       x := 'For the reciprocity model conditional on the total number of arcs,'+eol;
       outf.Write(pchar(x)^,length(x));
       x := 'the maximum likelihood estimator of the reciprocity parameter is '+eol;
       outf.Write(pchar(x)^,length(x));
       x := format('%8.4f',[alphaconditionalreciprocity])+eol;
       outf.Write(pchar(x)^,length(x));
     end
   end (* observations = 1 *)

   else begin (* observations >= 2 *)
     if (show = 1) then begin
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       x := 'Standard values for initial parameter values'+eol;
       outf.Write(pchar(x)^,length(x));
       x := '-------------------------------------------------'+eol;
       outf.Write(pchar(x)^,length(x));
      end; (* if show *)

(* first starting values for rate function intercept parameters *)
     for k := 1 to observations-1 do if (not skipi[k]) then begin
(* note that ndtot[k] = ObservedNetworkData[1,k], both could have been used here *)
       n := nni[k];
       case ModelCode of
      -1,1,2 : begin
              if symmetric (* each step reduces distance by 2 *)
              then alp[k] := n*(0.2+ObservedNetworkDistance[1,k])/(ndtot[k]+1)
              else alp[k] := n*(0.2+2*ObservedNetworkDistance[1,k])/(ndtot[k]+1);  // @@@  teruggezet TS
              (* motivation: These are very rough values.
                 They must be positive and at least log(ObservedDistance/n). *)
              if (alp[k] < 0.1) then alp[k] := 0.1;
              if (alp[k] > 100.0) then alp[k] := 100.0;  // @@@  teruggezet TS
            end; (* case -1,1,2 *)
          3 : begin
              if symmetric (* each step reduces distance by 2 *) then
                begin
                alp[k] := n*(0.2+ObservedNetworkDistance[1,k])/(ndtot[k]+1);
                if (alp[k] < 0.1) then alp[k] := 0.1;
                if (alp[k] > 100.0) then alp[k] := 100.0;
                end
               else
                begin
                alp[k] := (0.2+4*ObservedNetworkDistance[1,k])/(sqr(n) - StructElements(k));
                if (alp[k] < 0.001) then alp[k] := 0.001;
                if (alp[k] > 1.0) then alp[k] := 1.0;
                end;
            end; (* case -3 *)
      4,5   : begin
              if symmetric (* each step reduces distance by 2 *)
              then alp[k] := sqrt(n*(0.2+ObservedNetworkDistance[1,k])/(ndtot[k]+1))
              else alp[k] := sqrt(n*(0.2+2*ObservedNetworkDistance[1,k])/(ndtot[k]+1));
              if (alp[k] < 0.1) then alp[k] := 0.1;
              if (alp[k] > 100.0) then alp[k] := 100.0;
            end; (* case 4,5 *)
       end; (* case *)

     if (show = 1) then begin
       x := labels_l[k];
       outf.Write(pchar(x)^,length(x));
       for h := length(labels_l[k]) to 48 do outf.write(pchar(string(' '))^,1);
       x := format('%8.4f',[ alp[k]])+eol;
       outf.Write(pchar(x)^,length(x));
     end;  (* if show *)
    end else alp[k] := 1.0; (* for k *)

(* now starting values for objective function *)
     if (observations = 2)
     then begin
       if signed then begin
         alpf[1] := alpha[1];
         alpf[2] := prec[1];
         for k:=1 to 2 do begin
           if (alpf[k] < -3.0) then alpf[k] := -3.0;
           if (alpf[k] > 3.0) then alpf[k] := 3.0;
          end;
         if (show = 1) then begin
           for k:=1 to 2 do begin
             x := labels_f[k];
             outf.Write(pchar(x)^,length(x));
             for h := length(labels_f[k]) to 48 do outf.write(pchar(string(' '))^,1);
             x := format('%8.4f',[alpf[k]])+eol;
             outf.Write(pchar(x)^,length(x));
           end;
          end;
        end else begin
         alpf[1] := alpha[1];
         if (show = 1) then begin
//           writeln(outf);
           x := labels_f[1];
           outf.Write(pchar(x)^,length(x));
           for h := length(labels_f[k]) to 48 do outf.write(pchar(string(' '))^,1);
           x := format('%s%8.4f',[' ', alpf[1]])+eol;
           outf.Write(pchar(x)^,length(x));
           if symmetric
           then begin
           x := '(estimate under the (trivial) independent edges model)'+eol;
           outf.Write(pchar(x)^,length(x));
           end
           else begin
           x := '(estimate under the (trivial) independent ties model)'+eol;
           outf.Write(pchar(x)^,length(x));
           end;
          end; (* if show = 1 *)
        end; (* if not signed [s] *)

      end (* if observations = 2 *) else (* observations >= 3 *) begin
       if signed then begin
         sum := 0.0;
         nnoskip := 0;
         for k:=1 to observations-1 do if (not skipi[k]) then
           begin
           n := nni[k];
           sum := sum+alpha[k];
           inc(nnoskip);
           end;
         alpf[1] := sum/nnoskip;
         sum := 0.0;
         for k:=1 to observations-1 do if (not skipi[k]) then sum := sum+prec[k];
         alpf[2] := sum/nnoskip;
         for k:=1 to 2 do begin
           if (alpf[k] < -3.0) then alpf[k] := -3.0;
           if (alpf[k] > 3.0) then alpf[k] := 3.0;
          end;
         if (show = 1) then begin
           for k:=1 to 2 do begin
             x := labels_f[k];
             outf.Write(pchar(x)^,length(x));
             for h := length(labels_f[k]) to 48 do outf.write(pchar(string(' '))^,1);
             x := format('%8.4f',[alpf[k]])+eol;
             outf.Write(pchar(x)^,length(x));
           end;
          end;
       end else (* not signed *) begin
        sum := 0.0;
        nnoskip := 0;
        for k:=1 to observations-1 do if (not skipi[k]) then
          begin
          sum := sum+prec[k];
          inc(nnoskip);
          end;
        if (sum > 0.0)
        then begin
          alphaf1 := 0.0;
          for k:=1 to observations-1
          do alphaf1 := alphaf1 + (alpha[k]*prec[k])/sum;
        end
        else alphaf1 := 0.0;
        if (show = 1) then begin
          x := labels_f[1];
          outf.Write(pchar(x)^,length(x));
          for h := length(labels_f[1]) to 48 do outf.write(pchar(string(' '))^,1);
          x := format('%8.4f', [alpf[1]])+eol;
          outf.Write(pchar(x)^,length(x));

          x := ' - and zero for all other parameters'+eol;
          outf.Write(pchar(x)^,length(x));
{ The following output seemed superfluous and is skipped. T.S.
          if symmetric
          then begin
          x :='(weighted average of estimates under the (trivial) independent edges model);'+eol;
          lf.Write(pchar(x)^,length(x));
          end
          else begin
          x := '(weighted average of estimates under the (trivial) independent ties model);'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
          x := 'estimates and weights are as follows:'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'period        ';
          outf.Write(pchar(x)^,length(x));
          for k:=1 to observations-1 do begin
            x := format('%8d%s',[k,' ']);
            outf.Write(pchar(x)^,length(x));
          end;
          outf.Write(pchar(eol)^,length(eol));
          x := 'density par.est. ';
          outf.Write(pchar(x)^,length(x));
          for k:=1 to observations-1 do begin
            x := format('%8.4f%s',[alpha[k],' ']);
            outf.Write(pchar(x)^,length(x));
          end;
          outf.Write(pchar(eol)^,length(eol));
          x := 'weight           ';
          outf.Write(pchar(x)^,length(x));
          for k:=1 to observations-1 do begin
          x := format('%8.4f%s',[prec[k]/sum,' ']);
          outf.Write(pchar(x)^,length(x));
          end;
          x := ')'+eol;
          outf.Write(pchar(x)^,length(x));
}          
         end; (* if show = 1 *)
        end; (* if not signed [s] *)
     end; (* if observations >= 3 *)

     if (not signed) then begin
       if (alphaf1 < -3.0)
       then begin
         if (show = 1) then begin
         x := 'The initial parameter value is very low. It is truncated to -3.0.'+eol;
         outf.Write(pchar(x)^,length(x));
         end;
         alphaf1 := -3.0;
       end;
       if (alphaf1 > 3.0)
       then begin
         if (show = 1) then begin
         x := 'The initial parameter value is very high. It is truncated to 3.0.'+eol;
         outf.Write(pchar(x)^,length(x));
         end;
         alphaf1 := 3.0;
       end;
       alpf[1] := alphaf1;
      end; (* if not signed *)

(* Now rate function parameters *)
{ TS 18-jun-05:
Volgens mij kan het volgende weg.
  for i:=1 to ccmax
  do begin
    alp[i] := 0.0; alpf[i] := 0.0;
  end;
  if symmetric
  then begin
    if (ModelType <= 2)
    then ModelCode := 1
    else ModelCode := 4;
    if (ModelType < 0) then ModelCode := - ModelCode; (*[s]*)
  end
  else ModelCode := ModelType;
  alphaf2 := 0.0;

  (* First parameters for network change. *)
  (*  note that missing cells have the value 0=false *)
  if (observations = 1)
  then begin
    null := 0; asym := 0; mut := 0;
    for i:=n-1 downto 1
    do for j:=i+1 to n
    do if (not((i=j) or mis(i,j,1) or mis(j,i,1) or struct(i,j,1) or struct(j,i,1)))
    then begin
      k := 0;
      if ynm(i,j,1) then inc(k);
      if ynm(j,i,1) then inc(k);
      case k of
        0 : inc(null);
        1 : inc(asym);
        2 : inc(mut);
      end; (* case *)
    end; (* for i, j *)
    n1 := 2*null + asym;
    ntot := 2*(null+asym+mut);
  end
   else begin (* observations >=2 *)
    setlength(alpha,observations+1);
    setlength(prec,observations+1);
    ntot := 0;
    for k := 1 to observations-1
    do begin
      n11[k] := 0;
      n01[k] := 0;
      n10[k] := 0;
      n00[k] := 0;
      for i:=1 to n
      do for j:=1 to n
      do if (not ((i=j) or mis(i,j,k) or mis(i,j,k+1) or struct(i,j,k) or struct(i,j,k+1)))
      then begin
        if (ynm(i,j,k) and ynm(i,j,k+1)) then inc(n11[k]);
        if (ynm(i,j,k) and (not ynm(i,j,k+1))) then inc(n10[k]);
        if ((not(ynm(i,j,k))) and ynm(i,j,k+1)) then inc(n01[k]);
        if (not(ynm(i,j,k) or ynm(i,j,k+1))) then inc(n00[k]);
      end; (* for i,j if *)
      ntot := ntot + n00[k] + n01[k] + n10[k] + n11[k];
      if ((n00[k]+n01[k]) >= 1)
      then p01 := n11[k]/(n00[k]+n01[k])
      else p01 := 0.5;
      if ((n10[k]+n11[k]) >= 1)
      then p10 := n10[k]/(n10[k]+n11[k])
      else p10 := 0.5;
      truncate(p01);
      truncate(p10);
      alpha[k] := 0.5*log(p01/p10);
      if ((n00[k]+n01[k]) >= 1) then p00 := n00[k]/(n00[k]+n01[k]);
      if ((n10[k]+n11[k]) >= 1) then p11 := n11[k]/(n10[k]+n11[k]);
      truncate(p00);
      truncate(p11);
      (* truncation in order not to give one of the observations too much weight *)
      if ((n01[k]*n10[k]) >= 1)
      then prec[k] := 4.0/((p00/n01[k]) + (p11/n10[k]))
      else prec[k] := 1e-6;
      (* prec is 1/(estimation variance) *)
    end; (* for k *)
  end; (* if observations >= 2 *)

  if (observations = 1)
  then begin
    n0 := ntot - n1;
    if (n0 >= 1)
    then alphaf1 := log(n1/n0)
    else alphaf1 := 10.0;
    if ((null >= 1) and (asym >= 1) and (mut >= 1))
    then begin
      alphaf1 := log(asym/(2.0*null));
      alphaf2 := log((4.0*mut*null)/sqr(asym));
    end; (* if *)
  end

  else begin (* observations >= 2 *)
    if (observations = 2)
    then alphaf1 := alpha[1]
    else begin
      sum := 0.0;
      for k:=1 to observations-1 do sum := sum+prec[k];
      if (sum > 0.0)
      then begin
        alphaf1 := 0.0;
        for k:=1 to observations-1
        do alphaf1 := alphaf1 + (alpha[k]*prec[k])/sum;
      end
      else alphaf1 := 0.0;
    end; (* if observations >= 3 *)

    (*CS: a lot of logarithms added for the basic rate parameters' standard starting values;
          due to reparametrisation of the rate functions (april 2005): *)
    for i:=1 to observations-1
    do case ModelCode of
      1,2 : begin
              if symmetric (* each step reduces distance by 2 *)
              then alp[i] := log(0.2+ObservedNetworkDistance[1,i]/n)
              else alp[i] := log(0.2+2*ObservedNetworkDistance[1,i]/n);
              (* motivation: These are very rough values.
                 They must be positive and at least ObservedDistAnce/n. TS 01-jan-05 *)
              if (alp[i] < log(0.1)) then alpa_l[i] := log(0.1)
            end; (* case 1,2 *)
      3   : begin
              ntot :=n00[i] + n01[i] + n10[i] + n11[i];
              if (ntot >= 1)
              then alp[i] := log(ObservedNetworkDistance[1,i]/ntot)
              else alp[i] := (*1.0*)0.0;
              if (alp[i] < log(0.1)) then alp[i] := log(0.1)
            end; (* case 3 *)
      4   : begin
              ntot :=n00[i] + n01[i] + n10[i] + n11[i];
              if (ntot >= 1)
              then alp[i] := log(sqrt(0.01 + ObservedNetworkDistance[1,i]/ntot))
              else alp[i] := log(0.001);
              if (alp[i] < log(0.001)) then alp[i] := log(0.001);
            end; (* case 4 *)
    end; (* case *)
 }
  end; (* if observations >= 2 *)
  (* Now parameters for behavior change.
     TS 05-jan-05 *)
  setlength(n00,observations+1);
  setlength(n01,observations+1);
  setlength(n10,observations+1);
  setlength(n11,observations+1);

  for i:=1 to nza
  do begin
     if (round(range[nz+i]) = 2)
     then begin
     (* ML estimates under empty model, with some truncation *)
       p00 := 0.0; p01 := 0.0;
       p10 := 0.0; p11 := 0.0;
       for k:=1 to observations-1
       do begin
         n := nni[k];
         n11[k] := 1;
         n01[k] := 1;
         n10[k] := 1;
         n00[k] := 1;
        (* initialise at instead of 0:
           for regression to the mean and no problems with division by 0 *)
         for j:=1 to n do if (not (miszc(j,i,k) or miszc(j,i,k+1)))
         then begin
           if (zc[j,i,k] = zcmin[i])
           then if (zc[j,i,k+1] > zc[j,i,k])
           then inc(n01[k])
           else inc(n00[k]);
           if (zc[j,i,k] = zcmax[i])
           then if (zc[j,i,k+1] < zc[j,i,k])
           then inc(n10[k])
           else inc(n11[k]);
         end; (* for j *)
         p00 := p00 + n00[k];
         p01 := p01 + n01[k];
         p10 := p10 + n10[k];
         p11 := p11 + n11[k];
         sum := (n01[k]/(n00[k]+n01[k])) + (n10[k]/(n10[k]+n11[k]));
         if (sum > 0.9) then sum := 0.5;
         alp[NetworkEffects_l+(i-1)*ActionEffects_l+k] := (-log(1.0-sum));      // @@@ teruggezet TS
       end; (* for k *)
       aten := log((p01*(p10+p11))/(p10*(p01+p00)));
       if (aten < -2.0) then aten := -2.0;
       if (aten >  2.0) then aten :=  2.0;
     end (* range 2 *)

     else begin
     (* MoM estimates with some truncation for random walk without boundaries. *)
       dif := 0.0;
       dif2 := 0.0;
       for k:=1 to observations-1
       do begin
         n := nni[k];
         n00[k] := 0;
         n01[k] := 0;
         n11[k] := 0;
         for j:=1 to n
         do if (not (miszc(j,i,k) or miszc(j,i,k+1)))
         then begin
           n00[k] := n00[k] + round(zc[j,i,k+1]-zc[j,i,k]);
           n01[k] := n01[k] + abs(round(zc[j,i,k+1]-zc[j,i,k]));
           n11[k] := n11[k] + sqr(round(zc[j,i,k+1]-zc[j,i,k]));
         end; (* for j *)
         sum := (n/(n-1))*(n11[k]/n - sqr(n00[k]/n));
         if (sum < 0.1+(n01[k]/n)) then sum := 0.1+n01[k]/n;
         alp[NetworkEffects_l+(i-1)*ActionEffects_l+k] := (sum);      // teruggezet TS @@@
         dif := dif+n00[k];
         dif2 := dif2+n11[k];
       end; (* for k *)
       dif := dif/n;
       dif2 := (n/(n-1))*(dif2/n - sqr(dif));
       aten := dif/(dif2+1.0); (* the +1.0 is regression to the mean;
                                  this assignment is superseded if ... *)
       if (dif < 0.9*dif2)
       then aten := 0.5*log((dif+dif2)/(dif2-dif));
     end; (* range >= 3 *)
     if (abs(aten) > 3.0) then aten := 3.0*sign(aten);
     alpf[NetworkEffects_f + (i-1)*Actioneffects_f + 1] := aten;
     (* This should be the tendency effect of the i'th dependent behavioral variable *)
   end (* i *)

end; (* InitialAlphas *)


procedure DataSpecification.SetNumberOfPeriods(nn,hzc,hzz,hzzc,k: integer);
begin
  SetPeriods(nn,hzc,hzz,hzzc,k); (* see Eight *)
  SetLength(ObservedNetworkDistance,nxa+1,k+1);
  SetLength(ObservedActionDistance,nza+1,k+1);
  SetLength(ObservedNetworkData,nxa+1,k+1);
  SetLength(ObservedActionData,nza+1,k+1);
  SetLength(pzamean,nza+1,k+1);
  SetLength(ExistStruct,k+1);
  SetLength(ExistNewStruct,k+1);  // added 3.17s
  SetLength(outlier,nn+1);
  Setlength(nni,k+1);
  Setlength(skipi,k+1);
end;


procedure DataSpecification.FunctionsToBooleans(var FunctInd: BBarray);
var i: integer;
begin
 for i:= 1 to observations - 1 do FunctInd[i] := (not skipi[i]);
 for i:= observations to ccmax do FunctInd[i] := false;
 for i:= 1 to pmax do FunctInd[Functions[i]] := true;
end; (* FunctionsToBooleans *)


procedure DataSpecification.IndicateFixedThetaAndRC(var fix, thetarandom: pBVec);
 (*Indicates in array fix which parameters are fixed in estimation, and in array thetarandom
   which coefficients are regarded as random coefficients: random effects model.*)
var i,j,h,l,lf,lg: integer;
begin

(* initialise *)
  res := nil;
  re_position := nil;
  setlength(res,3,pmax+1); (*Links the fixed effects alpa-arrays with the corresponding actor-dependent random effects array.*)
  setlength(re_position,pmax+1); (*Links the fixed effects alpa-arrays with the corresponding actor-dependent random effects array.*)
  for i := 1 to pmax do
    begin
    fix[i] := false;
    test[i] := false;
    thetarandom[i] := false;
    end;
  h := 0;
  j := 0;

(* First networks *)
(* rate function *)
  if (observations >= 2) then
    begin
    if cconditional and (ConditioningNetworkVariable=1)
    then for i:=observations to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        fix[h] := fix_l[i];
        test[h] := test_l[i];
        end
      end
     else for i:=1 to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        fix[h] := fix_l[i];
        test[h] := test_l[i];
        end
      end;
    end; (* if observations >= 2 *)

  (* objective function *)

  lf := 0;
  for i:=1 to NetworkEffects_f do
    begin
    if Effects_f[i] then
      begin
      inc(h);
      inc(lf);
      fix[h] := fix_f[i];
      test[h] := test_f[i];
      thetarandom[h] := thetarandom_f[i];
      if thetarandom[h] then
        begin
        Inc(j);
        re_position[h] := j; (*MS: indicates the column of the random effects matrix.*)
        res[1,i] := j; (*MS: indicates the column of the random effects matrix.*)
        end;
      end;
    end;
  (* gratification function *)
  lg := 0;
  if (observations >= 2) then
    begin
    for i:=1 to NetworkEffects_g do (*Take into account that some (fixed, random) coefficients included in the <pname.mo> file are disabled: thus the actual number of effects could be smaller than <l>.*)
      begin
      if Effects_g[i] then
        begin
        inc(h);
        inc(lg);
        fix[h] := fix_g[i];
        test[h] := test_g[i];
        thetarandom[h] := thetarandom_g[i];
        if thetarandom[h] then
          begin
          Inc(j);
          re_position[h] := j; (*MS: indicates the column of the random effects matrix.*)
          res[2,i] := j; (*MS: indicates the column of the random effects matrix.*)
          end;
        end;
      end;
    end;

(* Then behavior *)
  (* rate functions *)
  if (observations >= 2) then
    begin
    for l := 1 to nza do
      begin
      if (Cconditional and (l=ConditioningActionVariable))
      then for i:= observations to ActionEffects_l do
        begin
        if Effects_l[NetworkEffects_l+(l-1)*ActionEffects_l+i] then
          begin
          inc(h);
          fix[h] := fix_l[NetworkEffects_l+(l-1)*ActionEffects_l+i];
          end
        end
       else for i:= 1 to ActionEffects_l do
        begin
        if Effects_l[NetworkEffects_l+(l-1)*ActionEffects_l+i] then
          begin
          inc(h);
          fix[h] := fix_l[NetworkEffects_l+(l-1)*ActionEffects_l+i];
          test[h] := test_l[NetworkEffects_l+(l-1)*ActionEffects_l+i];
          end
        end
      end
  end; (* if observations >= 2 *)

  (* objective functions *)
  for i:=NetworkEffects_f + 1 to MaxEffects_f do
    begin
    if Effects_f[i] then
      begin
      inc(h);
      inc(lf);
      fix[h] := fix_f[i];
      test[h] := test_f[i];
      thetarandom[h] := thetarandom_f[i];
      if thetarandom[h] then
        begin
        Inc(j);
        re_position[h] := j; (*MS: indicates the column of the random effects matrix.*)
        res[1,i] := j; (*MS: indicates the column of the random effects matrix.*)
        end;
      end;
    end;
  res[1,0] := lf; (*MS: number of random coefficients in objective function.*)
  (* gratification functions *)
  if (observations >= 2) then
    begin
    for i:= NetworkEffects_g + 1 to MaxEffects_g do (*Take into account that some (fixed, random) coefficients included in the <pname.mo> file are disabled: thus the actual number of effects could be smaller than <l>.*)
      begin
      if Effects_g[i] then
        begin
        inc(h);
        inc(lg);
        fix[h] := fix_g[i];
        test[h] := test_g[i];
        thetarandom[h] := thetarandom_g[i];
        if thetarandom[h] then
          begin
          Inc(j);
          re_position[h] := j; (*MS: indicates the column of the random effects matrix.*)
          res[2,i] := j; (*MS: indicates the column of the random effects matrix.*)
          end;
        end;
      end;
    end;
  res[2,0] := lg; (*MS: number of random coefficients in gratification function.*)
  nre := j; (*MS: total number of random coefficients.*)
  randomeffects := (nre > 0); (*MS: boolean indicator: random coefficients.*)

(* AA change *) if not (maxlike or bayesian) or (observations = 1) then 
    begin
    randomeffects := false;
    nre := 0;
    for i := 1 to h do thetarandom[i] := false;
    end;

end; (* IndicateFixedThetaAndRC *)


procedure DataSpecification.IndicateFixedAlpa(fix: pBVec);
 (* Changes arrays fix_l, fix_f, fix_g
    according to what is defined in array fix;
    fixing values for parameters not in the current model is unaffected. *)
var i,l,h: integer;
begin
  h:=0;
  if (observations >= 2)
  then begin
    if (cconditional and (ConditioningNetworkVariable=1))
    then for i:=observations to NetworkEffects_l
    do begin
      if Effects_l[i]
      then begin
        inc(h);
        fix_l[i] := fix[h]
      end
    end
    else for i:=1 to NetworkEffects_l
      do begin
      if Effects_l[i] (* this should always be the case for 1<=i<=(obs-1), unless skipi[i] *)
      then begin
        inc(h);
        fix_l[i] := fix[h]
      end
    end;

    for l := 1 to nza
    do begin
      if (Cconditional and (l=ConditioningActionVariable))
      then for i:= observations to ActionEffects_l
      do begin
        if Effects_l[NetworkEffects_l+(l-1)*ActionEffects_l+i]
        then begin
          inc(h);
          fix_l[NetworkEffects_l+(l-1)*ActionEffects_l+i]:=fix[h];
        end
      end
      else for i:= 1 to ActionEffects_l
      do begin
        if Effects_l[NetworkEffects_l+(l-1)*ActionEffects_l+i] (* this should always be the case for 1<=i<=(obs-1) *)
        then begin
          inc(h);
          fix_l[NetworkEffects_l+(l-1)*ActionEffects_l+i]:=fix[h];
        end
      end
    end
  end; (* if observations >= 2 *)

  for i:=1 to MaxEffects_f
  do if Effects_f[i]
  then begin
    inc(h);
    fix_f[i] := fix[h];
  end;

  if (observations >= 2)
  then for i:=1 to MaxEffects_g
  do if Effects_g[i]
  then begin
    inc(h);
    fix_g[i] := fix[h];
  end;
end; (* IndicateFixedAlpa *)



procedure DataSpecification.writestart(var wf: textfile; h: integer);
begin
 writeln(wf,'[SIENA ', SienaVersion, ' model definition file ',h,']');
(*<*)
 writeln(wf,observations,' ',n,' ',nz,' ',nzc,' ',nzz,' ',nzx,' ',nza,' [codes]');
(*>*)
 writeln(wf,'[All information between square brackets]');
 writeln(wf,'[is intended only for the human reader or for other programs]');
 writeln(wf,'[and is not used by SIENA.]');
end; (* writestart *)




procedure DataSpecification.WriteDefinitionModel(var lf : TFileStream);
  (* writes model definition to file .MO *)
  (* If this procedure is changed, then check if it is also necessary
     to change ChangeModelSpecification.LineChange in S_ChangeModel.  *)
var
  wf             : textfile;
  i,j,                       (* loop counters indicating variable number and effect number *)
  code, nee      : integer;
  fname          : string;
  Correspondence : SBYarray;
  Param          : SSIarray;
  labels         : SSarray;
  VertPar: BBarray;

begin

//  writeln(lf,'Write definition file .MO .');
  fname := pname+'.MO';
  assignfile(wf,fname);
  rewrite(wf);

  (* HEADER: *)
  writeln(wf,'[SIENA ',SienaVersion,' model specification file ',fname,']');
  writeln(wf,'[All information between square brackets]');
  writeln(wf,'[is intended only for the human reader or for other programs]');
  writeln(wf,'[and is not used by SIENA.]');
  writeln (wf);

  (* GENERAL INFORMATION ABOUT THE PROJECT (sizes of arrays): *)
  writeln(wf,'@1 [general information about SIENA project <',pname,'>:]');
  writeln(wf,observations,' [number of waves observed]');
  writeln(wf,nmax,' [number of actors]');
  writeln(wf,nxa,' [number of dependent network variables]');
  writeln(wf,nza,' [number of dependent behavioral variables]');
  writeln(wf,nz,' [number of constant actor covariates]');
  writeln(wf,nzc-nza,' [number of exogenous changing actor covariates]');
  writeln(wf,nzz,' [number of constant dyadic covariates]');
  writeln(wf,nzzc,' [number of exogenous changing dyadic covariates]');
  writeln(wf,nzx,' [flag indicating presence of file with composition change directives]');
  writeln(wf,'[Names of variables:]');
  for i := 1 to nz+nzc do writeln(wf,ZNames[i]);
  for i := 1 to nzz+nzzc do writeln(wf,ZZNames[i]);
  writeln(wf);

  (* DEPENDENT NETWORK VARIABLES: *)
  for i:= 1 to nxa
  do begin

    (* - rate function of dependent network variable #i *)
    writeln(wf,'@2.',i,'.1 [rate function effects for dependent network variable <#',i,'>:]');
    writeln(wf,NetworkEffects_l,' [number of these effects :]');
    writeln(wf,'[each effect is given in two rows]');
    writeln(wf,'[first row contains label of the effect,]');
    writeln(wf,'[second row contains flags for inclusion, fixing and testing,]');
    writeln(wf,'[the starting value and potential extra parameters for effect calculation.]');

    (* read effect labels into array labels_l: *)
    DefineModel_lnames(labels_l,Correspondence,Param);

    for j:=1 to NetworkEffects_l
    do begin
      writeln(wf,labels_l[j]);
{      if j<observations (* communicate non-exponents as basic rate parameters *)
      then writeln(wf,va(effects_l[j]),' ',va(fix_l[j]),' ',va(test_l[j]),' ',expo(alpa_l[j]):16:12,' 0')
      else} writeln(wf,va(effects_l[j]),' ',va(fix_l[j]),' ',va(test_l[j]),' ',alpa_l[j]:16:12,' 0')
    end;
    writeln(wf);

    (* - objective function of dependent network variable #i *)
    writeln(wf,'@2.',i,'.2 [objective function effects for dependent network variable <#',i,'>:]');
    writeln(wf,NetworkEffects_f,' [number of such effects :]');
    writeln(wf,'[first row contains label of the effect,]');
    writeln(wf,'[next three rows: functions in which effect can be included:]');
    writeln(wf,'[row 1: evaluation function, row 2: endowment function`, row 3: currently not used,]');
    writeln(wf,'[columns correspond to: (a) inclusion, (b) random effects, (c) fixing, (d) testing,]');
    writeln(wf,'[(e) starting value, (f) potential extra parameters for effect calculation.]');

    (* read labels into arrrays labels and labels_f: *)
    (* this is indispensable as soon as parameter values param_effect of effects were changed. *)
    Param := param_effect;
    Definemodel_fnames(labels,Correspondence,VertPar,Param); (* get names of labels *)
    for j:=1 to ccmax do labels_f[j] := labels[j];

    for j:=1 to NetworkEffects_f
    do begin
      writeln(wf,labels_f[j]);
      (* evaluation effect       *)
      writeln(wf,va(effects_f[j]),' ',va(thetarandom_f[j]),' ',va(fix_f[j]),' ',va(test_f[j]),' ',alpa_f[j]:16:12,' ',param_effect[j]);
      (* endowment effect 1    *)
      writeln(wf,va(effects_g[j]),' ',va(thetarandom_g[j]),' ',va(fix_g[j]),' ',va(test_g[j]),' ',alpa_g[j]:16:12,' 0');
      (* endowment effect 2 *)
      writeln(wf,va(false),' ',va(false),' ',va(false),' ',va(false),' ',0.0:16:12,' 0');
    end;
    writeln(wf);

  end; (* loop DEPENDENT NETWORK VARIABLES *)

  nee := NewNetworkEffects.NumberEffects;

  (* DEPENDENT BEHAVIOURAL VARIABLES: *)
  for i:= 1 to nza
  do begin

    (* - rate function of dependent behavioural variable #i *)
    writeln(wf,'@3.',i,'.1 [rate function effects for dependent behavioral variable <#',i,'>:]');
    writeln(wf,ActionEffects_l,' [number of these effects :]');
    writeln(wf,'[each effect is given in two rows]');
    writeln(wf,'[first row contains label of the effect,]');
    writeln(wf,'[second row contains flags for inclusion, fixing and testing,]');
    writeln(wf,'[the starting value and potential extra parameters for effect calculation.]');

    (* note that labels already were defined when treating the network rate function *)

    for j:=1 to ActionEffects_l
    do begin (* communicate non-exponents as basic rates *)
      writeln(wf,labels_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]);
{      if j<observations
      then writeln(wf,va(effects_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 va(fix_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 va(test_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 expo(alpa_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]):16:12,' 0')
      else} writeln(wf,va(effects_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 va(fix_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 va(test_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]),' ',
                 alpa_l[NetworkEffects_l+(i-1)*ActionEffects_l+j]:16:12,' 0');
    end;
    writeln(wf);

    (* - objective function of dependent behavioural variable #i *)
    writeln(wf,'@3.',i,'.2 [objective function effects for dependent behavioral variable <#',i,'>:]');
    writeln(wf,ActionEffects_f,' [number of such effects :]');
    writeln(wf,'[first row contains label of the effect,]');
    writeln(wf,'[next three rows: functions in which effect can be included:]');
    writeln(wf,'[row 1: evaluation function, row 2: endowment function, row 3: currently not used,]');
    writeln(wf,'[columns correspond to: (a) inclusion, (b) random effects, (c) fixing, (d) testing,]');
    writeln(wf,'[(e) starting value, (f) potential extra parameters for effect calculation.]');

    (* note that labels already were defined when treating the network objective function *)

    for j:=1 to ActionEffects_f do
      begin
      writeln(wf,labels_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]);
      (* evaluation effect       *)
      writeln(wf,va(effects_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]),' ',
                 va(thetarandom_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]),' ',
                 va(fix_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]),' ',
                 va(fix_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]),' ',
                 alpa_f[NetworkEffects_f+(i-1)*ActionEffects_f+j]:16:12,' ',
                 param_effect[nee+(i-1)*ActionEffects_f+j]);
      (* endowment effect     *)
      writeln(wf,va(effects_g[NetworkEffects_g+(i-1)*ActionEffects_g+j]),' ',
                      va(thetarandom_g[NetworkEffects_g+(i-1)*ActionEffects_g+j]),' ',
                      va(fix_g[NetworkEffects_g+(i-1)*ActionEffects_g+j]),' ',
                      va(fix_g[NetworkEffects_g+(i-1)*ActionEffects_g+j]),' ',
                      alpa_g[NetworkEffects_g+(i-1)*ActionEffects_g+j]:16:12,' 0');
      (* extra effect     *)
      writeln(wf,va(false),' ',va(false),' ',va(false),' ',va(false),' ',0.0:16:12,' 0');
      end;

    writeln(wf);

  end; (* loop DEPENDENT BEHAVIOURAL VARIABLES *)

  if (MaxLike or Bayesian) then Cconditional := false;

  (* ESTIMATION and MODEL TYPE OPTIONS: *)
  writeln(wf,'@4 [options : ]');

  (* estimation method *)
    code := 0;
    if Cconditional
    then
      begin
      if (ConditioningNetworkVariable = 1) then code := 1
       else
        begin
        if (ConditioningNetworkVariable >= 1)
        then code := 20 + ConditioningNetworkVariable;
        if (ConditioningActionVariable >= 1)
        then code := 20 + nxa + ConditioningActionVariable;
        end;
      end; (* if Cconditional *)
      if MaxLike then code := 2;
    if Bayesian then code := 3;
    if VertexPar then Inc(code,10);
    if (MaxDegree >= 2) and (MaxDegree < n) then code := code+100*MaxDegree;
    
  write(wf,' ',code:3,' [estimation method: 0 unconditional, 1 conditional');
  if (observations = 1)
   then writeln(wf,'; +10 for incid. vertex par.; > 100 for conditioning on maximal outdegree. ]')
  else (*666*) (*The following code, and possibly the entire unit, must be updated.*)
    begin
    if (nza <= 0) then writeln(wf,', 2 maximum likelihood, 3 Bayesian.]')
    else writeln(wf,' 2 maximum likelihood, 3 Bayesian; >= 21 conditional MoM, diverse conditioners.]');
    end;

  (*MS: further specifications for ML, Bayesian estimation and random effects models:*)
  {
  if initialFEestimate_input then write(wf,'   1')
  else write(wf,'   0');
  writeln(wf,' [initial estimate of parameter (estimation method 2 or 3): 0 no, 1 yes]');
  }
  writeln(wf,cdensity1:4,' [Bayesian: MCMC algorithms: 1 random walk M-H, 2 autoregressive M-H, 3 independence sampler]');
  writeln(wf,scalebeta:8:4,' [Bayesian: scale factor (proposal distribution): 0 calibrated during burn-in, other number: user input]');
  writeln(wf,cdensity2:4,' [actor-dependent random effects: MCMC algorithms: 0 random walk M-H, 2 autoregressive M-H, 3 independence sampler]');
  writeln(wf,scalere:8:4,' [actor-dependent random effects: scale factor (proposal distribution): 1 calibrated during burn-in, other number: user input]');

  (* Model Type *)
  if (observations = 1)
  then begin
   writeln(wf,Gibbscode:4,' [Codes for Gibbs/MH steps]')
  end
  else begin 
    i := abs(ModelType);
    if symmetric then begin
      if (OutputType = 2) then i := i+10;
     end else begin
      if ReciprocityModel then i := 6;
      if (StaySameType = 1) and (not ReciprocityModel) then i := 0;
      if (OutputType = 2) then i := i+10;
     end;
    if signed then i := -i;
    writeln(wf,i:4,' [type of network evolution model]');
   end;
  (* number of subphases in phase 2 *)
  writeln(wf,10*nextra + nsubphase:4,' [phase 2: number of subphases]');

  (* number of iterations in phase 3 *)
  writeln(wf,nrun3:4,' [phase 3: number of iterations]');

  (* observation=1 case: number of steps / n^2 per run *)
  writeln(wf,rstep0:5:1,
     ' [multiplication factor for number of iterations in 1-observation case and estimation method 2.]');

  (* initial value of gain parameter in estimation *)
  writeln(wf,firstgain:10:6,' [starting value of gain parameter (estimation methods 0, 1 or 2)]');

  (* starting value *)
  if InitialStandard
  then write(wf,'   1 ')
  else write(wf,'   0 ');
  writeln(wf,'[starting value of parameter: 0 current value, 1 standard]');

  (*MS*)
  (* period homogeneity test *)
  if (homogeneous>20)
  then write(wf,'   ',homogeneous-20)
  else write(wf,'   0'); (*Default: no test.*)
  write(wf,' [period homogeneity test: ');
  if  (observations <= 2) then writeln(wf,'disabled because there are less than 3 observations]')
  else writeln(wf,'0 no test, m>0 period number, m=',observations,' joint test]');

  (* actor homogeneity test *)
  if (homogeneous in [11..19])
  then write(wf,'   ',homogeneous-10)
  else write(wf,'   0'); (*Default: no test.*)
  writeln(wf,' [actor homogeneity test: 0 no test, 1 oudegree, 2 reciprocity]');
  if (homogeneous in [11..19])
  then
    begin
    for i := 1 to n do
      begin
      if outlier[i] then write(wf,' ',i);
      end;
    end
  else write(wf,'<empty>');
  writeln(wf,' [list of specified actors]');
  (*MS*)

  (* random seed for identical replication of runs *)
  writeln(wf,modelseed:4,' [random seed: 0 randomize, other value: user input]');

  (* Derivative estimation method for MoM: *)
  writeln(wf,'   ',d_estimator,' [derivative method: 0 finite differences method, 1 LR method]');

  (* only for observation=1 case: *)
  if VertexPar
  then begin
    writeln(wf,'[Method code >= 10: ',n,' + ',n,' incidental vertex parameters :]');
    for i:=1 to n do
    begin
      write(wf,vertex_a[i]:7:3,' ');
      if (i mod 10 = 0) then writeln(wf);
    end;
    if (n mod 10 <> 0) then writeln(wf);
    for i:=1 to n do
    begin
      write(wf,vertex_b[i]:7:3,' ');
      if (i mod 10 = 0) then writeln(wf);
    end;
    if (n mod 10 <> 0) then writeln(wf);
  end; (* if VertexPar *)

  closefile(wf);

end (* DataSpecification.WriteDefinitionModel *) ;


procedure DataSpecification.WriteDefinitionStats(var lf : TFileStream);
  (* writes definition of statistics simulation to file .SI *)
var
  wf      : textfile;
  i,j     : integer;
  fname   : string;
  FuncInd : BBarray;
begin
  fname := pname+'.SI';
  assignfile(wf,fname);
  rewrite(wf);

  (* HEADER: *)
  writeln(wf,'[SIENA ',SienaVersion,' simulation specification file ',fname,']');
  writeln(wf,'[All information between square brackets]');
  writeln(wf,'[is intended only for the human reader or for other programs]');
  writeln(wf,'[and is not used by SIENA.]');
  writeln (wf);

  writeln(wf,'@1 [statistics that can be generated]');
  writeln(wf,MaxFunctions,' [number of these statistics :]');
  writeln(wf,'[each statistic is given in two rows]');
  writeln(wf,'[first row contains label of the statistic,]');
  writeln(wf,'[second row contains flag for inclusion]');

  (* STATISTICS: *)

  DefineFunctionNames(lf);  (* This procedure must be in this unit because of SIENA04. TS *)
  FunctionsToBooleans(FuncInd);

  (* always include network rate intercept and density and next statistic: *)
  for i:=1 to observations-1 do FuncInd[i] := (not skipi[i]);
  FuncInd[Networkeffects_l+1] := true;
  FuncInd[Networkeffects_l+2] := true;
  (* always include behaviour rate intercept statistic: *)
  for i:= 1 to nza do
    for j:= observations-1 downto 1 do
      FuncInd[NetworkFunctions+(i-1)*(observations-1)+j] :=(not skipi[j]);
  (* always include behaviour intercept statistic: *)
  for i:= 1 to nza do
    FuncInd[NetworkFunctions+nza*actioneffects_l+(i-1)*actioneffects_f+1] := true;
  for i:=1 to MaxFunctions
  do begin
    writeln(wf,FDescriptions[i]);
    writeln(wf,va(FuncInd[i]));
  end;
  writeln(wf);

  (* SIMULATION OPTIONS: *)
  writeln(wf,'@2 [options : ]');
  (* number of simulations for straight simulation *)
  writeln(wf,niter, '   [Default number of simulations for straight simulation]');

  closefile(wf);
end; (* DataSpecification.WriteDefinitionStats *)

procedure DataSpecification.SienaToPajek(var lf: TFileStream);
(* This should be speeded up by using edge lists. *)
type intmat = array of array of integer;
var i,j,k,number,selected: integer;
    ms: intmat;
    nip: intvec;
    select: array of Boolean;
    x: string;
procedure MatrixSum(var matsum: Intmat);
var i,j,h,s: integer;
  datf: text;
  datfn: shortstring;
begin
  for i:=1 to n do for j:=1 to n do begin
    if (i = j) then matsum[i,i] := 0 else begin
      s := 0;
      for h:= 1 to observations do if ynm(i,j,h) then inc(s);
      matsum[i,j] := s;
     end;
   end;
  datfn :=  pname+'.sum';
  assign(datf,datfn);
  rewrite(datf);
  for i:=1 to n do begin
    for j:=1 to n do
      if (observations <= 9) then write(datf,' ',matsum[i,j]:1)
       else  write(datf,' ',matsum[i,j]:2);
    writeln(datf);
   end;
  closefile(datf);
end; (* MatrixSum *)
procedure WritePajekNet(k: integer);
var fname: shortstring;
  swf: textfile;
  i,j: integer;
  x: string;
begin (* WritePajekNet *)
  if (k < 10) then fname := pname +'0'+st(k)+'.net'
   else fname := pname +st(k)+'.net';
  assignfile(swf,fname);
  x := 'File '+fname+' .'+eol;
  lf.Write(pchar(x)^,length(x));
  rewrite(swf);
  writeln(swf,'*Vertices   ',selected:3);
  for i:=1 to selected do begin
    write(swf,i:3);
    if (i <= 9) then write(swf,' ');
    if (i <= 99) then write(swf,' ');
    write(swf,' "',nip[i],'" ');
    for j:=1 to nz do write(swf,z[nip[i],j]:8:4,' ');
    if (k >= 1) then begin
      if (k < observations) then
        for j:=1 to nzc do write(swf,zc[nip[i],j,k]:8:4,' ')
       else
        for j:=1 to nzc do write(swf,zc[nip[i],j,k-1]:8:4,' ');
     end;
    writeln(swf);
   end; (* for i *)
   if symmetric then writeln(swf,'*Edges ') else writeln(swf,'*Arcs ');
  for i:=1 to selected do
    for j:=1 to selected do if ((ms[nip[i],nip[j]] > 0) and ((not (symmetric)) or (i<j)))
      then writeln(swf,i:3,' ',j:3,' ',ms[nip[i],nip[j]]:3);
  writeln(swf);
  closefile(swf);
end (* WritePajekNet *);
begin (* SienaToPajek *);
  SetLength(ms,n+1,n+1);
  SetLength(nip,n+1);
  SetLength(select,n+1);
  for i:=1 to n do select[i] := true;
  number := 0;
  for i:=1 to n do
   if select[i] then begin
     inc(number);
     nip[number] := i;
    end;
  selected := number;
  lf.write(pchar(string('Write '))^,6);
  if (selected < n) then lf.write(pchar(string('selected '))^,9);
  x := format('%s%d%s',['data to ',observations,' Pajek .net file']);
  lf.Write(pchar(x)^,length(x));
  if (observations = 1) then lf.write(pchar(':'+eol)^,1+length(eol))
  else lf.Write(pchar('s:'+eol)^,2+length(eol));
  for k:=1 to observations do begin
    for i:=1 to n do for j:=1 to n do
      if ynm(i,j,k) then ms[i,j] := 1 else ms[i,j] := 0;
    WritePajekNet(k);
   end; (* k *)
  if (observations >= 2) then begin
    Matrixsum(ms);
    x := 'Write summed data to Pajek .net file.'+eol;
    lf.Write(pchar(x)^,length(x));
    WritePajekNet(0);
   end; (* if *)
  //writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
end; (* SienaToPajek *)


procedure DataSpecification.DefineModel_lNames
          (var modelnames: SSarray; var Correspondence: SBYarray;
           var param: SSIarray);
(* in this procedure, for each rate function effect...
   - the label of the effects is defined (array 'modelnames'),
   - the number of the corresponding network statistic
     under MoM estimation is identified (array 'correspondence'). *)

var i,j,count: integer;

begin

  (* initialise arrays *)
  for i:=1 to ccmax do
    begin
    modelnames[i] := '  ( not implemented l )';
    Correspondence[i] := 0;
    param[i] := 0;
    end;

  (* Basic rate parameters for the NETWORK RATE FUNCTION *)
  case observations of
    1 : exit;  (* no rate functions for one observation only *)
    2 : begin  (* one rate function intercept for the case of two observations *)
          modelnames[1] := 'basic network rate parameter';
          Correspondence[1] := 1
        end;
    else begin (* more than one rate function intercept for three or more observations *)
      for i := 1 to observations-1
      do begin
        modelnames[i] := 'constant network rate (period '+st(i)+')';
      end
    end (* else *)
  end; (* case *)
  count := observations-1;

  (* effects of network and covariates on network rate (and for type 2 also on the outdegree component) *)
  if symmetric then begin
  (* reproduced from modeltype 1 below, with some changes *)
          modelnames[count+1] := 'degree effect on rate   ';
          modelnames[count+2] := 'indegree effect on rate';
          modelnames[count+3] := 'reciprocity effect on rate';
          modelnames[count+4] := 'effect 1/degree on rate   ';
          count := count+4;
          for i:=1 to nz+nzc do
          begin
            modelnames[count+i] := 'effect '+ZNames[i]+' on rate';
          end;
          count := count + nz+nzc;
   end else (* not symmetric *)
  case ModelType of
    1 : begin
          modelnames[count+1] := 'outdegree effect on network rate';
          modelnames[count+2] := 'indegree effect on network rate';
          modelnames[count+3] := 'reciprocity effect on network rate';
          modelnames[count+4] := 'effect 1/outdegree on network rate';
          count := count+4;
          for i:=1 to nz+nzc do
          begin
            modelnames[count+i] := 'effect '+ZNames[i]+' on rate';
          end;
        end;
    2 : begin
          modelnames[observations  ] := 'outdegree effect            (xi)';
          modelnames[observations+1] := 'logarithmic outdegree effect (xi)';
          modelnames[observations+2] := 'factorial outdegree effect   (xi)';
          modelnames[observations+3] := 'effect 1/outdegree on rate   (nu)';
          for i:=1 to nz+nzc do
          begin
            modelnames[observations+3+i] := 'effect '+ZNames[i]+' on o.d. (xi)';
          end;
          for i:=1 to nz+nzc do
          begin
            modelnames[observations+3+nz+nzc+i] := 'effect '+ZNames[i]+' on rate (nu)';
          end
        end
  end; (* case ModelType *)

  for i:=1 to NetworkEffects_l do Correspondence[i] := i;

  (* BEHAVIOUR RATE FUNCTION(s) *)
  for j:=1 to nza do
    begin
      (* Basic rate parameter(s): *)
      for i:=1 to (observations-1) do
      begin
        modelnames[NetworkEffects_l+(j-1)*ActionEffects_l+i]
          := 'rate '+ZNames[nz+j]+' period '+st(i);
      end;
      (* Network effects on rate *)
      modelnames[NetworkEffects_l+(j-1)*ActionEffects_l+observations]
        := 'outdegree effect on rate '+Znames[nz+j];
      modelnames[NetworkEffects_l+(j-1)*ActionEffects_l+observations+1]
        := 'indegree effect on rate '+Znames[nz+j];
      modelnames[NetworkEffects_l+(j-1)*ActionEffects_l+observations+2]
        := 'reciprocated degree effect on rate '+Znames[nz+j];
      (* Covariate effects on rate *)
      for i:=1 to (nz+nzc) do
      begin
        modelnames[NetworkEffects_l+(j-1)*ActionEffects_l+observations+2+i]
          := 'effect '+ZNames[i]+' on rate '+ZNames[nz+j];
      end;
    end;
  for i:=1 to nza*ActionEffects_l do
    Correspondence[NetworkEffects_l + i] := NetworkFunctions + i;

  (* If this procedure is changed, then also check NetworkEffects_l and/or ActionEffects_l
     and/or MaxEffects_l in S_Data *)
end; (* DefineModel_lNames *)

procedure DataSpecification.DefineModel_fNames
          (var modelnames: SSarray; var Correspondence: SBYarray;
           var VertPar: BBarray; var param: SSIArray);
(* in this procedure, for each evaluation function effect...
   - the label of the effects is defined (array 'modelnames'),
   - the number of the corresponding network statistic
     under MoM estimation is identified (array 'correspondence'),
   - Param is the value of the default parameters,
   - Vertpar indicates for in the Vertexpar (observations = 1) case,
     which parameters must be excluded because they would lead to overparametrisation.
     (These are effects that depend on the degrees only, and therefore
      are not estimable if incidental vertex parameters are included,
      or if Gibbs sampling conditional on the degrees is done.)            *)

var i,j,h,count: integer;

begin
  (* initialise arrays *)
  for i:=1 to ccmax do begin
    modelnames[i] := '  ( not implemented f )';
    Correspondence[i] := 0;
    param[i] := 0;
    VertPar[i] := false;
  end;

 for i:=1 to NetworkEffects_f do
   modelnames[i] := NewNetworkEffects.EffectName(i);

 for i:=Networkeffects_f downto 1 do Correspondence[i] := NetworkEffects_l+i;
 count := NewNetworkEffects.NumberEffects;

{ if signed then begin
   modelnames[1] := 'outdegree positive ';
   modelnames[2] := 'outdegree negative ';
   modelnames[3] := 'reciprocity P/P';
   modelnames[4] := 'reciprocity 0/0';
   modelnames[5] := 'reciprocity N/N';
   modelnames[6] := 'balance PPP';
   modelnames[7] := 'balance PNN';
   modelnames[8] := 'balance NPN';
   modelnames[9] := 'balance NNP';
   modelnames[10] := 'anti-balance PPN';
   modelnames[11] := 'anti-balance PNP';
   modelnames[12] := 'anti-balance NPP';
   modelnames[13] := 'anti-balance NNN';
   for i:=1 to nzz do begin
     modelnames[13 + (i-1)*5 + 1] := ZZnames[i] + ' pos. ties';
     modelnames[13 + (i-1)*5 + 2] := ZZnames[i] + ' neg. ties';
     modelnames[13 + (i-1)*5 + 3] := ZZnames[i] + ' P/P dyads';
     modelnames[13 + (i-1)*5 + 4] := ZZnames[i] + ' 0/0 dyads';
     modelnames[13 + (i-1)*5 + 5] := ZZnames[i] + ' N/N dyads';
    end; (* for i *)
   count := 13+5*nzz;
   for i:=1 to (nz+nzc) do begin
     modelnames[count + (i-1)*6 + 1] := ZNames[i]+' pos. ties ego';
     modelnames[count + (i-1)*6 + 2] := ZNames[i]+' pos. ties alter';
     modelnames[count + (i-1)*6 + 3] := ZNames[i]+' pos. ties similarity';
     modelnames[count + (i-1)*6 + 4] := ZNames[i]+' neg. ties ego';
     modelnames[count + (i-1)*6 + 5] := ZNames[i]+' neg. ties alter';
     modelnames[count + (i-1)*6 + 6] := ZNames[i]+' neg. ties similarity';
    end; (* for i *)
     param[24] := 7;

   for i:=Networkeffects_f downto 1 do Correspondence[i] := observations+i-1;
 end (* signed *) else begin (* not signed *)
}


  for j:=1 to nza do (* now the behavioural objective function effects *)
    begin
    (* next: effects of network on behaviour objective function(s) *)
    modelnames[count + 1] := 'behavior '+ZNames[nz+j]+' linear shape (tendency)';
    modelnames[count + 2] := 'behavior '+ZNames[nz+j]+' average similarity';
    modelnames[count + 3] := 'behavior '+ZNames[nz+j]+' total similarity';
    modelnames[count + 4] := 'behavior '+ZNames[nz+j]+' indegree ';
    modelnames[count + 5] := 'behavior '+ZNames[nz+j]+' outdegree ';
    modelnames[count + 6] := 'behavior '+ZNames[nz+j]+' isolate ';
    modelnames[count + 7] := 'behavior '+ZNames[nz+j]+' ave. sim.  reciprocity ';
    modelnames[count + 8] := 'behavior '+ZNames[nz+j]+' tot. sim.  reciprocity ';
    modelnames[count + 9] := 'behavior '+ZNames[nz+j]+' ave. sim.  popularity alter ';
    case param_effect[count+10] of
      0: modelnames[count + 10] := 'behavior '+ZNames[nz+j]+' tot. sim.  popularity alter ';
      1: modelnames[count + 10] := 'behavior '+ZNames[nz+j]+' popularity alter ';
     else
         modelnames[count + 10] := 'behavior '+ZNames[nz+j]+' tot. sim.  popularity alter ';
     end; (* case *)
    modelnames[count + 11] := 'behavior '+ZNames[nz+j]+' ave. sim.  rec.  pop.(alter) ';
    modelnames[count + 12] := 'behavior '+ZNames[nz+j]+' tot. sim.  rec.  pop.(alter) ';
    modelnames[count + 13] := 'behavior '+ZNames[nz+j]+' average alter';
    modelnames[count + 14] := 'behavior '+ZNames[nz+j]+' average rec. alters';
    modelnames[count + 15] := 'behavior '+ZNames[nz+j]+' quadr. shape (effect from '+ZNames[nz+j]+')';
    modelnames[count + 16] := 'behavior '+ZNames[nz+j]+' dense triads <maybe wrong>';
    modelnames[count + 17] := 'behavior '+ZNames[nz+j]+' similarity in dense triads <maybe wrong>';
    modelnames[count + 18] := 'behavior '+ZNames[nz+j]+' reciprocated degree';
    modelnames[count + 19] := 'behavior '+ZNames[nz+j]+' ave. sim.  popularity ego ';

    (* update number of already-labelled effects *)
    count := count + 19;

    (* next: effects of actor characteristics on behaviour objective function(s)
       -- per dependent actor variable, there is one effect for each constant covariate
          and one effect for each of the (remaining) changing covariates *)
    (* effects of the constant covariates and the preceding changing covariates*)
    for h:=1 to nz+j-1 do
      begin
      modelnames[count+(2*h)-1] := 'behavior '+Znames[nz+j]+': effect from '+Znames[h];
      case param_effect[count+(2*h)] of
        0: modelnames[count+(2*h)] := 'behavior '+Znames[nz+j]+': influence interaction?'
                                        +'  '+Znames[h];
        1: modelnames[count+(2*h)] := 'behavior '+Znames[nz+j]+': av. sim.  '+Znames[h];
        2: modelnames[count+(2*h)] := 'behavior '+Znames[nz+j]+': tot. sim  '+Znames[h];
        3: modelnames[count+(2*h)] := 'behavior '+Znames[nz+j]+': av. alters  '+Znames[h];
        end; (* case *)
      end;
    (* effects of the subsequent changing covariates *)
    for h:=nz+j+1 to nz+nzc do
      begin
      modelnames[count+(2*h)-3] := 'behavior '+Znames[nz+j]+': effect from '+Znames[h];
      case param_effect[count+(2*h)-2] of
        0: modelnames[count+(2*h)-2] := 'behavior '+Znames[nz+j]+': influence interaction?'
                                        +'  '+Znames[h];
        1: modelnames[count+(2*h)-2] := 'behavior '+Znames[nz+j]+': av. sim.  '+Znames[h];
        2: modelnames[count+(2*h)-2] := 'behavior '+Znames[nz+j]+': tot. sim  '+Znames[h];
        3: modelnames[count+(2*h)-2] := 'behavior '+Znames[nz+j]+': av. alters  '+Znames[h];
        end; (* case *)
      end; (* for h *)

(* Added 27/10/8: *)
      case param_effect[count+(2*(nz+nzc))-1] of
        0: modelnames[count+(2*(nz+nzc))-1] := 'behavior '+Znames[nz+j]+': infl. interaction?'
                                        +'  '+Znames[nz+j]+' alter';
        1: modelnames[count+(2*(nz+nzc))-1] := 'behavior '+Znames[nz+j]+': ave. sim. to higher alters';
        2: modelnames[count+(2*(nz+nzc))-1] := 'behavior '+Znames[nz+j]+': tot. sim. to higher alters';
        3: modelnames[count+(2*(nz+nzc))-1] := 'behavior '+Znames[nz+j]+': ave. higher alters';
        end; (* case *)
      count := count+1;
(* end Added 27/10/8 *)

    for h:= 1 to ninta do
      begin
        case ni[j,h] of
        2: modelnames[count+2*(nz+nzc-1)+h] :=
                   'behavior '+Znames[nz+j]+
                   ': effect '+Znames[pv1[j,h]]
                                     +'  '+Znames[pv2[j,h]];
        3: modelnames[count+2*(nz+nzc-1)+h] :=
                   'behavior '+Znames[nz+j]+
                   ': i3 '+Alignstring(Znames[pv1[j,h]],5)
                    +'  '+Alignstring(Znames[pv2[j,h]],5)+'  '+Alignstring(Znames[pv3[j,h]],5);
           else modelnames[count+2*(nz+nzc-1)+h] :=
                   'behavior '+Znames[nz+j]+
                   ': unspecified interaction';
        end; (* case *)
      end; (* for h *)

   (* update current number of already-labelled statistics *)
    count := count + 2*(nz + nzc - 1) + ninta;
    end; (* for j:=1 to nza *)


  for i:=nza*Actioneffects_f downto 1 do
    Correspondence[NetworkEffects_f + i] :=
      NetworkFunctions + (nza*ActionEffects_l) + i;

(* If this procedure is changed, check NetworkEffects_f and ActionEffects_f  *)
end; (* DefineModel_fNames *)


procedure DataSpecification.DefineModel_gNames
          (var modelnames: SSarray; var Correspondence: SBYarray;
           var param: SSIarray);
(* in this procedure, the labelling of gratification functions' effects is defined, and
   for each such effect, the number of the corresponding network statistic used
   for estimating it in the Robbins Monro estimation procedure is identified *)

var i,j,count: integer;
   fmodelnames: SSarray;
   dumCorrespondence: SBYarray;
   dumVertPar: BBarray;
begin
  for i:=1 to ccmax do
    begin
    modelnames[i] := '  ( not implemented g )';
    Correspondence[i] := 0;
    param[i] := 0
    end;

  (* gratification function only for the case of more than one observation,
     and not yet for signed graphs *)
  if ((observations >= 2) and (not signed))    then
    begin
    (* First the network gratification function effects that model
       asymmetry of tie value between tie creation and tie dissolution: *)
    for i:=1 to NetworkEffects_g do
      if (NewNetworkEffects.Endowed(i) or (MaxLike or Bayesian)) then
        modelnames[i] := NewNetworkEffects.EffectName(i);

    for i:=Networkeffects_g downto 1 do
      Correspondence[i] := NetworkEffects_l + NetworkEffects_f + i;

    count := NetworkEffects_g;

 (* now the behavioural endowment function effects *)
    DefineModel_fNames(fmodelnames,dumCorrespondence,dumVertPar,param);
    for i:= nza*ActionEffects_f downto 1 do
      modelnames[count+i] :=  AnsiReplaceStr(fmodelnames[count+i],'behavior','dec. beh.');

    count := count + nz + nzc - 1;

  (* For the case TYPE=2, outdegree-effects on behaviour gratification function might need
   to be erased (as in objective function case above). This still needs to be checked!! *)

  (* A TEST OF THE NUMBER OF EFFECTS SHOULD BE INCLUDED TO CHECK
   WHETHER ACTIONEFFECTS-G IS REACHED *)

    end; (* observations>=2 *)

  for i:=nza*Actioneffects_g downto 1 do
      Correspondence[NetworkEffects_g + i]
        := NetworkFunctions + nza*(ActionEffects_l + ActionEffects_f) + i ;

  (* If this is changed, then also change NetworkEffects_g and ActionEffects_g  *)
end; (* DefineModel_gNames *)


procedure DataSpecification.ReadBehavior(const dim: byte);
(*MS: used when simulating data: read only first behavioral outcome variable.*)
var
  i: integer;
  h: byte;
  u: real;
  s: shortstring;
  f: textfile;
begin
s := CurrentProjectName + '.dac'; (*File with input.*)
AssignFile(f,s);
if FileExists(s) then reset(f);
for i := 1 to n do
  begin
  for h := 1 to dim do (*Read the line up to element <dim>, and store the <dim>-th value.*)
    begin
    read(f,u);
    if (h = dim) then (*Store the observed outcome.*)
      begin
      zc[i,1,dim] := round(u); (*Behavioral outcome variable at beginning of the period: <i> actor, <1> outcome variable, <dim> period.*)
      setya(i,1,round(u)); (*Behavioral outcome variable at present: <i> actor, <1> outcome variable, <g> value.*)
      end;
    end;
  readln(f)
  end;
closefile(f);
end; (*MS: ReadBehavior.*)


procedure DataSpecification.DefineFunctionNames(var outf: TFileStream);
var count,parcount,i,h,para: integer;
    ModelCode: integer;
begin
  (* initialize names *)
  for i:=0 to ccmax do
    FDescriptions[i] := ' (statistic not implemented)              ';

  (* special case: observations=1 *)
  if (observations = 1) then
    begin
    for i:=1 to NetworkEffects_f do
      FDescriptions[i] := NewNetworkEffects.FunctionName(i);
    count := NetworkEffects_f;

    end (* if observations = 1 *)

  (* now the case where observations >=2 *)
   else
    begin (* observations >= 2 *)
    ModelCode := ModelType;
    if symmetric then ModelCode := 1;
(* NETWORK RATE FUNCTION *)
    (* first statistics concern network rate function intercepts of all periods *)
    if (observations = 2)
      then FDescriptions[1] := 'Amount of network change'
      else for i:=1 to (observations-1)
      do FDescriptions[i] := 'Amount of network change in period '+st(i);
    (* remember current number of the statistic for ease of further labelling *)
    count := observations - 1;
    FDescriptions[count+1] := 'Amount of network change  outdegrees';
    FDescriptions[count+2] := 'Amount of network change  indegrees';
    FDescriptions[count+3] := 'Amount of network change  reciprocity';
    FDescriptions[count+4] := 'Amount of network change  (1/outdegrees)';
    (* update current number of already-labelled statistics *)
    count := count+4;
    (* next statistics contain network rate function effects based on actor covariates
       -- for each covariate one effect *)
    for i:=1 to (nz+nzc) do
      FDescriptions[count+i] := 'Amount of network change  '+Znames[i];
    (* update current number of already-labelled statistics *)

(* NETWORK OBJECTIVE FUNCTION  *)
    for i:=1 to NetworkEffects_f do
      FDescriptions[NetworkEffects_l + i] := NewNetworkEffects.FunctionName(i);

    if signed then
      begin

{
    (* NETWORK OBJECTIVE FUNCTION, 1st part *)
    (* next statistics contain network objective function effects based on the network alone
       -- currently, there are 13 such statistics in the type-(-1) model *)

    FDescriptions[count+1] := 'Number of positive ties';
    FDescriptions[count+2] := 'Number of negative ties';
    FDescriptions[count+3]  := 'Number of P/P dyads';
    FDescriptions[count+4]  := 'Number of 0/0 dyads';
    FDescriptions[count+5]  := 'Number of N/N dyads';
    FDescriptions[count+6]  := 'Number of PPP triples';
    FDescriptions[count+7]  := 'Number of PNN triples';
    FDescriptions[count+8]  := 'Number of NPN triples';
    FDescriptions[count+9]  := 'Number of NNP triples';
    FDescriptions[count+10] := 'Number of PPN triples';
    FDescriptions[count+11] := 'Number of PNP triples';
    FDescriptions[count+12] := 'Number of NPP triples';
    FDescriptions[count+13] := 'Number of NNN triples';

    count := count + 13;
    (* cumulative count: observations-1 + 13 *)

    (* next statistics contain network objective function effects based on dyadic covariates
       -- for each covariate 5 such effects *)
    for i:=1 to nzz do
    begin
      FDescriptions[count+5*(i-1)+1]:= 'Sum of pos. ties  '+ZZNames[i];
      FDescriptions[count+5*(i-1)+2]:= 'Sum of neg. ties  '+ZZNames[i];
      FDescriptions[count+5*(i-1)+3]:= 'Sum of P/P dyads  '+ZZNames[i];
      FDescriptions[count+5*(i-1)+4]:= 'Sum of 0/0 dyads  '+ZZNames[i];
      FDescriptions[count+5*(i-1)+5]:= 'Sum of N/N dyads  '+ZZNames[i];
    end;
    (* update current number of already-labelled statistics *)
    count := count+5*nzz;
    (* cumulative count: observations-1 + 13 +5*nzz *)

    (* next statistics contain network objective function effects based on actor covariates
       -- for each covariate 6 effects *)
    for i:=1 to (nz+nzc) do begin
      FDescriptions[count+6*(i-1)+1]:= 'Sum of pos. outdegrees  '+ZNames[i];
      FDescriptions[count+6*(i-1)+2]:= 'Sum of pos. indegrees   '+ZNames[i];
      FDescriptions[count+6*(i-1)+3]:= 'Sum of pos. ties  '+ZNames[i]+' similarity';
      FDescriptions[count+6*(i-1)+4]:= 'Sum of neg. outdegrees  '+ZNames[i];
      FDescriptions[count+6*(i-1)+5]:= 'Sum of neg. indegrees   '+ZNames[i];
      FDescriptions[count+6*(i-1)+6]:= 'Sum of neg. ties  '+ZNames[i]+' similarity';
     end; (* for i; actor covariates *)
    count := count + 6*(nz+nzc);
    (* cumulative count: observations-1 + 13 + 5*nzz + 6*(nz+nzc) *)
(* NETWORK RATE FUNCTION, 2nd part *)
    (* next statistics contain network rate function effects based on the network alone
       -- currently, there are 4 such effects *)
    FDescriptions[count+1] := 'Amount of network change  outdegrees';
    FDescriptions[count+2] := 'Amount of network change  indegrees';
    FDescriptions[count+3] := 'Amount of network change  reciprocity';
    FDescriptions[count+4] := 'Amount of network change  (1/(1+outdegrees))';
    (* update current number of already-labelled statistics *)
    count := count+4;
    (* cumulative count: observations-1 + 17 + 5*nzz + 6*(nz+nzc) *)

    (* next statistics contain network rate function effects based on actor covariates
       -- for each covariate one effect *)
    for i:=1 to (nz+nzc) do
      FDescriptions[count+i] := 'Amount of network change  '+Znames[i];
    (* update current number of already-labelled statistics *)
    count := count + nz + nzc;
    (* cumulative count: observations-1 + 13 + 5*nzz + 7*(nz+nzc) *)
    (* this should be the same as NetworkFunctions *)
(* tot hier *)
}
      end
     else (* not signed *)
      begin
      for i:=1 to NetworkEffects_f do
       if (NewNetworkEffects.Endowed(i) or (MaxLike or Bayesian)) then
        FDescriptions[NetworkEffects_l + NetworkEffects_f + i]
          := 'Lost ties: '+NewNetworkEffects.FunctionName(i);

(* BEHAVIOUR RATE FUNCTION *)
      count := NetworkFunctions;
    (* next statistics contain behaviour rate function intercepts
       -- per dependent actor variable one effect for each period *)
      for i:=1 to nza do
        begin
        if (observations = 2) then Fdescriptions[count+1] :=
          'Amount of behavioral change on '+ZNames[nz+i]
          else
        for h:=1 to (observations-1) do
          Fdescriptions[count+h] :=
          'Amount of behavioral change in period '+st(h)+' on '+ZNames[nz+i];
    (* update current number of already-labelled statistics *)
        count := count + (observations-1);
        Fdescriptions[count+1] := 'Amount of change on '+Znames[nz+i]+'  outdegree';
        Fdescriptions[count+2] := 'Amount of change on '+Znames[nz+i]+'  indegree';
        Fdescriptions[count+3] := 'Amount of change on '+Znames[nz+i]+'  reciprocity';
        for h:=1 to (nz+nzc)
          do Fdescriptions[count+3+h] := 'Amount of change on '+Znames[nz+i]+'  '+Znames[h];
        count := count + 3 + nz + nzc;
        end; (* for i: = 1 to nza *)

(* BEHAVIOUR OBJECTIVE FUNCTION *)
    (* next statistics contain behaviour objective function effects based on the network alone
       -- per dependent actor variable, there currently is 1 main effect and 11 network effects. *)

    count := nxa*NetworkFunctions + nza*ActionEffects_l;

    for i:=1 to nza do
      begin
      parcount := nxa*NetworkEffects_f + (i-1)*ActionEffects_f;
      FDescriptions[count + 1] := 'beh. '+Znames[nz+i];
      FDescriptions[count + 2] := 'beh. '+Znames[nz+i]+' average similarity';
      FDescriptions[count + 3] := 'beh. '+Znames[nz+i]+' total similarity';
      if symmetric
        then FDescriptions[count + 4] := 'beh. '+Znames[nz+i]+' degrees'
        else FDescriptions[count + 4] := 'beh. '+Znames[nz+i]+' indegrees';
      FDescriptions[count + 5] := 'beh. '+Znames[nz+i]+' outdegrees';
      FDescriptions[count + 6] := 'beh. '+Znames[nz+i]+' isolate';
      FDescriptions[count + 7] := 'beh. '+Znames[nz+i]+' ave. similarity  reciprocity';
      FDescriptions[count + 8] := 'beh. '+Znames[nz+i]+' tot. similarity  reciprocity';
      if symmetric
        then FDescriptions[count + 9] := 'beh. '+Znames[nz+i]+' ave. sim.  degrees'
        else FDescriptions[count + 9] := 'beh. '+Znames[nz+i]+' ave. sim.  indegrees(one-sided)';
      if symmetric
        then
        case param_effect[parcount+10] of
         0: FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+' tot. sim.  degrees';
         1: FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+'  av. degree alters';
         else FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+' tot. sim.  degrees';
        end (* case *)
        else
        case param_effect[parcount+10] of
         0: FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+' tot. sim.  indegrees (one-sided)';
         1: FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+'  av. indegrees alters';
         else FDescriptions[count + 10] := 'beh. '+Znames[nz+i]+' tot. sim.  indegrees (one-sided)';
         end; (* case *)
      if symmetric
        then FDescriptions[count + 11] := 'beh. '+Znames[nz+i]+' ave. sim.  rec.  degrees'
        else FDescriptions[count + 11] := 'beh. '+Znames[nz+i]+' ave. sim.  rec.  i.d. (one-sided)';
      if symmetric
        then FDescriptions[count + 12] := 'beh. '+Znames[nz+i]+' tot. sim.  rec.  degrees'
        else FDescriptions[count + 12] := 'beh. '+Znames[nz+i]+' tot. sim.  rec.  i.d.(one-sided)';
      FDescriptions[count + 13] := 'beh. '+Znames[nz+i]+' average alters';
      FDescriptions[count + 14] := 'beh. '+Znames[nz+i]+' average rec. alters';
      //'beh. '+Znames[nz+i]+' sim.  '+Znames[nz+i]+' alter';
      //'beh. '+Znames[nz+i]+' peripheral to dense triads';
      FDescriptions[count + 15] := 'beh. '+Znames[nz+i]+'  '+Znames[nz+i];
      FDescriptions[count + 16] := 'beh. '+Znames[nz+i]+' dense triads <<<maybe wrong>>>';
      FDescriptions[count + 17] := 'beh. '+Znames[nz+i]+' homogeneity of dense triads <<<maybe wrong>>>';
      //'beh. '+Znames[nz+i]+' similarity of peripherals to dense triads';
      FDescriptions[count + 18] := 'beh. '+Znames[nz+i]+' reciprocated degrees';
      if symmetric
        then FDescriptions[count + 19] := 'beh. '+Znames[nz+i]+' ave. sim.  degrees ego'
        else FDescriptions[count + 19] := 'beh. '+Znames[nz+i]+' ave. sim.  indegrees ego';


    (* update current number of already-labelled statistics *)
      count := count + 19;
      parcount := nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19;

(* BEHAVIOUR OBJECTIVE FUNCTION, 2nd part *)
    (* next statistics contain behaviour objective function effects based on actor covariates
       -- per dependent actor variable, there is one effect for each constant covariate
          and one effect for each of the (remaining) changing covariates *)
      (* effects of the constant covariates and the preceding changing covariates*)
      for h:=1 to nz+i-1 do
        begin
        Fdescriptions[count+(2*h)-1] := 'beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)];
        case para of
          0: FDescriptions[count+(2*h)] := 'beh. '+Znames[nz+i]+' influ. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)] := 'beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)] := 'beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)] := 'beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
      (* effects of the subsequent changing covariates *)
      for h:=nz+i+1 to nz+nza do
        begin
        Fdescriptions[count+(2*h)-3] := 'beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)-2];
        case para of
          0: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' influ. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
      for h:=nz+nza+1 to nz+nzc do
        begin
        Fdescriptions[count+(2*h)-3] := 'beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)-2];
        case para of
          0: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' influ. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)-2] := 'beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
(* Added 27/10/8: *)
      case param_effect[parcount+(2*(nz+nzc))-1] of
        0: FDescriptions[count+(2*(nz+nzc))-1] := 'beh. '+Znames[nz+i]+': infl. interaction?'
                                        +'  '+Znames[nz+i]+' alter';
        1: FDescriptions[count+(2*(nz+nzc))-1] := 'beh. '+Znames[nz+i]+': ave. sim. to higher alters';
        2: FDescriptions[count+(2*(nz+nzc))-1] := 'beh. '+Znames[nz+i]+': tot. sim. to higher alters';
        3: FDescriptions[count+(2*(nz+nzc))-1] := 'beh. '+Znames[nz+i]+': ave. higher alters';
        end; (* case *)
      count := count+1;
(* end Added 27/10/8 *)

      (* update current number of already-labelled statistics *)
      count := count + 2*(nz + nzc - 1);

      for h:= 1 to ninta do
        begin
        case ni[i,h] of
        2: FDescriptions[count+h] := 'beh. '+Znames[nz+i]+'  '+Znames[pv1[i,h]]
                                     +'  '+Znames[pv2[i,h]];
        3: FDescriptions[count+h] := 'beh. '+Znames[nz+i]+'  '+Znames[pv1[i,h]]
                                     +'  '+Znames[pv2[i,h]]+'  '+Znames[pv3[i,h]];
       else FDescriptions[count+h] := 'beh. '+Znames[nz+i]+': unspecified interaction';
        end; (* case *)
        end; (* for h *)
      count := count + ninta;              // added 3.17c
      end; (* for i:=1 to nza *)

    (* BEHAVIOUR GRATIFICATION FUNCTION *)
    (* same as beh. obj. fu., but for decrease in behavior *)

    count := NetworkFunctions + nza*(ActionEffects_l + ActionEffects_f);

    for i:=1 to nza do
      begin
      parcount := nxa*NetworkEffects_f + (i-1)*ActionEffects_f;
      (* Note that the parameters are for f and g effects together. *)
      FDescriptions[count + 1] := 'dec. beh. '+Znames[nz+i];
      FDescriptions[count + 2] := 'dec. beh. '+Znames[nz+i]+' average similarity';
      FDescriptions[count + 3] := 'dec. beh. '+Znames[nz+i]+' total similarity';
      if symmetric
        then FDescriptions[count + 4] := 'dec. beh. '+Znames[nz+i]+' degrees'
        else FDescriptions[count + 4] := 'dec. beh. '+Znames[nz+i]+' indegrees';
      FDescriptions[count + 5] := 'dec. beh. '+Znames[nz+i]+' outdegrees';
      FDescriptions[count + 6] := 'dec. beh. '+Znames[nz+i]+' isolate';
      FDescriptions[count + 7] := 'dec. beh. '+Znames[nz+i]+' ave. similarity  reciprocity';
      FDescriptions[count + 8] := 'dec. beh. '+Znames[nz+i]+' tot. similarity  reciprocity';
      if symmetric
        then FDescriptions[count + 9] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  degrees'
        else FDescriptions[count + 9] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  indegrees(one-sided)';
      if symmetric
        then
        case param_effect[parcount+10] of
         0: FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  degrees';
         1: FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+'  av. degree alters';
         else FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  degrees';
        end (* case *)
        else
        case param_effect[parcount+10] of
         0: FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  indegrees(one-sided)';
         1: FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+'  av. indegrees alters';
         else FDescriptions[count + 10] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  indegrees(one-sided)';
         end; (* case *)
      if symmetric
        then FDescriptions[count + 11] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  rec.  degrees'
        else FDescriptions[count + 11] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  rec.  i.d.(one-sided)';
      if symmetric
        then FDescriptions[count + 12] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  rec.  degrees'
        else FDescriptions[count + 12] := 'dec. beh. '+Znames[nz+i]+' tot. sim.  rec.  i.d.(one-sided)';
      FDescriptions[count + 13] := 'dec. beh. '+Znames[nz+i]+' average alters';
      FDescriptions[count + 14] := 'dec. beh. '+Znames[nz+i]+' average rec. alters';
      FDescriptions[count + 15] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[nz+i];
      FDescriptions[count + 16] := 'dec. beh. '+Znames[nz+i]+' dense triads <<<not implemented>>>';
      FDescriptions[count + 17] := 'dec. beh. '+Znames[nz+i]+' homogeneity of dense triads <<<not implemented>>>';
      FDescriptions[count + 18] := 'dec. beh. '+Znames[nz+i]+' reciprocated degrees';
      if symmetric
        then FDescriptions[count + 19] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  degrees ego'
        else FDescriptions[count + 19] := 'dec. beh. '+Znames[nz+i]+' ave. sim.  indegrees ego';

    (* update current number of already-labelled statistics *)
      count := count + 19;
      parcount := nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19;
    (* Note that the parameters are for f and g effects together. *)

      for h:=1 to nz do
        begin
        Fdescriptions[count+(2*h)-1] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)];
        case para of
          0: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' infl. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
      (* effects of the preceding changing covariates *)
      for h:=nz+1 to nz+i-1 do
        begin
        Fdescriptions[count+(2*h)-1] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)];
        case para of
          0: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' infl. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)] := 'dec. beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
      (* effects of the subsequent changing covariates *)
      for h:=nz+i+1 to nz+nza do
        begin
        Fdescriptions[count+(2*h)-3] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)-2];
        case para of
          0: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' infl. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end;
      for h:=nz+nza+1 to nz+nzc do
        begin
        Fdescriptions[count+(2*h)-3] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[h];
        para := param_effect[parcount+(2*h)-2];
        case para of
          0: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' infl. int. possible '
                                          +'  '+Znames[h];
          1: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' average similarity '
                                          +'  '+Znames[h];
          2: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' total similarity '
                                          +'  '+Znames[h];
          3: FDescriptions[count+(2*h)-2] := 'dec. beh. '+Znames[nz+i]+' average alters '
                                          +'  '+Znames[h];
          end; (* case *)
        end; (* for h *)


(* Added 27/10/8: *)
      case param_effect[parcount+(2*(nz+nzc))-1] of
        0: FDescriptions[count+(2*(nz+nzc))-1] := 'beh. '+Znames[nz+i]+': infl. interaction?'
                                        +'  '+Znames[nz+i]+' alter';
        1: FDescriptions[count+(2*(nz+nzc))-1] := 'dec. beh. '+Znames[nz+i]+': av. sim. to higher alters';
        2: FDescriptions[count+(2*(nz+nzc))-1] := 'dec. beh. '+Znames[nz+i]+': tot. sim to higher alters';
        3: FDescriptions[count+(2*(nz+nzc))-1] := 'dec. beh. '+Znames[nz+i]+': av. alters (trunc >= mean)';
        end; (* case *)
      count := count+1;
(* end Added 27/10/8 *)


      (* update current number of already-labelled statistics *)
      count := count + 2*(nz + nzc - 1);

      for h:= 1 to ninta do
        begin
         case ni[i,h] of
           2: FDescriptions[count+h] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[pv1[i,h]]
                                     +'  '+Znames[pv2[i,h]];
           3: FDescriptions[count+h] := 'dec. beh. '+Znames[nz+i]+'  '+Znames[pv1[i,h]]
                                     +'  '+Znames[pv2[i,h]]+'  '+Znames[pv3[i,h]];
           else FDescriptions[count+h] := 'dec. beh. '+Znames[nz+i]+': unspecified interaction';
         end; (* case *)
        end; (* for h *)
      count := count + ninta;              // added 3.17c
      end; (* for i:=1 to nza *)
    end; (* if not signed *)
  end; (* if observations >= 2 *)

  (* Give all names equal lengths 60 : *)
  for i:=1 to count do Fdescriptions[i]:=AlignString(Fdescriptions[i],60);

(* If this is changed, then also change MaxFunctions  *)
end; (* DefineFunctionNames *)

procedure DataSpecification.WritePData(var halterr: Boolean; var outf, lf:TFileStream);
(* Writes data to project data files *)
var swf: textfile;
   fname, x: string;
   i,j,h,k,hmax: integer;
begin (* WritePData *)
  halterr := false;
  if (observations >= 100) then begin
    x:= format('%d%s',[observations,' observations is not implemented in ReadPDefinition.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many observations in ReadPdefinition in S_DAT.',outf,lf);
    halterr := true;
    exit;
   end;

(* for multi: node numbers and skips *)
  fname := pname + '.mul';
  assignfile(swf,fname);
  rewrite(swf);
  for i:=1 to observations do
    begin
    write(swf,nni[i]:4,' ');
    if skipi[i] then writeln(swf,'0') else writeln(swf,'1');
    end; (* for i *)
  closefile(swf);

(* Networks *)
  (* digraph data *)
  for k:=1 to observations
  do begin
    n := nni[k];
    if (k < 10)
    then fname := pname +'.d0'+st(k)
    else fname := pname +'.d'+st(k);
    WriteData(fname,k);
  end; (* k *)

  (* missing data *)
  for k:=1 to observations
  do begin
    n := nni[k];
    if (k < 10)
    then fname := pname +'.m0'+st(k)
    else fname := pname +'.m'+st(k);
    WriteMisData(fname,k);
  end; (* for k *)
  (* structurally determined data *)
  for k:=1 to observations
  do begin
    n := nni[k];
    if (k < 10)
    then fname := pname +'.s0'+st(k)
    else fname := pname +'.s'+st(k);
   WriteStructData(fname,k);
  end; (* for k *)

  n := nmax;
(* Individual non-changing covariates *)
  if (nz > 0)
  then begin
  (* covariate data *)
    fname := pname + '.dav';
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to nz
      do write(swf,z[i,j]:8:4,' ');
      writeln(swf)
    end; (* for i *)
    closefile(swf);
  (* missing data *)
    fname := pname + '.miv';
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to nz
      do if misz(i,j)
      then write(swf,'1')
      else write(swf,'0');
      writeln(swf)
    end; (* for i *)
    closefile(swf)
  end; (* if nz *)

(* Individual changing covariates *)
  if (nzc > 0)
  then begin
  (* covariate data *)
    fname := pname + '.dac';
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to nzc
      do begin
        if (j <= nza)
        then hmax := observations
        else hmax := observations-1;
        for h:=1 to hmax
        do write(swf,zc[i,j,h]:8:4,' ');
      end;
      writeln(swf)
    end; (* for i *)
    closefile(swf);
  (* missing data *)
    fname := pname + '.mac';
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to nzc
      do begin
        if (j <= nza)
        then hmax := observations
        else hmax := observations-1;
        for h:=1 to hmax
        do if miszc(i,j,h)
        then write(swf,'1')
        else write(swf,'0');
      end;
      writeln(swf)
    end; (* for i *)
    closefile(swf)
  end; (* if nzc *)

(* Non-changing dyadic covariates *)
  for h:=1 to nzz
  do begin
  (* covariate data *)
    if (h < 10)
    then fname := pname + '.z0'+st(h)
    else fname := pname + '.z'+st(h);
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to n
      do write(swf,zz[i,j,h]:7:3,' ');
      writeln(swf)
    end; (* for i *)
    closefile(swf);
  (* missing data *)
    if (h < 10)
    then fname := pname + '.n0'+st(h)
    else fname := pname + '.n'+st(h);
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n
    do begin
      for j:=1 to n
      do if miszz(i,j,h)
      then write(swf,'1')
      else write(swf,'0');
      writeln(swf)
    end; (* for i *)
    closefile(swf)
  end; (* for h *)

(* Changing dyadic covariates *)
  for h:=1 to nzzc
  do
    begin
    if (h < 10)
    then fname := pname + '.c0'+st(h)
    else fname := pname + '.c'+st(h);
    assignfile(swf,fname);
    rewrite(swf);
    for k:=1 to observations-1 do
      begin
      for i:=1 to n do
        begin
        for j:=1 to n
        do write(swf,zzc[i,j,h,k]:7:3,' ');
        writeln(swf)
        end; (* for i *)
      end; (* for k *)
    closefile(swf);
    end; (* for h *)
{  (* missing data *)
of: neem aan allemaal dezelfde missings voor versch. tijdstippen
    if (h < 10)
    then fname := pname + '.q0'+st(h)
    else fname := pname + '.q'+st(h);
    assignfile(swf,fname);
    rewrite(swf);
    for k := 1 to observations-1 do
      begin
        for i:=1 to n do
          begin
          for j:=1 to n
          do if miszzc(i,j,k,h)
          then write(swf,'1')
           else write(swf,'0');
          writeln(swf)
        end; (* for i *)
      end; (* for k *)
    closefile(swf)
  end; (* for h *)
}

  (* MH: Exogenous events *)
  if (exo) then begin
    fname := pname + '.dex';
    assignfile(swf,fname);
    rewrite(swf);
    for i:=1 to n do begin
      for j:=1 to 2 do write(swf,zx[i,j]:12:8,' ');  // changed 3.17u
      writeln(swf);
    end; (* for i *)
    closefile(swf);
  end; (* if exo *)
end; (* DataSpecification.WritePData *)


procedure DataSpecification.ReadDefinition1
            (var halterr: Boolean; var outf, lf: TFileStream);
(* reads model definition files.
  readdefinition1 reads section 1 of file .MO.
  readdefinition2 defines effects according to S_Effects,
  and reads other sections of file .MO, and file .SI.
  They are separated because definition of effects requires data information
  (symmetry!), so ReadPData must be executed between readdefinition1 and readdefinition2.
    *)
var
  wf         : textfile;
  loop       : integer;
  fname, x   : string;
  section, n1: integer;

  procedure nextsection(var n: integer);
  (* CS: searches for next section of .IN file in SIENA 2 format *)
  var ch   : char;
      code : integer;
      x : string;
  begin
{$I-}
    repeat
      read(wf,ch);
      if ch='['
      then repeat
        read(wf,ch)
      until ch=']'
    until ((ch='@') or eof(wf));
    if ch='@'
    then begin
      readln(wf,ch);
      val(ch,n,code);
    end
    else begin
      x := 'A SIENA 2 section cannot be read from basic information file.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Basic information file does not have SIENA 2 format.'+eol;
      lf.Write(pchar(x)^,length(x));
      halterr := true
    end
{$I+}
  end; (* nextsection *)

begin (* readdefinition1 *)
//  Heading(2,lf,'Read model definition 1.');
  halterr := false;

  (* FIRST READ THE .MO FILE, section 1: *)
  fname := pname+'.MO';
  if not FileExists(fname)
  then begin
    Halterror('Model definition file not found: <'+pname+'.MO> does not exist.',outf,lf);
    if (pname = 'siena') then begin
    x := 'Probably you forgot to give a project name.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;
    halterr := true;
    exit;
  end;
  assignfile(wf,fname);
  reset(wf);

  nextsection(section);
  if halterr then exit;
  if section<>1
  then begin
    x := format('%s%d%s',['Unexpected section encountered: @',section,' instead of @1 (general information).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    exit
  end;
{$I-}
  readln(wf,observations);  (* number of waves                                           *)
  readln(wf,n1);            (* number of actors                                          *)
  readln(wf,nxa);           (* number of dependent network variables                     *)
  readln(wf,nza);           (* number of dependent actor variables                       *)
  readln(wf,nz);            (* number of constant actor covariates                       *)
  readln(wf,nzc);           (* number of exogenous changing actor covariates             *)
  readln(wf,nzz);           (* number of constant dyadic covariates                      *)
  readln(wf,nzzc);          (* number of exogenous changing dyadic covariates            *)
  readln(wf,nzx);           (* MH: 0-1 flag nzx indicates presence exogenous events file *)
  readln(wf);
  for loop := 1 to nz+nzc+nza do readln(wf,ZNames[loop]);
  for loop := 1 to nzz+nzzc do readln(wf,ZZNames[loop]);
{$I+}

  if (IOResult > 0)
  then begin
    x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%s%s',['Error in reading basic model codes (section @1 of ',fname,' file).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Misspecification of basic model codes.',outf,lf);
    halterr := true;
    closefile(wf);
    exit;
  end;

  n := n1;

  (* some checks of allowed size: *)
  (* waves *)
  if (observations < 1)
  then begin
    HaltError('Number of observations given as '+st(observations)+
       '; this must be at least 1.',outf,lf);
    halterr := true;
  end;
  if (observations > 99)
  then begin
    HaltError('Number of observations given as '+st(observations)+
       '; this must be less than 100.',outf,lf);
     (* this is because of internal data file names.
        Not an important reason, but let them call us
        before working with more than 99 observation moments. *)
    halterr := true;
  end;

  (* individual covariates *)
  if (nza+nz+nzc > nzmax) then begin
    x := format('%s%d%s%d%s%d%s%d%s',['The total number of individual covariates is ',
      nza,'+',nz,'+',nzc,' = ',nza+nz+nzc,
         ', larger than constant nzmax defined in unit EIGHT.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many individual covariates.', outf,lf);
    halterr := true;
  end;

  (* dyadic covariates *)
  if (nzz+nzzc > nzzmax) then begin
    x := format('%s%d%s%d%s%d%s',[ 'The total number of individual covariates is ',
      nzz,'+',nzzc,' = ',nzz+nzzc,
         ', larger than constant nzzmax defined in unit EIGHT.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many dyadic covariates.', outf,lf);
    halterr := true;
  end;

  if halterr then
    begin
    closefile(wf);
    exit;
    end;

  (* MH: indication exogenous events *)
  if (nzx >= 1) then exo := true;

  (* number of dependent network variables *)
  if nxa<>1
  then begin
    HaltError('Number of dependent network variables must be one!', outf,lf);
    halterr := true;
    closefile(wf);
    exit
  end;

(* CS: internally, the number of dependent actor variables must be included in the number
       of changing covariates -- and there, the first nza variables are the dependent ones. *)
  nzc := nzc+nza;
  closefile(wf);
  SetNumberOfPeriods(n,nzc,nzz,nzzc,observations);
//  x := 'readdefinition1 integers: '+inttostr(n)+', '+inttostr(nzc)+', '+inttostr(nzz)+', '+inttostr(nzzc)+', '+
//  inttostr(observations)+'. '+eol;
//    outf.Write(pchar(x)^,length(x));
end; (* readdefinition1 *)

procedure DataSpecification.ReadDefinition2
            (var moname: shortstring; var halterr: Boolean; var outf, lf: TFileStream);
(* reads model definition files.
  readdefinition1 reads section 1 of file .MO.
  readdefinition2 defines effects according to S_Effects,
  and reads other sections of file .MO, and file .SI.
  They are separated because definition of effects requires data information
  (symmetry!), so ReadPData must be executed between readdefinition1 and readdefinition2.
    *)
type
  TSomeInts = 1..250;
  TIntSet = set of TSomeInts;
var
  myset1,
  myset2,
  myset3     : TIntSet;
  wf         : textfile;
  h,i,j,k,
  dd,uu,
  ff,tt,
  di,
  number,
  loop       : integer;
  rd         : real;
  fname,
  dumstri    : string;
  nsubphas,
  observ1,
  n1,nz1,
  nzc1,nzz1,
  nzx1,nza1,             (* MH: nzx1 for exogenous events *)
  section,
  listnumber : integer;
  cc         : char;
  next       : boolean;
  string19   : string[19];
  x          : string;

  procedure nextsection(var n: integer);
  (* CS: searches for next section of .IN file in SIENA 2 format *)
  var ch   : char;
      code : integer;
      x : string;
  begin
{$I-}
    repeat
      read(wf,ch);
      if ch='['
      then repeat
        read(wf,ch)
      until ch=']'
    until ((ch='@') or eof(wf));
    if ch='@'
    then begin
      readln(wf,ch);
      val(ch,n,code);
    end
    else begin
      x := 'A SIENA 2 section cannot be read from basic information file.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Basic information file does not have SIENA 2 format.'+eol;
      lf.Write(pchar(x)^,length(x));
      halterr := true
    end
{$I+}
  end; (* nextsection *)

begin (* readdefinition2 *)
//  Heading(2,lf,'Read model definition 2.');
  halterr := false;

  (* FIRST READ THE .MO FILE: *)
  fname := moname+'.MO';
  if not FileExists(fname)
  then begin
    Halterror('Model definition file not found: <'+fname+'.MO> does not exist.',outf,lf);
    if (fname = 'siena') then begin
    x := 'Probably you forgot to give a project name.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;
    halterr := true;
    exit;
  end;
  assignfile(wf,fname);
  reset(wf);

  nextsection(section);
(* section 1 was read in ReadDefinition1 *)

  variable_rc := nil;
  SetLength(variable_rc,nxa + nza + 1); (*MS*)

  (* read effects for network variables' dynamics *)
  for loop := 1 to nxa
  do begin
    (* find section for network rate function: *)
    nextsection(section);
    if section<>2
    then begin
      x := format('%s%d%s%d%s%s',['Unexpected section encountered: @',section,' instead of @2.',
      loop,'.','1.'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      closefile(wf);
      exit
    end;
{$I-}
    (* CS: if the exact numbering of @2.x.y - sections is to be checked, this should be inserted here *)
    readln(wf,number); (* in this row of the .MO file, the first number should be NetworkEffects_l *)
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf); (* four additional lines with commentary *)
    (* NETWORK RATE FUNCTION SPECIFICATION: *)
      for i:=1 to number
      do begin
        readln(wf,labels_l[i]);           (* first row: effect label *)
        readln(wf,dd,ff,tt,alpa_l[i]);    (* in this row, one more number is contained: effect parameter -- unused for rate function, up to now *)
        effects_l[i] := (dd = 1);
        (* superseded below, after cconditional is known *)
        fix_l[i] := (ff = 1);
        test_l[i] := (tt = 1);       
      end;
{$I+}
      if (IOResult > 0)
      then begin
        x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s%s%s   ',
          ['Error in reading network rate function (section @2.',loop,'.1 of ',fname,' file).'])+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Model misspecified.',outf,lf);
        halterr := true;
        closefile(wf);
        exit;
      end;

   (* find section for network objective function: *)
    nextsection(section);
    if section<>2
    then begin
      x := format('%s%d%s%d%s%s',['Unexpected section encountered: @',section,' instead of @2.',
      loop,'.','2.'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      closefile(wf);
      exit
    end;
{$I-}
    (* CS: if the exact numbering of @2.x.y - sections is to be checked, this should be inserted here *)
    readln(wf,number); (* in this row of the .MO file, the first number should be NetworkEffects_f *)
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf); (* five additional lines with commentary *)

    (* NETWORK OBJECTIVE FUNCTION SPECIFICATION: *)

    for i:=1 to number
    do begin
      readln(wf,labels_f[i]); (* first row: effect label *)

      (* evaluation effect *)
      readln(wf,dd,uu,ff,tt,alpa_f[i],param_effect[i]);
      effects_f[i] := (dd = 1);
      thetarandom_f[i] := (uu = 1);
      variable_rc[loop] := variable_rc[loop] or (uu = 1); (*MS: keep track of whether or not, for dependent variable <loop> (corresponding to network or behavior), there are some random coefficients.*)
      fix_f[i] := (ff = 1);
      test_f[i] := (tt = 1);

      (* endowment effect *)
      readln(wf,dd,uu,ff,tt,rd); (* in this row, one more number is contained: effect parameter -- unused for endowment function, up to now *)
        effects_g[i] := (dd = 1);
        thetarandom_g[i] := (uu = 1);
        variable_rc[loop] := variable_rc[loop] or (uu = 1); (*MS: keep track of whether or not, for dependent variable <loop> (corresponding to network or behavior), there are some random coefficients.*)
        fix_g[i] := (ff = 1);
        test_g[i] := (tt = 1);
        alpa_g[i] := rd;

      (* reinforcement effect *)
      readln(wf); (* CS: no network reinforcement function, up to now *)

    end; (* i, the effects-loop *)
{$I-}
    if (IOResult > 0)
    then begin
      x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%s%s',
        ['Error in reading network objective function (section @2.',loop,'.2 of ',fname,' file).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Model misspecified.',outf,lf);
      halterr := true;
      closefile(wf);
      exit;
    end
  end; (* loop (network variables from 1 to nxa) *)

  NewNetworkEffects := TEffects.Create;  (* This is destroyed in Siena_7 *)
  if signed then DefineSignedNetworkEffects(NewNetworkEffects)
   else DefineNetworkEffects(NewNetworkEffects);
(* The definition of NewNetworkEffects must not be done much later,
   because it is indirectly used below in NetworkEffects_f. *)

  (* read effects for behavioural variables' dynamics *)
  for loop := 1 to nza
  do begin

    (* find section for behavioural rate function: *)
    nextsection(section);
    if section<>3
    then begin
      x := format('%s%d%s%d%s%s',['Unexpected section encountered: @',section,' instead of @3.',
      loop,'.','1.'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      closefile(wf);
      exit
    end;
{$I-}
    (* CS: if the exact numbering of @3.x.y - sections is to be checked, this should be inserted here *)
    readln(wf,number); (* in this row of the .MO file, the first number should be ActionEffects_l *)
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf); (* four additional lines with commentary *)

    (* BEHAVIOURAL RATE FUNCTION SPECIFICATION: *)
    for i:=1 to number
    do begin
      readln(wf,labels_l[NetworkEffects_l+(loop-1)*ActionEffects_l+i]); (* effect label *)
      readln(wf,dd,ff,tt,alpa_l[NetworkEffects_l+(loop-1)*ActionEffects_l+i]); (* in this row, one more number is contained: effect parameter -- unused for rate function, up to now *)
      effects_l[NetworkEffects_l+(loop-1)*ActionEffects_l+i] := (dd = 1);
      (* superseded below, after cconditional is known *)
      fix_l[NetworkEffects_l+(loop-1)*ActionEffects_l+i] := (ff = 1);
      test_l[NetworkEffects_l+(loop-1)*ActionEffects_l+i] := (tt = 1)
    end;
{$I+}
    if (IOResult > 0)
    then begin
      x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%s%s',
       [ 'Error in reading behavioral rate function (section @3.',loop,'.1 of ',fname,' file).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Model misspecified.',outf,lf);
      halterr := true;
      closefile(wf);
      exit;
    end;

    (* find section for behavioural objective function: *)
    nextsection(section);
    if section<>3
    then begin
      x := format('%s%d%s%d%s%s',['Unexpected section encountered: @',section,
      ' instead of @3.',loop,'.','2.'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      closefile(wf);
      exit
    end;
{$I-}
    (* CS: if the exact numbering of @2.x.y - sections is to be checked, this should be inserted here *)
    readln(wf,number); (* in this row of the .MO file, the first number should be ActionEffects_f *)
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf);
    readln(wf); (* five additional lines with commentary *)

    (* BEHAVIOURAL OBJECTIVE FUNCTION SPECIFICATION: *)

    for i:=1 to number
    do begin
      listnumber := NetworkEffects_f+(loop-1)*ActionEffects_f+i;
      readln(wf,labels_f[listnumber]); (* effect label *)

      (* evaluation effect *)
      readln(wf,dd,uu,ff,tt,alpa_f[listnumber],param_effect[listnumber]);
         (* in this row, one more number is contained: effect parameter --
            used for behavioural evaluation function from version 3.11 *)
      effects_f[listnumber] := (dd = 1);
      thetarandom_f[listnumber] := (uu = 1);
      variable_rc[nxa + loop] := variable_rc[nxa + loop] or (uu = 1); (*MS: keep track of whether or not, for dependent variable <loop> (corresponding to network or behavior), there are some random coefficients.*)
      fix_f[listnumber] := (ff = 1);
      test_f[listnumber] := (tt = 1);

      (* endowment effect  *)
      readln(wf,dd,uu,ff,tt,rd); (* in this row, one more number is contained: effect parameter -- unused for endowment function, up to now *)
      listnumber := NetworkEffects_g+(loop-1)*ActionEffects_g+i;
      effects_g[listnumber] := (dd = 1);
      thetarandom_g[listnumber] := (uu = 1);
      variable_rc[nxa + loop] := variable_rc[nxa + loop] or (uu = 1); (*MS: keep track of whether or not, for dependent variable <loop> (corresponding to network or behavior), there are some random coefficients.*)
      fix_g[listnumber] := (ff = 1);
      test_g[listnumber] := (tt = 1);
      alpa_g[listnumber] := rd;

      (* extra endowment effect decrease *)
      readln(wf,dd,uu,ff,tt,rd); (* in this row, one more number is contained: effect parameter -- unused for reinforcement function, up to now *)
      (* Currently, this is not used. *)

    end; (* i, the effects-loop *)
{$I+}
    if (IOResult > 0)
    then begin
      x := format('%s%d',['IOResult = ', IOResult])+eol;
      outf.Write(pchar(x)^,length(x));
      x:= format('%s%d%s%s%s',['Error in reading behavioral objective function (section @3.',
      loop,'.2 of ',fname,' file).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Model misspecified.',outf,lf);
      halterr := true;
      closefile(wf);
      exit;
    end
  end; (* loop (behavioural variables from 1 to nza) *)

  (* find section for options: *)
  nextsection(section);
  if section<>4
  then begin
    x:= format('%s%d%s',['Unexpected section encountered: @',section,' instead of @4.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    closefile(wf);
    exit
  end;
{$I-}

 (* estimation method *)
  readln(wf,dd);
  maxdegree := dd div 1000;
  MakeStep3 := (maxdegree >= 1);
  if MakeStep3 then dd := dd - 1000;
  maxdegree := dd div 100;
  if (maxdegree >= 1) then dd := dd - 100*maxdegree;
  if (maxdegree <= 1) then maxdegree := high(integer);
  CConditional := ((dd=1) or (dd>=21));
  if (exo and Cconditional)
  then begin
  (* MH: only unconditional estimation with exogenous events *)
    CConditional := false;
    //writeln(lf);
    lf.Write(pchar(eol)^,length(eol));
    x := 'Conditional moment estimation is requested for a network of changing composition.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'This is not possible and the unconditional moment estimation procedure will be used.'+eol;
    lf.Write(pchar(x)^,length(x));
   /// writeln(lf);
    lf.Write(pchar(eol)^,length(eol));
  end;
  if Cconditional
  then begin
    if dd=1 then dd:=21;
    if ((dd - 20) > (nxa+nza))
    then begin
      x := format('%s%d%s',['Conditional estimation is requested for conditioning variable ',dd-20,','])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'which is higher than the total number of dependent variables.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'This is changed to 1, the first dependent network variable.'+eol;
      outf.Write(pchar(x)^,length(x));
      ConditioningNetworkVariable := 1;
      ConditioningActionVariable := 0;
    end (* if *)
    else begin
      ConditioningNetworkVariable := 0;
      ConditioningActionVariable := 0;
      if ((dd - 20) <= nxa)
      then ConditioningNetworkVariable := dd -20
      else ConditioningActionVariable := dd - nxa - 20;
    end (* else *)
  end;

  MaxLike := ((dd = 2) and (observations >= 2));
  Bayesian := ((dd = 3) and (observations >= 2));
  if (MaxLike or Bayesian) then
    begin
    CConditional := false;
    continuous := true;
    if exo then
      begin
      if Maxlike then begin
        x := 'Maximum likelihood ';
        outf.Write(pchar(x)^,length(x));
      end
        else begin
          x := 'Bayesian ';
          outf.Write(pchar(x)^,length(x));
        end;
      x := 'estimation was requested together with changing composition.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'This combination is not implemented;'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'estimation method is changed to Method of Moments.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Likelihood-based inference with changing composition'+eol;
      outf.Write(pchar(x)^,length(x));
       x := 'can be carried out by using structural zeros instead of'+eol;
      outf.Write(pchar(x)^,length(x));
       x := 'a joiners/leavers file.'+eol;
      outf.Write(pchar(x)^,length(x));
      MaxLike := false;
      Bayesian := false;
      end;
    end;

  if (not cconditional) then
    begin
    ConditioningNetworkVariable := 0;
    ConditioningActionVariable := 0;
    end;

  VertexPar := ((dd in [10,11]) and (observations = 1));

 (*666*) (*The following code, and possibly the entire unit, must be updated.*)
(*MS: further specifications for ML, Bayesian estimation and random effects models:*)
 {
  readln(wf,dd);
  initialFEestimate_input := (dd = 1);
  }
  initialFEestimate_input := false;
  readln(wf,cdensity1);
  if not(cdensity1 in [1, 2, 3]) then cdensity1 := 1;
  readln(wf,scalebeta);
  if (scalebeta < 0) then scalebeta := 0;
  readln(wf,cdensity2);
  if not(cdensity2 in [1, 2, 3]) then cdensity2 := 3;
  readln(wf,scalere);
  if (scalere < 0) then scalere := 0;

  (* model type *)
  if (observations = 1)
  then begin
    readln(wf,dd);
    Gibbs := dd;
  end
  else begin(* observations >= 2 *)
    readln(wf,dd);
    ModelType := dd;
    if (abs(dd) >= 10)
    then OutputType := 2
    else OutputType := 1;
    ModelType := (abs(dd) mod 10);
    if (ModelType = 0)
    then begin
      ModelType := 1;
      StaySameType := 1;
    end
    else begin
      StaySameType := 2;
    end;
   ReciprocityModel := (not symmetric) and (ModelType = 6);
   if (ReciprocityModel and (MaxLike or Bayesian)) then begin
     x := 'The Reciprocity Model can not be used in conjunction with '+eol;
     outf.Write(pchar(x)^,length(x));
     x := 'the likelihood-based algorithms.'+eol;
     outf.Write(pchar(x)^,length(x));
     ModelType := 1;
     StaySameType := 2;
     ReciprocityModel := false;
    end;
   if ((ModelType >= 3) and (MaxLike or Bayesian)) then begin
     wwriteln(outf,'The likelihood-based algorithms only can be applied with ');
     wwriteln(outf,'the standard actor-oriented model. ');
     wwriteln(outf,'This is now restored.');
     ModelType := 1;
     StaySameType := 2;
    end;
   if ReciprocityModel then begin
     ModelType := 1;
     StaySameType := 1;
//     ModelType := 3;  (* superseded in ReviseDefinition @@@ *)
//     StaySameType := 2;
   (* MOdeltype = 1, StaySametype = 1
      also works if there are only out-degree and reciprocity effects,
      i.e., a pure reciprocity model without covariates *)
    end;
  end; (* if observations >= 2 *)
  if ((Gibbs <= 0) or (Gibbs >= 23)) and (Gibbs <> 26) then Gibbs := 14;
  Continuous := continuous or ((Gibbs >= 8));
  Gibbscode := Gibbs;
  GibbsConnected := (Gibbscode >= 21);
  if (Gibbs >= 21) then Gibbs := Gibbs - 10;
  if (Gibbs >= 11) then Gibbs := Gibbs - 10;
  if (Gibbs in [8,9,10,19,20,21,22]) then
    begin
    //writeln(lf);
    lf.Write(pchar(eol)^,length(eol));
    x :='Conditioning on marginals supersedes (un)conditional definition.'+eol;
    lf.Write(pchar(x)^,length(x));
    Cconditional := true;
    end;

// 1234 signed moet worden gedefinieerd! stond eerst hier uit modeltype.

  (* number of subphases in phase 2 *)
  (* changed 3.181 *)
  readln(wf,nsubphas);
  nextra := nsubphas div 10;
  nsubphase := nsubphas mod 10;
//  if ((nsubphase >= 0) and (nsubphase <= 9)) then nsubphase := nsubphas
//    else nsubphase := 4;

  (* number of iterations in phase 3 *)
  readln(wf,nrun3);
  if (nrun3 < 10) then nrun3 := 10;

  (* number of steps / n^2 per run *)
  readln(wf,rd);
  rstep0 := abs(rd);
  if (rstep0 < 0.00001) then rstep0 := 1.0;
  rstep := rstep0;
(* rstep is changed during burnin;
   if the program stops during this phase, the use of rstep0 ensures that
   the value in the MO file does not change. *)

  (* initial value of gain parameter in estimation *)
  readln(wf,firstgain);
  if (firstgain < 1e-9) then firstgain := 1e-9;

  (* starting value *)
  readln(wf,dd);
  InitialStandard := Odd(dd);

(* CS: observations=1 case model options concerning VertexPar-case are NOT included in .MO file;
   TS: but see below. *)

  (*MS*)
  myset1 := [0, 1, 2]; (*Set of admissible tests.*)
  myset2 := [1, 2]; (*Set of admissible parameter coordinates (restricted for homogeneous actors test, unrestricted for homogeneous periods test).*)
  myset3 := [1..observations]; (*Set of admissible periods; <observations> indicates that a joint test is desired.*)
  for i := 1 to n do outlier[i] := false; (*Default.*)

  (*Period homogeneity test.*)
  readln(wf,period);
  if (period > 0) then (*Homogeneous period test desired.*)
    begin
    homogeneous := 2;
    if (observations = 2) then
      begin
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := 'W A R N I N G: homogeneous period test: a single period: test nonsensical...cancelled.'+eol;
      outf.Write(pchar(x)^,length(x));
     // writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      homogeneous := 0;
      period := 0;
      end
    else (*observations > 2.*)
      begin
      if (period in myset3) then homogeneous := 2
      else
        begin
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        x := 'W A R N I N G: homogeneous period test: inadmissible period...cancelled.'+eol ;
        outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        homogeneous := 0;
        period := 0;
        end;
      end;
    end;

  if ((homogeneous = 2) and (period = observations)) then (*Joint period homogeneity test.*)
    begin
    {
    Code below is ready (needs not be checked), but is disabled because it needs to be completed in the Siena 3 version.
    for i := 2 to (observations - 1) do
      begin
      (*Remarks:
      - <observations - 1> periods: hence we need <observations - 2> dummies.
      - first period is reference period.*)
      Inc(nzc); (*Increment the total number of actor covariates.*)
      Znames[nzc] := 'dummy: period ' + inttostr(i); (*Define label.*)
      end;
    for i := 1 to n do (*Specify dummies.*)
      begin
      for j := 1 to (observations - 2) do (*For dummy <1> to dummy <observations - 2> do:*)
        begin
        for h := 1 to (observations - 1) do (*Exogeneous covariate: the last observation point is redundant and not stored internally.*)
          begin
          if (h = (j + 1)) then zc[i,j,h] := 1
          else zc[i,j,h] := 0;
          end;
        end;
      end;
    homogeneous := 0; (*Joint tests makes no use of the code of single-period tests.*)
    period := 0;
    }
    end;

  (*Actor homogeneity test.*)
  readln(wf,number);
  if (observations >= 2) and (number > 0) then
    begin (*Homogeneous actor(s) test desired.*)
    if (homogeneous=2) then
      begin
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := 'W A R N I N G: simultaneous homogeneity tests (period and actor(s)) inadmissible...default: homogeneous period test selected.'+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      readln(wf) (* skip line of test actors *)
      end
     else
      begin
      if not(number in myset2) then
        begin
        number := 1; (*Default: out-degree.*)
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        x := 'W A R N I N G: homogeneous actor(s) test: inadmissible parameter...default: out-degree selected.'+eol;
        outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        end;
      homogeneous := number + 10;
      i := 0;
      next := false;
      while not (next or eoln(wf)) do
        begin (* read list of outlier candidates. *)
        number := 0;
        while not (next or eoln(wf)) do
          begin
          read(wf,cc); (* read one character *)
          if cc in ['0'..'9'] then number :=10*number+strtoint(cc); (* update actor number by current decimal digit *)
          if cc in [' ','['] then next := true (* if comment begins or if space occurs, the current number is finished *)
          end;
        if cc=' ' then next := false; (* if number was finished by space, enable reading of next number *)
        if (number in [1..n]) then
          begin
          i := 1;
          outlier[number] := true
          end
        end; (* actor list is read *)
      if (i = 0) then
        begin
        homogeneous := 0;
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        x :='W A R N I N G: homogeneous actors test: no admissible outlier candidate...cancelled.'+eol;
        outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        end;
      end;
    readln(wf) (* finish line of test actors *)
    end
   else readln(wf) (* skip line of test actors *);

  if (homogeneous = 2)
  then begin (*Find the "desired" parameter coordinate in <test> vectors.*)
    number := 0;
    j := 0;
    k := 0;
    for i := 1 to MaxEffects_l
    do begin
      if Effects_l[i] and (number = 0) then inc(k);
      if test_l[i]
      then begin
        number := i;
        test_l[i] := false
      end
    end;
    j := k;
    if (number = 0)
    then begin
      for i := 1 to MaxEffects_f
      do begin
        if Effects_f[i] and (number = 0) then inc(k);
        if test_f[i]
        then begin
          number := MaxEffects_l+i;
          test_f[i] := false;
        end;
      end;
    end;
    if (number=0)
    then begin
      for i := 1 to MaxEffects_g
      do begin
        if Effects_g[i] and (number = 0) then inc(k);
        if test_g[i]
        then begin
          number := MaxEffects_l+MaxEffects_f+i;
          test_g[i] := false;
        end;
      end;
    end;
    if (number = 0)
    then begin
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := 'W A R N I N G: homogeneous period test: no parameter selected...default parameter selected: out-degree.'+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      homogeneous := 21; (*Out-degree.*)
    end
    else if (number in [1..observations-1])
    then begin
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := 'W A R N I N G: homogeneous period test: rate parameters inadmissible...cancelled.'+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      period := 0;
      homogeneous := 0;
    end
    else begin
      homogeneous := k + 20;
      homogeneous := homogeneous - j; (*Will be "corrected" in unit <S_est>.*)
    end;
  end; (* if homogeneous=2 *)

  for i := 0 to ccmax
  do begin (*Consistent specification: to be tested parameter coordinates must be fixed in score-type tests.*)
    if test_l[i] then fix_l[i] := true;
    if test_f[i] then fix_f[i] := true;
    if test_g[i] then fix_g[i] := true;
  end;
  test_CovRE[0] := false; (*MS: testing restrictions on the variance-covariance matrix of actor-dependent random effects.*)
 (*MS*)

  (* Model-specific random seed: *)
  readln(wf,modelseed); (* 0 = randomize, all other values to be taken as seed *)
  seedsown := false;

  readln(wf,h); (*MS: MoM derivatives method.*)
  if not(h in [0, 1]) then h := 1; (*Default.*)
  // other options discontinued, version 3.17
  d_estimator := h;

  (*TS: I added the VertexPar option to ReadDefinition, 1-jan-05 *)
  VertexParNew := false;
  if VertexPar then begin
  (* The following construction allows that the last lines with the
     values of the incidental vertex parameters are absent; in that case,
     the earlier initialisation with values 0.0 remains unaffected. *)
      string19 := ''; cc := ' ';
      for i:= 1 to 19
      do begin
        if not eof(wf)
        then read(wf,cc)
        else VertexParNew := true;
        string19 := string19 + cc;
      end;
      if (string19 = '[Method code >= 10:')
      then begin
        readln(wf);
        if eof(wf) then VertexParNew := true;
        for i:=1 to n
        do if (not eof(wf))
           then read(wf,vertex_a[i])
           else VertexParNew := true;
        readln(wf);
        for i:=1 to n
        do if (not eof(wf))
           then read(wf,vertex_b[i])
           else VertexParNew := true;
        readln(wf);
      end (* if string19 *)
     else VertexParNew := true
    end; (* if VertexPar *)

{$I+}
  if (IOResult > 0)
  then begin
    x := format('%s%d', ['IOResult = ', IOResult])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%s%s',
      ['Error in reading model options (section @4. of ',fname,' file).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Model misspecified.',outf,lf);
    halterr := true;
    closefile(wf);
    exit;
  end;

  closefile(wf); (* Model definition file .MO was read *)
//  writeln(lf, 'File ',fname,' was read.');

  (* NOW READ THE .SI FILE: *)
  fname := pname+'.SI';
  if not FileExists(fname)
  then begin
    x := 'File '+fname+' in SIENA 2 format is not found.'+eol;
    lf.Write(pchar(x)^,length(x));
    Halterror('No file specifying statistics for simulation.',outf,lf);
    halterr:=true;
    closefile(wf);
    exit;
  end;
  assignfile(wf,fname);
  reset(wf);
  h:=0; (* counts statistics *)

  (* STATISTICS' CALCULATION-FLAGS: *)
  nextsection(section);
  if halterr then exit;
  if section<>1
  then begin
    x := format('%s%d%s',['Unexpected section encountered: @',section,
        ' instead of @1 (statistics).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    closefile(wf);
    exit
  end;
{$I-}
  readln(wf,number);  (* should contain the total number of possible statistics *)
  readln(wf);
  readln(wf);
  readln(wf);  (* skip three more lines containing but commenting text *)

(*CS: check the following as soon as multiple network variables are allowed! *)
  (* read names of statistics and indicators of whether they are included: *)
  for i:=1 to number
  do begin
    readln(wf,FDescriptions[i]); (* statistic label *)
    readln(wf,dd);
(* Changed version 3.17c: *)
    if (i <= observations - 1) and (not skipi[i]) then dd:=1;
    if ((dd = 1) and (h < pmax))
    then begin
      inc(h);
      Functions[h] := i;
    end
  end;

{$I+}
  if (IOResult > 0)
  then begin
    x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%s%s',
      ['Error in reading statistics specification (section @1. of ',fname,' file).'])+eol;
      outf.Write(pchar(x)^,length(x));
    Halterror('Model misspecified.',outf,lf);
    halterr := true;
    closefile(wf);
    exit;
  end;

  NumberFunctions := h;
  if (h > pmax)
  then begin
    HaltError('The number of statistics is '+st(h)+
              ', greater than constant pmax = '+st(pmax)+'.',outf,lf);
    halterr := true;
    closefile(wf);
    exit;
  end;
  for i:= h+1 to pmax do Functions[i] := 0;

 (* MODEL OPTIONS: *)
  nextsection(section);
  if halterr then exit;
  if section<>2
  then begin
    x := format('%s%d%s',['Unexpected section encountered: @',section,
    ' instead of @2 (options).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    closefile(wf);
    exit
  end;
{$I-}
  (* number of simulations in straight simulation *)
  readln(wf,niter);
{$I+}
  if (IOResult > 0)
  then begin
    x := format('%s%d',['IOResult = ', IOResult])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%s%s',
      ['Error in reading simulation options (section @2. of ',fname,' file).'])+eol;
      outf.Write(pchar(x)^,length(x));
    Halterror('Model misspecified.',outf,lf);
    halterr := true;
    closefile(wf);
    exit;
  end;

  if (niter < 0) then niter := 10;

  closefile(wf); (* statistics specification file for simulation was read *)
//  writeln(lf, 'File ',fname,' was read.');

  (* this procedure must be followed by procedure ReviseDefinition
     defined in S_Base. *)
end (* DataSpecification.ReadDefinition2 *) ;


procedure DataSpecification.ReadPData
      (var halterr: boolean; var outf, lf:TFileStream);
(* Reads data from project data files *)
var
  swf   : textfile;
  fname : string;
  i,j,
  h,k,
  hmax  : integer;
  dd    : char;
  x: string;

begin (* ReadPData *)
  halterr := false;
//  heading(2,lf,'Read project data.');
//  write(lf,'Read data for ',observations,' observation');
//  if (observations <= 1) then writeln(lf,'.') else writeln(lf,'s.');

(* for multi: node numbers and skips *)
  fname := pname + '.mul';
  assignfile(swf,fname);
  reset(swf);
  nmax := 0;
  for i:=1 to observations do
    begin
    read(swf,nni[i],k);
    skipi[i] := (k <= 0);
    if nni[i] > nmax then nmax := nni[i];
    end; (* for i *)
  closefile(swf);
  if (nmax <> n) then
    begin
    Halterror('Inconsistency for n with file '+fname+'.',outf,lf);
    halterr := true;
    exit;
    end;

(* Networks *)
  (* digraph data *)
  valmin := 0;
  valplus := 1;
  for k:=1 to observations
  do begin (* digraph at time k *)
    if (k < 10)
    then fname := pname +'.d0'+st(k)
    else fname := pname +'.d'+st(k);
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
    end;
    ReadData(fname,n,k,valmin,valplus);
  end; (* digraph at time k *)
  signed := (valmin <= -1);
//  x := 'readpdata integers: '+inttostr(n)+', '+inttostr(k)+', '+inttostr(valmin)
//        +', '+inttostr(valplus)+'. '+eol;
//    outf.Write(pchar(x)^,length(x));
// //   xxx := xxx+eol;
//    outf.Write(pchar(xxx)^,length(xxx));


  if (valmin <= -2) or (valplus >= 2)
  then valued :=TRUE;
  //CSCS begin
  //CSCS  writeln(outf,'Network data files contain numbers ranging from ',valmin,' to ',valplus,'.');
  //CSCS  Halterror('Incorrect range in network files.',outf,lf);
  //CSCS  halterr := true;
  //CSCS end;

  (* missing data *)
  for k:=1 to observations
  do begin
    if (k < 10)
    then fname := pname +'.m0'+st(k)
    else fname := pname +'.m'+st(k);
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
    end;
    ReadMisData(fname,n,k);
  end; (* for k *)
  (* structurally determined data *)
  for k:=1 to observations
  do begin
    if (k < 10)
    then fname := pname +'.s0'+st(k)
    else fname := pname +'.s'+st(k);
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
    end;
    ReadStructData(fname,n,k);
  end; (* for k *)


(* Individual non-changing covariates *)
  if (nz > 0)
  then begin
  (* covariate data *)
    fname := pname + '.dav';
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit
    end;
    assignfile(swf,fname);
    reset(swf);
    for i:=1 to n
    do begin
      for j:=1 to nz do read(swf,z[i,j]);
      readln(swf)
    end; (* for i *)
    closefile(swf);
  (* missing data *)
    fname := pname + '.miv';
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit
    end;
    assignfile(swf,fname);
    reset(swf);
    for i:=1 to n
    do begin
      for j:=1 to nz
      do begin
        read(swf,dd);
        setmisz(i,j,dd='1')
      end;
      readln(swf)
    end; (* for i *)
    closefile(swf)
  end; (* if nz *)

(* Individual changing covariates *)
  if (nzc > 0)
  then begin
  (* covariate data *)
    fname := pname + '.dac';
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
    end;
    assignfile(swf,fname);
    reset(swf);
    for i:=1 to n do
      begin
      for j:=1 to nzc do
        begin
        if (j <= nza)
        then hmax := observations
        else hmax := observations-1;
        for h:=1 to hmax do read(swf,zc[i,j,h]);
        end;
      readln(swf);
      end; (* for i *)
    closefile(swf);
  (* missing data *)
    fname := pname + '.mac';
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname)
    then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
    end;
    assignfile(swf,fname);
    reset(swf);
    for i:=1 to n
    do begin
      for j:=1 to nzc
      do begin
        if (j <= nza)
        then hmax := observations
        else hmax := observations-1;
        for h:=1 to hmax
        do begin
          read(swf,dd);
          if dd='1'
          then setmiszc(i,j,h,true)
          else setmiszc(i,j,h,false)
        end
      end;
      readln(swf)
    end; (* for i *)
    closefile(swf)
  end; (* if nzc *)

(* Non-changing dyadic covariates *)
   for h:=1 to nzz
   do begin
   (* covariate data *)
     if (h < 10)
     then fname := pname + '.z0'+st(h)
     else fname := pname + '.z'+st(h);
//     writeln(lf, 'Reading file ',fname);
     if not FileExists(fname)
     then begin
       Halterror('File '+fname+' is not found.',outf,lf);
       halterr := true;
       exit
     end;
     assignfile(swf,fname);
     reset(swf);
     for i:=1 to n
     do begin
       for j:=1 to n do read(swf,zz[i,j,h]);
       readln(swf)
     end; (* i *)
     closefile(swf);
   (* missing data *)
     if (h < 10)
     then fname := pname + '.n0'+st(h)
     else fname := pname + '.n'+st(h);
//     writeln(lf, 'Reading file ',fname);
     if not FileExists(fname)
     then begin
       Halterror('File '+fname+' is not found.',outf,lf);
       halterr := true;
       exit
     end;
     assignfile(swf,fname);
     reset(swf);
     for i:=1 to n
     do begin
       for j:=1 to n
       do begin
         read(swf,dd);
         if dd='1'
         then setmiszz(i,j,h,true)
         else setmiszz(i,j,h,false);
         readln(swf)
         end
      end; (* i *)
     closefile(swf)
   end; (* h *)

(* Changing dyadic covariates *)
   for h:=1 to nzzc do
     begin
     if (h < 10)
     then fname := pname + '.c0'+st(h)
     else fname := pname + '.c'+st(h);
//     writeln(lf, 'Reading file ',fname);
     if not FileExists(fname)
     then
       begin
       Halterror('File '+fname+' is not found.',outf,lf);
       halterr := true;
       exit
       end;
     assignfile(swf,fname);
     reset(swf);

     for k:=1 to observations-1 do
       begin
       for i:=1 to n do
         begin
         for j:=1 to n do read(swf,zzc[i,j,h,k]);
         readln(swf)
         end; (* for i *)
       end; (* for k *)
      closefile(swf);
      end; (* for h *)
   (* missing data *)
{ not implemented - this was copied from the nonchangings:
     if (h < 10)
     then fname := pname + '.n0'+st(h)
     else fname := pname + '.n'+st(h);
     writeln(lf, 'Reading file ',fname);
     if not FileExists(fname)
     then begin
       Halterror('File '+fname+' is not found.',outf,lf);
       halterr := true;
       exit
     end;
     assignfile(swf,fname);
     reset(swf);
     for i:=1 to n
     do begin
       for j:=1 to n
       do begin
         read(swf,dd);
         if dd='1'
         then setmiszz(i,j,h,true)
         else setmiszz(i,j,h,false);
         readln(swf)
         end
      end; (* i *)
     closefile(swf)
   end; (* h *)
}

(* MH: Exogenous events *)
  if (exo) then begin
    fname := pname + '.dex';
//    writeln(lf, 'Reading file ',fname);
    if not FileExists(fname) then begin
      Halterror('File '+fname+' is not found.',outf,lf);
      halterr := true;
      exit;
     end;
    assignfile(swf,fname);
    reset(swf);
    for i:=1 to n do begin
      for j:=1 to 2 do read(swf,zx[i,j]);
      readln(swf);
    end; (* for i *)
    closefile(swf);
  end; (* if exo *)

  Calculations(false, outf);
//  writeln(outf);
//  if ((observations >=2 ) and (StaySameType = 1) and (ModelType <> 3)) then begin
//    writeln(outf,'Note: According to variable <StaySameType>, at RunSteps it is allowed');
//    writeln(outf,'that no change occurs.');
//    writeln(outf);
//   end;
//writeln(outf,'PD9 ',TimeToStr(Now));
end; (* DataSpecification.ReadPData *)

procedure DataSpecification.ReadGraphicOptions(var logfile,
                                 outfile : TFileStream;
                             var halterr : boolean);
(* CS 7-apr-2005: This procedure reads - if it exists - a graphic options file <pname.vis>
   from the current directory, and sets the graphical output options of SIENA accordingly. *)

var section,
    dumint1,
    dumint2,
    count,
    loop       : integer;
    infile     : textfile;
    infilename,
    namestring,
    typestring, x : string;

  procedure nextsection(var n              : integer;
                        var infile         : textfile;
                        var outfile        : TFileStream;
                        var halterr        : boolean);
  (* CS: searches for next section of .MLI file *)
  var ch   : char;
      code : integer;
      x    : string;
  begin
    repeat
      read(infile,ch);
      if ch='['
      then repeat
        read(infile,ch)
      until ch=']'
    until ((ch='@') or eof(infile));
    if ch='@'
    then begin
      readln(infile,ch);
      val(ch,n,code);
    end
    else begin
      x := format('%s%s%s',['A section cannot be read from file <',pname,'.vis>.'])+eol;
      outfile.Write(pchar(x)^,length(x));
      errorstring:='A section of the projectlist file is missing.';
      halterr := true
    end
  end; (* nextsection *)

  procedure killspacesLR (var st : string);
  (* CS: eliminates leading and ending spaces from a string *)
  begin
    while copy(st,1,1)=' ' do st:=copy(st,2,length(st)-1);
    while copy(st,length(st),1)=' ' do st:=copy(st,1,length(st)-1)
  end; (* killspacesLR *)

begin (* ReadGraphicOptions *)
  SoniaGraphics := false;
  PajekGraphics := false;
  infilename := pname+'.vis'; (* extension for file with visualisation commands *)
  if fileexists(infilename)
  then begin
    x := format('%s%s%s',['Reading file with visualization options <',infilename,'>.'])+eol;
    logfile.Write(pchar(x)^,length(x));
    x := format('%s%s%s',['Reading file with visualization options <',infilename,'>.'])+eol;
    outfile.Write(pchar(x)^,length(x));
    assignfile(infile,infilename);
    reset(infile);

    {$I-}
    nextsection(section,infile,outfile,halterr);
    {$I+}
    if halterr then exit;
    if section<>1
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',section,
      ' instead of @1 (basic graphic options).'])+eol;
      outfile.Write(pchar(x)^,length(x));
      errorstring:='Unexpected section in file with visualization options.';
      halterr := true;
      exit
    end;
    {$I-}
    readln(infile,dumint1);
    readln(infile,dumint2);
    {$I+}
    if (IOResult > 0)
    then begin
      x := format('%s%d',['IOResult = ',IOResult])+eol;
      outfile.Write(pchar(x)^,length(x));
      errorstring:='Graphic options could not be read.';
      x := errorstring+eol;
      outfile.write(pchar(x)^,length(x));
      halterr := true;
      exit;
    end;
    SoniaGraphics:=(dumint1<>0);
    PajekGraphics:=(dumint2<>0);

    if SoniaGraphics
    then begin
      {$I-}
      nextsection(section,infile,outfile,halterr);
      {$I+}
      if halterr then exit;
      if section<>2
      then begin
        x := format('%s%d%s',['Unexpected section encountered: @',section,' instead of @2 (SoNIA options).'])+eol;
        outfile.Write(pchar(x)^,length(x));
        errorstring:='Unexpected section in file with visualization options.';
        halterr := true;
        exit
      end;
      {$I-}
      readln(infile,nSoniaVars);
      {$I+}
      if nSoniaVars<=0
      then begin
        nSoniaVars:=0; (* for the case that a fool entered a negative number *)
        x := 'Zero variables to be visualized -- SoNIA file will be generated for network dynamics only.'+eol;
        outfile.Write(pchar(x)^,length(x));
      end;
      setlength(SoniaVar,nSoniaVars+1);
      setlength(SoniaType,nSoniaVars+1);
      setlength(nSoniaCat,nSoniaVars+1);
      setlength(SoniaCatBoundary,nSoniaVars+1,nzmax+2*(nxa+nza)+1); (* nzmax+2*(nxa+nza) = largest possible number of to-be-visualised variables *)
      setlength(SoniaCatLabel,nSoniaVars+1,nzmax+2*(nxa+nza)+1);
      for count:= 1 to nSoniaVars
      do begin
        SoniaVar[count]:=0;
        {$I-}
        readln(infile);
        readln(infile,namestring);
        readln(infile,typestring);
        {$I+}
        if (IOResult > 0)
        then begin
          x := format('%s%d',['IOResult = ',IOResult])+eol;
          outfile.Write(pchar(x)^,length(x));
          errorstring:='SoNIA visualization commands for variable '+st(count)+' could not be read.';
          x := errorstring+eol;
          outfile.Write(pchar(x)^,length(x));
          halterr := true;
          exit;
        end;

        (* first check the name of the to-be-visualised variable *)
        killspacesLR(namestring);
        if copy(namestring,1,11)='$objective_'
        then begin
          namestring:=copy(namestring,12,length(namestring)-11);
          if namestring='network'
          then SoniaVar[count]:=-1
          else for loop := 1 to nza
          do if UpperCase(copy(namestring,1,12))=UpperCase(Znames[nz+loop]) then SoniaVar[count]:=-1-loop;
        end
        else begin
          if copy(namestring,1,9)='$entropy_'
          then begin
            namestring:=copy(namestring,10,length(namestring)-9);
            if namestring='network'
            then SoniaVar[count]:=-(1+nza)-1
            else for loop := 1 to nza
            do if UpperCase(copy(namestring,1,12))=UpperCase(Znames[nz+loop]) then SoniaVar[count]:=-(1+nza)-1-loop;
          end
          else begin
            for loop := 1 to nz+nzc
            do if UpperCase(copy(namestring,1,12))=UpperCase(Znames[loop]) then SoniaVar[count]:=loop;
          end
        end;
        if SoniaVar[count]=0 (* no valid variable name in .vis-file *)
        then begin
          errorstring:='SoNIA visualization requested for non-existing variable '+namestring+'.';
          x := errorstring+eol;
          outfile.Write(pchar(x)^,length(x));
          halterr := true;
          exit;
        end;

        (* now for the type ("visual dimension") *)
        killspacesLR(typestring);
        typestring:=UpperCase(typestring);
        if typestring='EMPTY'
        then SoniaType[count]:=0
        else if typestring='COLOR'
        then SoniaType[count]:=1
        else if typestring='BORDERCOLOR'
        then SoniaType[count]:=2
        else if typestring='SHAPE'
        then SoniaType[count]:=3
        else if typestring='SIZE'
        then SoniaType[count]:=4
        else if typestring='BORDERWIDTH'
        then SoniaType[count]:=5
        else begin
          errorstring:='SoNIA visualization requested for non-existing type '+typestring+'.';
          x := errorstring+eol;
          outfile.Write(pchar(x)^,length(x));
          halterr := true;
          exit;
        end;

        (* now for the categories: *)
        if (SoniaType[count] in [1,2,3]) (* types that require categories *)
        then begin
          {$I-}
          readln(infile,nSoniaCat[count]);
          {$I+}
          if ((IOResult>0) or (nSoniaCat[count]<=0)) (* no valid category number in .vis-file *)
          then begin
            errorstring:='Wrong number of categories for SoNIA visualization (variable '+namestring+').';
            x := errorstring +eol;
            outfile.Write(pchar(x)^,length(x));
            halterr := true;
            exit;
          end;
          if ((nSoniaCat[count]>0) and (SoniaVar[count]<0)) (* entropy and objective values must have '0' as cat. numbers *)
          then begin
            errorstring:='SoNIA categories requested for entropy or objective function (variable '+namestring+').';
            x := errorstring +eol;
            outfile.Write(pchar(x)^,length(x));
            halterr := true;
            exit;
          end;
          if ((nSoniaCat[count]<>2) and (SoniaType[count]=3)) (* 'shape' variables must have two categories *)
          then begin
            errorstring:='SoNIA categories requested for entropy or objective function (variable '+namestring+').';
            x := errorstring+eol;
            outfile.Write(pchar(x)^,length(x));
            halterr := true;
            exit;
          end;
          if ((nSoniaCat[count]>0) and (SoniaType[count] in [0,4,5])) (* 'size', 'borderwidht' and 'empty' variables must not have categories *)
          then begin
            errorstring:='SoNIA categories requested for entropy or objective function (variable '+namestring+').';
            x := errorstring+eol;
            outfile.Write(pchar(x)^,length(x));
            halterr := true;
            exit;
          end;
          (* now read category definitions *)
          for loop := 1 to nSoniaCat[count]
          do begin
            {$I-}
            readln(infile,namestring);
            {$I+}
            val(namestring,SoniaCatBoundary[count,loop,1],dumint1);
            while copy(namestring,1,2)<>'<=' do namestring:=copy(namestring,2,length(namestring)-1);
            namestring:=copy(namestring,3,length(namestring)-2);
            while copy(namestring,1,1)<>'<' do namestring:=copy(namestring,2,length(namestring)-1);
            namestring:=copy(namestring,2,length(namestring)-1);
            val(namestring,SoniaCatBoundary[count,loop,2],dumint2);
            while copy(namestring,1,1)<>'=' do namestring:=copy(namestring,2,length(namestring)-1);
            namestring:=copy(namestring,2,length(namestring)-1);
            killspacesLR(namestring);
            SoniaCatLabel[count,loop]:=namestring;
            if ((IOResult > 0) {or (dumint1>0) or (dumint2>0)} or (SoniaCatBoundary[count,loop,2]<=SoniaCatBoundary[count,loop,1]))
            then begin
              errorstring:='Wrong category specification for SoNIA visualization (variable '+Znames[SoniaVar[count]]+').';
              x := errorstring+eol;
              outfile.Write(pchar(x)^,length(x));
              halterr := true;
              exit;
            end;
          end
        end (* categories *)

      end (* for count:=1 to SoniaVars *)

    end; (* if SoniaGraphics *)

  end;
 // writeln(outfile);
  outfile.Write(pchar(eol)^,length(eol));
end; (* DataSpecification.ReadGraphicOptions *)

procedure DataSpecification.Calculations(writeoutput: boolean; var outf: TFileStream);
var
  i,j,h,k,c,
  hmax, ncomp,
(* AA  start *)
  tempa, tempb, mi, ini, ji, maxk,
  difh, uph, downh, nmish, datah, readgibbs,
  nrmean        : integer;
  minz, maxz, minzk, maxzk,
  tempra, temprb,
  sum           : real;
  x             : string;
  temparr, tempbrr: array of real;
  sv: array of integer;
(* AA end *)

begin
  MaxRuns := 0;
  tempa := 0;
  tempb := 0;
  anynetworkmissings := false;
  tempra := 0.0;
  temprb := 0.0;
  SetLength(uponly,observations);
  SetLength(downonly,observations);

(* Versions up to 3r had a Order(n^3) algorithm for calculating various counts here.
   This was now changed to an order(n) algorithm.
   The earlier algorithm still is left, temporarily, commented away,
   for possibilities of checking. *)

  uponly[0] := false;
  downonly[0] := false;
  if (observations >= 2) then begin
    symmetric := AllSymmetric(observations);
    for h:=1 to observations-1 do if (not skipi[h]) then
      begin
      n := nni[h];
      SumDifMat(h,difh,uph,downh,datah);
      uponly[h] := (uph = 0);
      if (uph = 0) then uponly[0] := true; (* any uponly *)
      downonly[h] := (downh = 0);
      if (downh = 0) then downonly[0] := true; (* any downonly *)
      nmish := n*(n-1) - datah;
{      difh := 0;
      nmish := 0;
      datah := 0;
      writeln(outf,'1111 ',difh,' ',datah,' ',nmish);
      for i:=1 to n do for j:=1 to n do
       begin
        if (mis(i,j,h) <> mis(j,i,h)) then symmetric := false;
        if (ynm(i,j,h) <> ynm(j,i,h)) then symmetric := false;
        if ((i <> j) and (not (mis(i,j,h) or mis(i,j,h+1))))
          then begin
            inc(difh, abs(ynvm(i,j,h) - ynvm(i,j,h+1)));
            inc(datah);
           end;
        if ((i <> j) and (mis(i,j,h) or mis(i,j,h+1)))
          then inc(nmish);
        for k:=1 to n do
        if ((i <> j) and (i <> k) and (j <> k)) and (not(mis(i,k,h) or mis(j,k,h)))
        then begin
          inc(tempb);
          if (ynm(i,k,h) <> ynm(j,k,h)) then inc(tempa);
        end;
       end; (* for i, j *)
      writeln(outf,'2222 ',difh,' ',datah,' ',nmish);
 }
      if (difh+nmish > Maxruns) then MaxRuns := difh+nmish;
      ObservedNetworkDistance[1,h] := difh;
      ObservedNetworkData[1,h] := datah;
      if (nmish >= 1) then anynetworkmissings := true;

      for i:= 1 to n do
        begin
        mi := MissingIn(i,h);
        ini := Indm(i,h);
        ji := IndMis(i,h);
        tempra := tempra + 2*ini*(n - 1.0 - mi - ini)
         + 2*ji*(mi + 2*ini - ji - n + 1.0);
        end;
      temprb := temprb + n*(n-1.0)*(n-2.0) + ((-2.0*n) + 3)*Totalmissings(h);
      for i:=1 to n do temprb := temprb + sqr(MissingIn(i,h));
      end (* for h if *)
     else
      begin
      uponly[h] := false;
      downonly[h] := false;      
      end;
{
   writeln(outf,'aaa tempa ',tempa);
   writeln(outf,'aaa tempb ',tempb);
   writeln(outf,'aaa tempra ',tempra:20:1);
   writeln(outf,'bbb temprb ',temprb:20:1);

    for i:=1 to n do for j:=1 to n do
     begin
        if (mis(i,j,observations) <> mis(j,i,observations)) then symmetric := false;
        if (ynm(i,j,observations) <> ynm(j,i,observations)) then symmetric := false;
     end;  (* for i, j *)
}
   end; (* if observations >= 2 *)


  if (uponly[0] and (CallingProgram = 1)) then
    begin
    if (observations = 2) and writeoutput then
      begin
      x :='All network changes are upward.'+eol ;
      outf.Write(pchar(x)^,length(x));
      x :='This will be respected in the simulations.'+eol ;
      outf.Write(pchar(x)^,length(x));
      outf.Write(pchar(eol)^,length(eol));
      end
     else
      begin
      x :='All network changes are upward for the following periods: '+eol;
      outf.Write(pchar(x)^,length(x));
      c := 0;
      for h:=1 to observations-1 do if uponly[h] then
          begin
          x :=inttostr(h)+' => '+inttostr(h+1)+'; ';
          inc(c);
          if (c >= 10) then
            begin x := x+eol; c := 0; end;
          outf.Write(pchar(x)^,length(x));
          end;
      outf.Write(pchar(eol)^,length(eol));
      x :='This will be respected in the simulations.'+eol ;
      outf.Write(pchar(x)^,length(x));
      end;
    uponly[0] := true;
    for h:=1 to observations-1 do if (not uponly[h]) then uponly[0] := false;
    end;
(* now uponly[0] = all uponly[h] are true *)
  if writeoutput then
    begin
    if uponly[0] then
      begin
      x := 'All changes are only up: no outdegree parameter.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    end;

  if (downonly[0] and (CallingProgram = 1) and writeoutput) then
    begin
    if (observations = 2) then
      begin
      x :='All network changes are downward.'+eol ;
      outf.Write(pchar(x)^,length(x));
      x :='This will be respected in the simulations.'+eol ;
      outf.Write(pchar(x)^,length(x));
      outf.Write(pchar(eol)^,length(eol));
      end
     else
      begin
      x :='All network changes are downward for the following periods: '+eol;
      outf.Write(pchar(x)^,length(x));
      c := 0;
      for h:=1 to observations-1 do if downonly[h] then
          begin
          x :=inttostr(h)+' => '+inttostr(h+1)+'; ';
          inc(c);
          if (c >= 10) then
            begin x := x+eol; c := 0; end;
          outf.Write(pchar(x)^,length(x));
          end;
      outf.Write(pchar(eol)^,length(eol));
      x :='This will be respected in the simulations.'+eol ;
      outf.Write(pchar(x)^,length(x));
      outf.Write(pchar(eol)^,length(eol));
      end;
    end;
  if (downonly[0] and (CallingProgram = 1)) then
    begin
    downonly[0] := true;
    for h:=1 to observations-1 do if (not downonly[h]) then downonly[0] := false;
    end;
(* now downonly[0] = all downonly[h] are true *)
  if downonly[0] and writeoutput then
    begin
    x := 'All changes are only down: no outdegree parameter.'+eol;
    outf.Write(pchar(eol)^,length(eol));
    end else if writeoutput then outf.Write(pchar(eol)^,length(eol));


  if (anynetworkmissings and (CallingProgram = 1) and writeoutput)
    then begin
    x :='There are missing data for the network variable.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;
  if (observations = 1) then
    begin
    k := 0;
    for i:=1 to n do inc(k,Outdm(i,1));
(* AA change *)    densty := k/(n*(n-1)); (* Modified later if there are structurally determined values *)
    symmetric := true;
    asymmetric := true;
    for i:=1 to n do
      begin
      if (Outdm(i,1) <> Recidm(i,1)) then symmetric := false;
      if (recidm(i,1) > 0) then asymmetric := false;
      if (Outdm(i,1) + Indm(i,1) <> (n-1)) then asymmetric := false;
      end;
    if not StructSym(1) then
      begin
        symmetric := false;
        asymmetric := false;
      end;
    end; (* if observations = 1 *)

  for h:=1 to observations do ExistStruct[h] := StructureExists(h);
  SetChangeStruct(observations); 

  if (observations = 1)
  then MaxRuns := trunc(50*(n+1.0)*n)
  else MaxRuns := 20*MaxRuns;

  if (observations = 1) then balmean := 0.0 else balmean := tempra/temprb;

  n := nmax;
  (* means of dyadic covariates into zzmean[.] *)
  for i :=1 to nzz
  do begin
    zzmean[i] := 0.0;
    nrmean := 0;
    for j:=1 to n
    do for h:=1 to n
    do if ((j <> h) and not miszz(j,h,i))
    then begin
      zzmean[i] := zzmean[i] + zz[j,h,i];
      inc(nrmean)
    end;
    zzmean[i] := zzmean[i]/nrmean;
  end;


  (* constant actor covariates
     NOTE that constant covariates have been centered before
          they were written to the internal data files already! *)
  for i:=1 to nz do
  begin
   (* determine valid range of constant covariates *)
    minz:= MaxInt;
    maxz:=-MaxInt;
    for j:=1 to n do
    if not misz(j,i)
    then begin
      if z[j,i]<minz then minz:=z[j,i];
      if z[j,i]>maxz then maxz:=z[j,i];
    end;
    range[i] := maxz-minz;
(* AA start *)
    if range[i] < 1e-40 then
      begin
      range[i]:=1e-40;
      poszvar[i] := false;
      end else poszvar[i] := true;
(* AA end *)

    (* determine similarity means of constant covariates *)
    zsmean[i] := 0.0; (* note that for calling the similarity function right below, we need this array
                         to be set to zero if we want to arrive at the right similarity averages *)
    nrmean := 0;
    sum := 0;
    for j:=1 to n do
    if not misz(j,i)
    then for h:=1 to n
    do if ((h <> j) and (not misz(h,i)))
    then begin
      sum := sum + similarity(z[j,i],z[h,i],i);          // @@@ maak lineair
      inc(nrmean)
    end;
    zsmean[i] := sum/nrmean;
  end; (* for i *)

 (* now for the changing actor covariates *)
  for i:=1 to nzc do
  begin
    if (i <= nza)
    then hmax := observations
    else hmax := observations-1;

   (* determine range; and mean for dependent actor variables  *)
(* AA start *)
    poszvar[nz+i] := (i <= nza);
    minz:= MaxInt;
    maxz:=-MaxInt;
    zamean[i] := 0.0;
    for k:=1 to hmax do
     if skipi[k] then // added 260809
      begin
      if (i <= nza) then pzamean[i,k] := 0.0;
      end
     else
      begin
      n := nni[k];
      sum := 0.0;
      nrmean := 0;
      minzk:= MaxInt;
      maxzk:=-MaxInt;
      for j:=1 to n do if not miszc(j,i,k) then
        begin
        if (i <= nza) then
          begin
          inc(nrmean);
          sum := sum+zc[j,i,k];
          end;
        if zc[j,i,k]>maxzk then maxzk:=zc[j,i,k];
        if zc[j,i,k]<minzk then minzk:=zc[j,i,k];
        end; (* j *)
      if (i <= nza) then
        begin
        if (nrmean <= 0) then
          begin
          pzamean[i,k] := 0.0;
          x := format('%s%d%s%d%s',['WARNING: no valid data for dependent actor variable ',i,
            ', period ',k,'.'])+eol;
          outf.Write(pchar(x)^,length(x));
          end
         else
          begin
          pzamean[i,k] := sum/nrmean;
          zamean[i] := zamean[i]+sum/nrmean;
          end;
        end (* if i <= nza *)
       else
        begin
        if (minzk + 1e-40 < maxzk) then poszvar[nz + i] := true;
        end;
      if (minzk < minz) then minz := minzk;
      if (maxzk > maxz) then maxz := maxzk;
(* AA end *)

      end; (* k *)
    if (i <= nza)
    then begin
      zcmin[i]:=round(minz);
      zcmax[i]:=round(maxz);
      zamean[i] := zamean[i]/hmax;

//      if (zcmax[i] - zcmin[i] < 3.9) then   // special version
//        begin
//        write(outf,'Note: the maximum of the allowed range of actor variable ',ZNames[nz+i],
//                      ' is extended from the observed value of ',round(zcmax[i]));
//        zcmax[i] := zcmin[i] + 4;
//        writeln(outf,' to ',round(zcmax[i]),'.');
//        end;

    end
    else begin
      zcmin[i]:=minz;
      zcmax[i]:=maxz
    end;
    range[i+nz]:=zcmax[i]-zcmin[i];
    if range[i+nz]<1e-40 then range[i+nz]:=1e-40;
   (* determine distance *)
    if (i <= nza) then for k:=1 to observations-1 do
      begin
      n := nni[k];
      difh := 0;
      datah := 0;
      for j:=1 to n do
       if (not (miszc(j,i,k) or miszc(j,i,k+1))) then
        begin
        difh := difh + abs(round(zc[j,i,k]) - round(zc[j,i,k+1]));
        inc(datah);
        end;
      ObservedActionDistance[i,k] := difh;
      ObservedActionData[i,k] := datah;
      end; (* if i <= nza then for k *)
   (* determine similarity means *)
    zcsglobalmean[i] := 0.0;
    zsmean[nz+i] := 0.0;       // required for SimOnce
    nrmean := 0;
    for k:=1 to hmax do if not skipi[k] then
    begin
      n := nni[k];
      for j:=1 to n do
      if not miszc(j,i,k)
      then for h:=1 to n
      do if not (miszc(h,i,k)or (j = h))
      then begin
        zcsglobalmean[i] := zcsglobalmean[i] + similarity(zc[j,i,k],zc[h,i,k],nz+i);          // @@@ maak lineair
        inc(nrmean)
      end (* j *)
    end; (* k *)
    if (nrmean >= 1) then
      zcsglobalmean[i] := zcsglobalmean[i]/nrmean
      else zcsglobalmean[i] := 0.0;

  end; (* for i *)

(* Dependent actor variables: *)
  hmax := observations-1;
  for h := 1 to nza do
    begin
      for k:=1 to hmax do
        begin
          n := nni[k];
          for i:=n downto 1 do
            begin
//          matignore[i,h,k] := miszc[i,h,k] or miszc[i,h,k+1];
            end;
        end;
    end;

 (* changing covariates : *)
  for j:=1 to nzc do zsmean[nz+j] := zcsglobalmean[j];

  if ((nza >= 1) or (ModelType = 3))
  then begin
    if ((StaySameType <> 2) and (nza >= 1)) then
    begin
      x := 'With dependent behavior variables, StaySameType must be 2.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    StaySameType := 2;
  end;

(* Determine maximum number of components *)
  ncomp := 1;
  SetLength(ComponentIndicator,nmax+1);
  for k:=1 to observations do
    if (((k <= 1) or (k < observations)) and ExistStruct[k]) then
      begin
      n := nni[k];
      Component(j,true,ComponentIndicator,k);
      if (j >= ncomp) then ncomp := j;
      end;
  Setlength(ComponentNumber,1+ncomp,n+1);


(* AA start *)
(* Calculation of compzvar; recalculation of densty.  *)
  if (observations <= 1) then maxk := 1 else maxk := observations-1;
  for k:=1 to maxk do if ExistStruct[k] then
    begin
    setlength(sv,valplus - valmin + 1);
    StructuralDataValues(k, sv);
(* sv[h-valmin]  now contains the number of
   structurally determined values equal to h *)
    h := 0;
    for i:=valmin to valplus do if (i-valmin <> 0) then inc(h,sv[i]);
(* h = number of structurally determined values not equal to 0 *)
    if (observations <= 1) then
      begin
      densty := (densty * n * (n-1) - h)/(n*(n-1) - nstruct(1));
      end;
    if (h <= 0) then
(* no structurally determined values different from 0 *)
      begin
      Component(ncomp,false,ComponentIndicator,k);
      setlength(temparr,1+ncomp);
      setlength(tempbrr,1+ncomp);
      for h:=1 to nz+nzc do
       if (not anymisz(h,n,nz,observations-1)) then
        begin
        if (k <= 1) then compzvar[h] := true;
        for i:=ncomp downto 1 do
          begin
          temparr[i] := 0.0;
          tempbrr[i] := 0.0;
          end;
        n := nni[k];
        for i:=n downto 1 do
          if (h <= nz) then
          begin
          temparr[ComponentIndicator[i]] :=
            temparr[ComponentIndicator[i]] + z[i,h];
          tempbrr[ComponentIndicator[i]] :=
            tempbrr[ComponentIndicator[i]] + sqr(z[i,h]);
          end else if (h > (nz+nza)) then
          begin
          temparr[ComponentIndicator[i]] :=
            temparr[ComponentIndicator[i]] + zc[i,h-nz,k];
          tempbrr[ComponentIndicator[i]] :=
            tempbrr[ComponentIndicator[i]] + sqr(zc[i,h-nz,k]);
          end;
        if ((h > nz) and (h <= nz+nza))  then compzvar[h] := false
         else for i:=ncomp downto 1 do
           if ((tempbrr[i]*ComponentNumber[i,0]) > sqr(temparr[i]) + 1e-10)
          then compzvar[h] := false;
          // this is a dependent actor variable or
          // there is positive variance of z[.,h] in component i.
        end;
      end (* if (h <= 0) *)
       else for h:=nz+nzc downto 1 do compzvar[h] := false;
    end (* for k if ExistStruct[k] *)
     else for h:=nz+nzc downto 1 do compzvar[h] := (not poszvar[h]);

   for h:=nz+nzc downto 1 do if anymisz(h,nmax,nz,observations-1)
     then
         begin
         compzvar[h] := false;
         poszvar[h] := true;
         end;

   if (callingprogram = 1) then
     begin
     tempa := 0;
     for h:=1 to nz do if compzvar[h] then Inc(tempa);
     case tempa of
       1: begin
          h := 0;
          repeat
           inc(h)
           until compzvar[h] or (h >= nz);
          x := 'Note: variable '+Znames[h]+' is constant within components.'+eol;
          outf.Write(pchar(x)^,length(x));
          wwriteln(outf,'');
          end;
       2..nzmax: begin
          x := 'Note: the following variables are constant within components:';
          outf.Write(pchar(x)^,length(x));
          for h:=1 to nz do if compzvar[h] then
            begin
            x := eol+ZNames[h];
            outf.Write(pchar(x)^,length(x));
            end;
          x := '.'+eol;
          outf.Write(pchar(x)^,length(x));
          wwriteln(outf,'');
          end;
      end; (* case *)
    end;
(* AA finish *)

  ActivePeriods := 0;
  SetLength(ActiveObservationRank,observations);
  ActiveObservationRank[0] := 0;
  for h:=1 to observations-1 do if (not skipi[h]) then
    begin
    inc(ActivePeriods);
    ActiveObservationRank[h] := ActivePeriods;
    end else ActiveObservationRank[h] := 999;

  d_estAB := false;
  d_estA := false;
  d_estB := false;
end; (*  DataSpecification.Calculations *)


procedure DataSpecification.AdaptModel;
    (* Adapts model; only for use
       when running special versions of Siena04. *)
       var x: string;
begin
  if (observations >= 2) then
    begin
    maxlike := true; // special version
//    rstep0 := 15.0;  // special version
//    firstgain := 0.2;
//    modelseed := 1234;
//    nrun3 := 2000;  // special version
    end;
  InitialStandard := false; // special version
  x := 'Model adaptation: maximum likelihood and current initial value.'+eol;
  outf.Write(pchar(x)^,length(x));
end; (* AdaptModel *)


procedure DataSpecification.Halterror(s: string; var f, lf: TFileStream);
var x : string;
begin
 // writeln(lf); writeln(lf, s);
 // writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
  lf.Write(pchar(s)^,length(s));
  lf.Write(pchar(eol)^,length(eol));
  x := 'This is an error. Program stops.'+eol;
  lf.Write(pchar(x)^,length(x));
  //writeln(f); writeln(f,s);
  f.Write(pchar(eol)^,length(eol));
  f.Write(pchar(s)^,length(s));
  f.Write(pchar(eol)^,length(eol));
  x := 'This is an error.'+eol;
  f.Write(pchar(x)^,length(x));
  x :='Program terminated.'+eol;
  f.Write(pchar(x)^,length(x));
  //writeln(f);
  f.Write(pchar(eol)^,length(eol)) ;
  x := format('%s%s%s',['Depending on the type of error, the file ',pname,'.log'])+eol;
  f.Write(pchar(x)^,length(x));
  x := 'may contain some extra information for error diagnosis.'+eol;
  f.Write(pchar(x)^,length(x));
  //writeln(f);
  f.Write(pchar(eol)^,length(eol));

  errorstring := s;
{$if not defined(batch) and not defined(silent)}
  ShowMessage('An error occurred: '+s+' The program stops. Look in the output file.');
//{$ifend}
{$elseif defined(batch)}
  writeln('An error occurred: '+s+' The program stops. Look in the output file.');
{$ifend}
end; (* Halterror *)

procedure DataSpecification.FinishDat(var outfile: TFileStream);
begin
//  FinishEight(outfile);  This leads to errors since version 3.17c, I do not know why. abcdefg
  Finalize(ComponentNumber);
  if Assigned(NewNetworkEffects) then FreeAndNil(NewNetworkEffects);

//abcd  closefile(checkfile);
end;

procedure unpack(const pa: integer; var nia, p1a, p2a, p3a: integer);
(* interprets parameter value pa for behavior interactions.
   nia = number of interacting variables
   p1a,p2a,p3a = index numbers of interacting variables *)
begin
  nia := 0;
  p1a := 0;
  p2a := 0;
  p3a := 0;
  if (pa <= 100) then exit;
  p1a := pa mod 100;
  p2a := (pa div 100) mod 100;
  p3a := (pa div 10000);
  if (p3a >= 1) then
    begin
    if (p1a <= 0) or (p2a <= 0) then nia := 0 else nia := 3;
    end
   else
    begin
    if (p1a <= 0) or (p2a <= 0) then nia := 0 else nia := 2;
    end;
end; (* unpack *)


begin (* S_Dat *)
  SienaVersion := '3.181 (08 April 10)'; (* TS *)
(* NOTE mimean etc. not initialized in s_desc;
   structactive niet goed gebruikt; describe is ad hoc  *)
  oldmofile := false; (* if this is true, then the siena 2-style mo file is used;
                           please indicate this by o in the sienaversion *)
(* changes 3.181: further permission of reals for dyadic covariates;
                  (did not work well for pajek input and for changing dyadic covariates)
           extra options for SimOnce ("GenerateMany" in s_sim);
           condition i <> h added in Siena02, s_desc, transitivity ratio;
           multiple estimation runs with nextra in s_est;
           option Gibbscode 26 for ergm of undirected graphs
           under condition of connectedness added;
           3.17w: multiply balance effect by n-2;
           repair error making it impossible to estimate more than one endowment effect;
           centering of actor covariates only for non-constant actor covariates;
           some changes multigroup projects: correction for multiperiod case in s_base.setalpha;
           3.17v: further error correction in indegree-popularity (sqrt);
           3.17u: correction various inconsistencies in treatment of joiners-leavers;
           correct errors in indegree-popularity (sqrt),
           and out-in, in-out, and in-in assortativity effects;
           in Spatialstatistics, the sums over actors i and j
           are taken only over structurally active actors;
           sums in actionstatistics only over structurally active actors, not joiners/leavers;
           imputation (last observation carried forward) for changing actor covariates;
           ERGM: #-cliques works only for nondirected.
           3.17t: various ERGM effects revived:
                     cliques; indirect connections; direct & indirect connections;
                     histogram possibility added (for research purposes,
                     depending on the variable histograms set at the bottom of s_est).
              3.17s: use new structurally determined values at end of runepoch;
              3.17r: change calculation of target statistics
                     for the case that structurally determined values
                     change between waves;
                     reappearance of bug repair of version 3.17i;
              3.17q: correct transitive triplets effect and score function initialization
                     which led to errors for endowment effects;
                     scale indegree and outdegree popularity effects differently;
                     localize dist to distfd in FiniteDifferences in s_est;
              3.17p: popularity/activity effects renamed in accordance to tutorial
                     and extended.
              3.17o: sum of degrees needed a sqrt for nondirected;
                     closefile(simdatf) later;
              3.17n: error in interaction which led to program crash.
              3.17m: errors in calculation of degrees, siena01. Nobody had ever noticed!!!!
              3.17i: bug repair related to wrong counting of network distances
                     in the case of structurally fixed ties;
                     due to an oversight, this bug repair has been absent in
                     versions 3.17m-3.17q!!!, and reappeared again in 3.17r.
              3.17g: small bugs corrected that occurred in Siena01;
                     popularity alter effect on behavior;
              3.17f: effects behavior reciprocated degree and
                     behavior ave. sim. x popularity ego.
              3.17e: corrected endowment effects for behavior.
                     run through the elements of UpdateComps in CalcComponents_fa in Eight
                     only once, to aviod multiple counting; this leads to
                     other values of behavioral statistics.
              3.17d: multi-group projects.    endowment effects for behavior.
              3.17c: rectified bookkeeping for nza > 1. Some extra effects.
              3.17a: rectified use of anymis in S_dat,
                    and an error with matmiszc in eight. *)
(* change 3.17: add checkmaxima in s_start.Preparations
                because there was a risk of undiscovered range check errors *)
(* change 3.16: add some effects, special treatment of variables constant in components *)
(* 3.15 : linux compatible (Ruth Ripley);
          bug fixed occurring in Siena01 when there are more dyadic covariates than actor covariates;
          allow uponly and downonly;
          assortativity effects added;
          testing rate effects allowed in IndicateFixedThetaAndRC;
          it was unclear to me (t.s.) why this was excluded there. *)
(* 3.14 : interaction effects between triadic effects and covariates *)
(*        missing for dyadic covariate discarded in calculation of effects. *)
(*        Bug repaired for `same covariate' effect. *)
(* 3.13 : No more automatic switching to unconditional estimation in S_Est.*)
(* 3.12a: bug repair for structurally determined values in ERGM,
          model code 11 works only;  *)
(* 3.12: user-defined interactions for dependent behavior variables;
         adaptation to compatibility with Lazarus;
(* 3.11: new ERGM options: fixed out-degrees;
         for ERGM case: more real arithmetic in S_Effects to avoid errors coming out of
         checkrunningstatistics for large networks;
         diverse small changes in input/output;
         a window is visible in Siena01 which is helpful for large data sets
         to inform the user that something is happening. *)
(* many changes in 3 beta 3; after that, 3.1 *)
(* 3 beta 2 corrected an error in Digraph-CalcStatistics, used only for ERGM *)
(* 3 beta 1 is the first external release of Siena 3 *)
(* 3.bet17s is a special version: look for "special version" in S.Start and S.Dat *)
(* TS changes in version 1.83: missing codes; only in s_start,
   and in s_dat for this version number and procedure halterror *)
(* MH changes in version 1.84: Estimation of networks of changing
   composition; major changes in reading times of composition change
   in S_Dat and S_Start, Markov simulation in S_Base *)
(* TS changes in version 1.85: changing actor-bound covariates;
   no readln after reading lines of network data files in S_Start *)
(* MH changes in version 1.86: default dichotomization *)
(* TS changes in version 1.87: error correction for changing covariates,
   other format output matrices: no writeln within row *)
(* TS changes in version 1.92: some effects added, program SIENA02 added,
   error message on screen. *)
(* MH changes in version 1.92: undocumented options for composition change
   (for private use), reprogramming of procedure InitialiseActiveActors. *)
(* TS changes in version 1.93:
    for changing covariates subtract global rather than periodwise mean;
    change procedure Statistics so that it does not affect ynm;
    the earlier situation may have led to problems for data with many missings.
   version 1.94:
    inversion step in estimation for 1 observation.   (skipped again in Siena 3 )
   version 1.95:
    vertex parameters in estimation for 1 observation;
    procedure ReviseDefinition split off from ReadDefinition;
   version 1.96:
    option Gibbs = 6 (conditioning on degrees) added.
   version 1.97:
    error in Gibbscode = 5 (for graphs or tournaments) corrected.
   version 1.98:
    modeltype 2 defined in S_Base; output about out-degrees added;
    nmax increased to 500, therefore some array definitions were transferred
    from S_Dat to Eight and shifted to dynamic memory.
    version 1.981:
     repaired little bug (in eff_l3) making it impossible to estimate effects
     of 1/out-degree on rate.
   version 2.0:
   extension to joint dynamics of networks and behavior.
   2.0b: bug repair 1.98 => 1.981 incorporated.
   2.0c: gratification effects for behavior added.
   2.0d: put closefile in Sim and Est before form.close.
   2.0e: new models and deviance calculation for Observations = 1.
         bug repair for contr_fn case h=1, observations = 1, symmetric,
         which led to incorrect standard errors.
   2.0f: the parameters alpa_* taken out of contr_* and moved into
         utilitygain and runstep;
         ModelType 3 added;
   2.0g: start with procedures towards ML estimation;
   2.0j: bug repair for pseudolikelihood estimation
         and for dissimilarity effects for observations = 1;

   MS:  - new unit S_test added, containing a Neyman/Basawa
           test statistic for goodness-of-fit tests.
        - multiple sequences with identical model enabled; the number
           of sequences is given by the global constant <sequences> in unit S_Est.

   2.0k: effects of group position included into behaviour and network
         objective functions, effects of changing behaviour included into
         behaviour gratification function (a start of modelling reinforcement
         learning); further, some reordering of the effects/statistics was
         done and some minor bugs were eliminated. The similarity and
         dissimilarity effects were re-defined, such that estimates of models
         including such effects do not always correspond to those of earlier
         SIENA 2.0* versions (CS).
   2.0m: integration of 2.0k with TS' work on the p*-part and the ML estimation
         option.
   2.0n: inclusion of unit S_ML with first ML estimation procedure,
         missing value handling introduced on covariates and behaviourals,
         as well as range checks on the covariates.
   2.0p: TS: cleaning up a few things, repair of a bug in the p* version.
   MS:   five new units added:
         - new unit S_mcmc containing Markov-chain Monte Carlo-related routines,
           such as routines to assess the accuracy of the Monte Carlo integration, building
           blocks for Metropolis-Hastings algorithms, routines to sample from probability density /
           mass functions, evaluating standard probability density / mass functions, and so on.
         - new unit S_mcmc_I containing routines to conduct likelihood-based inference: m.l. and Bayesian models.
         - new unit S_mcmc_II containing routines to estimate hierarchical / non-hierarchical Bayesian models
           with network objective and network gratification functions.
         - new unit S_re containing routines for estimating
           (a) frequentist random effects models with actor-dependent random effects ("heterogeneous preferences")
               in both the network objective and network gratification function.
           (b) Bayesian hierarchical models with actor-dependent random effects ("heterogeneous preferences")
               in both the network objective and network gratfication function.
            In addition some output-related routines, related to generating R, Matlab, and Gnuplot source code.
         - new unit S_matrices containing basic matrix routines and other basic routines.
   2.0r: Lot of changes by TS. Main structural changes:
         repair conditional estimation;
         double random number generators used in simstats
         to get more precise alignment of common random numbers;
         other values for scale and epsilon;
         new initial values behavior change.
   2.0s: TS: non-directed networks also for longitudinal models
         with MoM estimation.
         In S_ML, where ChangeTie(i,j) is called, this has to be supplemented
         with if symmetric then ChangeTie(j,i) before non-directed networks
         also are possible for ML estimation.
         Explained variation (entropy) added.
   2.0t: TS: some errors in the definition of symmetric effects corrected.
   2.0u: MS: in case of omnibus generalised score tests, single-parameter score tests are computed
         in addition to the omnibus score test.
   2.0v: MS: some bugs removed:
         (1) There was a bug in <ReviseDefinition> which made it impossible to
             estimate the Snijders (2001) model with enforced changes.
         (2) In unconditional moment estimation, the simulation continued until time >= 1,
             including the first change after 1, which is undesirable.
         (3) The variable <cconditional> was manipulated in <S_test>, which is undesired.
         (4) When one or more parameter coordinates were fixed, the standard errors of the non-fixed
             parameter coordinates were wrong.
         (5) In case of omnibus score tests, single-parameter score tests
             were carried out incorrectly.
   2.0w: MS: three additions:
         (1) Derivatives: new method to estimate the matrix of partial derivatives in the MoM framework (see paper),
             including a variance reduction method; the variance reduction method is used by default,
             because it seems to cut down the variance by factor 10 to 100; however, the initial, plain
             estimate of the partial derivatives is printed to the check-file.
             Note that the new method relies heavily on the complete-data likelihood, which means that
             for every class of specification for which the complete-data likelihood takes a different
             form, the method must be adapted. Note furthermore that any permutation of parameters / statistics
             must be done consistently, that is, in the derivatives procedures as well.
             The new method is implemented at the moment for the following cases:
             - conditional and unconditional moment estimation
             - constant and non-constant rate functions
             - network and behavior evolution models
             - enforced and non-enforced changes
             - actors sets with changing composition
             - phase 1 en phase 3.
             Comments:
             - The estimation of ERGMs is unaffected.
             - The method is not implemented for some special cases:
               * undirected networks
               * modeltype 2 and 3 in the case of directed networks.
               * actor / period homogeneity tests; NOTE: in this special case, it is
                 possible (although less convenient) to design dummy variables so that
                 you can use the new method.
             - The method is implemented for phase 1 as well; but the Crude Monte Carlo Method is default in phase 1,
               because when the number of iterations is very small, the crude method seems to be more stable.
             - The file <S_phase3> was added, containing routines needed in phase 1 and 3;
               since those routines rely heavily on variables in <S_base>, the file
               is attached to <S_base> using the <$I> compiler directive.
             In cases where the new method is not implemented, the crude method is used, situated in the procedures <CrudeMethod1,2>.
             NOTE: it is always possible to use the crude method and override the Siena default option
                   by submitting as last argument of <Initialise_d_est> the value <true>!
         (2) In the maximum likelihood framework, the efficient score test of Rao (1948)
             is implemented; one thing that must be taken care of is that <ScoreTest> gets the correct arguments in <S_est>.
         (3) SoNIA input files are generated in the likelihood-based framework (MLE and Bayesian),
             but limited to the network case...
   2.1:  CS: network objective function extended by various interaction effects of covariate similarity
         and covariate-ego.
   2.1b: CS: effects on behavioural rate function introduced, in analogy to network rate function.
   2.1c: CS: simplification of the rate functions as exponential transformations of linear combinations of
         common 'effects'. In this context, (a) the basic rate parameters now are internally stored as
         exponents, and (b) it seemed a good idea to carry out an..
         ELIMINATION OF TWO DOWNWARD COMPATIBILITY ROUTINES for the sake of code legibility and simplicity
         of model and programming. These simplifications concern the following:
         - old parametrisations of the rate function (as described in the 1.98 manual) are no longer
           supported (effects of outdegree, indegree, reciprocity and 1/outdegree).
         - SIENA-1 model definition files with the extensions .mo1 and .mo3 are no longer supported.
   2.2a: CS & MS; integration of versions 2.1c and 2.0w
   2.2a1: TS ; adaptation of the code to signed graphs. changes marked by [s] (one line),
          or by [[s] .. (several lines) .. [s]] between comment signs;
          a few other nonessential changes. 21-jun-05
   2.2a2: TS: decreased scale values in S_Est. 22-jun-05
   2.2b:  MS: debugging of efficient derivative estimates;
          CS: debugging of behaviour endowment function
   3.0:   TS: data structures as edge lists in Eight and Digraph;
              effects defined in S_Effects;
              more efficient random choice in RanGen.
   3.0a   MS:
          with regard to (generalized) score tests:
          - one-sided tests added
          - one-step estimator added
          - ML score test completed, incl. one-sided tests and one-step estimator.
   3.0b   MS:
          - Bayesian framework extended to more than two observations.
          - Bayesian framework extended to behavior.
          - random effects framework extended to more than two observations.
          - random effects framework extended to behavior.
   3.0c   MS: two bugs removed:
          - one bug in procedure <Scores> in unit <S_ml> which affected the ML estimation of
          rate parameters in network and behavior models and sometimes led to divergence.
          - one bug in procedure <Stepprob> in unit <S_ml> which affected the treatment of
          missings in the ML estimation.

*)

  errorstring := '';
end.


